(***********************************************************************
 * File: Cache.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 *******************************************************************
 * This file specifies the System Cache and TLB
 * 
 ******************************************************************)
Require Import Omega.
Require Import Arith.
Require Import List.
Require Export Maps.
Require Import Maps_proof.
Require Export Memory.
Require Export Page.

(********************)
(* Cache Definition *)
(********************)

Definition cache := mapping vadd page.

(* Caches are finite mappings *) 
Parameter max_cache : nat.
Parameter max_cache_pos : max_cache > 0.

Definition is_full_cache (c:cache)  :=
  eq_nat_dec (map_size c) max_cache.

(* Selects an entry to delete from a cache *)
(* Parameter map_drop_cache : cache -> cache. *)
(* Parameter map_drop_cache_behavior : forall (c:cache), *)
(*   (c = map_empty vadd page /\ map_drop_cache c = map_empty vadd page) \/ *)
(*   (exists va_del, map_valid_index vadd_eq c va_del /\ *)
(*     map_drop_cache c = map_drop vadd_eq c va_del). *)

Definition cache_value (c:cache) (va:vadd) : exc page vadd :=
  map_apply vadd_eq c va.

(* Definition cache_add (c:cache) (va:vadd) (pg:page): cache := *)
(*   If is_full_cache c then *)
(*     if map_valid_index_fix vadd_eq c va then *)
(*       map_add vadd_eq c va pg *)
(*     else *)
(*       map_add vadd_eq (map_drop_cache c) va pg *)
(*   else *)
(*     map_add vadd_eq c va pg. *)

Inductive cache_distinct : cache -> Prop :=
| empty_cache_distinct : cache_distinct (map_empty vadd page)
| cons_cache_distinct : forall (c:cache) (va:vadd) (pg:page),
  cache_distinct c ->
  ~(exists pg', Value _ pg' = cache_value c va) ->
  cache_distinct ((Item va pg)::c).

Definition cache_flush : cache -> cache :=
  fun _ => map_empty vadd page.

Definition cache_add (c1 c2:cache) (va:vadd) (pg:page) :=
  Value _ pg = map_apply vadd_eq c2 va /\
  (forall va' pg', va' <> va -> Value _ pg' = map_apply vadd_eq c2 va' ->
    Value _ pg' = map_apply vadd_eq c1 va') /\
  map_size c2 <= max_cache /\
  cache_distinct c2.



(******************)
(* TLB Definition *)
(******************)

Definition tlb := mapping vadd madd.

(* TLB are finite mappings *) 
Parameter max_tlb : nat.
Parameter max_tlb_pos : max_tlb > 0.

Definition is_full_tlb (t:tlb) :=
  eq_nat_dec (map_size t) max_tlb.

(* Selects an entry and deletes it from a TLB *)
(* Parameter map_drop_tlb : tlb -> tlb. *)
(* Parameter map_drop_tlb_behavior : forall (t:tlb), *)
(*   (t = map_empty vadd madd /\ map_drop_tlb t = map_empty vadd madd) \/ *)
(*   (exists va_del, map_valid_index vadd_eq t va_del /\ *)
(*     map_drop_tlb t = map_drop vadd_eq t va_del). *)

Definition tlb_value (t:tlb) (va:vadd) : exc madd vadd :=
  map_apply vadd_eq t va.

(* Definition tlb_add (t:tlb) (va:vadd) (ma:madd): tlb := *)
(*   If is_full_tlb t then *)
(*     if map_valid_index_fix vadd_eq t va then *)
(*       map_add vadd_eq t va ma *)
(*     else *)
(*       map_add vadd_eq (map_drop_tlb t) va ma *)
(*   else *)
(*     map_add vadd_eq t va ma. *)

Inductive tlb_distinct : tlb -> Prop :=
| empty_tlb_distinct : tlb_distinct (map_empty vadd madd)
| cons_tlb_distinct : forall (t:tlb) (va:vadd) (ma:madd),
  tlb_distinct t ->
  ~(exists ma', Value _ ma' = tlb_value t va) ->
  tlb_distinct ((Item va ma)::t).

Definition tlb_flush : tlb -> tlb :=
  fun _ => map_empty vadd madd.

Definition tlb_add (t1 t2:tlb) (va:vadd) (ma:madd) :=
  Value _ ma = map_apply vadd_eq t2 va /\
  (forall va' ma', va' <> va -> Value _ ma' = map_apply vadd_eq t2 va' ->
    Value _ ma' = map_apply vadd_eq t1 va') /\
  map_size t2 <= max_tlb /\
  tlb_distinct t2.

(* Auxiliary lemmas *)
Lemma cache_add_size : forall (old_cache new_cache:cache) (va:vadd) (pg:page),
  cache_add old_cache new_cache va pg ->
  map_size old_cache <= max_cache ->
  map_size new_cache <= max_cache.
Proof.
  intros old_cache new_cache va pg H_new_cache H_old_size.
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
Qed.  

Lemma cache_map_add_distinct : forall (old_cache new_cache:cache) (va:vadd) (pg:page),
  new_cache = map_add vadd_eq old_cache va pg ->
  cache_distinct old_cache -> 
  cache_distinct new_cache.
Proof.
  intros old_cache new_cache va pg H_new_cache H_old_cache_distinct.
  subst new_cache.
  induction old_cache.
  simpl.
  repeat constructor.
  unfold cache_value.
  intros [pg' H_pg'_map].
  simpl in H_pg'_map; discriminate H_pg'_map.
  simpl.
  induction (vadd_eq (item_index a) va).
  inversion H_old_cache_distinct.
  rewrite <- H in a0.
  simpl in a0.
  subst va0. 
  constructor; auto.
  replace a with (Item (item_index a) (item_info a)).
  constructor.
  apply IHold_cache.
  inversion_clear H_old_cache_distinct; auto.
  unfold cache_value.
  intros [pg' H_pg'_map].
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  inversion H_old_cache_distinct.
  apply H2.
  exists pg'.
  unfold cache_value.
  rewrite <- H in H_pg'_map.
  simpl in H_pg'_map; auto.
  induction a; simpl; auto.
Qed.

Lemma cache_map_drop_distinct : forall (old_cache new_cache:cache) (va:vadd),
  new_cache = map_drop vadd_eq old_cache va ->
  cache_distinct old_cache -> 
  cache_distinct new_cache.
Proof.
  intros old_cache new_cache va H_new_cache H_old_cache_distinct.
  subst new_cache.
  induction old_cache.
  constructor.
  inversion_clear H_old_cache_distinct.
  simpl.
  induction (vadd_eq va va0).
  subst va0; auto.
  constructor; auto.
  intros [pg' H_pg'_map].
  unfold cache_value in *.
  apply H0.
  exists pg'.
  rewrite map_drop_apply_distinct in H_pg'_map; auto.
Qed.

Lemma cache_add_distinct : forall (old_cache new_cache:cache) (va:vadd) (pg:page),
  cache_add old_cache new_cache va pg ->
  cache_distinct old_cache -> 
  cache_distinct new_cache.
Proof.
  intros old_cache new_cache va pg H_new_cache H_old_cache_distinct.
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
Qed.


Lemma tlb_add_size : forall (old_tlb new_tlb:tlb) (va:vadd) (ma:madd),
  tlb_add old_tlb new_tlb va ma ->
  map_size old_tlb <= max_tlb ->
  map_size new_tlb <= max_tlb.
Proof.
  intros old_tlb new_tlb va ma H_new_tlb H_old_size.
  inversion_clear H_new_tlb as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
Qed.

Lemma tlb_map_add_distinct : forall (old_tlb new_tlb:tlb) (va:vadd) (ma:madd),
  new_tlb = map_add vadd_eq old_tlb va ma ->
  tlb_distinct old_tlb -> 
  tlb_distinct new_tlb.
Proof.
  intros old_tlb new_tlb va ma H_new_tlb H_old_tlb_distinct.
  subst new_tlb.
  induction old_tlb.
  simpl.
  repeat constructor.
  unfold tlb_value.
  intros [ma' H_ma'_map].
  simpl in H_ma'_map; discriminate H_ma'_map.
  simpl.
  induction (vadd_eq (item_index a) va).
  inversion H_old_tlb_distinct.
  rewrite <- H in a0.
  simpl in a0.
  subst va0. 
  constructor; auto.
  replace a with (Item (item_index a) (item_info a)).
  constructor.
  apply IHold_tlb.
  inversion_clear H_old_tlb_distinct; auto.
  unfold tlb_value.
  intros [ma' H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  inversion H_old_tlb_distinct.
  apply H2.
  exists ma'.
  unfold tlb_value.
  rewrite <- H in H_ma'_map.
  simpl in H_ma'_map; auto.
  induction a; simpl; auto.
Qed.

Lemma tlb_map_drop_distinct : forall (old_tlb new_tlb:tlb) (va:vadd),
  new_tlb = map_drop vadd_eq old_tlb va ->
  tlb_distinct old_tlb -> 
  tlb_distinct new_tlb.
Proof.
  intros old_tlb new_tlb va H_new_tlb H_old_tlb_distinct.
  subst new_tlb.
  induction old_tlb.
  constructor.
  inversion_clear H_old_tlb_distinct.
  simpl.
  induction (vadd_eq va va0).
  subst va0; auto.
  constructor; auto.
  intros [ma' H_ma'_map].
  unfold tlb_value in *.
  apply H0.
  exists ma'.
  rewrite map_drop_apply_distinct in H_ma'_map; auto.
Qed.

Lemma tlb_add_distinct : forall (old_tlb new_tlb:tlb) (va:vadd) (ma:madd),
  tlb_add old_tlb new_tlb va ma ->
  tlb_distinct old_tlb -> 
  tlb_distinct new_tlb.
Proof.
  intros old_tlb new_tlb va ma H_new_tlb H_old_tlb_distinct.
  inversion_clear H_new_tlb as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
Qed.
