(***********************************************************************
 * File: Cache_Leakage_indistinguishability.v
 * $Revision: $
 * $Date:  $
 * $LastChangedBy:  $
 ***********************************************************************
 * Cache non leakage isolation.
 **********************************************************************)

Require Import List.
Require Import State_equivalence.
Require Import Exec_invariant.
Require Import Write_Isolation.
Require Import Isolation.
Require Import Execution_isolation.
Require Import Leak_traces.


Section Cache_Non_leakage.

Variable ctxt : context.

(* An instance of observation obs: os |-> [vadd] and cache_test *)

Definition observation : Set := mapping os_ident (list vadd).
Definition empty_observation (o:observation) : Prop :=
  o = map_empty os_ident (list vadd).

(* Add new entry to the observation mapping *)
Definition add_observation (o:observation) (osi:os_ident) (va:vadd) : observation :=
  match map_apply os_ident_eq o osi with
    | Error _ => map_add os_ident_eq o osi (va::nil)
    | Value l => map_add os_ident_eq o osi (va::l)
  end.

(* cache_to_obs: *)
(* Returns the list of vadd corresponding to each non-active os in the cache.
   The parameter aos is the identifier of the active os *)
Fixpoint cache_to_obs (aos:os_ident) (cache:list (item vadd page)) : observation :=
  match cache with
    | nil => map_empty os_ident (list vadd)
    | (entry::cs) => match page_owned_by (item_info entry) with
                      | Os_PO osi => if os_ident_eq aos osi then cache_to_obs aos cs
                        else add_observation (cache_to_obs aos cs) osi (item_index entry)
                      | _ => cache_to_obs aos cs
                     end
  end.
                      
(* cache_test *)
(* Returns the observations corresponding to cache entries in s. *)
Definition cache_test (s:state) (s':state) (a:action) : observation :=
  cache_to_obs (st_active_os s') (map2list (st_cache s')).

(* Lemma cache_test empty *)
(* Given a valid state and a non active osi victim, we prove that the
   active os cannot obtain information about osi_v using the leakage
   function cache_test. *)
Lemma cache_test_empty : forall (s s':state) (a:action),
  valid_state ctxt s' -> cache_test s s' a = map_empty os_ident (list vadd).
Proof.
   intros s s' a H_valid.
  unfold cache_test, map_empty.
  clear a s; rename s' into s.
  unfold map2list.

  generalize H_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_untosexmd
    H_va_has_pa H_tlb_valid; unfold valid_cache in H_cache_valid;
      destruct H_cache_valid as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_included in H_cache_included.
  
  assert (forall (va : vadd) (pg : page),
    Value vadd pg = cache_value (st_cache s) va ->
    (os_accessible ctxt va ->
      page_owned_by pg = Os_PO (st_active_os s)) /\
    (~ os_accessible ctxt va -> page_owned_by pg = Hyp_PO)) as H_cache_owner.
  intros va pg H_pg_value; elim valid_state_cache_owner with ctxt s va pg; auto.
  
  induction (st_cache s); auto.

  assert (map_size c <= max_cache) as H_c_size.
  inversion_clear H_cache_size; try simpl in H;
    try omega.
  assert (cache_distinct c) as H_c_distinct.
  inversion_clear H_cache_distinct; auto.
  assert (forall (va : vadd) (pg : page),
    Value vadd pg = map_apply vadd_eq c va ->
    exists ma : madd,
      va_mapped_to_ma s (st_active_os s) va ma /\
      page_type (st_memory s) ma = Some RW /\
      Value madd pg = map_apply madd_eq (st_memory s) ma) as H_c_included.
  intros.
  apply H_cache_included.
  cut (item_index a <> va).
  intro.
  simpl.
  case (vadd_eq va (item_index a)); auto.
  intro; elim H0; auto.
  intro H_a_eq_va.
  inversion H_cache_distinct.
  subst c0.
  subst a.
  simpl in H_a_eq_va.
  subst va0.
  apply H3.
  exists pg; auto.

  assert (forall (va : vadd) (pg : page),
    Value vadd pg = cache_value c va ->
    (os_accessible ctxt va ->
      page_owned_by pg = Os_PO (st_active_os s)) /\
    (~ os_accessible ctxt va -> page_owned_by pg = Hyp_PO)) as H_c_owner.
  intros va pg H_pg_value.
  apply H_cache_owner; auto.
  simpl.
  cut (item_index a <> va).
  intro.
  simpl.
  case (vadd_eq va (item_index a)); auto.
  intro; elim H; auto.
  intro H_a_eq_va.
  inversion H_cache_distinct.
  subst c0.
  subst a.
  simpl in H_a_eq_va.
  subst va0.
  apply H2.
  exists pg; auto.
  

  simpl.
  remember (page_owned_by (item_info a)) as a_owner.
  induction a_owner; try (apply IHc; auto).
  cut (st_active_os s = o).
  intro; subst o.
  case (os_ident_eq (st_active_os s) (st_active_os s)); auto.
  intro H_abs; elim H_abs; auto.
 
  elim H_cache_owner with (item_index a) (item_info a).
  unfold os_accessible.
  case (ctxt_vadd_accessible ctxt (item_index a)).
  intros; rewrite H in Heqa_owner; congruence.
  intros; rewrite H0 in Heqa_owner; congruence.
  simpl.
  case (vadd_eq (item_index a) (item_index a)); auto.
  intro H_abs; contradiction H_abs; auto.
Qed.

End Cache_Non_leakage.

Section Cache_Leakage_Indistinguishability.

Variable ctxt : context.
Variable osi : os_ident.

(* Non leakage over traces *)
CoInductive leak_same_os_actions : leak_state_stream observation -> leak_state_stream observation -> Prop :=
  Os_action: forall (str1 str2:leak_state_stream observation)
    (A:leak_action observation) (s1 s2:leak_state observation),
    leak_same_os_actions str1 str2 -> 
    osi_action osi (leak_state_state s1) (A (leak_state_obs s1)) ->
    osi_action osi (leak_state_state s2) (A (leak_state_obs s2)) ->
    leak_same_os_actions (Cons (Leak_Ex_State s1 A cache_test) str1) (Cons (Leak_Ex_State s2 A cache_test) str2)
  | Not_os_action1: forall (str1 str2:leak_state_stream observation)
    (A:leak_action observation) (s1 s2:leak_state observation),
    leak_same_os_actions str1 str2 -> 
    ~(osi_action osi (leak_state_state s1) (A (leak_state_obs s1))) ->
    leak_same_os_actions (Cons (Leak_Ex_State s1 A cache_test) str1) str2
  | Not_os_action2: forall (str1 str2:leak_state_stream observation)
    (A:leak_action observation) (s1 s2:leak_state observation),
    leak_same_os_actions str1 str2 -> 
    ~(osi_action osi (leak_state_state s2) (A (leak_state_obs s2))) ->
    leak_same_os_actions str1 (Cons (Leak_Ex_State s2 A cache_test) str2). 


CoInductive leak_osi_equivalent_traces : leak_state_stream observation -> leak_state_stream observation -> Prop :=
  eq_trace_osi_action: forall (str1 str2:leak_state_stream observation)
    (A1 A2:leak_action observation) (s1 s2:leak_state observation),
    leak_osi_equivalent_traces str1 str2 -> 
    osi_action osi (leak_state_state s1) (A1 (leak_state_obs s1)) ->
    osi_action osi (leak_state_state s2) (A2 (leak_state_obs s2)) ->
    osi_equivalent ctxt (leak_state_state s1) (leak_state_state s2) osi ->
    leak_osi_equivalent_traces (Cons (Leak_Ex_State s1 A1 cache_test) str1) (Cons (Leak_Ex_State s2 A2 cache_test) str2)
  | eq_trace_not_osi_action1: forall (str1 str2:leak_state_stream observation)
    (A:leak_action observation) (s1 s2:leak_state observation),
    leak_osi_equivalent_traces str1 str2 -> 
    ~(osi_action osi (leak_state_state s1) (A (leak_state_obs s1))) ->
    (* osi_equivalent ctxt s1 (ex_state (hd str1)) osi -> *)
    leak_osi_equivalent_traces (Cons (Leak_Ex_State s1 A cache_test) str1) str2
  | eq_trace_not_osi_action2: forall (str1 str2:leak_state_stream observation)
    (A:leak_action observation) (s1 s2:leak_state observation),
    leak_osi_equivalent_traces str1 str2 -> 
    ~(osi_action osi (leak_state_state s2) (A (leak_state_obs s2))) ->
    (* osi_equivalent ctxt s1 (ex_state (hd str2)) osi -> *)
    leak_osi_equivalent_traces str1 (Cons (Leak_Ex_State s2 A cache_test) str2). 

CoInductive cache_leak_trace : leak_state_stream observation -> Prop :=
  Cache_leak_trace : 
           forall (s:leak_exec_state observation)(str:leak_state_stream observation),
           cache_leak_trace str ->
           l_ex_function s = cache_test ->
	   leak_system_transition ctxt s (hd str) ->
           cache_leak_trace (Cons s str).

(* cache_leak_run: a leak_trace that is a run where 
  the observation function is always cache_test *)             
Definition cache_leak_run (str:leak_state_stream observation): Prop := 
  leak_init_state ctxt empty_observation (hd str) /\ 
  cache_leak_trace str.

Lemma cache_leak_run_inv : 
          forall ostr, cache_leak_run ostr -> cache_leak_run (tl ostr).
Proof.
intros ostr; unfold cache_leak_run; simpl.
case ostr; simpl; unfold leak_init_state.
intros s str1 (H_valids,H_trace).
inversion H_trace.
split.
inversion H3.
unfold leak_valid_state; simpl; auto.
assert (valid_state ctxt (leak_state_state (l_ex_state (hd str1)))) as H_validstHdstr1.
(* Proof that the state of (hd str1) is a valid state *)
apply valid_state_invariant with s1 a;
   destruct H_valids as [valids H_emptyobs]; simpl; auto.
split; auto.
(* Proof that the observation of (hd str1) is empty *)
generalize (cache_test_empty ctxt s1 (leak_state_state (l_ex_state (hd str1))) 
                                  a H_validstHdstr1); intro.
   rewrite H2 in H5;
     unfold empty_observation; rewrite <- H10; simpl; auto.
(* str1 is an observable trace *)
assumption.
Qed.

Theorem cache_leakage_indistinguishability : 
   forall (str1 str2:leak_state_stream observation),
     cache_leak_run str1 -> cache_leak_run str2 ->
     leak_same_os_actions str1 str2 -> 
     osi_equivalent ctxt (leak_state_state (l_ex_state (hd str1))) 
                         (leak_state_state (l_ex_state (hd str2))) osi ->
     leak_osi_equivalent_traces str1 str2.
Proof.
cofix str1.
  intros str2 str3 H_run2 H_run3 H_R H_first_osi_eq.
  inversion_clear H_R in H_run2 H_run3 H_first_osi_eq .  
 (* osi is the active_os in s1 and s2 *)
  simpl in H_first_osi_eq.
  constructor; auto.
  apply str1; auto.
 (* Proof that str0 is a cache_leak_run *)
  unfold cache_leak_run, leak_run, leak_init_state in H_run2, H_run3.
  simpl in H_run2, H_run3. destruct H_run2; destruct H_run3. 
  unfold cache_leak_run, leak_run, leak_init_state.
  inversion_clear H3.
  unfold leak_system_transition in H8; simpl in H8; split; auto.
  inversion H8; auto.
 (* generalize H7; intros (validst,(Ao1a,(validexec,obs)));auto.*)
(* Proof that state of hd str1 is valid *)
  assert (valid_state ctxt (leak_state_state (l_ex_state (hd str0)))) as H_validstHdrstr0.
  apply valid_state_invariant with s a.
  destruct H2. unfold leak_valid_state in H2;auto.
  unfold s' in H3; auto.
  (* Proof of validity of observable trace state *)
   split; auto.
 (* Proof that the observation of (hd str0) is empty *)
  destruct H2; auto.
generalize (cache_test_empty ctxt s s' a H_validstHdrstr0); intro.
   rewrite H9 in H15; auto.
  (* Proof that str4 is an obs_run *)
   unfold cache_leak_run, leak_run, leak_init_state in H_run2,H_run3.
   simpl in H_run2, H_run3. destruct H_run2; destruct H_run3.    
   unfold cache_leak_run, leak_run, leak_init_state.
   assert (valid_state ctxt (leak_state_state (l_ex_state (hd str4)))) as H_validstHdrstr4.
  apply valid_state_invariant with (leak_state_state s2) (A (leak_state_obs s2)).
  destruct H4. unfold leak_valid_state in H4;auto.
  inversion H5.
  unfold leak_system_transition in H10; simpl in H10; auto.
  inversion H10;auto.
  unfold s0,a,s',o in H11. rewrite H13 in H11; assumption.
  (* Proof of validity of observable trace state *)
   split. split;auto.
 (* Proof that the observation of (hd str0) is empty *)
  destruct H4; auto.
  generalize 
   (cache_test_empty ctxt (leak_state_state s2) 
                          (leak_state_state (l_ex_state (hd str4)))
                          (A (leak_state_obs s2)) H_validstHdrstr4); intro.
   inversion H5.
   unfold leak_system_transition in H12; simpl in H12; auto.
   inversion H12;auto.
   unfold s0,a,s',o',o in H14. rewrite H15,H7 in H14;auto.
   unfold empty_observation. symmetry;auto.
  (* Proof that str4 is a cache_leak_trace *)
   inversion H5;auto.
  (* Proofs that are osi_equivalent *)
  unfold cache_leak_run, leak_run, leak_init_state in H_run2,H_run3;
   simpl in H_run2, H_run3; destruct H_run2; destruct H_run3.
  destruct H2; destruct H4; unfold leak_valid_state in H2,H4;auto.    
  generalize full_isolation_proof; unfold full_isolation; intro H_full_isolation_proof.
  apply H_full_isolation_proof with 
     (leak_state_state s1)(leak_state_state s2) (A (leak_state_obs s1)); auto;
       inversion H3;inversion H5;
         unfold leak_system_transition in H12,H17; simpl in H12,H17; auto;
            inversion H12;inversion H17;auto; 
              unfold s3,a,s',o,o' in H18; rewrite H20 in H18;
               unfold s4,a0,s'0,o'0,o1 in H24;rewrite H26 in H24;
                auto.
  assert (A (leak_state_obs s2) = A (leak_state_obs s1)) as H_Aobss2eqAobsso1;
    inversion H6;inversion H7.
      rewrite H31; rewrite H32;auto.
  rewrite <- H_Aobss2eqAobsso1;auto.
 (* a is not an osi action wrt to s1 *)
 constructor; auto.
  apply str1; auto.
 (* Proof that str0 is a cache_leak_run *)
  apply (cache_leak_run_inv (Cons (Leak_Ex_State s1 A cache_test) str0));auto. 
 (* Proof that the states of hd str0 and hd str3 are osi-equivalents *)
  apply osi_equivalent_trans with (leak_state_state s1);
    simpl in H_first_osi_eq; auto. (* one equality out *)
  apply osi_equivalent_symm.
  inversion H_run2.
  unfold leak_init_state in H1; simpl in H1.
  inversion H2;auto.
  inversion H7.
  apply not_osi_action_state_eq with (A (leak_state_obs s1));
   unfold s0,a,s',o in H8; simpl in H8; rewrite H10 in H8; simpl in H8; auto.
 destruct H1; unfold leak_valid_state in H1; auto.
 (* a is not an osi action wrt to s2 *)
  constructor; auto.
  apply str1; auto.
(* Proof that str4 is a cache_leak_run *)
  apply (cache_leak_run_inv (Cons (Leak_Ex_State s2 A cache_test) str4));auto. 
(* Proof that the states of hd str2 and hd str4 are osi-equivalents *)
  apply osi_equivalent_trans with (leak_state_state s2);
    simpl in H_first_osi_eq; auto. (* one equality out *)
  inversion H_run3.
  unfold leak_init_state in H1; simpl in H1.
  inversion H2;auto.
  inversion H7.
  apply not_osi_action_state_eq with (A (leak_state_obs s2));
   unfold s0,a,s',o in H8; simpl in H8; rewrite H10 in H8; simpl in H8; auto.
 destruct H1; unfold leak_valid_state in H1; auto.
Qed.

End Cache_Leakage_Indistinguishability.







