(***********************************************************************
 * File: Execution_isolation.v
 * $Revision: 404 $
 * $Date: 2010-12-30 23:08:20 -0200 (Thu, 30 Dec 2010) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * Definition of equivalence between states (for the AOS and a given
   osi). Proofs of equivalence properties and hypervisor behaviour
   related to equivalence.
 **********************************************************************) 

Section Cache_coherence.

Require Import Actions.
Require Import Exec.
Require Import Exec_invariant.
Require Import State_equivalence.

Set Implicit Arguments.

Variable ctxt : context. 

Definition c (s:state) (va:vadd) : exc page vadd :=
  cache_value (st_cache s) va.

Definition m (s:state) (va:vadd) : exc page vadd :=
  match map_apply os_ident_eq (st_oss s) (st_active_os s) with
    | Error _ => Error page va
    | Value aos =>
  match map_apply os_ident_eq (st_hypervisor s) (st_active_os s) with
    | Error _ => Error page va
    | Value p2m =>
  match map_apply padd_eq p2m (os_current_page aos) with
    | Error _ => Error page va
    | Value pt_madd =>
  match map_apply madd_eq (st_memory s) pt_madd with
    | Error _ => Error page va
    | Value pt_page =>
  match page_content pt_page with
    | PT_c mp =>
      match map_apply vadd_eq mp va with
        | Error _ => Error page va
        | Value ma =>
      match map_apply madd_eq (st_memory s) ma with
        | Error _ => Error page va
        | Value pg =>
      match page_content pg with
        | RW_c _ => Value vadd pg
        | _ => Error page va
      end                  
      end
      end
    | _ => Error page va
  end
  end
  end
  end
  end.

Definition in_dom_of (f:state -> vadd -> exc page vadd) (s:state) (va:vadd) : Prop :=
  match f s va with
    | Value _ => True
    | Error _ => False
  end.

Definition cache_coherence (s s':state) : Prop :=
  forall va, os_accessible ctxt va ->

  (* Hit / Hit *)
  (in_dom_of c s va -> in_dom_of c s' va -> c s va = c s' va) /\
  (* Miss / Miss *)
  (~ in_dom_of c s va -> ~ in_dom_of c s' va -> m s va = m s' va) /\
  (* Miss / Hit *)
  (~ in_dom_of c s va -> in_dom_of c s' va -> m s va = c s' va) /\
  (* Hit / Miss *)
  (in_dom_of c s va -> ~ in_dom_of c s' va -> c s va = m s' va).


(* A version of cache_included *)
Lemma c_eq_m : forall s va,
  valid_state ctxt s -> in_dom_of c s va -> 
  c s va = m s va.
Proof.
  intros s va H_valid H_in_dom.

  unfold in_dom_of in H_in_dom.
  remember (c s va) as pg_value.
  induction pg_value; try contradiction.
  rename v into pg; clear H_in_dom.

  generalize H_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].

  unfold valid_cache in H_cache_valid.
  destruct H_cache_valid as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_pg_type H_pg_map]]] ; auto.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  unfold hypervisor_mapping in H_p2m_map.
  unfold m.

  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map.
  rewrite <- H_pg_map.
  unfold page_type in H_pg_type.
  rewrite <- H_pg_map in H_pg_type.
  induction (page_content pg); try discriminate H_pg_type; auto.
Qed.

Lemma m_eq_m' : forall s s' osi va,
  valid_state ctxt s -> valid_state ctxt s' ->
  st_active_os s = osi -> st_active_os s' = osi ->
  osi_equivalent ctxt s s' osi ->
  os_accessible ctxt va ->
  m s va = m s' va.
Proof.
  intros s s' osi va H_s_valid H_s'_valid H_osi_active H_osi_active' H_eq H_va_accessible.

  unfold osi_equivalent in H_eq.
  destruct H_eq as [H_active_eq [H_oss_eq [H_hyper_eq H_pt_page_eq]]].
  unfold m.
  rewrite H_osi_active.
  rewrite H_osi_active'.
  rewrite <- H_oss_eq.
  remember (map_apply os_ident_eq (st_oss s) osi) as aos_value.
  induction aos_value; auto.
  rename v into aos.

  remember (map_apply os_ident_eq (st_hypervisor s) osi) as p2m_value.
  induction p2m_value.
  rename v into p2m; rename Heqp2m_value into H_p2m_map.

  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_valid_virtmap H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi aos p2m; auto.
  intros pt_addr H_pt_addr_map [H_pt_type H_pt_owner].
  rewrite <- H_pt_addr_map.
  unfold page_type in H_pt_type.
  remember (map_apply madd_eq (st_memory s) pt_addr) as pt_page_value.
  induction pt_page_value; try discriminate.
  rename v into pt_page; rename Heqpt_page_value into H_pt_page_map.
  remember (page_content pt_page) as pt_content.
  induction pt_content; try discriminate.
  rename m0 into mp.
  clear H_valid_currpg H_pt_owner H_pt_type.

  unfold osi_equivalent_pt_pages in H_pt_page_eq.
  destruct H_pt_page_eq as [H_pt_page_eq H_pt_page_eq'].
  unfold osi_pt_pages_eq in H_pt_page_eq.
  elim H_pt_page_eq with (os_current_page aos) pt_page mp; auto.
  2:exists pt_addr; try exists p2m; auto.
  clear H_pt_page_eq.
  intros pt_page' H_pt_map' H_mp'.
  destruct H_mp' as [mp' H_mp'_map H_pt_page_eq].
  unfold padd_page in H_pt_map'.
  destruct H_pt_map' as [pt_addr' H_pt_addr'_map H_pt_page'_map].
  unfold hypervisor_mapping_madd in H_pt_addr'_map.
  destruct H_pt_addr'_map as [p2m' H_p2m'_map H_pt_addr'_map].
  rewrite <- H_p2m'_map.
  rewrite <- H_pt_addr'_map.
  rewrite <- H_pt_page'_map.
  rewrite <- H_mp'_map.

  unfold osi_pt_pages_eq in H_pt_page_eq'.
  elim H_pt_page_eq' with (os_current_page aos) pt_page' mp'; auto.
  2: exists pt_addr'; try exists p2m'; auto.
  clear H_pt_page_eq'.
  intros pt_page2 H_pt_map2 H_mp2.
  destruct H_mp2 as [mp2 H_mp2_map H_pt_page_eq'].
  unfold padd_page in H_pt_map2.
  destruct H_pt_map2 as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [p2m2 H_p2m2_map H_pt_addr2_map].
  rewrite <- H_p2m_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  rewrite <- H_pt_addr_map in H_pt_addr2_map; injection H_pt_addr2_map;
    intro; subst pt_addr2; clear H_pt_addr2_map.
  rewrite <- H_pt_page_map in H_pt_page2_map; injection H_pt_page2_map;
    intro; subst pt_page2; clear H_pt_page2_map.
  rewrite <- Heqpt_content in H_mp2_map; injection H_mp2_map;
    intro; subst mp2; clear H_mp2_map.
  
  remember (map_apply vadd_eq mp va) as ma_value.
  induction ma_value.
  rename v into ma; rename Heqma_value into H_ma_map.

  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_virtual_mapping in H_valid_virtmap.
  elim H_valid_virtmap with osi pt_addr mp pt_page va ma; auto.
  Focus 2.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi aos p2m; auto.
  intros pt_addr2 H_pt_addr2_map [H_pt_addr_type H_pt_addr_owner].
  rewrite <- H_pt_addr_map in H_pt_addr2_map; injection H_pt_addr2_map;
    intro; subst pt_addr2; clear H_pt_addr2_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  clear H_valid_currpg.

  intros pg H_pg_map [H_pg_owner H_abs].
  clear H_abs H_valid_virtmap.
  rewrite <- H_pg_map.
  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_valid_currpg.
  unfold va_has_valid_pa, hypervisor_mapping_madd in H_va_has_pa.
  destruct H_va_has_pa as [H_va_has_pa H_valid_cache_tlb].
  clear H_valid_cache_tlb.
  elim H_va_has_pa with osi (os_current_page aos) pt_addr pt_page mp va ma; auto.
  2: exists p2m; auto.
  intros pa [p2m2 H_p2m2_map H_ma_map_pa].
  rewrite <- H_p2m_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  clear H_va_has_pa H_valid_virtmap.

  remember (page_content pg) as pg_content.
  induction pg_content.
  rename o into value; rename Heqpg_content into H_pg_content.
  (* RW page *)
  elim H_pt_page_eq with va pa ma pg; auto.
  2: unfold page_type; rewrite <- H_pg_map, <- H_pg_content; auto.
  2: exists p2m; auto.
  unfold hypervisor_mapping_madd.
  intros ma' H_ma'_map [[p2m2 H_p2m2_map H_ma'_map_pa] H_pg_map'].
  rewrite <- H_p2m'_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  rewrite <- H_ma'_map.
  rewrite <- H_pg_map'.
  rewrite <- H_pg_content; auto.

  (* PT page *)
  remember (map_apply vadd_eq mp' va) as ma_value.
  induction ma_value; auto.
  rename v into ma'; rename Heqma_value into H_ma'_map.
  remember (map_apply madd_eq (st_memory s') ma') as pg'_value.
  induction pg'_value; auto.
  rename v into pg'; rename Heqpg'_value into H_pg'_map.
  remember (page_content pg') as pg'_content.
  induction pg'_content; auto.
  generalize H_s'_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_valid_virtmap H_valid_currpg.
  unfold va_has_valid_pa, hypervisor_mapping_madd in H_va_has_pa.
  destruct H_va_has_pa as [H_va_has_pa H_valid_cache_tlb].
  clear H_valid_cache_tlb.
  elim H_va_has_pa with osi (os_current_page aos) pt_addr' pt_page' mp' va ma'; auto.
  2: exists p2m'; auto.
  intros pa' [p2m2 H_p2m2_map H_ma'_map_pa].
  rewrite <- H_p2m'_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  clear H_va_has_pa.

  elim H_pt_page_eq' with va pa' ma' pg'; auto.
  2: unfold page_type; rewrite <- H_pg'_map, <- Heqpg'_content; auto.
  2: exists p2m'; auto.
  unfold hypervisor_mapping_madd.
  intros ma2 H_ma2_map [[p2m2 H_p2m2_map H_ma2_map_pa] H_pg_map2].
  rewrite <- H_p2m_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  rewrite <- H_ma_map in H_ma2_map; injection H_ma2_map;
    intro; subst ma2; clear H_ma2_map.
  rewrite <- H_pg_map in H_pg_map2; injection H_pg_map2;
    intro; subst pg'; clear H_pg_map2.
  rewrite <- Heqpg_content in Heqpg'_content; discriminate.

  (* Other content *)
  remember (map_apply vadd_eq mp' va) as ma_value.
  induction ma_value; auto.
  rename v into ma'; rename Heqma_value into H_ma'_map.
  remember (map_apply madd_eq (st_memory s') ma') as pg'_value.
  induction pg'_value; auto.
  rename v into pg'; rename Heqpg'_value into H_pg'_map.
  remember (page_content pg') as pg'_content.
  induction pg'_content; auto.
  generalize H_s'_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_valid_virtmap H_valid_currpg.
  unfold va_has_valid_pa, hypervisor_mapping_madd in H_va_has_pa.
  destruct H_va_has_pa as [H_va_has_pa H_valid_cache_tlb].
  clear H_valid_cache_tlb.
  elim H_va_has_pa with osi (os_current_page aos) pt_addr' pt_page' mp' va ma'; auto.
  2: exists p2m'; auto.
  intros pa' [p2m2 H_p2m2_map H_ma'_map_pa].
  rewrite <- H_p2m'_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  clear H_va_has_pa.

  elim H_pt_page_eq' with va pa' ma' pg'; auto.
  2: unfold page_type; rewrite <- H_pg'_map, <- Heqpg'_content; auto.
  2: exists p2m'; auto.
  unfold hypervisor_mapping_madd.
  intros ma2 H_ma2_map [[p2m2 H_p2m2_map H_ma2_map_pa] H_pg_map2].
  rewrite <- H_p2m_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  rewrite <- H_ma_map in H_ma2_map; injection H_ma2_map;
    intro; subst ma2; clear H_ma2_map.
  rewrite <- H_pg_map in H_pg_map2; injection H_pg_map2;
    intro; subst pg'; clear H_pg_map2.
  rewrite <- Heqpg_content in Heqpg'_content; discriminate.

  (* va does not exists in mp *)
  remember (map_apply vadd_eq mp' va) as ma_value.
  induction ma_value; auto.
  rename v into ma'; rename Heqma_value0 into H_ma'_map.
  remember (map_apply madd_eq (st_memory s') ma') as pg'_value.
  induction pg'_value; auto.
  rename v into pg'; rename Heqpg'_value into H_pg'_map.
  remember (page_content pg') as pg'_content.
  induction pg'_content; auto.
  rename o into value; rename Heqpg'_content into H_pg'_content.

  generalize H_s'_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_valid_virtmap H_valid_currpg.
  unfold va_has_valid_pa, hypervisor_mapping_madd in H_va_has_pa.
  destruct H_va_has_pa as [H_va_has_pa H_valid_cache_tlb].
  clear H_valid_cache_tlb.
  elim H_va_has_pa with osi (os_current_page aos) pt_addr' pt_page' mp' va ma'; auto.
  2: exists p2m'; auto.
  intros pa' [p2m2 H_p2m2_map H_ma'_map_pa].
  rewrite <- H_p2m'_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  clear H_va_has_pa.
  elim H_pt_page_eq' with va pa' ma' pg'; auto.
  2: unfold page_type; rewrite <- H_pg'_map, <- H_pg'_content; auto.
  2: exists p2m'; auto.
  intros ma2 H_ma2_map [[p2m2 H_p2m2_map H_ma2_map_pa] H_pg_map2].
  rewrite <- H_p2m_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  rewrite <- Heqma_value in H_ma2_map; discriminate.

  (* osi does not have an hypervisor mapping *)
  remember (map_apply os_ident_eq (st_hypervisor s') osi) as p2m'_value.
  induction p2m'_value; auto.
  rename v into p2m'; rename Heqp2m'_value into H_p2m'_map.
  remember (map_apply padd_eq p2m' (os_current_page aos)) as pt_addr_value.
  induction pt_addr_value; auto.
  rename v into pt_addr'; rename Heqpt_addr_value into H_pt_addr'_map.
  remember (map_apply madd_eq (st_memory s') pt_addr') as pt_page_value.
  induction pt_page_value; auto.
  rename v into pt_page'; rename Heqpt_page_value into H_pt_page'_map.
  remember (page_content pt_page') as pt_page'_content.
  induction pt_page'_content; auto.
  rename m0 into mp'; rename Heqpt_page'_content into H_pt_page'_content.
  remember (map_apply vadd_eq mp' va) as ma_value.
  induction ma_value; auto.
  rename v into ma'; rename Heqma_value into H_ma'_map.
  remember (map_apply madd_eq (st_memory s') ma') as pg'_value.
  induction pg'_value; auto.
  rename v into pg'; rename Heqpg'_value into H_pg'_map.
  remember (page_content pg') as pg'_content.
  induction pg'_content; auto.
  rename o into value; rename Heqpg'_content into H_pg'_content.
  
  generalize H_s'_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
    H_valid_untosexmd H_valid_hyper H_injective_hyper H_valid_virtmap H_valid_currpg.
  unfold va_has_valid_pa, hypervisor_mapping_madd in H_va_has_pa.
  destruct H_va_has_pa as [H_va_has_pa H_valid_cache_tlb].
  clear H_valid_cache_tlb.
  elim H_va_has_pa with osi (os_current_page aos) pt_addr' pt_page' mp' va ma'; auto.
  2: exists p2m'; auto.
  intros pa' [p2m2 H_p2m2_map H_ma'_map_pa].
  rewrite <- H_p2m'_map in H_p2m2_map; injection H_p2m2_map;
    intro; subst p2m2; clear H_p2m2_map.
  clear H_va_has_pa.
  unfold osi_equivalent_pt_pages in H_pt_page_eq.
  destruct H_pt_page_eq as [H_pt_page_eq H_pt_page_eq'].
  unfold osi_pt_pages_eq in H_pt_page_eq'.
  elim H_pt_page_eq' with (os_current_page aos) pt_page' mp'; auto.
  2: exists pt_addr'; try exists p2m'; auto.
  clear H_pt_page_eq'.
  intros pt_page2 H_pt_map2 H_mp2.
  destruct H_mp2 as [mp2 H_mp2_map H_pt_page_eq'].
  elim H_pt_page_eq' with va pa' ma' pg'; auto.
  2: unfold page_type; rewrite <- H_pg'_map, <- H_pg'_content; auto.
  2: exists p2m'; auto.
  intros ma2 H_ma2_map [[p2m2 H_p2m2_map H_ma2_map_pa] H_pg_map2].
  rewrite <- Heqp2m_value in H_p2m2_map; discriminate.
Qed.


(* Lemma 3 *)
Lemma lemma_cache_coherence : forall s1 s2 osi,
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  st_active_os s1 = osi -> st_active_os s2 = osi ->
  osi_equivalent ctxt s1 s2 osi ->
  cache_coherence s1 s2.
Proof.
  intros s1 s2 osi H_s1_valid H_s2_valid H_osi_active1 H_osi_active2
    H_eq va H_va_accessible.
  
  assert (m s1 va = m s2 va) as H_m1_eq_m2.
  subst osi.
  apply m_eq_m' with (st_active_os s1); auto.
  repeat split; auto.
  intros H_in_dom1 H_in_dom2.
  rewrite c_eq_m; auto.
  rewrite c_eq_m; auto.
  intros H_not_in_dom1 H_in_dom2.
  rewrite c_eq_m; auto.
  intros H_in_dom1 H_not_in_dom2.
  rewrite c_eq_m; auto.
Qed.
  
  
(* TLB Coherence *)
Definition t (s:state) (va:vadd) : exc page vadd :=
  match map_apply vadd_eq (st_tlb s) va with
    | Error _ => Error page va
    | Value ma =>
      match map_apply madd_eq (st_memory s) ma with
        | Error _ => Error page va
        | Value pg =>
          match page_content pg with
            | RW_c _ => Value vadd pg
            | _ => Error page va
          end                  
      end
  end.

Definition tlb_coherence (s s':state) : Prop :=
  forall va, os_accessible ctxt va ->

  (* Hit / Hit *)
  (in_dom_of t s va -> in_dom_of t s' va -> t s va = t s' va) /\
  (* Miss / Miss *)
  (~ in_dom_of t s va -> ~ in_dom_of t s' va -> m s va = m s' va) /\
  (* Miss / Hit *)
  (~ in_dom_of t s va -> in_dom_of t s' va -> m s va = t s' va) /\
  (* Hit / Miss *)
  (in_dom_of t s va -> ~ in_dom_of t s' va -> t s va = m s' va).

(* A version of cache_included *)
Lemma t_eq_m : forall s va,
  valid_state ctxt s -> in_dom_of t s va -> 
  t s va = m s va.
Proof.
  intros s va H_valid H_in_dom.

  unfold in_dom_of in H_in_dom.
  remember (t s va) as pg_value.
  induction pg_value; try contradiction.
  rename v into pg; clear H_in_dom.

  unfold t in Heqpg_value.
  remember (map_apply vadd_eq (st_tlb s) va) as ma_value.
  induction ma_value; try congruence.
  rename v into ma; rename Heqma_value into H_ma_map.
  remember (map_apply madd_eq (st_memory s) ma) as pg_value.
  induction pg_value; try congruence.
  rename v into pg'; rename Heqpg_value0 into H_pg_map.
  remember (page_content pg') as pg_content_value.
  induction pg_content_value; try congruence.
  rename o into val; rename Heqpg_content_value into H_pg_content.
  injection Heqpg_value; intro; subst pg'; clear Heqpg_value.

  generalize H_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].
  unfold valid_tlb in H_tlb_valid.
  destruct H_tlb_valid as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  unfold tlb_included, va_mapped_to_ma,
    os_current_vadd_mapping_madd, hypervisor_mapping,
      vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va ma as [pt_addr [aos H_aos_map
    [p2m H_p2m_map H_pt_addr_map]] [mp [pt_page H_pt_page_map H_mp_map] H_ma_vm_map]]; auto.

  unfold m.

  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_vm_map.
  rewrite <- H_pg_map.
  rewrite <- H_pg_content; auto.
Qed.


Lemma lemma_tlb_coherence : forall s1 s2 osi,
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  st_active_os s1 = osi -> st_active_os s2 = osi ->
  osi_equivalent ctxt s1 s2 osi ->
  tlb_coherence s1 s2.
Proof.
  intros s1 s2 osi H_s1_valid H_s2_valid H_osi_active1 H_osi_active2
    H_eq va H_va_accessible.
  
  assert (m s1 va = m s2 va) as H_m1_eq_m2.
  subst osi.
  apply m_eq_m' with (st_active_os s1); auto.
  repeat split; auto.
  intros H_in_dom1 H_in_dom2.
  rewrite t_eq_m; auto.
  rewrite t_eq_m; auto.
  intros H_not_in_dom1 H_in_dom2.
  rewrite t_eq_m; auto.
  intros H_in_dom1 H_not_in_dom2.
  rewrite t_eq_m; auto.
Qed.

(* END *)


End Cache_coherence.
  
