(***********************************************************************
 * File: Cache_invariant.v
 * $Revision: 482 $
 * $Date: 2011-03-05 09:41:24 -0200 (Sat, 05 Mar 2011) $
 * $Author: jdcampo $
 *******************************************************************
 * In this module the invariance of secure caches and tlbs wrt the
 * actions is proved.
 ******************************************************************)
Require Import Omega.
Require Import List.
Require Export State_proof.
Require Export State_effects.
Require Export Exec.

Set Undo 20000.

Set Implicit Arguments.

Section Secure_Cache.
Variable ctxt : context.

Lemma valid_cache_add : forall (s s':state) (va:vadd) (pg:page) (ma:madd)
  (new_cache:cache) (new_tlb:tlb),
  va_mapped_to_ma s (st_active_os s) va ma ->
  (exists val, RW_c val = page_content pg) ->
  cache_add (st_cache s) new_cache va pg ->    
  Value _ pg = map_apply madd_eq (st_memory s) ma ->
  valid_cache s ->

  s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) (st_memory s) new_cache new_tlb ->

  valid_cache s'.
Proof.
  intros s s' va pg ma new_cache new_tlb H_ma_map H_ptype H_new_cache H_pg_map H_s_valid H_s'.
  destruct H_s_valid as [H_cache_size [H_cache_distinct H_cache_included]].
  destruct H_ptype as [val H_ptype].
  rewrite H_s'; simpl.
  unfold valid_cache in *; repeat split; simpl.
  apply cache_add_size with (st_cache s) va pg; auto.
  apply cache_add_distinct with (st_cache s) va pg; auto.

  (* Cache included proof *)
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
  unfold cache_included in *.
  simpl.
  intros va' pg' H_pg'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].

  induction (vadd_eq va va').
  subst va'.
  assert (pg = pg').
  rewrite <- H_pg_new_map in H_pg'_map; injection H_pg'_map; auto.
  subst pg'; clear H_pg'_map.
  exists ma.
  repeat split; auto.
  exists pt_addr.
  exists aos; auto.
  exists p2m; auto.
  exists mp; auto.
  exists pt_page; auto.
  unfold page_type.
  rewrite <- H_pg_map, <- H_ptype; auto.

  (* va <> va' *)
  destruct H_cache_included with va' pg' as [ma' [H_ma'_map [H_pg'_ptype H_pg'_map']]].
  apply H_old_entries; auto.
  exists ma'; auto.
Qed.

Lemma map_apply_fix_distinct : forall (s:state) (c:cache) (va:vadd) (ma:madd) (pg:page),
  cache_distinct c -> 
  Value _ pg = cache_value (fix_cache_synonym s c ma) va ->
  Value _ pg = cache_value c va.
Proof.
  intros s c va ma pg H_distinct.
  unfold cache_value.
  generalize pg; clear pg.

  induction H_distinct; auto.
  intros pg0 H_pg0_map.

  simpl.
  induction (vadd_eq va va0).
  subst va0.
  assert (exists e, map_apply vadd_eq (fix_cache_synonym s c ma) va = Error _ e) as [e H_va_error].
  assert (exists it, it = map_apply vadd_eq (fix_cache_synonym s c ma) va) as [it H_it_map].
  induction (map_apply vadd_eq (fix_cache_synonym s c ma) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  rename v into pg'.
  elim H.
  exists pg'.
  apply IHH_distinct; auto.
  exists e; auto.

  simpl in H_pg0_map.
  induction (va_mapped_to_ma_cache s va).
  induction (madd_eq ma a).
  rewrite H_va_error in H_pg0_map; discriminate.
  simpl in H_pg0_map.
  induction (vadd_eq va va); auto.
  contradiction b0; auto.
  rewrite H_va_error in H_pg0_map; discriminate.
  
  simpl in H_pg0_map.
  induction (va_mapped_to_ma_cache s va0).
  induction (madd_eq ma a).
  rewrite IHH_distinct; auto.
  simpl in H_pg0_map.
  induction (vadd_eq va va0); auto.
  contradiction b; auto.
  rewrite IHH_distinct; auto.
Qed.
  
Lemma valid_state_included_tlb : forall (s:state) (va:vadd) (ma:madd),
  valid_state ctxt s -> Value _ ma = tlb_value (st_tlb s) va ->
  va_mapped_to_ma s (st_active_os s) va ma.
Proof.
  intros s va ma H_s_valid H_ma_map.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid [H_tlb_size [H_tlb_disctinct H_tlb_included]]]].
  unfold tlb_included in H_tlb_included.
  apply H_tlb_included; auto.
Qed.

Lemma valid_state_included_cache : forall (s:state) (va:vadd) (pg:page),
  valid_state ctxt s -> Value _ pg = cache_value (st_cache s) va ->
  exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
    Value _ pg = map_apply madd_eq (st_memory s) ma.
Proof.
  intros s va pg H_s_valid H_pg_map.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_disctinct H_cache_included]] H_valid_tlb]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_ptype H_pg_map_mem]]]; auto.
  exists ma; auto.
Qed.

Lemma valid_state_included_tlb_mem : forall (s:state) (va:vadd) (ma:madd),
  valid_state ctxt s -> Some ma = va_mapped_to_ma_cache s va ->
  va_mapped_to_ma s (st_active_os s) va ma.
Proof.
  intros s va ma H_s_valid H_ma_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) va).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  unfold va_mapped_to_ma_cache in H_ma_map.
  rewrite <- H_it_map in H_ma_map.
  induction it.
  apply valid_state_included_tlb; auto.
  unfold tlb_value.
  injection H_ma_map; intro; subst v; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.

  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  assert (exists pt_addr, Value _ pt_addr =  map_apply padd_eq v0 (os_current_page v))
    as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v0 (os_current_page v)); try discriminate.
  exists v1; auto.
  rewrite <- H_pt_addr_map in H_ma_map.
  exists pt_addr.
  exists v; auto.
  exists v0; auto.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  assert (exists mp, PT_c mp = page_content v1) as [mp H_mp_map].
  induction (page_content v1); try discriminate.
  exists m; auto.
  rewrite <- H_mp_map in H_ma_map.
  exists mp.
  exists v1; auto.
  induction (map_apply vadd_eq mp va); try discriminate.
  injection H_ma_map; intro; subst v2; auto.
Qed.

Lemma valid_state_included_cache_mem : forall (s:state) (va:vadd) (pg:page),
  valid_state ctxt s -> Some pg = va_mapped_to_pg_cache s va ->
  exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
    Value _ pg = map_apply madd_eq (st_memory s) ma.
Proof.
  intros s va pg H_s_valid H_pg_map.
  unfold va_mapped_to_pg_cache in H_pg_map.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map in H_pg_map.
  induction it.
  apply valid_state_included_cache; auto.
  unfold cache_value.
  injection H_pg_map; intro; subst v; auto.

  assert (exists ma, Some ma = va_mapped_to_ma_cache s va) as [ma H_ma_map].
  induction (va_mapped_to_ma_cache s va); try discriminate.
  exists a; auto.
  rewrite <- H_ma_map in H_pg_map.
  exists ma.
  apply valid_state_included_tlb_mem; auto.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  injection H_pg_map; intro; subst v; auto.
Qed.

Lemma valid_state_included_tlb_cache : forall (s:state) (va:vadd) (ma:madd),
  valid_state ctxt s -> va_mapped_to_ma s (st_active_os s) va ma ->
  Some ma = va_mapped_to_ma_cache s va.
Proof.
  intros s va ma H_s_valid H_ma_map.
  unfold va_mapped_to_ma_cache.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) va).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  induction it.
  rewrite va_mapped_same_ma with s va ma v; auto.
  apply valid_state_included_tlb; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map; auto.
Qed.

Lemma valid_state_included_cache_cache : forall (s:state) (va:vadd) (ma:madd) (pg:page),
  valid_state ctxt s -> va_mapped_to_ma s (st_active_os s) va ma ->
  Value _ pg = map_apply madd_eq (st_memory s) ma ->
  Some pg = va_mapped_to_pg_cache s va.
Proof.
  intros s va ma pg H_s_valid H_ma_map H_pg_map.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with s va v; auto.
  intros ma' H_ma'_map H_v_map.
  assert (ma=ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'.
  rewrite <- H_pg_map in H_v_map.
  injection H_v_map; intro; subst v; auto.

  rewrite <- valid_state_included_tlb_cache with s va ma; auto.
  rewrite <- H_pg_map; auto.
Qed.

Lemma valid_state_cache_owner : forall (s:state) (va:vadd) (pg:page),
  valid_state ctxt s -> Value _ pg = cache_value (st_cache s) va ->
  (os_accessible ctxt va -> page_owned_by pg = Os_PO (st_active_os s)) /\
  (~os_accessible ctxt va -> page_owned_by pg = Hyp_PO).
Proof.
  intros s va pg H_s_valid H_pg_map.
  unfold os_accessible.
  assert (exists b, b = ctxt_vadd_accessible ctxt va) as [b H_va_accessible].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  rewrite <- H_va_accessible.

  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_distinct H_cache_included]] H_valid_tlb]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_ptype H_pg_map_mem]]]; auto.
  generalize H_ma_map; intro H_ma_map'.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg'' H_pg''_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype' H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  rewrite <- H_pg_map_mem in H_pg''_map; injection H_pg''_map;
    intro; subst pg''; clear H_pg''_map.

  induction b; split; intro.
  apply H_va_accessible_owner; auto.
  unfold os_accessible; auto.
  contradiction H; auto.
  discriminate H.
  apply H_va_not_accessible_owner; auto.
  unfold os_accessible.
  intro H_abs; rewrite <- H_va_accessible in H_abs; discriminate.
Qed.


Lemma fix_not_mapped : forall (s:state) (va:vadd) (ma:madd) (pg:page),
  valid_state ctxt s ->
  Value _ pg = cache_value (fix_cache_synonym s (st_cache s) ma) va ->
  ~(va_mapped_to_ma s (st_active_os s) va ma).
Proof.
  intros s va ma pg H_s_valid H_pg_map H_ma_map.
  unfold cache_value in H_pg_map.
  
  induction (st_cache s); try discriminate.
  simpl in H_pg_map.
  induction a.
  rename item_index into va'.
  rename item_info into pg'.
  assert (exists it, it = va_mapped_to_ma_cache s va') as [it H_it_map].
  induction (va_mapped_to_ma_cache s va').
  exists (Some a); auto.
  exists None; auto.
  rewrite <-  H_it_map in H_pg_map.
  induction it.
  induction (madd_eq ma a).
  apply IHc; auto.
  simpl in H_pg_map.
  induction (vadd_eq va va').
  subst va'.
  apply b.
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem; auto.
  apply IHc; auto.
  apply IHc; auto.  
Qed.  


Lemma cache_fix_less_elements : forall (s:state) (ma:madd),
  map_size (fix_cache_synonym s (st_cache s) ma) <= map_size (st_cache s).
Proof.
  intros s ma.
  
  induction (st_cache s); auto.
  simpl.
  induction a.
  induction (va_mapped_to_ma_cache s item_index); auto.
  induction (madd_eq ma a); auto.
  simpl; omega.
Qed.

Lemma cache_fix_distinct : forall (s:state) (ma:madd),
  cache_distinct (st_cache s) ->
  cache_distinct (fix_cache_synonym s (st_cache s) ma).
Proof.
  intros s ma H_cache_distinct.

  induction H_cache_distinct.
  constructor.
    
  simpl.
  induction (va_mapped_to_ma_cache s va); auto.
  induction (madd_eq ma a); auto.
  constructor; auto.
  intros [pg' H_pg'_map].
  apply H.
  exists pg'.

  apply map_apply_fix_distinct with s ma; auto.
Qed.


Lemma valid_cache_write : forall (s s':state) (va:vadd) (pg:page) (ma:madd)
  (new_cache:cache) (new_mem:system_memory) (new_tlb:tlb),
  va_mapped_to_ma s (st_active_os s) va ma ->
  page_type (st_memory s) ma = Some RW ->
  (exists val, page_content pg = RW_c val) ->
  
  cache_add (fix_cache_synonym s (st_cache s) ma) new_cache va pg ->    
  new_mem = map_add madd_eq (st_memory s) ma pg ->
  valid_state ctxt s ->

  s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) new_mem new_cache new_tlb ->

  valid_cache s'.
Proof.
  intros s s' va pg ma new_cache new_mem new_tlb H_ma_map H_ptype H_pg_content
    H_new_cache H_new_mem H_s_valid.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_distinct H_cache_included]] H_valid_tlb]].
  repeat split.
  rewrite H; simpl.
  apply cache_add_size with (fix_cache_synonym s (st_cache s) ma) va pg; auto.
  apply Le.le_trans with (map_size (st_cache s)); auto.
  apply cache_fix_less_elements; auto.

  rewrite H; simpl.
  apply cache_add_distinct with (fix_cache_synonym s (st_cache s) ma) va pg; auto.
  apply cache_fix_distinct; auto.

  destruct H_pg_content as [val H_pg_content].
  subst new_mem.
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
  unfold cache_included in *; simpl.
  intros va' pg' H_pg'_map.
  induction (vadd_eq va' va).
  subst va'.
  rewrite H in H_pg'_map; simpl in H_pg'_map.
  rewrite <- H_pg_new_map in H_pg'_map; injection H_pg'_map;
    intro; subst pg'; clear H_pg'_map.
  exists ma.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  repeat split.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  rewrite H; simpl.
  exists pt_addr.
  exists aos; auto.
  exists p2m; auto.
  exists mp; auto.
  exists pt_page; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr; subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page_map in H_ptype.
  rewrite <- H_mp_map in H_ptype.
  discriminate H_ptype.
  unfold page_type.
  rewrite H; simpl;
    rewrite lemma_map_add_apply; auto.
  induction (page_content pg); try discriminate; auto.
  rewrite H; simpl;
    rewrite lemma_map_add_apply; auto.

  (* va' <> va *)
  destruct H_cache_included with va' pg' as [ma' [H_ma'_map [H_pg'_content H_pg'_map_mem]]].
  rewrite H in H_pg'_map; simpl in H_pg'_map.
  assert (Value _ pg' = map_apply vadd_eq (fix_cache_synonym s (st_cache s) ma) va')
    as H_pg'_map_s.
  apply H_old_entries; auto.
  apply map_apply_fix_distinct with s ma; auto.

  assert (ma <> ma').
  intro H_ma_eq_ma'.
  subst ma'.
  apply fix_not_mapped with s va' ma pg'; auto.
  rewrite H in H_pg'_map; simpl in H_pg'_map.
  apply H_old_entries; auto.

  exists ma'.
  repeat split.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma'_map.
  destruct H_ma'_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]].
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  rewrite H; simpl.
  exists pt_addr.
  exists aos; auto.
  exists p2m; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr; subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page_map in H_ptype.
  rewrite <- H_mp_map in H_ptype.
  discriminate H_ptype.
  unfold page_type.
  rewrite H; simpl;
    rewrite lemma_map_add_apply_distinct; auto.

  rewrite H; simpl; rewrite lemma_map_add_apply_distinct; auto.
Qed.

Lemma valid_cache_drop : forall (s s':state) (va:vadd) (osi:os_ident)
  (new_oss:oss_map) (new_cache:cache) (new_mem:system_memory) (new_tlb:tlb),

  (forall osi os, Value _ os = map_apply os_ident_eq (st_oss s) osi ->
    (exists2 os', Value _ os' = map_apply os_ident_eq new_oss osi &
      (os_current_page os = os_current_page os'))) ->
  (new_cache = if (os_ident_eq (st_active_os s) osi) then
                map_drop vadd_eq (st_cache s) va 
              else (st_cache s)) ->
  (exists2 pt_addr, os_current_vadd_mapping_madd s osi pt_addr &
    (exists ma, update_pt_page pt_addr va ma (st_memory s) new_mem) \/
    drop_va_pt_page pt_addr va (st_memory s) new_mem) ->

  valid_state ctxt s ->

  s' = State (st_active_os s) new_oss (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) new_mem new_cache new_tlb ->

  valid_cache s'.
Proof.
  intros s s' va osi new_oss new_cache new_mem new_tlb
    H_same_current_page H_new_cache [pt_addr H_pt_addr_map H_update_or_drop]
    H_s_valid H_s'.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_distinct H_cache_included]] H_valid_tlb]].
  repeat split.

  (* cache size *)
  rewrite H_s', H_new_cache; simpl.
  apply Le.le_trans with (map_size (st_cache s)); auto.
  assert (exists bool, bool = map_valid_index_fix vadd_eq (st_cache s) va) as [bool H_valid_index].
  induction (map_valid_index_fix vadd_eq (st_cache s) va); [exists true | exists false]; auto.
  induction bool.
  induction (os_ident_eq (st_active_os s) osi); auto.
  apply lt_le_weak.
  apply map_drop_valid_size; auto.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite map_drop_invalid_size; auto.

  (* cache distinct *)
  rewrite H_s', H_new_cache; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  apply cache_map_drop_distinct with (st_cache s) va; auto.

  (* cache included *)
  unfold cache_included.
  intros va' pg' H_pg'_map.
  unfold cache_included in H_cache_included.
  rewrite H_s', H_new_cache in H_pg'_map; simpl in H_pg'_map.
  induction (vadd_eq va va').
  subst va'.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite map_drop_apply in H_pg'_map; discriminate.
  destruct H_cache_included with va pg' as [ma [H_ma_map [H_ptype H_pg'_map_mem]]]; auto.
  exists ma.

  assert (va_mapped_to_ma s' (st_active_os s') va ma) as H_ma_map'; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in *.
  destruct H_ma_map as [pt_addr' [aos H_aos_map [p2m H_p2m_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  exists pt_addr'.
  destruct H_same_current_page with (st_active_os s) aos as [aos' H_aos'_map H_same_current_page']; auto.
  exists aos'.
  rewrite H_s'; auto.
  exists p2m.
  rewrite H_s'; auto.
  rewrite <- H_same_current_page'; auto.
  exists mp; auto.
  exists pt_page; auto.
  assert (pt_addr <> pt_addr') as H_pt_addr_ne_pt_addr'.
  intro H_pt_addr_eq_pt_addr'; subst pt_addr'.
  destruct H_pt_addr_map as [system H_system_map [p2m' H_p2m'_map H_pt_addr_map]].
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page'_owner]]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page'_owner.
  rewrite <- H_pt_page_map in H_pt_page'_owner.
  destruct H_valid_currpg with osi system p2m' as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner.
  rewrite H_pt_page_owner in H_pt_page'_owner.
  injection H_pt_page'_owner; intro; contradiction b; auto.
  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  repeat split; auto.
  unfold page_type.
  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.
  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.

  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.
  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.

  rename H_pg'_map into H_pg'_map'.
  assert (Value vadd pg' = map_apply vadd_eq (st_cache s) va') as H_pg'_map.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite map_drop_apply_distinct in H_pg'_map'; auto.
  clear H_pg'_map'.
  destruct H_cache_included with va' pg' as [ma [H_ma_map [H_ptype H_pg'_map_mem]]]; auto.
  
  assert (va_mapped_to_ma s' (st_active_os s') va' ma) as H_ma_map'; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in *.
  destruct H_ma_map as [pt_addr' [aos H_aos_map [p2m H_p2m_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  exists pt_addr'.
  destruct H_same_current_page with (st_active_os s) aos as [aos' H_aos'_map H_same_current_page']; auto.
  exists aos'.
  rewrite H_s'; auto.
  exists p2m.
  rewrite H_s'; auto.
  rewrite <- H_same_current_page'; auto.
  induction (os_ident_eq osi (st_active_os s)).
  (* osi is active *)
  subst osi.
  induction (madd_eq pt_addr pt_addr').
  (* pt_addr' is the changed pt_addr *)
  subst pt_addr'.
  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  exists (map_add vadd_eq old_mp va ma').
  exists (Page (PT_c (map_add vadd_eq old_mp va ma')) (page_owned_by pt_page')); auto.
  rewrite lemma_map_add_apply; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_old_mp_map.
  injection H_old_mp_map; intro; subst old_mp; clear H_old_mp_map; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  exists (map_drop vadd_eq old_mp va).
  exists (Page (PT_c (map_drop vadd_eq old_mp va)) (page_owned_by pt_page')); auto.
  rewrite lemma_map_add_apply; auto.
  rewrite map_drop_apply_distinct; auto.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_old_mp_map.
  injection H_old_mp_map; intro; subst old_mp; clear H_old_mp_map; auto.

  (* pt_addr' is not the changed pt_addr *)
  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  exists mp; auto.
  exists pt_page; auto.
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  exists mp; auto.
  exists pt_page; auto.
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  (* osi is not active *)
  exists mp; auto.
  exists pt_page; auto.
  assert (pt_addr <> pt_addr') as H_pt_addr_ne_pt_addr'.
  intro H_pt_addr_eq_pt_addr'; subst pt_addr'.
  destruct H_pt_addr_map as [system H_system_map [p2m' H_p2m'_map H_pt_addr_map]].
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page'_owner]]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page'_owner.
  rewrite <- H_pt_page_map in H_pt_page'_owner.
  destruct H_valid_currpg with osi system p2m' as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner.
  rewrite H_pt_page_owner in H_pt_page'_owner.
  injection H_pt_page'_owner; contradiction.

  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  exists ma; repeat split; auto.
  unfold page_type.
  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.
  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.

  destruct H_update_or_drop as [[ma' H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.
  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_pt_addr_eq_ma.
  subst ma.
  unfold page_type in H_ptype.
  rewrite <- H_pt_page'_map in H_ptype.
  rewrite <- H_old_mp_map in H_ptype; discriminate.
Qed.


Lemma valid_cache_unchanged : forall (s s':state)
  (new_oss:oss_map) (new_hyper:hypervisor) (new_em : exec_mode)
  (new_activity:os_activity),

  (forall osi os, Value _ os = map_apply os_ident_eq (st_oss s) osi ->
    (exists os', (Value _ os' = map_apply os_ident_eq new_oss osi) /\
      (os_current_page os = os_current_page os')) /\
    (forall pt_addr, hypervisor_mapping_madd (st_hypervisor s) osi (os_current_page os) pt_addr ->
      hypervisor_mapping_madd new_hyper osi (os_current_page os) pt_addr)) ->

  valid_cache s ->

  s' = State (st_active_os s) new_oss new_em new_activity
         new_hyper (st_memory s) (st_cache s) (st_tlb s) ->

  valid_cache s'.
Proof.
  intros s s' new_oss new_hyper new_em new_activity H_same_current_page H_valid_cache H_s'.
  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  repeat split; try (rewrite H_s'; assumption).
  unfold cache_included in *.
  intros va pg H_pg_map.

  destruct H_cache_included with va pg as [ma [H_ma_map [H_ptype H_pg_map_mem]]].
  rewrite H_s' in H_pg_map; auto.
  exists ma.
  repeat split; try (rewrite H_s'; assumption).

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in *; simpl.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  destruct H_same_current_page with (st_active_os s) aos as [H_same_current_page' H_same_pt_addr]; auto.
  destruct H_same_current_page' as [aos' [H_aos'_map H_same_current_page']].
  rewrite H_s'; simpl.
  exists pt_addr.
  exists aos'; auto.
  unfold hypervisor_mapping_madd in H_same_pt_addr.
  destruct H_same_pt_addr with pt_addr as [p2m' H_p2m'_map H_pt_addr_map'].
  exists p2m; auto.
  exists p2m'; auto.
  rewrite <- H_same_current_page'; auto.
  exists mp; auto.
  exists pt_page; auto.
Qed.


Lemma valid_empty_cache : forall (s:state),
  map_empty vadd page = (st_cache s) ->
  valid_cache s.
Proof.
  intros s H_cache_empty.
  unfold valid_cache.
  rewrite <- H_cache_empty.
  repeat split.
  simpl; generalize max_cache_pos;
    intro H_max_cache_pos; omega.
  constructor.

  unfold cache_included.
  rewrite <- H_cache_empty.
  simpl.
  intros va pg H_abs; discriminate H_abs.
Qed.

Lemma valid_empty_tlb : forall (s:state),
  map_empty vadd madd = (st_tlb s) ->
  valid_tlb s.
Proof.
  intros s H_tlb_empty.
  unfold valid_tlb.
  rewrite <- H_tlb_empty.
  repeat split.
  simpl; generalize max_tlb_pos;
    intro H_max_tlb_pos; omega.
  constructor.

  unfold tlb_included.
  rewrite <- H_tlb_empty.
  simpl.
  intros va pg H_abs; discriminate H_abs.
Qed.


(* TLB validity lemmas *)
Lemma valid_tlb_add : forall (s s':state) (va:vadd) (ma:madd)
  (new_tlb:tlb) (new_mem:system_memory) (new_cache:cache),

  tlb_add (st_tlb s) new_tlb va ma ->

  (exists pt_addr, os_current_vadd_mapping_madd s (st_active_os s) pt_addr /\
    (update_pt_page pt_addr va ma (st_memory s) new_mem \/
      (map_apply madd_eq (st_memory s) pt_addr = map_apply madd_eq (st_memory s') pt_addr
        /\ va_mapped_to_ma s (st_active_os s) va ma))) ->
  
  valid_tlb s ->
  
  s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
         (st_hypervisor s) new_mem new_cache new_tlb ->
         
  valid_tlb s'.
Proof.
  intros s s' va ma new_tlb new_mem new_cache H_new_tlb H_newmem
    H_tlb_valid H_s'.
  destruct H_tlb_valid as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  
  unfold valid_tlb.
  repeat split; rewrite H_s'; simpl;
    inversion_clear H_new_tlb as [H_pg_new_map [H_old_entries
      [H_new_size H_new_distinct]]]; auto.
  
  (* tlb included proof *)
  destruct H_newmem as [pt_addr [H_pt_addr_map H_newmem]].
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  unfold tlb_included in *.
  simpl.
  intros va' ma' H_ma'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  exists pt_addr.
  exists aos; auto.
  exists p2m; auto.

  destruct H_newmem as [H_newmem | [H_same_pt_addr H_newmem]].
  unfold update_pt_page in H_newmem.
  destruct H_newmem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_newmem]].
  exists (map_add vadd_eq old_mp va ma).
  exists (Page (PT_c (map_add vadd_eq old_mp va ma)) (page_owned_by pt_page')); auto.
  rewrite H_newmem.
  rewrite lemma_map_add_apply; auto.
  induction (vadd_eq va va').
  subst va'.
  rewrite lemma_map_add_apply; auto.
  rewrite <- H_pg_new_map in H_ma'_map; injection H_ma'_map;
    intro; subst ma'; auto.
  
  rewrite lemma_map_add_apply_distinct; auto.
  assert (Value _ ma' = map_apply vadd_eq (st_tlb s) va') as H_ma'_map_s.
  apply H_old_entries; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va' ma' as [pt_addr' [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]]; auto.
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map;
    intro; subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map;
    intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map; injection H_pt_page'_map;
    intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_old_mp_map in H_mp_map; injection H_mp_map;
    intro; subst mp; clear H_mp_map; auto.

  induction (vadd_eq va va').
  subst va'.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_newmem.
  destruct H_newmem as [pt_addr' [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]]; auto.
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map;
    intro; subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map;
    intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map.
  exists mp.
  exists pt_page; auto.
  rewrite H_s' in H_same_pt_addr; simpl in H_same_pt_addr.
  rewrite <- H_same_pt_addr; auto.
  rewrite <- H_pg_new_map in H_ma'_map; injection H_ma'_map;
    intro; subst ma'; clear H_ma'_map; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va' ma' as [pt_addr' [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]]; auto.
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map;
    intro; subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map;
    intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map.

  exists mp; auto.
  exists pt_page; auto.
  rewrite H_s' in H_same_pt_addr; simpl in H_same_pt_addr.
  rewrite <- H_same_pt_addr; auto.
Qed.


Lemma valid_tlb_drop : forall (s s':state) (va:vadd) (osi:os_ident)
  (new_oss:oss_map) (new_cache:cache) (new_mem:system_memory) (new_tlb:tlb),

  (forall osi os, Value _ os = map_apply os_ident_eq (st_oss s) osi ->
    (exists os', Value _ os' = map_apply os_ident_eq new_oss osi /\
      (os_current_page os = os_current_page os'))) ->
  (new_tlb = if (os_ident_eq (st_active_os s) osi) then
                map_drop vadd_eq (st_tlb s) va 
              else (st_tlb s)) ->
  (exists2 pt_addr, os_current_vadd_mapping_madd s osi pt_addr &
    (exists ma, update_pt_page pt_addr va ma (st_memory s) new_mem) \/
    drop_va_pt_page pt_addr va (st_memory s) new_mem) ->

  valid_state ctxt s ->

  s' = State (st_active_os s) new_oss (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) new_mem new_cache new_tlb ->

  valid_tlb s'.
Proof.
  intros s s' va osi new_oss new_cache new_mem new_tlb
    H_same_current_page H_new_tlb [pt_addr H_pt_addr_map H_update_or_drop]
    H_s_valid H_s'.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache [H_tlb_size [H_tlb_distinct H_tlb_included]]]].
  repeat split.

  (* tlb size *)
  rewrite H_s', H_new_tlb; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  apply Le.le_trans with (map_size (st_tlb s)); auto.
  assert (exists bool, bool = map_valid_index_fix vadd_eq (st_tlb s) va) as [bool H_valid_index].
  induction (map_valid_index_fix vadd_eq (st_tlb s) va); [exists true | exists false]; auto.
  induction bool.
  apply lt_le_weak.
  apply map_drop_valid_size; auto.
  rewrite map_drop_invalid_size; auto.

  (* tlb distinct *)
  rewrite H_s', H_new_tlb; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  apply tlb_map_drop_distinct with (st_tlb s) va; auto.

  (* tlb included *)
  unfold tlb_included.
  intros va' ma' H_ma'_map.
  unfold tlb_included in H_tlb_included.
  rewrite H_s', H_new_tlb in H_ma'_map; simpl in H_ma'_map.
  assert (Value _ ma' = map_apply vadd_eq (st_tlb s) va' ->
    va_mapped_to_ma s' (st_active_os s') va' ma') as H_old_ma'_map.

  intro H_old_ma'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va' ma' as [pt_addr' [aos H_aos_map [p2m H_p2m_map H_pt_addr'_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]]; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  exists pt_addr'.
  destruct H_same_current_page with (st_active_os s) aos as [aos' [H_aos'_map H_same_current_page']]; auto.
  exists aos'.
  rewrite H_s'; auto.
  exists p2m.
  rewrite H_s'; auto.
  rewrite <- H_same_current_page'; auto.
  induction (os_ident_eq osi (st_active_os s)).
  (* osi is active *)
  subst osi.
  induction (madd_eq pt_addr pt_addr').
  (* pt_addr' is the changed pt_addr *)
  subst pt_addr'. 
  assert (va <> va') as H_va_ne_va'.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.
  intro H_va_eq_va'; subst va'; rewrite map_drop_apply in H_ma'_map; discriminate.
 
  destruct H_update_or_drop as [[ma_up H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  exists (map_add vadd_eq old_mp va ma_up).
  exists (Page (PT_c (map_add vadd_eq old_mp va ma_up)) (page_owned_by pt_page')); auto.
  rewrite lemma_map_add_apply; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_old_mp_map.
  injection H_old_mp_map; intro; subst old_mp; clear H_old_mp_map; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  exists (map_drop vadd_eq old_mp va).
  exists (Page (PT_c (map_drop vadd_eq old_mp va)) (page_owned_by pt_page')); auto.
  rewrite lemma_map_add_apply; auto.
  rewrite map_drop_apply_distinct; auto.
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_old_mp_map.
  injection H_old_mp_map; intro; subst old_mp; clear H_old_mp_map; auto.

  (* pt_addr' is not the changed pt_addr *)
  destruct H_update_or_drop as [[ma_up H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  exists mp; auto.
  exists pt_page; auto.
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  exists mp; auto.
  exists pt_page; auto.
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  (* osi is not active *)
  exists mp; auto.
  exists pt_page; auto.
  assert (pt_addr <> pt_addr') as H_pt_addr_ne_pt_addr'.
  intro H_pt_addr_eq_pt_addr'; subst pt_addr'.
  destruct H_pt_addr_map as [system H_system_map [p2m' H_p2m'_map H_pt_addr_map]].
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page'_owner]]; auto.
  rewrite <- H_pt_addr'_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page'_owner.
  rewrite <- H_pt_page_map in H_pt_page'_owner.
  destruct H_valid_currpg with osi system p2m' as
    [pt_addr'' H_pt_addr''_map [H_ptype' H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr''_map.
  injection H_pt_addr''_map; intro; subst pt_addr''; clear H_pt_addr''_map H_ptype'.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner.
  rewrite H_pt_page_owner in H_pt_page'_owner.
  injection H_pt_page'_owner; contradiction.

  destruct H_update_or_drop as [[ma_up H_new_mem] | H_new_mem].
  unfold update_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  unfold drop_va_pt_page in H_new_mem.
  destruct H_new_mem as [pt_page' H_pt_page'_map [old_mp H_old_mp_map H_new_mem]].
  rewrite H_s', H_new_mem; simpl.
  rewrite lemma_map_add_apply_distinct; auto.

  induction (os_ident_eq (st_active_os s) osi); auto.
  induction (vadd_eq va va').
  subst va'; rewrite map_drop_apply in H_ma'_map; discriminate.
  rewrite map_drop_apply_distinct in H_ma'_map; auto.
Qed.


Lemma valid_tlb_unchanged : forall (s s':state)
  (new_oss:oss_map) (new_hyper:hypervisor) (new_em : exec_mode)
  (new_activity:os_activity), 

  (forall osi os, Value _ os = map_apply os_ident_eq (st_oss s) osi ->
    (exists os', Value _ os' = map_apply os_ident_eq new_oss osi /\
    (os_current_page os = os_current_page os')) /\
    (forall pt_addr, hypervisor_mapping_madd (st_hypervisor s) osi (os_current_page os) pt_addr ->
      hypervisor_mapping_madd new_hyper osi (os_current_page os) pt_addr)) ->

  valid_tlb s ->

  s' = State (st_active_os s) new_oss new_em new_activity
         new_hyper (st_memory s) (st_cache s) (st_tlb s) ->

  valid_tlb s'.
Proof.
  intros s s' new_oss new_hyper new_em new_activity H_same_current_page H_valid_tlb H_s'.
  destruct H_valid_tlb as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  repeat split; try (rewrite H_s'; assumption).
  unfold tlb_included in *.
  intros va ma H_ma_map.

  assert (va_mapped_to_ma s (st_active_os s) va ma) as H_ma_map_mem.
  apply H_tlb_included.
  rewrite H_s' in H_ma_map; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map_mem.
  destruct H_ma_map_mem as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map_mem]].
  destruct H_same_current_page with (st_active_os s) aos as [H_same_current_page' H_same_pt_addr]; auto.
  destruct H_same_current_page' as [aos' [H_aos'_map H_same_current_page']].
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map_mem.
  rewrite H_s'; simpl.
  exists pt_addr.
  exists aos'; auto.
  unfold hypervisor_mapping_madd in H_same_pt_addr.
  destruct H_same_pt_addr with pt_addr as [p2m' H_p2m'_map H_pt_addr_map'].
  exists p2m; auto.
  exists p2m'; auto.
  rewrite <- H_same_current_page'; auto.
  exists mp; auto.
  exists pt_page; auto.
Qed.


Lemma exec_valid_cache_tlb : forall (s s':state) (a:action),
  valid_state ctxt s ->
  valid_exec ctxt s a s'->

  valid_cache s' /\ valid_tlb s'.
Proof.
  intros s s' a H_s_valid H_valid_exec.

  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].

  induction a; inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post];
      unfold pre_action, post_action in *.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.

  (* Read *)
  unfold pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.

  generalize H_ma_map; intro H_ma_map'.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page v ma as
    [pg'' H_pg''_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype' H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.

  unfold post_read, post_all_read in H_post.
  assert (Some ma = va_mapped_to_ma_cache s v) as H_ma_cache_map.
  unfold va_mapped_to_ma_cache.
  unfold va_mapped_to_ma in H_ma_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v0); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  induction it.
  assert (va_mapped_to_ma s (st_active_os s) v v0).
  apply valid_state_included_tlb; auto.
  rewrite va_mapped_same_ma with s v ma v0; auto.
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map'; auto.

  assert (Some pg = va_mapped_to_pg_cache s v) as H_pg_map_cache.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v0); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with s v v0; auto.
  intros ma' H_ma'_map H_pg_map'.
  assert (ma = ma').
  apply va_mapped_same_ma with s v; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg_map'; injection H_pg_map'; intro; subst v0; auto.
  rewrite <- H_ma_cache_map.
  rewrite <- H_pg_map; auto.

  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  intros [H_triv [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache, H_new_tlb.
  clear H_triv.

  assert (s' = State (st_active_os s) (st_oss s)
    (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
       (st_memory s) new_cache new_tlb) as H_s'.

  rewrite <- H_new_cache.
  rewrite <- H_new_tlb.
  replace (st_oss s) with (st_oss s').
  replace (st_active_os s) with (st_active_os s').
  replace (st_memory s) with (st_memory s').
  replace (st_hypervisor s) with (st_hypervisor s').
  replace (st_exec_mode s) with (st_exec_mode s').
  replace (st_aos_activity s) with (st_aos_activity s').
  destruct s'; auto.
  
  elim H_assign with OS_Activity; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Exec_Mode; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Hypervisor_SC; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with System_Memory; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Guest_OSs; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.

  split.
  apply valid_cache_add with s v pg ma new_cache new_tlb; auto.
  unfold page_type in H_ptype; rewrite <- H_pg_map in H_ptype.
  induction (page_content pg); try discriminate; auto.
  exists o; auto.
  apply valid_tlb_add with s v ma new_tlb (st_memory s) new_cache; auto.
  exists pt_addr; auto.
  split.
  unfold os_current_vadd_mapping_madd.
  exists aos; auto.
  exists p2m; auto.
  right.
  split; auto.
  rewrite H_s'; auto.

  (* Read hyper *)
  unfold pre_read, pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.

  generalize H_ma_map; intro H_ma_map'.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page v ma as
    [pg'' H_pg''_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype' H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.

  unfold post_read, post_read_hyper, post_all_read in H_post.
  assert (Some ma = va_mapped_to_ma_cache s v) as H_ma_cache_map.
  unfold va_mapped_to_ma_cache.
  unfold va_mapped_to_ma in H_ma_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v0); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  induction it.
  assert (va_mapped_to_ma s (st_active_os s) v v0).
  apply valid_state_included_tlb; auto.
  rewrite va_mapped_same_ma with s v ma v0; auto.
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map'; auto.

  assert (Some pg = va_mapped_to_pg_cache s v) as H_pg_map_cache.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v0); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with s v v0; auto.
  intros ma' H_ma'_map H_pg_map'.
  assert (ma = ma').
  apply va_mapped_same_ma with s v; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg_map'; injection H_pg_map'; intro; subst v0; auto.
  rewrite <- H_ma_cache_map.
  rewrite <- H_pg_map; auto.

  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  intros [H_triv [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache, H_new_tlb.
  clear H_triv.

  assert (s' = State (st_active_os s) (st_oss s)
    (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
       (st_memory s) new_cache new_tlb) as H_s'.

  rewrite <- H_new_cache.
  rewrite <- H_new_tlb.
  replace (st_oss s) with (st_oss s').
  replace (st_active_os s) with (st_active_os s').
  replace (st_memory s) with (st_memory s').
  replace (st_hypervisor s) with (st_hypervisor s').
  replace (st_exec_mode s) with (st_exec_mode s').
  replace (st_aos_activity s) with (st_aos_activity s').
  destruct s'; auto.
  
  elim H_assign with OS_Activity; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Exec_Mode; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Hypervisor_SC; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with System_Memory; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_assign with Guest_OSs; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.

  split.
  apply valid_cache_add with s v pg ma new_cache new_tlb; auto.
  unfold page_type in H_ptype; rewrite <- H_pg_map in H_ptype.
  induction (page_content pg); try discriminate.
  exists o; auto.
  apply valid_tlb_add with s v ma new_tlb (st_memory s) new_cache; auto.
  exists pt_addr; auto.
  split.
  unfold os_current_vadd_mapping_madd.
  exists aos; auto.
  exists p2m; auto.
  right.
  split; auto.
  rewrite H_s'; auto.

  (* Write *)
  unfold pre_write in H_pre.
  destruct H_pre as [H_os_running [H_v_accessible [ma H_ma_map H_ptype]]].
  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v1; auto.

  unfold post_write, post_all_write in H_post.
  destruct H_post as [ma' H_ma'_map [new_mem H_update [pg' H_pg'_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  assert (ma = ma').
  apply va_mapped_same_ma with s v; auto.
  unfold va_mapped_to_ma_cache in H_ma'_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v1); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map in H_ma'_map.
  induction it.
  injection H_ma'_map; intro; subst v1.
  apply valid_state_included_tlb; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  assert (exists pt_addr, Value _ pt_addr = map_apply padd_eq v2 (os_current_page v1)) as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v2 (os_current_page v1)); try discriminate.
  exists v3; auto.
  exists pt_addr.
  exists v1; auto.
  exists v2; auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  assert (exists mp, PT_c mp = (page_content v3)) as [mp H_mp_map].
  induction (page_content v3); try discriminate.
  exists m; auto.
  rewrite <- H_mp_map in H_ma'_map.
  exists mp.
  exists v3; auto.
  induction (map_apply vadd_eq mp v); try discriminate.
  injection H_ma'_map; intro; subst v4; auto.
  subst ma'.

  assert (pg = pg').
  unfold va_mapped_to_pg_cache in H_pg'_map.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v1); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map in H_pg'_map.
  induction it.
  injection H_pg'_map; intro; subst v1.
  elim valid_state_included_cache with s v pg'; auto.
  intros ma'' H_ma''_map H_pg'_map'.
  replace ma'' with ma in H_pg'_map'.
  rewrite <- H_pg_map in H_pg'_map'.
  injection H_pg'_map'; intro; subst pg'; auto.
  apply va_mapped_same_ma with s v; auto.

  rewrite <- H_ma'_map in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map.
  injection H_pg'_map; auto.
  subst pg'.

  split.
  apply valid_cache_write with s v (Page (RW_c (Some v0)) (page_owned_by pg))
    ma new_cache new_mem new_tlb; auto.
  simpl.
  exists (Some v0); auto.

  inversion_clear H_update.
  rewrite <- H_pg_map in H; injection H; intro; subst x.
  rewrite H0; auto.

  apply valid_tlb_add with s v ma new_tlb new_mem new_cache; auto.

  generalize H_ma_map; intro H_ma_map'.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].
  exists pt_addr; auto.
  split.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping.
  exists aos; auto.
  exists p2m; auto.
  right.
  split; auto.
  rewrite H_s'; simpl.
  inversion_clear H_update.
  rewrite <- H_pg_map in H; injection H; intro; subst x.
  rewrite H0.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite <- H_ma_eq_pt_addr in H_pt_page_map.
  rewrite <- H_pg_map in H_pt_page_map.
  injection H_pt_page_map; intro H_pt_page_eq_pg.
  rewrite H_pt_page_eq_pg in H_mp_map.
  unfold page_type in H_ptype.
  rewrite <- H_pg_map in H_ptype.
  induction (page_content pg); discriminate.

  (* Write hyper *)
  unfold pre_write, pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running [ma H_ma_map [H_ptype H_ma_owner]]].
  assert (exists pg , Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v1; auto.

  unfold post_write, post_write_hyper, post_all_write in H_post.
  destruct H_post as [ma' H_ma'_map [new_mem H_update [pg' H_pg'_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  assert (ma = ma').
  apply va_mapped_same_ma with s v; auto.
  unfold va_mapped_to_ma_cache in H_ma'_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v1); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map in H_ma'_map.
  induction it.
  injection H_ma'_map; intro; subst v1.
  apply valid_state_included_tlb; auto.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  assert (exists pt_addr, Value _ pt_addr = map_apply padd_eq v2 (os_current_page v1)) as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v2 (os_current_page v1)); try discriminate.
  exists v3; auto.
  exists pt_addr.
  exists v1; auto.
  exists v2; auto.
  rewrite <- H_pt_addr_map in H_ma'_map.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  assert (exists mp, PT_c mp = (page_content v3)) as [mp H_mp_map].
  induction (page_content v3); try discriminate.
  exists m; auto.
  rewrite <- H_mp_map in H_ma'_map.
  exists mp.
  exists v3; auto.
  induction (map_apply vadd_eq mp v); try discriminate.
  injection H_ma'_map; intro; subst v4; auto.
  subst ma'.

  assert (pg = pg').
  unfold va_mapped_to_pg_cache in H_pg'_map.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v1); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map in H_pg'_map.
  induction it.
  injection H_pg'_map; intro; subst v1.
  elim valid_state_included_cache with s v pg'; auto.
  intros ma'' H_ma''_map H_pg'_map'.
  replace ma'' with ma in H_pg'_map'.
  rewrite <- H_pg_map in H_pg'_map'.
  injection H_pg'_map'; intro; subst pg'; auto.
  apply va_mapped_same_ma with s v; auto.

  rewrite <- H_ma'_map in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map.
  injection H_pg'_map; auto.
  subst pg'.

  split.
  apply valid_cache_write with s v (Page (RW_c (Some v0)) (page_owned_by pg))
    ma new_cache new_mem new_tlb; auto.
  simpl.
  exists (Some v0); auto.
  inversion_clear H_update.
  rewrite <- H_pg_map in H; injection H; intro; subst x.
  rewrite H0; auto.

  apply valid_tlb_add with s v ma new_tlb new_mem new_cache; auto.
  generalize H_ma_map; intro H_ma_map'.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map' as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map']].
  exists pt_addr; auto.
  split.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping.
  exists aos; auto.
  exists p2m; auto.
  right.
  split; auto.
  rewrite H_s'; simpl.
  inversion_clear H_update.
  rewrite <- H_pg_map in H; injection H; intro; subst x.
  rewrite H0.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite <- H_ma_eq_pt_addr in H_pt_page_map.
  rewrite <- H_pg_map in H_pt_page_map.
  injection H_pt_page_map; intro H_pt_page_eq_pg.
  rewrite H_pt_page_eq_pg in H_mp_map.
  unfold page_type in H_ptype.
  rewrite <- H_pg_map in H_ptype.
  induction (page_content pg); discriminate.

  (* New untrusted *)
  unfold pre_new_untrusted, hypervisor_mapping_madd in H_pre.
  destruct H_pre as [H_os_untrusted [ma [p2m H_p2m_map H_ma_map]
    [H_ptype [H_ma_owner [H_v_accessible [H_hcall H_hyper_running]]]]]].
  unfold post_new_untrusted, post_all_new, hypervisor_mapping_madd,
    os_current_vadd_mapping_madd in H_post.
  destruct H_post as [ma' [p2m' H_p2m'_map H_ma'_map]
    [pt_addr [system H_system_map [p2m'' H_p2m''_map H_pt_addr_map]]
      [new_mem H_new_mem [o' H_o'_map H_s']]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  unfold hypervisor_mapping in H_p2m''_map.
  rewrite <- H_p2m_map in H_p2m''_map.
  injection H_p2m''_map; intro; subst p2m''; clear H_p2m''_map.
  rewrite <- H_system_map in H_o'_map.
  injection H_o'_map; intro; subst o'; clear H_o'_map.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.

  split.
  apply valid_cache_drop with s v o (map_add os_ident_eq (st_oss s) o (OS None (os_current_page system)))
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_cache s) v)
    else
      (st_cache s))
    new_mem
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_tlb s) v)
    else
      (st_tlb s)); auto.
  intros osi os H_os_map.
  induction (os_ident_eq o osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  left.
  exists ma; auto.
  rewrite H_s'; auto.

  apply valid_tlb_drop with s v o (map_add os_ident_eq (st_oss s) o (OS None (os_current_page system)))
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_cache s) v)
    else
      (st_cache s))
    new_mem
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_tlb s) v)
    else
      (st_tlb s)); auto.
  intros osi os H_os_map.
  induction (os_ident_eq o osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  left.
  exists ma; auto.

  (* New trusted *)
  unfold pre_new_trusted, hypervisor_mapping_madd in H_pre.
  destruct H_pre as [ma [p2m H_p2m_map H_ma_map]
    [H_ptype [H_ma_owner [H_v_accessible [H_hcall H_hyper_running]]]]].
  unfold post_new_trusted, post_all_new, hypervisor_mapping_madd,
    os_current_vadd_mapping_madd in H_post.
  destruct H_post as [ma' [p2m' H_p2m'_map H_ma'_map]
    [pt_addr [system H_system_map [p2m'' H_p2m''_map H_pt_addr_map]]
      [new_mem H_new_mem [o' H_o'_map H_s']]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  unfold hypervisor_mapping in H_p2m''_map.
  rewrite <- H_p2m_map in H_p2m''_map.
  injection H_p2m''_map; intro; subst p2m''; clear H_p2m''_map.
  rewrite <- H_system_map in H_o'_map.
  injection H_o'_map; intro; subst o'; clear H_o'_map.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.

  split.
  apply valid_cache_drop with s v (st_active_os s) (map_add os_ident_eq (st_oss s) (st_active_os s) (OS None (os_current_page system)))
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  induction (os_ident_eq (st_active_os s) osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.
  
  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  left.
  exists ma; auto.
  rewrite H_s'.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  apply valid_tlb_drop with s v (st_active_os s) (map_add os_ident_eq (st_oss s) (st_active_os s) (OS None (os_current_page system)))
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  induction (os_ident_eq (st_active_os s) osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  left.
  exists ma; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  (* New hyper *)
  unfold pre_new_hyper, hypervisor_mapping_madd in H_pre.
  destruct H_pre as [H_hyper_running [H_ptype
    [H_ma_owner [H_os H_v_not_accessible]]]].
  destruct H_os with (st_active_os s) as [p2m H_p2m_map].
  unfold running_os_not_hypercall in H_running_nothypcall.
  destruct H_running_nothypcall as [aos H_aos_map H_running_nothypcall].
  exists aos; auto.
  unfold post_new_hyper, post_all_new, hypervisor_mapping_madd,
    os_current_vadd_mapping_madd in H_post.
  
  destruct H_post as [aos H_aos_map [pt_addr [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr_map]]
    [new_mem H_update H_s']]].
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_aos_map in H_aos'_map.
  injection H_aos'_map; intro; subst aos'; clear H_aos'_map.

  split.
  apply valid_cache_drop with s v (st_active_os s) (st_oss s) 
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists aos; auto.
  exists p2m; auto.
  left.
  exists m; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  apply valid_tlb_drop with s v (st_active_os s) (st_oss s)
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists aos; auto.
  exists p2m; auto.
  left.
  exists m; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  (* Del untrusted *)
  unfold pre_del_untrusted, pre_all_del, hypervisor_mapping,
    vadd_mapping in H_pre.
  destruct H_pre as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_p2m_map,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  destruct H_pre_all_del as [mp [pt_page H_pt_page_map H_mp_map] [ma H_ma_map]].
  unfold post_del_untrusted, post_all_del, hypervisor_mapping in H_post.
  destruct H_post as [system' H_system'_map [p2m' H_p2m'_map
    [pt_addr' H_pt_addr'_map [new_mem H_new_nem]]]].
  rewrite <- H_system_map in H_system'_map;
    injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map;
    injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.

  split.
  apply valid_cache_drop with s v o (map_add os_ident_eq (st_oss s) o (OS None (os_current_page system)))
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_cache s) v)
    else
      (st_cache s))
    new_mem
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_tlb s) v)
    else
      (st_tlb s)); auto.
  intros osi os H_os_map.
  induction (os_ident_eq o osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.

  apply valid_tlb_drop with s v o (map_add os_ident_eq (st_oss s) o (OS None (os_current_page system)))
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_cache s) v)
    else
      (st_cache s))
    new_mem
    (if (os_ident_eq (st_active_os s) o) then
      (map_drop vadd_eq (st_tlb s) v)
    else
      (st_tlb s)); auto.
  intros osi os H_os_map.
  induction (os_ident_eq o osi).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  simpl; auto.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.

  (* Del trusted *)
  unfold pre_del_trusted in H_pre.
  unfold pre_del_trusted, current_vadd_mapping_madd,
    os_current_vadd_mapping_madd, hypervisor_mapping,
    pre_all_del, vadd_mapping in H_pre.
  destruct H_pre as [pt_addr [system H_system_map
    [p2m H_p2m_map H_pt_addr_map]] [[mp [pt_page H_pt_page_map H_mp_map] [ma H_ma_map]]
    [H_os_running [H_os_trusted H_v_accessible]]]].
  unfold post_del_trusted, current_vadd_mapping_madd,
    os_current_vadd_mapping_madd, hypervisor_mapping, post_all_del in H_post.
  destruct H_post as [pt_addr' [system' H_system'_map [p2m' H_p2m'_map H_pt_addr'_map]]
    [new_mem H_new_mem H_s']].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.

  split.
  apply valid_cache_drop with s v (st_active_os s) (st_oss s)
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  apply valid_tlb_drop with s v (st_active_os s) (st_oss s)
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  (* Del hyper *)
  unfold post_del_hyper, current_vadd_mapping_madd,
    os_current_vadd_mapping_madd, hypervisor_mapping,
    post_all_del in H_post.
  destruct H_post as [pt_addr [system H_system_map
    [p2m H_p2m_map H_pt_addr_map]] [new_mem H_new_mem H_s']].

  split.
  apply valid_cache_drop with s v (st_active_os s) (st_oss s) 
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  apply valid_tlb_drop with s v (st_active_os s) (st_oss s)
    (map_drop vadd_eq (st_cache s) v) new_mem (map_drop vadd_eq (st_tlb s) v); auto.
  intros osi os H_os_map.
  exists os; auto. 
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  exists pt_addr.
  unfold os_current_vadd_mapping_madd.
  exists system; auto.
  exists p2m; auto.
  right; auto.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.

  (* Swith *)
  unfold post_switch in H_post.
  rewrite H_post; split.
  apply valid_empty_cache; auto.
  apply valid_empty_tlb; auto.

  (* Lswith trusted *)
  unfold post_lswitch_trusted, post_all_lswitch in H_post.
  destruct H_post as [system H_system_map H_s'].
  rewrite H_s'.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.
  split; [apply valid_empty_cache | apply valid_empty_tlb]; auto.

  (* Lswitch untrusted *)
  unfold post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as [system H_system_map H_s'].
  rewrite H_s'.
  induction (os_ident_eq (st_active_os s) o).
  split; [apply valid_empty_cache | apply valid_empty_tlb]; auto.
  destruct H_cache_valid as [H_cache_size [H_cache_distinct H_cache_included]].
  destruct H_tlb_valid as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  unfold valid_cache, valid_tlb, cache_included, tlb_included in *;
    repeat split; simpl; auto.
  intros va pg H_pg_map.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_ptype H_pg_map_mem]]]; auto.
  exists ma.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  split.
  exists pt_addr.
  exists aos.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  simpl.
  exists p2m; auto.
  simpl.
  exists mp; auto.
  exists pt_page; auto.
  auto.
  intros va ma H_ma_map.
  elim H_tlb_included with va ma; auto.
  intros pt_addr H_pt_addr_map [mp H_mp_map H_ma_map'].
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [aos H_aos_map [p2m' H_p2m'_map H_pt_addr_map]].
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  exists pt_addr.
  exists aos.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  simpl.
  exists p2m'; auto.
  simpl.
  exists mp; auto.

  (* Hcall *)
  unfold post_hcall in H_post.
  destruct H_post as [system H_system_map H_s'].
  split.
  apply valid_cache_unchanged with s
    (map_add os_ident_eq (st_oss s) (st_active_os s) (OS (Some h) (os_current_page system)))
    (st_hypervisor s) EXMD_svc OSA_waiting; auto.

  intros osi os H_os_map.
  split; auto.
  induction (os_ident_eq osi (st_active_os s)).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS (Some h) (os_current_page system)); auto.
  split; simpl; auto.
  rewrite <- H_system_map in H_os_map; injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.
  
  apply valid_tlb_unchanged with s
    (map_add os_ident_eq (st_oss s) (st_active_os s) (OS (Some h) (os_current_page system)))
    (st_hypervisor s) EXMD_svc OSA_waiting; auto.
  intros osi os H_os_map.
  split; auto.
  induction (os_ident_eq osi (st_active_os s)).
  subst osi.
  rewrite lemma_map_add_apply.
  exists (OS (Some h) (os_current_page system)); auto.
  split; simpl; auto.
  rewrite <- H_system_map in H_os_map; injection H_os_map; intro; subst os; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists os; auto.

  (* Ret ctrl *)
  unfold post_ret_ctrl in H_post.
  split.
  apply valid_cache_unchanged with s (st_oss s) (st_hypervisor s) EXMD_svc OSA_waiting; auto.

  intros osi os H_os_map.
  split; auto.
  exists os; auto.

  unfold state_components_assign, state_components_update in H_post.
  replace (st_active_os s) with (st_active_os s').
  replace (st_oss s) with (st_oss s').
  replace (st_hypervisor s) with (st_hypervisor s').
  replace (st_memory s) with (st_memory s').
  replace (st_cache s) with (st_cache s').
  replace (st_tlb s) with (st_tlb s').

  elim H_post with Exec_Mode.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  unfold state_component_interp.
  intros [H_triv [H_exec H_activity]].
  rewrite <- H_exec, <- H_activity.
  destruct s'; auto.
  elim H_post with System_TLB; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with System_Cache; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with System_Memory; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Hypervisor_SC; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Guest_OSs; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  
  apply valid_tlb_unchanged with s (st_oss s) (st_hypervisor s) EXMD_svc OSA_waiting; auto.
  intros osi os H_os_map.
  split; auto.
  exists os; auto.

  unfold state_components_assign, state_components_update in H_post.
  replace (st_active_os s) with (st_active_os s').
  replace (st_oss s) with (st_oss s').
  replace (st_hypervisor s) with (st_hypervisor s').
  replace (st_memory s) with (st_memory s').
  replace (st_cache s) with (st_cache s').
  replace (st_tlb s) with (st_tlb s').

  elim H_post with Exec_Mode.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  unfold state_component_interp.
  intros [H_triv [H_exec H_activity]].
  rewrite <- H_exec, <- H_activity.
  destruct s'; auto.
  elim H_post with System_TLB; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with System_Cache; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with System_Memory; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Hypervisor_SC; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Guest_OSs; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  elim H_post with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.

  (* Chmod *)
  unfold post_chmod in H_post.
  unfold state_components_assign, state_components_update in H_post.
  assert (s' = State (st_active_os s) (st_oss s)
    (if (ctxt_oss ctxt (st_active_os s)) then EXMD_svc else EXMD_usr)
    OSA_running (st_hypervisor s) (st_memory s) (st_cache s) (st_tlb s)) as H_s'.
  replace (st_active_os s) with (st_active_os s').
  replace (st_oss s) with (st_oss s').
  replace (st_hypervisor s) with (st_hypervisor s').
  replace (st_memory s) with (st_memory s').
  replace (st_cache s) with (st_cache s').
  replace (st_tlb s) with (st_tlb s').
  unfold trusted_aos in H_post.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]].
  replace (st_active_os s) with (st_active_os s') in H_trusted.
  rewrite H_trusted.
  elim H_post with Exec_Mode.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  unfold state_component_interp.
  intros [H_triv [H_exec H_activity]].
  rewrite <- H_exec, <- H_activity.
  destruct s'; auto.
  elim H_post with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  assert (ctxt_oss ctxt (st_active_os s) = false) as H_untrusted2.
  induction (ctxt_oss ctxt (st_active_os s)); auto.
  elim H_untrusted; auto.
  replace (st_active_os s) with (st_active_os s') in H_untrusted2.
  rewrite H_untrusted2.
  elim H_post with Exec_Mode.
  intros [[H_abs H_triv] H_interp]; elim H_abs; auto.
  unfold state_component_interp.
  intros [H_triv [H_exec H_activity]].
  rewrite <- H_exec, <- H_activity.
  destruct s'; auto.
  elim H_post with Active_OS; auto.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; elim H_abs; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with System_TLB; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with System_Cache; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with System_Memory; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with Hypervisor_SC; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with Guest_OSs; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.
  destruct H_post as [[H_trusted H_post] | [H_untrusted H_post]];
    elim H_post with Active_OS; intros [H_triv H_interp];
      unfold state_component_interp in H_interp; auto;
        elim H_triv; split; intro; discriminate.

  split.
  apply valid_cache_unchanged with s (st_oss s) (st_hypervisor s)
    (if (ctxt_oss ctxt (st_active_os s)) then EXMD_svc else EXMD_usr)    
    OSA_running; auto.

  intros osi os H_os_map.
  split; auto.
  exists os; auto.

  apply valid_tlb_unchanged with s (st_oss s) (st_hypervisor s)
    (if (ctxt_oss ctxt (st_active_os s)) then EXMD_svc else EXMD_usr)    
    OSA_running; auto.
  intros osi os H_os_map.
  split; auto.
  exists os; auto.

  (* Page pin trusted *)
  unfold post_page_pin_trusted, post_all_pin,
    hypervisor_mapping in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content
    [p2m H_p2m_map [system H_system_map H_s']]]]]].

  destruct H_cache_valid as [H_valid_cache_size [H_cache_distinct H_cache_included]].
  destruct H_tlb_valid as [H_valid_tlb_size [H_tlb_distinct H_tlb_included]].

  repeat split;
    rewrite H_s'; auto.

  (* cache included *)
  unfold cache_included in *; simpl.
  intros va pg' H_pg'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_cache_included.
  destruct H_cache_included with va pg' as [ma' [[pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]] [H_pg'_content H_pg'_map_mem]]]; auto.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  induction (madd_eq ma ma').
  subst ma'.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg2 H_pg2_map [H_acc H_not_acc]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  unfold os_accessible in *.
  rewrite <- H_pg_map in H_pg2_map.
  injection H_pg2_map; intro; subst pg2; clear H_pg2_map.
  induction (ctxt_vadd_accessible ctxt va).
  rewrite H_acc in H_pg_owner; auto; discriminate.
  rewrite H_not_acc in H_pg_owner; auto; discriminate.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_add padd_eq p2m p ma).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_pin_trusted, pre_all_pin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [[p2m' H_p2m'_map
    H_p_not_mapped] [ma2 [pg2 H_pg2_map [H_pg2_owner H_pg2_content]]]]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_p_eq_curr in H_pt_addr_map.
  rewrite <- H_pt_addr_map in H_p_not_mapped.
  discriminate.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.

  unfold page_type.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  (* tlb included *)
  unfold tlb_included in *; simpl.
  intros va ma' H_ma'_map_tlb.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va ma' as [pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]]; auto.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_add padd_eq p2m p ma).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_pin_trusted, pre_all_pin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [[p2m' H_p2m'_map
    H_p_not_mapped] [ma2 [pg2 H_pg2_map [H_pg2_owner H_pg2_content]]]]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_p_eq_curr in H_pt_addr_map.
  rewrite <- H_pt_addr_map in H_p_not_mapped.
  discriminate.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.

  (* Page unpin trusted *)
  unfold post_page_unpin_trusted, post_all_unpin,
    hypervisor_mapping in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map
    [system H_system_map H_s']]].

  destruct H_cache_valid as [H_valid_cache_size [H_cache_distinct H_cache_included]].
  destruct H_tlb_valid as [H_valid_tlb_size [H_tlb_distinct H_tlb_included]].

  repeat split;
    rewrite H_s'; auto.

  (* cache included *)
  unfold cache_included in *; simpl.
  intros va pg' H_pg'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_cache_included.
  destruct H_cache_included with va pg' as [ma' [[pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]] [H_pg'_content H_pg'_map_mem]]]; auto.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  induction (madd_eq ma ma').
  subst ma'.
  unfold pre_page_unpin_trusted, pre_all_unpin, hypervisor_mapping in H_pre.
  destruct H_pre as [H_os_running [H_trusted [system' H_system'_map [H_p_not_curr
    [p2m' H_p2m'_map [ma2 H_ma2_map [H_ptype H_no_va_mapped_to_ma]]]]]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.
  unfold no_va_mapped_to_ma in H_no_va_mapped_to_ma.
  elim H_no_va_mapped_to_ma with pt_addr pt_page va mp; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_drop padd_eq p2m p).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite map_drop_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [system' H_system'_map [H_p_not_curr H_pre]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold injective_hyper_mappings, injective_mapping in H_injective_hyper.
  assert (p = (os_current_page system)) as H_p_current.
  apply H_injective_hyper with (st_active_os s) p2m; auto.
  rewrite <- H_pt_addr_map.
  rewrite <- H_ma_eq_pt_addr; auto.
  unfold pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [system' H_system'_map [H_p_not_curr H_pre]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'.
  apply H_p_not_curr; auto.
  unfold page_type; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  
  (* tlb included *)
  unfold tlb_included in *; simpl.
  intros va ma' H_ma'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va ma' as [pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map_mem]]; auto.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_drop padd_eq p2m p).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite map_drop_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [system' H_system'_map [H_p_not_curr H_pre]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold injective_hyper_mappings, injective_mapping in H_injective_hyper.
  assert (p = (os_current_page system)) as H_p_current.
  apply H_injective_hyper with (st_active_os s) p2m; auto.
  rewrite <- H_pt_addr_map.
  rewrite <- H_ma_eq_pt_addr; auto.
  unfold pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running [H_trusted [system' H_system'_map [H_p_not_curr H_pre]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'.
  apply H_p_not_curr; auto.

  (* Page pin untrusted *)
  unfold post_page_pin_untrusted, post_all_pin,
    hypervisor_mapping in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content
    [p2m H_p2m_map [system H_system_map H_s']]]]]].

  destruct H_cache_valid as [H_valid_cache_size [H_cache_distinct H_cache_included]].
  destruct H_tlb_valid as [H_valid_tlb_size [H_tlb_distinct H_tlb_included]].

  repeat split;
    rewrite H_s'; auto.

  (* cache included *)
  unfold cache_included in *; simpl.
  intros va pg' H_pg'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_cache_included.
  destruct H_cache_included with va pg' as [ma' [[pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]] [H_pg'_content H_pg'_map_mem]]]; auto.
  induction (os_ident_eq o (st_active_os s)).
  (* o active *)
  subst o.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  induction (madd_eq ma ma').
  subst ma'.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg2 H_pg2_map [H_acc H_not_acc]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  unfold os_accessible in *.
  rewrite <- H_pg_map in H_pg2_map.
  injection H_pg2_map; intro; subst pg2; clear H_pg2_map.
  induction (ctxt_vadd_accessible ctxt va).
  rewrite H_acc in H_pg_owner; auto; discriminate.
  rewrite H_not_acc in H_pg_owner; auto; discriminate.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_add padd_eq p2m p ma).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_pin_untrusted, pre_all_pin in H_pre.
  destruct H_pre as [H_untrusted [H_hcall [H_hyper_running [[p2m' H_p2m'_map
    H_p_not_mapped] [ma2 [pg2 H_pg2_map [H_pg2_owner H_pg2_content]]]]]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_p_eq_curr in H_pt_addr_map.
  rewrite <- H_pt_addr_map in H_p_not_mapped.
  discriminate.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.
  unfold page_type; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  
  (* o not active *)
  induction (madd_eq ma ma').
  subst ma'.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg2 H_pg2_map [H_acc H_not_acc]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system' p2m' as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  unfold os_accessible in *.
  rewrite <- H_pg_map in H_pg2_map.
  injection H_pg2_map; intro; subst pg2; clear H_pg2_map.
  induction (ctxt_vadd_accessible ctxt va).
  rewrite H_acc in H_pg_owner; auto; discriminate.
  rewrite H_not_acc in H_pg_owner; auto; discriminate.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists system'.
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.
  unfold page_type; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  
  (* tlb included *)
  unfold tlb_included in *; simpl.
  intros va ma' H_ma'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va ma' as [pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map_mem]]; auto.
  induction (os_ident_eq o (st_active_os s)).
  (* o active *)
  subst o.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.

  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_add padd_eq p2m p ma).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_p_eq_curr.
  unfold pre_page_pin_untrusted, pre_all_pin in H_pre.
  destruct H_pre as [H_untrusted [H_hcall [H_hyper_running [[p2m' H_p2m'_map
    H_p_not_mapped] [ma2 [pg2 H_pg2_map [H_pg2_owner H_pg2_content]]]]]]].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_p_eq_curr in H_pt_addr_map.
  rewrite <- H_pt_addr_map in H_p_not_mapped.
  discriminate.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.
  rewrite lemma_map_add_apply_distinct; auto.
  (* o not active *)
  exists pt_addr.
  exists system'; auto.
  exists p2m'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  rewrite H_ma_eq_pt_addr in H_pg_map. 
  rewrite <- H_pt_page_map in H_pg_map.
  injection H_pg_map; intro; subst pg.
  rewrite <- H_mp_map in H_pg_content; discriminate.

  (* Page unpin untrusted *)
  unfold post_page_unpin_untrusted, post_all_unpin,
    hypervisor_mapping in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map
    [system H_system_map H_s']]].
  destruct H_cache_valid as [H_valid_cache_size [H_cache_distinct H_cache_included]].
  destruct H_tlb_valid as [H_valid_tlb_size [H_tlb_distinct H_tlb_included]].

  repeat split;
    rewrite H_s'; auto.

  (* cache included *)
  unfold cache_included in *; simpl.
  intros va pg' H_pg'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_cache_included.
  destruct H_cache_included with va pg' as [ma' [[pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]] [H_pg'_content H_pg'_map_mem]]]; auto.

  induction (os_ident_eq o (st_active_os s)).
  (* o active *)
  subst o.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold pre_page_unpin_untrusted, pre_all_unpin,
    hypervisor_mapping in H_pre.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  destruct H_pre as [H_untrusted [H_hcall [H_hyper_running
    [system' H_system'_map [H_p_not_curr [p2m' H_p2m'_map
      [ma2 H_ma2_map [H_ptype H_no_va_mapped_to_ma]]]]]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.

  induction (madd_eq ma ma').
  subst ma'.
  unfold no_va_mapped_to_ma in H_no_va_mapped_to_ma.
  elim H_no_va_mapped_to_ma with pt_addr pt_page va mp; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_drop padd_eq p2m p).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite map_drop_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold injective_hyper_mappings, injective_mapping in H_injective_hyper.
  assert (p = (os_current_page system)) as H_p_current.
  apply H_injective_hyper with (st_active_os s) p2m; auto.
  rewrite <- H_pt_addr_map.
  rewrite <- H_ma_eq_pt_addr; auto.
  apply H_p_not_curr; auto.
  unfold page_type; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  (* o not active *)
  induction (madd_eq ma ma').
  subst ma'.
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg2 H_pg2_map [H_acc H_not_acc]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) system' p2m' as
    [pt_addr' H_pt_addr'_map [H_pt_page_type H_pt_page_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  destruct H_valid_hyper with o p p2m as
    [[ma2 H_ma2_map H_ma_owner] | H_abs]; auto.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg2_map in H_ma_owner.
  unfold os_accessible in *.
  induction (ctxt_vadd_accessible ctxt va).
  rewrite H_acc in H_ma_owner; auto.
  injection H_ma_owner; intro; elim b; auto.
  rewrite H_not_acc in H_ma_owner; auto; discriminate.
  rewrite <- H_ma_map in H_abs; discriminate H_abs.

  exists ma'.
  repeat split.
  exists pt_addr.
  exists system'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) (os_current_page system') p2m' as
    [[pt_addr2 H_pt_addr2_map H_pt_addr_owner] | H_abs]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr2_map.
  injection H_pt_addr2_map; intro; subst pt_addr2; clear H_pt_addr2_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  destruct H_valid_hyper with o p p2m as
    [[ma2 H_ma2_map H_ma_owner] | H_abs]; auto.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_eq_pt_addr in H_ma_owner.
  rewrite <- H_pt_page_map in H_ma_owner.
  rewrite H_pt_addr_owner in H_ma_owner.
  injection H_ma_owner; intro; apply b; auto.
  rewrite <- H_ma_map in H_abs; discriminate H_abs.
  rewrite <- H_pt_addr_map in H_abs; discriminate H_abs.
  unfold page_type; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  (* tlb included *)
  unfold tlb_included in *; simpl.
  intros va ma' H_ma'_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; simpl.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_tlb_included.
  destruct H_tlb_included with va ma' as [pt_addr [system'
    H_system'_map [p2m' H_p2m'_map H_pt_addr_map]]
  [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map_mem]]; auto.

  induction (os_ident_eq o (st_active_os s)).
  (* o active *)
  subst o.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold pre_page_unpin_untrusted, pre_all_unpin,
    hypervisor_mapping in H_pre.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  destruct H_pre as [H_untrusted [H_hcall [H_hyper_running
    [system' H_system'_map [H_p_not_curr [p2m' H_p2m'_map
      [ma2 H_ma2_map [H_ptype H_no_va_mapped_to_ma]]]]]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.
  exists pt_addr.
  exists (OS None (os_current_page system)).
  rewrite lemma_map_add_apply; auto.
  exists (map_drop padd_eq p2m p).
  rewrite lemma_map_add_apply; auto.
  simpl.
  rewrite map_drop_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold injective_hyper_mappings, injective_mapping in H_injective_hyper.
  assert (p = (os_current_page system)) as H_p_current.
  apply H_injective_hyper with (st_active_os s) p2m; auto.
  rewrite <- H_pt_addr_map.
  rewrite <- H_ma_eq_pt_addr; auto.
  apply H_p_not_curr; auto.
  (* o not active *)
  exists pt_addr.
  exists system'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists p2m'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp; auto.
  exists pt_page; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma_eq_pt_addr.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) (os_current_page system') p2m' as
    [[pt_addr2 H_pt_addr2_map H_pt_addr_owner] | H_abs]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr2_map.
  injection H_pt_addr2_map; intro; subst pt_addr2; clear H_pt_addr2_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  destruct H_valid_hyper with o p p2m as
    [[ma2 H_ma2_map H_ma_owner] | H_abs]; auto.
  rewrite <- H_ma_map in H_ma2_map.
  injection H_ma2_map; intro; subst ma2; clear H_ma2_map.
  unfold maddress_owner in H_ma_owner.
  rewrite H_ma_eq_pt_addr in H_ma_owner.
  rewrite <- H_pt_page_map in H_ma_owner.
  rewrite H_pt_addr_owner in H_ma_owner.
  injection H_ma_owner; intro; apply b; auto.
  rewrite <- H_ma_map in H_abs; discriminate H_abs.
  rewrite <- H_pt_addr_map in H_abs; discriminate H_abs.
Qed.

End Secure_Cache.