(*******************************************************************
 * File: Exec.v
 * $Revision: 476 $
 * $Date: 2011-02-23 18:19:58 -0200 (Wed, 23 Feb 2011) $
 * $Author: jdcampo $
 *******************************************************************
 * This module specifies the semantics of action execution
 * 
 ******************************************************************)
Require Import List.
Require Export Env.
Require Import State_valid.
Require Export State_proof.
Require Export State_effects.
Require Export Actions.

(** *** Pre and post-conditions *)

(** The following definitions specify the preconditions (properties on
the initial state [s]) and post-condition (properties relating [s] to
the final state [s']) of the actions. *)

Definition same_state (s s':state) : Prop :=
  exists new_cache, exists new_tlb,
    valid_cache s' /\ valid_tlb s' /\ 
    state_components_assign System_Cache System_TLB new_cache new_tlb s s'.

(** **** Silent *)
(** represents an action (some OS sub-specified action, i.e., silent) 
in which the properties relevant to our state abstraction don't change  *)

Definition pre_silent (s:state) := True.

Definition post_silent := same_state.

(** **** Read *)

(** Read in the memory : 
    the virtual address can be mapped to any kind of page *)

(** For all different read actions, we can make some general pre/post *)
Definition pre_all_read (va:vadd) (s:state) :=
  exists ma, va_mapped_to_ma s (st_active_os s) va ma /\
    page_type (st_memory s) ma = Some RW.

Definition post_all_read (va:vadd) (s s':state) :=
  forall ma pg, Some ma = va_mapped_to_ma_cache s va ->
    Some pg = va_mapped_to_pg_cache s va ->
    exists2 new_cache, cache_add (st_cache s) new_cache va pg &
    exists2 new_tlb, tlb_add (st_tlb s) new_tlb va ma &
      state_components_assign System_Cache System_TLB new_cache new_tlb s s'.

(** [read] 
corresponds to an OS reading virtual address va from its address space.*)
Definition pre_read va ctxt s := 
              pre_all_read va s 
   (** va is not reserved for the Hypervisor *)
           /\ os_accessible ctxt va
   (** guest OS is executing *)
           /\ os_running s.

Definition post_read := post_all_read.

(** [read_hyper]
corresponds to the Hypervisor reading virtual address va.*)

Definition pre_read_hyper va s := 
              pre_all_read va s
   (** Hypervisor is executing *)
           /\ hyper_running s.
   (** Here, we could add for example, not allowing the Hypervisor to read 
      a page  allocated to an OS which is not a page table... 
      At the moment, we are allowing the Hypervisor 
      to read all the memory that belongs to the Active OS. *)

Definition post_read_hyper := post_all_read.

(** **** Write *)
Fixpoint fix_cache_synonym (s:state) (c:cache) (ma:madd) : cache :=
  match c with
    | nil => nil
    | Item va' pg::c' =>
      match (va_mapped_to_ma_cache s va') with
        | Some ma' =>
          if (madd_eq ma ma') then
            (fix_cache_synonym s c' ma)
          else
            (Item va' pg)::(fix_cache_synonym s c' ma)
        | _ => (fix_cache_synonym s c' ma)
      end
  end.

(** va is mapped in the Active Memory Mapping *)
Definition post_all_write va val s s' :=
  exists2 ma:madd,
     Some ma = va_mapped_to_ma_cache s va &
     exists2 new_mem, update_rw_page ma val (st_memory s) new_mem &
       exists2 pg, Some pg = va_mapped_to_pg_cache s va &
       let new_pg := Page (RW_c (Some val)) (page_owned_by pg) in
         exists2 new_cache, cache_add (fix_cache_synonym s (st_cache s) ma) new_cache va new_pg &
           exists2 new_tlb, tlb_add (st_tlb s) new_tlb va ma &

           s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
           new_mem new_cache new_tlb.

(** [write] represents an OS writing value val in virtual address va 
from its address space.*)

Definition pre_write va (val : value) ctxt s :=
   (** guest OS is executing *)
                os_running s
   (** va is not reserved for the Hypervisor *)
             /\ os_accessible ctxt va
   (** va is mapped to Machine address ma in the Active Memory Mapping *)
             /\ exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
   (** Check that the page being written is not a page table 
       This can be discarded without impact in the case of VLX*)
                page_type (st_memory s) ma = Some RW.
   (** We do not require that the page is owned by the active OS. This
   property can be proved in the presence of a valid state *)

Definition post_write := post_all_write.


(** [write_hyper]
represents the Hypervisor writing value val in virtual address va. *)

Definition pre_write_hyper va (val:value) s :=
  (** The Hypervisor is executing *)
             hyper_running s
   (** va is mapped to Machine address ma in the Active Memory Mapping *)
          /\ exists2 ma, va_mapped_to_ma s (st_active_os s) va ma &
             page_type (st_memory s) ma = Some RW 
  (* machine address ma belongs to the Hypervisor *)
          /\ maddress_owner (st_memory s) ma = Hyp_PO.
  
(** Here we restricted the Hypervisor to write memory values in pages 
of his own. *)

Definition post_write_hyper := post_all_write.


(** **** New *)

(** All the new actions correspond to adding a new pair [(va, pa)] 
to the given page table, 
where [va] is a virtual address and [pa] is a physical one 
*)

(** Postcondition shared by all the 'new' actions *)
Definition post_all_new osi (va:vadd) (pa:padd) hcall (s s':state) :=
  (* The mapping va->ma is added to the PT in pt_addr.
     hcall changes to None with an untrusted OS, otherwise it remain unchanged *)
  exists2 ma,
     hypervisor_mapping_madd (st_hypervisor s) osi pa ma &
     exists2 pt_addr, os_current_vadd_mapping_madd s osi pt_addr &
           exists2 newmem, update_pt_page pt_addr va ma (st_memory s) newmem &
           exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
           let newos := OS hcall (os_current_page system) in
           let newoss := (map_add os_ident_eq (st_oss s) osi newos) in
           let new_cache := if (os_ident_eq (st_active_os s) osi) then
                             (map_drop vadd_eq (st_cache s) va)
                           else
                             (st_cache s)
              in               
           let new_tlb := if (os_ident_eq (st_active_os s) osi) then
                             (map_drop vadd_eq (st_tlb s) va)
                           else
                             (st_tlb s)
              in               
             s' = State (st_active_os s) newoss (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    newmem new_cache new_tlb.

(** [new_trusted] corresponds to a new action done by a trusted OS. *)
Definition pre_new_trusted (va:vadd) (pa:padd) (ctxt:context) (s:state) :=
  (** pa should be mapped to a madd owned and pinned by the active OS. *)        
  exists2 ma, hypervisor_mapping_madd (st_hypervisor s) (st_active_os s) pa ma &
                 page_type (st_memory s) ma <> None
              /\ maddress_owner (st_memory s) ma = Os_PO (st_active_os s)
        (** va is not reserved for the Hypervisor ; As before this can be discarded *)
              /\ os_accessible ctxt va
              /\ trusted_aos ctxt s
              /\ os_running s.

Definition post_new_trusted va pa s s' :=
  (** The Active memory mapping in s' maps va to ma, 
  where pa is mapped to ma by the hypervisor *)
      post_all_new (st_active_os s) va pa None s s'.

(** [new_untrusted] corresponds to a new action done by the Hypervisor
for an untrusted OS that had asked for it. *)
Definition pre_new_untrusted osi va pa ctxt s := 
       ctxt_oss ctxt osi = false (* untrusted osi *) 
  (** pa should be mapped to a madd owned and pinned by osi. *)
    /\ exists2 ma, hypervisor_mapping_madd (st_hypervisor s) osi pa ma &
         page_type (st_memory s) ma <> None
    /\ maddress_owner (st_memory s) ma = Os_PO osi
       (** va is not reserved for the Hypervisor ; As before this can be discarded *)
    /\ os_accessible ctxt va
       (** Untrusted OS has a 'new' hypercall *)
    /\ get_os_hypercall s osi (Some (Hyperv_call_new va pa))
    /\ hyper_running s.

Definition post_new_untrusted osi va pa s s' :=
  (** The current memory mapping in [s'] maps [va] to [ma], 
  where pa is mapped to ma by hypervisor *)
  post_all_new osi va pa None s s'.

(** [new_hyper] corresponds  to the action of the Hypervisor adding a 
new mapping to the Active Memory Mapping for its own needs. *)

Definition pre_new_hyper va ma ctxt s :=
   (** Hypervisor is running *)
          hyper_running s
          /\ page_type (st_memory s) ma <> None
          /\ maddress_owner (st_memory s) ma = Hyp_PO
          /\ (forall osi, (exists os, Value _ os = map_apply os_ident_eq (st_oss s) osi) ->
            exists p2m, Value _ p2m = map_apply os_ident_eq (st_hypervisor s) osi)
   (** va is reserved for the Hypervisor *)
          /\ ~(os_accessible ctxt va).

Definition post_new_hyper va ma s s':=
   (** The Active memory mapping in s' maps va to pa*)
   exists2 aos, Value _ aos = map_apply os_ident_eq (st_oss s) (st_active_os s) &
     exists2 pt_addr, os_current_vadd_mapping_madd s (st_active_os s) pt_addr &
       exists2 newmem, update_pt_page pt_addr va ma (st_memory s) newmem &
           let new_cache := (map_drop vadd_eq (st_cache s) va) in
           let new_tlb := (map_drop vadd_eq (st_tlb s) va) in
             s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    newmem new_cache new_tlb.

(** **** Switch *)

(** [switch] corresponds to a changing of Active OS by the Hypervisor. 
At the moment this happens only when some OS returns control to the 
Hypervisor, because we haven't specified an interruption mechanism yet.  *)

Definition pre_switch osi s :=
    (** Hypervisor is running *)
           hyper_running s
    (** There shouldn't be any pending request for the new OS *)
          /\ exists2 system:os, 
                   Value os_ident system = map_apply os_ident_eq (st_oss s) osi & 
                   (os_hcall system) = None.

Definition post_switch osi s s' :=
   (** new active OS *)
   (** Hypervisor still running after switching. *)
  let new_cache := cache_flush (st_cache s) in
  let new_tlb := tlb_flush (st_tlb s) in
    s' = State osi (st_oss s) (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
    (st_memory s) new_cache new_tlb.

(** **** Local switch *)

(** [lswitch] corresponds to a process context switch inside some 
Active OS executed by the Hypervisor on behalf of  the untrusted 
OSes and by the trusted OSes.   *)

Definition pre_all_lswitch pa osi s := 
   (** The new page table address belongs to the OS *)
       exists2 ma, hypervisor_mapping_madd (st_hypervisor s) osi pa ma &
                   page_type (st_memory s) ma = Some PT
                /\ maddress_owner (st_memory s) ma = Os_PO osi.

Definition post_all_lswitch osi pa hcall s s' :=
   (** The new current page table address is pa *)
       exists2 aos, Value _ aos = map_apply os_ident_eq (st_oss s) osi &
                let aos' := OS hcall pa in
                let new_oss := map_add os_ident_eq (st_oss s) osi aos' in
                let new_cache := if (os_ident_eq (st_active_os s) osi) then
                                      cache_flush (st_cache s) 
                                 else
                                      (st_cache s)
                   in               
                let new_tlb := if (os_ident_eq (st_active_os s) osi) then
                                      tlb_flush (st_tlb s) 
                                 else
                                      (st_tlb s)
                   in               
                  s' = State (st_active_os s) new_oss (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                  (st_memory s) new_cache new_tlb.

Definition pre_lswitch_trusted pa ctxt s := 
           os_running s /\ trusted_aos ctxt s /\ pre_all_lswitch pa (st_active_os s) s.

Definition post_lswitch_trusted pa s s' := 
        post_all_lswitch (st_active_os s) pa None s s'.

Definition pre_lswitch_untrusted osi pa ctxt s := 
              ctxt_oss ctxt osi = false (* untrusted osi *) 
           /\ exists2 aos, Value _ aos = map_apply os_ident_eq (st_oss s) osi &  
              os_hcall aos = Some (Hyperv_call_lswitch pa)
           /\ hyper_running s
           /\ pre_all_lswitch pa osi s.

Definition post_lswitch_untrusted osi pa s s' := 
        post_all_lswitch osi pa None s s'.


(** **** Delete mapping *)


(** All the 'delete' actions
corresponds to removing an existing  pair [(va, pa)] from the Active 
memory mapping, where [va] is a virtual address and [pa] is a machine one 
*)

Definition pre_all_del (pt_addr:madd) va s :=
  (** va is defined in the Active memory mapping *)
  exists2 map, vadd_mapping pt_addr (st_memory s) map &
                 exists ma, Value _ ma = map_apply vadd_eq map va.

Definition post_all_del va (pt_addr:madd) s s':=
  exists2 newmem, drop_va_pt_page pt_addr va (st_memory s) newmem &
           let new_cache := (map_drop vadd_eq (st_cache s) va) in
           let new_tlb := (map_drop vadd_eq (st_tlb s) va) in
             s' = State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    newmem new_cache new_tlb.

(** [del_untrusted] corresponds to a delete action done by the Hypervisor
for an untrusted OS that had asked for it. *)
Definition pre_del_untrusted osi va ctxt s :=
  ctxt_oss ctxt osi = false (* untrusted osi *) 
  (** If the Active OS is untrusted 
      then there is a request to the Hypervisor *)
/\ get_os_hypercall s osi (Some (Hyperv_call_del va))
/\ exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
      exists2 p2m, hypervisor_mapping osi (st_hypervisor s) p2m &
         exists2 pt_addr, Value _ pt_addr = map_apply padd_eq p2m (os_current_page system) &
                     pre_all_del pt_addr va s 
/\ hyper_running s
  (** va is not reserved for the Hypervisor ; As before this can be discarded *)
/\ os_accessible ctxt va.

Definition post_del_untrusted osi va s s' :=
  exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
     exists2 p2m, hypervisor_mapping osi (st_hypervisor s) p2m &
        exists2 pt_addr, Value _ pt_addr = map_apply padd_eq p2m (os_current_page system) &
        let system' := OS None (os_current_page system) in
          exists2 newmem, drop_va_pt_page pt_addr va (st_memory s) newmem &
          let newoss := (map_add os_ident_eq (st_oss s) osi system') in
          let new_cache := if (os_ident_eq (st_active_os s) osi) then
                             (map_drop vadd_eq (st_cache s) va)
                           else
                             (st_cache s)
            in
          let new_tlb := if (os_ident_eq (st_active_os s) osi) then
                           (map_drop vadd_eq (st_tlb s) va)
                         else
                           (st_tlb s)
              in             
            s' = State (st_active_os s) newoss (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                   newmem new_cache new_tlb.

(** [del_trusted] corresponds to a delete action done by a trusted OS. *)
Definition pre_del_trusted va ctxt s :=
  exists2 pt_addr, current_vadd_mapping_madd s pt_addr &
                   pre_all_del pt_addr va s 
  (** If the Active OS is trusted then there is no request to the Hypervisor *)
           /\ os_running s /\ trusted_aos ctxt s
  (** va is not reserved for the Hypervisor ; As before this can be discarded *)
           /\ os_accessible ctxt va.

Definition post_del_trusted va s s' :=
  exists2 pt_addr, current_vadd_mapping_madd s pt_addr &
       post_all_del va pt_addr s s'.

(** [del_hyper]
 corresponds  to the action of the Hypervisor removing one of its own
mapping to the Active Memory Mapping. *)

Definition pre_del_hyper va ctxt s :=
  exists2 pt_addr, current_vadd_mapping_madd s pt_addr &
                   pre_all_del pt_addr va s 
           /\ hyper_running s
  (** va is reserved for the Hypervisor. *)
           /\ ~(os_accessible ctxt va).

Definition post_del_hyper va s s' :=
  exists2 pt_addr, current_vadd_mapping_madd s pt_addr &
       post_all_del va pt_addr s s'.
 
(** **** Hypercall *)

(** [hcall]
corresponds  to an untrusted OS making an hypercall of type hc. 
This is the case when the untrusted OSes need the Hypervisor to run 
some service that needs privilege on their behalf. 
So this transition changes the execution mode from usr to svc 
and sets the required service *)            

Definition pre_hcall ctxt s :=
  os_running s /\ ~(trusted_aos ctxt s).

Definition post_hcall hc s s' :=
  exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) (st_active_os s) &
  let newos := OS (Some hc) (os_current_page system) in
  let newoss := map_add os_ident_eq (st_oss s) (st_active_os s) newos in
    s' = State (st_active_os s) newoss EXMD_svc OSA_waiting (st_hypervisor s)
    (st_memory s) (st_cache s) (st_tlb s).

(** **** Changing control *)

(** [ret_ctrl] corresponds to Hypervisor getting back the control.
This can be seen both like a timer interruption 
or a call from an OS when it is going idle.
So this transition changes the execution mode from user to supervisor mode *)

Definition pre_ret_ctrl s := os_running s.

Definition post_ret_ctrl s s' :=
  state_components_assign Exec_Mode OS_Activity EXMD_svc OSA_waiting s s'.

(** [chmod] corresponds  to the Hypervisor changing the execution mode, 
particularly to let an OS execute. 
It could be the case after a context switch or any Hypervisor execution
was carried out by the Hypervisor on behalf of an untrusted OS. *)            

Definition pre_chmod s :=
              hyper_running s
   (** There is no request to the Hypervisor pending in s*)
           /\ get_os_hypercall s (st_active_os s) None.

Definition post_chmod ctxt s s' :=
   ((trusted_aos ctxt s) /\ state_components_assign Exec_Mode OS_Activity EXMD_svc OSA_running s s')
\/ (~(trusted_aos ctxt s) /\ state_components_assign Exec_Mode OS_Activity EXMD_usr OSA_running s s').


(** **** Page classification/registration *)

(** [page_pin] corresponds to the registration of a page 
(and classification with type [t]) by a trusted OS 
or by the Hypervisor for the untrusted OS. 
The Hypervisor doesn't use it for its own purposes. *)

Definition pre_all_pin osi pa s :=
(** There is at least an emtpy madd to assign to pa. *)
  (exists2 p2m, Value _ p2m = map_apply os_ident_eq (st_hypervisor s) osi &
    map_apply padd_eq p2m pa = Error _ pa) /\
  (exists ma, exists2 pg, Value _ pg = map_apply madd_eq (st_memory s) ma &
    page_owned_by pg = No_Owner /\ page_content pg = Other_c).

Definition post_all_pin osi pa type s s' :=
(** Get a free madd (it exists because of pre). *)
  exists ma, 
     exists2 pg, Value _ pg = map_apply madd_eq (st_memory s) ma &
     page_owned_by pg = No_Owner /\ page_content pg = Other_c /\
 (** The new page is registered with type type and owner osi in s' *)
     let newpage := new_page (Some type) (Os_PO osi) in
       (** Update the new page in system memory *)
       let newmem := map_add madd_eq (st_memory s) ma newpage in
     (** pa is added to the hypervisor mapping for osi *)
         exists2 p2m, hypervisor_mapping osi (st_hypervisor s) p2m &
            let newp2m := map_add padd_eq p2m pa ma in
            let newhyper := map_add os_ident_eq (st_hypervisor s) osi newp2m in
            exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
            let newos := OS None (os_current_page system) in
              let newoss := map_add os_ident_eq (st_oss s) osi newos in
                s' = State (st_active_os s) newoss (st_exec_mode s)
                           (st_aos_activity s) newhyper newmem
                           (st_cache s) (st_tlb s).

Definition pre_page_pin_untrusted osi pa type ctxt s :=
  (** osi is not trusted *)
     (ctxt_oss ctxt) osi = false
  (** If osi is untrusted then the corresponding request is pending *)
  /\ get_os_hypercall s osi (Some (Hyperv_call_pin pa type))
  /\ hyper_running s
  /\ pre_all_pin osi pa s.

Definition post_page_pin_untrusted osi pa type s s' :=
  post_all_pin osi pa type s s'.

Definition pre_page_pin_trusted ctxt pa s :=
  os_running s /\ trusted_aos ctxt s /\ pre_all_pin (st_active_os s) pa s.

Definition post_page_pin_trusted pa type s s' := 
  post_all_pin (st_active_os s) pa type s s'.


(** [page_unpin] corresponds to the de-registration 
(and declassification) of previously a registered/classified page. *)
Definition no_va_mapped_to_ma osi ma s : Prop :=
  forall pt_addr pt_page va mp,
    Value _ pt_page = map_apply madd_eq (st_memory s) pt_addr ->
    page_owned_by pt_page = Os_PO osi -> page_content pt_page = PT_c mp ->
    map_apply vadd_eq mp va <> Value _ ma.

Definition pre_all_unpin osi pa s :=
  (** pa is not the osi current page physical address *)
  exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
  pa<>(os_current_page system) /\
  (** The page is registered in s *)
  exists2 p2m, hypervisor_mapping osi (st_hypervisor s) p2m &
     exists2 ma, Value _ ma = map_apply padd_eq p2m pa &
        ((page_type (st_memory s) ma <> None) /\
          (no_va_mapped_to_ma osi ma s)).

Definition post_all_unpin osi pa s s' :=
  (** The page is not registered with in s' *)
  exists2 p2m, 
     hypervisor_mapping osi (st_hypervisor s) p2m &
     exists2 ma, 
       Value _ ma = map_apply padd_eq p2m pa &
       (** Delete pa from s' hypervisor mapping *)
       let newp2m := map_drop padd_eq p2m pa in 
       let newhyper := map_add os_ident_eq (st_hypervisor s) osi newp2m in
       (** Free the page in system memory *)
       let newpage := new_page None No_Owner in
       let newmem := map_add madd_eq (st_memory s) ma newpage in
         exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
            let newos := OS None (os_current_page system) in
              let newoss := map_add os_ident_eq (st_oss s) osi newos in
                s' = State (st_active_os s) newoss (st_exec_mode s)
                           (st_aos_activity s) newhyper newmem
                           (st_cache s) (st_tlb s).

Definition pre_page_unpin_untrusted osi pa ctxt s :=
     (ctxt_oss ctxt) osi = false (* untrusted osi *)
 (** If the OS is untrusted then the corresponding request is pending  *)
  /\ get_os_hypercall s osi (Some (Hyperv_call_unpin pa))
  /\ hyper_running s
  /\ pre_all_unpin osi pa s.

Definition post_page_unpin_untrusted osi pa s s' :=
  post_all_unpin osi pa s s'.

Definition pre_page_unpin_trusted pa ctxt s :=
  os_running s /\ trusted_aos ctxt s /\ pre_all_unpin (st_active_os s) pa s.

Definition post_page_unpin_trusted pa s s' := 
  post_all_unpin (st_active_os s) pa s s'.


(** *** Valid transitions *)
Section Transitions.
Variable ctxt : context.

(** Get the precondition of an action *)
Definition pre_action (act:action) : state -> Prop := 
  match act with
  | Silent => pre_silent
  | Read va => pre_read va ctxt
  | Read_hyper va => pre_read_hyper va
  | Write va val => pre_write va val ctxt
  | Write_hyper va val => pre_write_hyper va val
  | New_untrusted osi va pa => pre_new_untrusted osi va pa ctxt
  | New_trusted va pa => pre_new_trusted va pa ctxt
  | New_hyper va ma => pre_new_hyper va ma ctxt
  | Del_untrusted osi va => pre_del_untrusted osi va ctxt
  | Del_trusted va => pre_del_trusted va ctxt
  | Del_hyper va => pre_del_hyper va ctxt
  | Switch osi => pre_switch osi
  | Lswitch_trusted pa => pre_lswitch_trusted pa ctxt
  | Lswitch_untrusted osi pa => pre_lswitch_untrusted osi pa ctxt
  | Hcall c => pre_hcall ctxt
  | Ret_ctrl => pre_ret_ctrl
  | Chmod => pre_chmod
  | Page_unpin_trusted pa => pre_page_unpin_trusted pa ctxt
  | Page_pin_trusted pa pt => pre_page_pin_trusted ctxt pa
  | Page_unpin_untrusted osi pa => pre_page_unpin_untrusted osi pa ctxt
  | Page_pin_untrusted osi pa pt => pre_page_pin_untrusted osi pa pt ctxt
end.

(** The family of post-conditions of actions *)
Definition post_action (act:action): state -> state -> Prop := 
  match act with
  | Silent => post_silent
  | Read va => post_read va
  | Read_hyper va => post_read_hyper va
  | Write va val => post_write va val
  | Write_hyper va val => post_write_hyper va val
  | New_untrusted osi va pa => post_new_untrusted osi va pa 
  | New_trusted va pa => post_new_trusted va pa
  | New_hyper va ma => post_new_hyper va ma
  | Del_untrusted osi va => post_del_untrusted osi va
  | Del_trusted va => post_del_trusted va
  | Del_hyper va => post_del_hyper va
  | Switch osi => post_switch osi
  | Lswitch_trusted pa => post_lswitch_trusted pa
  | Lswitch_untrusted osi pa => post_lswitch_untrusted osi pa
  | Hcall c => post_hcall c
  | Ret_ctrl => post_ret_ctrl
  | Chmod => post_chmod ctxt
  | Page_unpin_trusted pa => post_page_unpin_trusted pa
  | Page_pin_trusted pa pt => post_page_pin_trusted pa pt
  | Page_unpin_untrusted osi pa => post_page_unpin_untrusted osi pa
  | Page_pin_untrusted osi pa pt => post_page_pin_untrusted osi pa pt
end.

(** [exec] specifies a transition from one state to another 
     executing a given action. *)

(* First alternative: only valid transitions *)
Inductive valid_exec : state ->  action -> state -> Prop :=
  Valid_exec : forall (s s':state),forall (act: action),
                pre_action act s ->
                post_action act s s' ->
                valid_exec s act s'.

(* Second alternative: all possible transitions *)
Inductive all_exec : state ->  action -> state -> Prop :=
 |Valid_pre_exec : forall (s s':state), forall (act: action),
                   pre_action act s ->
                   post_action act s s' ->
                   all_exec s act s'
 |Non_valid_exec : forall (s:state), forall (act: action),
                   ~ pre_action act s ->
                   all_exec s act s.
End Transitions.
