(***********************************************************************
 * File: State_valid.v
 * $Revision: 455 $
 * $Date: 2011-01-25 20:29:37 -0200 (Tue, 25 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This file provides mild conditions under which hypervisor calls terminate
 **********************************************************************)

Require Export State_proof.
Require Export State_effects.
Require Export Exec.
Require Import Cache_invariant.

Section Prop_hcalls.

Variable ctxt : context.

(* no hypercall is pending in a state *)
Definition noHcall (s:state) : Prop :=
  forall osi:os_ident,
    exists2 tos:os, 
      Value os_ident tos = map_apply os_ident_eq (st_oss s) osi &
      os_hcall tos = None.


(* if the hypervisor is not running, then no hypercall is pending *)
Definition osRunningNoHcall (s:state) : Prop :=
  ~(hyper_running s) -> noHcall s.


(* Auxiliar lemmas *)

Lemma osRunningNoHcall_Chmod : 
  forall (s s':state),
  noHcall s ->
  valid_exec ctxt s Chmod s' ->
  osRunningNoHcall s'.
Proof.
  intros s s'; unfold osRunningNoHcall, noHcall; intros not_hcall_s valid_exec_chmod. 
  inversion_clear valid_exec_chmod as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_chmod in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
Qed.


Lemma post_Chmod_os_running : 
             forall(s s' : state), 
             valid_exec ctxt s Chmod s' ->                       
             ~ hyper_running s'.
Proof.
  unfold not, hyper_running; intros s s' valide hyperr.
  inversion_clear valide as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold post_action, post_chmod in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((eq1, eq2), Hinterp) | (noteq, Hinterp)]; try discriminate.
  elim eq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexecm, Hosact).
  rewrite Hosact in hyperr; discriminate.
  unfold state_components_assign, state_components_update in Hassign.
  destruct Hassign with OS_Activity as [((eq1, eq2), Hinterp) | (noteq, Hinterp)]; try discriminate.
  elim eq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexecm, Hosact).
  rewrite Hosact in hyperr; discriminate.
Qed.


(* The only action that produces a hypercall is Hcall *)
Lemma noHcall_inv : 
  forall (s s':state) (a: action), 
  (forall (hc: Hyperv_call), a <> Hcall hc) ->
  noHcall s ->
  valid_state ctxt s ->
  valid_exec ctxt s a s'->
  noHcall s'.
Proof.
  unfold noHcall; intros s s' a a_not_Hcall no_hcall_s H_s_valid valide.
  induction a; inversion_clear valide as [s_vexec s'_vexec act_vexec H_pre H_post]; unfold pre_action in H_pre; unfold post_action in H_post.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_oss System_Cache Guest_OSs.
  rewrite <- H_oss; try (intro; discriminate); auto.


  (* Read *)
  unfold pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  unfold post_read, post_all_read in H_post.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  assert (forall cmp'' : state_component,
    IF cmp'' <> System_Cache /\ cmp'' <> System_TLB
    then state_component_interp s cmp'' =
      state_component_interp s' cmp''
    else state_component_interp s' System_Cache = new_cache /\
      state_component_interp s' System_TLB = new_tlb) as H_interp.
  apply H_assign.
  elim H_interp with Guest_OSs;
    try (intros [H_abs H_oss]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_oss]; 
    rewrite <- H_oss; auto.

  (* Read_hyper *)
  unfold pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  unfold post_read, post_all_read in H_post.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  assert (forall cmp'' : state_component,
    IF cmp'' <> System_Cache /\ cmp'' <> System_TLB
    then state_component_interp s cmp'' =
      state_component_interp s' cmp''
    else state_component_interp s' System_Cache = new_cache /\
      state_component_interp s' System_TLB = new_tlb) as H_interp.
  apply H_assign.
  elim H_interp with Guest_OSs;
    try (intros [H_abs H_oss]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_oss]; 
    rewrite <- H_oss; auto.
  
 (* Write *)
  unfold post_write, post_all_write in H_post.
  destruct H_post as (ma, va_ma, (newm, Hupd, (pg, Hpg,
    (new_cache, H_new_cache, (new_tlb, H_new_tlb, H_s'))))).
  rewrite H_s'; auto.

  (* Write_hyper *)
  unfold post_write_hyper, post_all_write in H_post.
  destruct H_post as (ma, va_ma, (newm, Hupd, (pg, Hpg,
    (new_cache, H_new_cache, (new_tlb, H_new_tlb, H_s'))))).
  rewrite H_s'; auto.

  (* New_untrusted *)
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (ma, hypma, (Hpt, (Hmaown, (Hacc, (Hhcall, Hyper)))))).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os', Hval, Hcall).
  destruct no_hcall_s with o as (os'', Hvalos'', hcall_os'').
  assert (os' = os'') as eqos.
  rewrite <- Hval in Hvalos''; injection  Hvalos''; auto.
  rewrite eqos in Hcall; rewrite Hcall in hcall_os''; discriminate.

  (* New_trusted *)  
  unfold post_new_trusted, post_all_new in H_post.
  destruct H_post as (ma, Hhyp, (pta, osc, (Hnewmen, Hupd, (os', Hval_os', H_s')))).
  rewrite H_s'; simpl.
  intro osi; destruct (no_hcall_s osi) as (os'', Hvalos'', Hhcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hvalos''.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.

  (* New_hyper *)
  unfold post_new_hyper, post_all_new in H_post.
  destruct H_post as (aos, Heq, (pt_addr, H_os_current, (Hnewmen, H_update, H_s'))). 
  rewrite H_s'; auto.

  (* Del_untrusted *)
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.

  (* Del_trusted *)
  unfold post_del_trusted, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr, (Hnewmen, Hdrop, H_s')).
  rewrite H_s'; auto.

  (* Del_hyper *)
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr, (Hnewmen, Hdrop, H_s')). 
  rewrite H_s'; auto.

  (* Switch *)
  unfold post_switch in H_post.
  unfold state_component_assign, state_component_update in H_post.
  rewrite H_post; auto.

  (* Lswitch_trusted *)
  unfold post_lswitch_trusted, post_all_lswitch in H_post.
  destruct H_post as (os', Hval, H_s'). 
  rewrite H_s'; simpl.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None p); auto.
  exists os''; auto.
  rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.

  (* Lswitch_untrusted *)
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (os', Hval, (Hhcall, Hrest))).
  destruct no_hcall_s with o as (os'', Hval_o, os'_not_hcall).
  rewrite <- Hval in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hhcall; discriminate.
  
  (* Hcall *)
  elim (a_not_Hcall h); auto.
  
  (* Ret_ctrl *)
  unfold post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update  in H_post.
  destruct H_post with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.

  (* Chmod *)
  unfold post_chmod in H_post.
  unfold state_components_assign, state_components_update  in H_post.
  destruct H_post as [(Htrusted, Hassign) | (Huntrusted, Hassign)].
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.
  destruct Hassign with Guest_OSs as [(Heqcomp1 , Hinterp) | (Heqcomp2, Hinterp)].
  unfold state_component_interp in Hinterp; rewrite <- Hinterp; auto.
  elim Heqcomp2; split; discriminate.

  (* Pin_trusted *)
  unfold post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as (ma, (pg, val, (pgown, Hother))).
  destruct Hother as (pgcon, (p2m, hypp2m, (os', valos', eqs'))).
  rewrite eqs'; clear eqs'; simpl.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.

  (* Unpin_trusted *)
  unfold post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as (p2m, hypp2m, (ma', valma', (os', valos', eqs'))).
  rewrite eqs'; clear eqs'; simpl.
  intro osi; destruct no_hcall_s with osi as (os'', Hval', os''_not_hcall).
  elim (os_ident_eq (st_active_os s) osi); intro eq_aos_osi.
  rewrite eq_aos_osi; rewrite lemma_map_add_apply; auto.
  exists (OS None (os_current_page os')); auto.
  exists os''. rewrite Hval'.
  symmetry; rewrite lemma_map_add_apply_distinct; auto.
  auto.
  
  (* Pin_untrusted *)
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.

  (* Unpin_untrusted *)
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, Hrest)).
  unfold get_os_hypercall in Hhcall.
  destruct Hhcall as (os, Hva, Hos_hcall).
  destruct no_hcall_s with o as (os', Hval_o, os'_not_hcall).
  rewrite <- Hva in Hval_o.
  simplify_eq Hval_o; intro eq_os'_os; rewrite eq_os'_os in os'_not_hcall.
  rewrite os'_not_hcall in Hos_hcall; discriminate.
Qed.


Lemma running: forall (s:state),
  os_running s -> ~(hyper_running s).
Proof.
  unfold hyper_running, os_running, not; intros.
  rewrite H in H0; discriminate.
Qed.


(* osRunningNoHcall is an one-step invariant, for all action a<>Chmod *) 
Lemma OS_runnig_noHypercallState_inv : 
  forall (s s':state) (a: action),
  a <> Chmod ->
  valid_state ctxt s ->
  osRunningNoHcall s ->
  valid_exec ctxt s a s' ->
  osRunningNoHcall s'.
Proof.
  intros s s' a a_not_Chmod H_s_valid.
  unfold osRunningNoHcall; intros osRunningNoHcall_s valid_exec_a .
  generalize (noHcall_inv s s' a); intro noHcall_inv_s_s'.
  induction a; inversion valid_exec_a as [s_vexec s'_vexec act_vexec H_pre H_post]; unfold pre_action in H_pre; unfold post_action in H_post.

  (* Silent *)
  intro H_hyper_running.
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold noHcall.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_TwoComps H_interp H_oss H_act System_Cache Guest_OSs OS_Activity.
  rewrite <- H_oss; try (intro; discriminate); auto.
  assert (os_running s) as os_running.
  unfold os_running, hyper_running in *.
  rewrite <- H_act in H_hyper_running; try (intro; discriminate); auto.
  induction (st_aos_activity s); auto; try contradiction H_hyper_running; auto.
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  elim noHcall_inv_s_s' with osi; auto. 
  unfold not; intros; discriminate.

  (* Read *)
  unfold pre_read in H_pre.
  destruct H_pre as (H_pre, (H_v_accessible, os_running)). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto. 
  unfold not; intros; discriminate.

  (* Read_hyper *)
  unfold pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  unfold post_read, post_all_read in H_post.
  unfold post_read, post_all_read in H_post.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with OS_Activity;
    try (intros [H_abs H_act]; contradiction H_abs;
      split; intro; discriminate).
  unfold hyper_running, state_component_interp; intros [H_triv H_act]; 
    rewrite <- H_act; auto.

  (* Write *)
  unfold pre_write in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.

  (* Write_hyper *)
  unfold post_all_write in H_post.
  elim H_post.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (va_mapped_to_ma s (st_active_os s) v ma) as H_ma_map.
  apply valid_state_included_tlb_mem with ctxt; auto.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  simpl in H_s'.
  unfold hyper_running; rewrite H_s'; auto.

  (* New_untrusted *)
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (ma, hypma, (Hpt, (Hmaown, (Hacc, (Hhcall, hyperr)))))).
  intro not_hyperr_s'.
  unfold post_new_untrusted, post_all_new in H_post.
  destruct H_post as (ma', hypma', (ptadr, oscurr, (newmen', Hupd, (os', Hval, H_s')))).
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite H_s' in not_hyperr_s'; simpl in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* New_trusted *)  
  unfold pre_new_trusted in H_pre.
  destruct H_pre as (ma, Hhyp, (pgt, (Howner, (Hacc, (Htrusted, os_running))))).
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.

  (* New_hyper *)
  unfold pre_new_hyper in H_pre.
  destruct H_pre as (hyperr, Hrest).
  intro not_hyperr_s'.
  unfold post_new_hyper, post_all_new in H_post.
  destruct H_post as (aos, Heq, (pt_addr, H_os_current, (Hnewmen, H_update, H_s'))). 
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite H_s' in not_hyperr_s'; simpl in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Del_untrusted *)
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (os', Hval', (Hp2m, Hhyp, (ptadr, Hval'', (Hpre_all, (hyperr, Hacc))))))).
  intro not_hyperr_s'.
  unfold post_del_untrusted in H_post.
  destruct H_post as (os'', Hval''', (Hp2m', Hhyp', (ptadr', Hval4, (Hnewmen, Hdrop, H_s')))).
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite H_s' in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Del_trusted *)
  unfold pre_del_trusted in H_pre.
  destruct H_pre as (ptaddr, Hcurr, (Hpre_all, (os_running, Hrest))). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.

  (* Del_hyper *)
  unfold pre_del_hyper in H_pre.
  destruct H_pre as (ptaddr, Hcurr, (Hpre_all, (os_running, Hrest))). 
  intro not_hyperr_s'.  
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as (Hpta, Hcurr', (Hnewmen, Hdrop, H_s')). 
  unfold hyper_running in os_running, not_hyperr_s'.
  rewrite H_s' in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Switch *)
  unfold pre_switch in H_pre.
  destruct H_pre as (hyperr, Hrest). 
  intro not_hyperr_s'.  
  unfold post_switch in H_post.
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite H_post in not_hyperr_s'.
  elim not_hyperr_s'; auto.

  (* Lswitch_trusted *)
  unfold pre_lswitch_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.

  (* Lswitch_untrusted *)
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (os', Hval, (Hhcall, (hyperr, Hrest)))).
  intro not_hyperr_s'.  
  unfold post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as (os'', Hval', H_s'). 
  unfold hyper_running in hyperr, not_hyperr_s'.
  rewrite H_s' in not_hyperr_s'.
  elim not_hyperr_s'; auto.
  
  (* Hcall *)
  unfold post_hcall in H_post.
  destruct H_post as (os', H_val, eqs').
  intro not_hyperr_s'.  
  unfold hyper_running in not_hyperr_s'.
  cut ((st_aos_activity s')=OSA_waiting); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | rewrite eqs'; simpl; auto].
  
  (* Ret_ctrl *)
  unfold post_ret_ctrl in H_post.
  intro not_hyperr_s'.  
  unfold state_components_assign, state_components_update in H_post.
  destruct H_post with OS_Activity as [((Heq1, Heq2), Hinterp) | (Heqcomp2, Hinterp)]; try discriminate.
  unfold state_component_interp in Hinterp.
  unfold hyper_running in not_hyperr_s'.
  elim Heq2; auto.
  unfold state_component_interp in Hinterp; destruct Hinterp as (Hexcm, Hact).
  unfold hyper_running in not_hyperr_s'.
  rewrite Hact in not_hyperr_s'; elim not_hyperr_s'; auto.

  (* Chmod *)
  elim a_not_Chmod; auto.

  (* Pin_trusted *)
  unfold pre_page_pin_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.

  (* Unpin_trusted *)
  unfold pre_page_unpin_trusted in H_pre.
  destruct H_pre as (os_running, Hrest). 
  generalize (osRunningNoHcall_s (running s os_running)); intros.
  apply noHcall_inv_s_s'; auto.
  unfold not; intros; discriminate.
  
  (* Pin_untrusted *)
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (hyperr, Hrest))).
  intro not_hyperr_s'.  
  unfold post_page_pin_untrusted, post_all_pin in H_post.
  destruct H_post as (ma, (pg, Hval, (Hown, (Hpage, (Hp2m, Hhyp, (os', Hval', eqs')))))). 
  unfold hyper_running in hyperr, not_hyperr_s'.
  cut ((st_aos_activity s')=(st_aos_activity s)); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | auto].
  rewrite eqs'; simpl; auto.

  (* Unpin_untrusted *)
  unfold post_page_unpin_untrusted, post_all_unpin in H_post.
  destruct H_post as (Hp2m, Hhyp, (ma', Hval, (os', Hval', eqs'))). 
  intro not_hyperr_s'.  
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as (Huntrusted, (Hhcall, (hyperr, Hrest))).
  unfold hyper_running in hyperr, not_hyperr_s'.
  cut ((st_aos_activity s')=(st_aos_activity s)); 
  [intro st_aos_activity_s'; rewrite st_aos_activity_s' in not_hyperr_s'; elim not_hyperr_s'; auto | auto].
  rewrite eqs'; simpl; auto.
Qed.

End Prop_hcalls.


Require Export Traces.
Require Export Exec.
Require Export Streams.

Section Prop_hcall_traces.

Variable ctxt : context.

(* A  co-inductive definition of trace where the Chmod action requires no hcall is pending *) 
CoInductive trace_imp : state_stream-> Prop :=
  Trace_imp_chmod : forall (s:exec_state)(str:state_stream),
           trace_imp str ->
           (ex_action s) = Chmod ->
           (* no hcall is pending *)  
	   noHcall (ex_state s) ->
	   system_transition ctxt s (hd str) ->
           trace_imp (Cons s str)
  | Trace_imp_not_chmod : forall (s:exec_state) (str:state_stream),
           trace_imp str ->
           (ex_action s) <> Chmod ->
	   system_transition ctxt s (hd str) ->
           trace_imp (Cons s str).


(* trace_imp <= trace *) 
Theorem incl_trace_imp_in_trace : 
  forall s: state_stream, trace_imp s -> trace ctxt s.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 trace_imp_sstr0.
  destruct str0.
  inversion trace_imp_sstr0; constructor; auto.
Qed.


(* Condition about an initial state of an implementation of the system (for trace_imp) *)
Definition init_state_imp : exec_state -> Prop :=
  fun execst => osRunningNoHcall (ex_state execst) /\ valid_state ctxt (ex_state execst).

(* run_imp: a trace starting from an initial state without hypercalls, if an OS is running *)             
Definition run_imp (str:state_stream): Prop := 
  init_state_imp (hd str) /\ trace_imp str.


(* Auxiliar definitions *)

(* the hypervisor is not running in (hd str)*)
Definition state_stream_imp_not_hyper_running (str:state_stream) := 
  ~ hyper_running (ex_state (hd str)).

Definition exist_state_stream_imp_not_hyper_running (str:state_stream) := 
  Exists state_stream_imp_not_hyper_running str.

(* if the hypervisor is not running, then no hypercall is pending in (hd str) *)
Definition state_stream_imp_osRunningNoHcall (str:state_stream) := 
  osRunningNoHcall (ex_state (hd str)).

(* no hypercall is pending in (hd str) *)
Definition state_stream_imp_noHcall (str:state_stream) := 
  noHcall (ex_state (hd str)).

Definition exist_state_stream_imp_noHcall (str:state_stream) := 
  Exists state_stream_imp_noHcall str.


(* osRunningNoHcall is an invariant for trace_imp *)
Lemma osRunningNoHcall_inv : 
             forall str, run_imp str -> 
             ForAll state_stream_imp_osRunningNoHcall str.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 run_str0.
  destruct str0.
  inversion run_str0 as [Hinit Htrace_cons].
  simpl in Hinit.
  unfold init_state_imp in Hinit.
  destruct Hinit as [H_init H_valid].
  inversion Htrace_cons as [es st Htrace_tl Hact noHcall_hd Hvalid_exec | es st Htrace_tl Hact Hvalid_exec].
  (* case: Chmod action *)
  unfold system_transition in Hvalid_exec.
  rewrite Hact in Hvalid_exec.
  (* using auxiliar lemma osRunningNoHcall_Chmod *)
  generalize (osRunningNoHcall_Chmod ctxt (ex_state e)(ex_state (hd str0)) noHcall_hd Hvalid_exec).
  intro H_osRunningNoHcall_Chmod.
  constructor; auto.
  unfold state_stream_imp_osRunningNoHcall; simpl; auto.
  (* apply coinduction hypothesis *)
  simpl; apply str.
  unfold run_imp, init_state_imp; split.
  split; auto.
  elim run_inv with ctxt (Cons e str0); simpl; auto.
  constructor; unfold init_state; auto.
  apply incl_trace_imp_in_trace; auto.
  auto.

  (* case: action <> Chomd *)
  unfold system_transition in Hvalid_exec.
  (* using auxiliar lemma OS_runnig_noHypercallState_inv *)
  generalize (OS_runnig_noHypercallState_inv ctxt (ex_state e) (ex_state (hd str0)) (ex_action e) Hact H_valid H_init Hvalid_exec).
  intro H_osRunningNoHcall_hds.
  constructor.
  unfold state_stream_imp_osRunningNoHcall; simpl.
  assumption.
  (* apply coinduction hypothesis *)
  simpl; apply str.
  unfold run_imp, init_state_imp; split.
  split; auto.
  elim run_inv with ctxt (Cons e str0); simpl; auto.
  constructor; unfold init_state; auto.
  apply incl_trace_imp_in_trace; auto.
  auto.
Qed.
  
(* If every execution of the hypervisor is finite, then the hypervisor calls terminate (every hypercall is executed) *)
Theorem hcalls_are_executed : 
             forall str, run_imp str -> 
             ForAll (fun str' => exist_state_stream_imp_not_hyper_running str') str ->
             ForAll (fun str' => exist_state_stream_imp_noHcall str') str.
Proof.
(* proof by coinduction *)
  cofix str.
  intros str0 run_str0 H_str0.
  destruct str0.
  constructor.
  (* proof for the first state *)
  inversion run_str0 as [Hinit Htrace_cons].
  (* using auxiliar lemma osRunningNoHcall_inv *)
  generalize (osRunningNoHcall_inv (Cons e str0) run_str0); intro H_osRunningNoHcall_inv.
  inversion H_str0 as [Hexists_hd Hexists_tl].
  (* induction in exist_state_stream_imp_not_hyper_running *)
  induction Hexists_hd as [str1 | str1 Exists].
  destruct str1.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_e0 H_osRunningNoHcall_tl].
  unfold state_stream_imp_osRunningNoHcall, osRunningNoHcall  in H_osRunningNoHcall_e0.
  apply Here; unfold state_stream_imp_noHcall; auto.
  (* inductive case for exist_state_stream_imp_not_hyper_running *)
  apply Further.
  apply IHExists; auto. 
  inversion run_str0 as [H_init_hd_str1]; auto.
  unfold run_imp; split.
  unfold init_state_imp.
  split.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_hd H_osRunningNoHcall_tl_str1]. 
  inversion H_osRunningNoHcall_tl_str1 as [H_osRunningNoHcall_hd_tl H_osRunningNoHcall_tl_tl_str1].
  unfold state_stream_imp_osRunningNoHcall in H_osRunningNoHcall_hd_tl; auto.
  unfold init_state_imp in H_init_hd_str1.
  destruct H_init_hd_str1 as [H_init_hd_str1 H_init_hd_valid].
  elim run_inv with ctxt str1; simpl; auto.
  constructor; unfold init_state; auto.
  apply incl_trace_imp_in_trace; auto.
  inversion H; auto.
  unfold init_state_imp.
  split.
  inversion H_osRunningNoHcall_inv as [H_osRunningNoHcall_str1 H_osRunningNoHcall_tl_str1].
  inversion H_osRunningNoHcall_tl_str1; auto.
  unfold init_state_imp in Hinit.
  destruct Hinit as [Hinit H_init_valid].
  elim run_inv with ctxt str1; simpl; auto.
  constructor; unfold init_state; auto.
  apply incl_trace_imp_in_trace; auto.
  inversion Htrace_cons; auto.
  inversion H_osRunningNoHcall_inv; auto.
  inversion Hexists_tl; auto.
  (* apply coinduction hypothesis *)
  simpl; apply str; auto.
  unfold run_imp in  run_str0; simpl; elim run_str0; intros.
  unfold run_imp; split.
  unfold init_state_imp; split.
  (* using auxiliar lemma osRunningNoHcall_inv *)
  generalize (osRunningNoHcall_inv (Cons e str0) run_str0); intro osRunningNoHcall_cons_str0.
  inversion osRunningNoHcall_cons_str0 as [H_osRunningNoHcall_hd H_osRunningNoHcall_tl].
  inversion H_osRunningNoHcall_tl as [H_osRunningNoHcall_hd_tl H_osRunningNoHcall_tl_tl].
  simpl in H_osRunningNoHcall_tl_tl; auto.
  elim run_inv with ctxt (Cons e str0); simpl; auto.
  constructor; unfold init_state; simpl.
  unfold init_state_imp in H; simpl in H.
  destruct H as [H H_valid]; auto.
  apply incl_trace_imp_in_trace; auto.
  inversion H0; auto.
  inversion H_str0; auto.
Qed.


End Prop_hcall_traces.