(***********************************************************************
 * File: Leak_traces.v
 * $Revision: 408 $
 * $Date: 2011-01-03 13:23:41 -0200 (lun 03 de ene de 2011) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * Trace definition with leakage.
 **********************************************************************)

Require Import List.
Require Export Streams.
Require Export State_proof.
Require Export State_effects.
Require Export Actions.
Require Export Exec.
Require Import Exec_invariant.

Set Implicit Arguments.

Section Leak_Traces.

Variable ctxt : context.
Variable observation : Set.
Variable empty_observation : observation -> Prop.

Definition leak_action : Set := observation -> action.
Definition leak_function : Set := state -> state -> action -> observation.
Record leak_state : Set :=
  Leak_State {
    leak_state_state:state;
    leak_state_obs:observation
  }.

Definition leak_valid_state (s:leak_state) : Prop :=
  valid_state ctxt (leak_state_state s).


Inductive leak_valid_exec : leak_state -> leak_action -> leak_function -> leak_state -> Prop :=
Leak_valid_exec : forall (ls ls':leak_state) (f:leak_function) (A:leak_action),
  let s := (leak_state_state ls) in
  let s' := (leak_state_state ls') in
  let o := (leak_state_obs ls) in
  let o' := (leak_state_obs ls') in
  let a := A o in
    valid_exec ctxt s a s' ->
    f s s' a = o' ->
    leak_valid_exec ls A f ls'.


Record leak_exec_state : Set :=
  Leak_Ex_State
  {l_ex_state : leak_state;
   l_ex_action  : leak_action;
   l_ex_function : leak_function
  }.

(* A valid initial state for the trace has a valid state and an empty
   observation *)
(* JD: Revisar esto. *)
Definition leak_init_state (execst:leak_exec_state) : Prop :=
  let s := l_ex_state execst in
  let o := leak_state_obs s in
  leak_valid_state s /\ empty_observation o.

Definition leak_system_transition (src tgt : leak_exec_state) : Prop :=
   let s := l_ex_state src in
   let A := l_ex_action src in
   let f := l_ex_function src in
   let s' := l_ex_state tgt in
   leak_valid_exec s A f s'.

Definition leak_state_prop := leak_exec_state -> Prop.

(* Invariants *)

Definition leak_leads_to 
  (P Q : leak_state_prop) : Prop := 
   forall s s': leak_exec_state,
    P s ->  
    leak_system_transition s s' ->  
    Q s'.

Definition leak_invariant: leak_state_prop -> Prop := fun P => leak_leads_to P P.

(* Streams of states, formulas and temporal operators *)
Definition leak_state_stream := (Stream leak_exec_state).
Definition leak_stream_prop := leak_state_stream->Prop.
Definition leak_binary_state_prop := leak_state -> leak_state -> Prop.
Definition leak_lift_unary (P:leak_state_prop) : leak_stream_prop := fun str => P (hd str).
Definition leak_lift_binary (P:leak_binary_state_prop) : leak_stream_prop := 
    fun str => P (l_ex_state (hd str)) (l_ex_state (hd (tl str))).

(****** Traces and runs with leakage *********)

(* A trace is a stream of execution states (si,Ai,fi) for i=1.., s.t. si -Ai,fi-> s(i+1),
   where -> is leak_valid_exec  *)

(* A  co-inductive fdefinition of trace *) 
CoInductive leak_trace : leak_state_stream-> Prop :=
  Leak_trace : forall (s:leak_exec_state)(str:leak_state_stream),
           leak_trace str ->
	   leak_system_transition s (hd str) ->
           leak_trace (Cons s str).


(* leak_run: a leak_trace starting from an initial state *)             
Definition leak_run (str:leak_state_stream): Prop := 
  leak_init_state (hd str) /\ leak_trace str.

(* The following lemmas shall be extensively used when proving invariants 
   (safety properties) of platform runs *)

(* Runs and valid state *)

Definition leak_valid_state_stream (str:leak_state_stream) := 
  leak_valid_state (l_ex_state (hd str)).

Lemma state_of_run_is_valid: 
  forall str, leak_run str -> leak_valid_state_stream str.
Proof.
  intros str; case str.
  intros e s H_run; unfold leak_valid_state_stream; simpl.
  inversion H_run.
  unfold leak_init_state in H; simpl in H.
  destruct H.
  assumption.
Qed.


(***********************)
(******  Safety *********)
(***********************)

(* A Property is safe for a run of the system if it holds for all
   state of the execution trace *)

Definition leak_safe (P:leak_state -> leak_state -> Prop) : Prop := 
  forall str:leak_state_stream, leak_run str -> ForAll (leak_lift_binary P) str.

End Leak_Traces. 

