(***********************************************************************
 * File: Maps_proof.v
 * $Revision: 408 $
 * $Date: 2011-01-03 13:23:41 -0200 (Mon, 03 Jan 2011) $
 * $LastChangedBy: jdcampo $
 * ********************************************************************
 * 
 ************************************************************************)

Require Export Maps.

Set Implicit Arguments.

(* This module describes properties (and eventually proofs) of mappings *)

Section Mapping_Properties.

(* The type of index for accessing the map *)

Variable index : Set.
Variable index_eq : forall x y : index, {x = y} + {x <> y}.

(* The type of information associated to each index *)
Variable info : Set.

Definition injective_mapping 
  (mp:mapping index info) : Prop :=
  forall id1,
   forall id2,
      map_apply index_eq mp id1 = map_apply index_eq mp id2 -> id1 = id2.  


(* Obtaining an element in index I, after having added VAL in I, gives
   VAL as result *)
Lemma lemma_map_add_apply :
  forall (map:mapping index info) (i:index) (val:info),
    map_apply index_eq (map_add index_eq map i val) i = Value index val.
Proof.
  intros map i val.
  induction map.
  simpl.
  elim (index_eq i i).
  intro; reflexivity.
  intro H_i_ne_i; cut False; (contradiction || apply H_i_ne_i; reflexivity).

  simpl.
  case (index_eq (item_index a) i).
  intro H_aindex_eq_i.
  simpl.
  elim (index_eq i i).
  intro; reflexivity.

  intro H_i_ne_i; cut False; (contradiction || apply H_i_ne_i; reflexivity).

  intro H_aindex_ne_i.
  simpl.
  elim (index_eq i (item_index a)).
  intro H_abs; cut False.
  intro; contradiction.
  apply H_aindex_ne_i; symmetry; assumption.


  intro; assumption.
Qed.

(* Obtaining an element in index I, after having added VAL in I', does
   not change the result. *)
Lemma lemma_map_add_apply_distinct :
  forall (map: mapping index info) (i i':index) (val:info),
    i <> i' -> 
    map_apply index_eq (map_add index_eq map i val) i' = 
    map_apply index_eq map i'.
Proof.
  intros map i i' val H_i_ne_i'.
  induction map.

  simpl.
  elim (index_eq i' i).
  intro H_abs; cut False; (contradiction || auto).
  auto.

  simpl.
  case (index_eq (item_index a) i).
  intro H_aindex_eq_i.
  simpl.
  elim (index_eq i' i).
  intro H_asb; cut False; (contradiction || auto).
  intro H; clear H.
  
  case (index_eq i' (item_index a)).
  intros H_abs.
  rewrite H_aindex_eq_i in H_abs.
  symmetry in H_abs; cut False; (contradiction || auto).
  intro H_i'_ne_aindex.
  reflexivity.

  intro H_aindex_ne_i.
  simpl.
  case (index_eq i' (item_index a)).
  intro H_i'_eq_aindex.
  reflexivity.

  intro H_i'_ne_aindex.
  assumption.
Qed.

Lemma map_drop_apply : forall (map:mapping index info) (i:index),
    map_apply index_eq (map_drop index_eq map i) i = Error info i.
Proof.
  intros map i.
  induction map.
  trivial.
  simpl.

  destruct (index_eq i (item_index a));auto.
  simpl.
  destruct (index_eq i (item_index a)). 
  contradiction.
  assumption.
Qed.

Lemma map_drop_apply_distinct : forall (map:mapping index info) (i i':index),
  i <> i' -> map_apply index_eq (map_drop index_eq map i) i' = map_apply index_eq map i'.
Proof.
  intros map i i' H_i_ne_i'.
  induction map.

  trivial.

  simpl map_drop.
  case (index_eq i (item_index a)).
  intro H_i_index_a.
  destruct (index_eq (item_index a) i').
  rewrite <- H_i_index_a in e.
  contradiction.
  rewrite IHmap.
  simpl.
  destruct (index_eq i' (item_index a)).
  cut False; (contradiction || auto).

  trivial.
  
  intro H_i_ne_index_a.
  simpl.
  destruct (index_eq i' (item_index a)).
  trivial.
  assumption.
Qed.

Lemma map_error_index_inv : forall (map:mapping index info) (i i':index),
  i<>i'->  map_apply index_eq map i <> Error info i'.
Proof.
  intros map i i' H_i_ne_i'.

  induction map.

  intro H_abs;inversion H_abs;auto.

  unfold map_apply.
  case (index_eq i (item_index a)).
  intro H_i_a.
  intro H_abs;inversion H_abs.

  intro H_no_i_a.
  assumption.
Qed.

Lemma map_add_valid_size : forall (map:mapping index info) (i:index) (val:info),
  map_valid_index_fix index_eq map i = true ->
  map_size (map_add index_eq map i val) = map_size map.
Proof.
  intros map i val.
  induction map.
  simpl; intro H_abs; discriminate.
  intros H_i_valid.
  unfold map_valid_index_fix in H_i_valid.
  unfold map_add.
  induction (index_eq (item_index a) i).
  simpl; auto.
  simpl.
  unfold map_add in IHmap.
  rewrite IHmap; auto.
Qed.
  
Lemma map_add_invalid_size : forall (map:mapping index info) (i:index) (val:info),
  map_valid_index_fix index_eq map i = false ->
  map_size (map_add index_eq map i val) = (map_size map) + 1.
Proof.
  intros map i val.
  induction map.
  simpl; auto.
  intros H_i_invalid.
  unfold map_valid_index_fix in H_i_invalid.
  unfold map_add.
  induction (index_eq (item_index a) i).
  discriminate H_i_invalid.
  simpl; auto.
Qed.

Lemma map_drop_invalid_size : forall (map:mapping index info) (i:index),
  map_valid_index_fix index_eq map i = false ->
  map_size (map_drop index_eq map i) = (map_size map).
Proof.
  intros map i.
  induction map.
  simpl; auto.
  intros H_i_invalid.
  unfold map_valid_index_fix in H_i_invalid.
  unfold map_drop.
  induction (index_eq (item_index a) i).
  discriminate H_i_invalid.
  unfold map_valid_index_fix in IHmap.
  induction (index_eq i (item_index a)).
  absurd (i=(item_index a)); auto.
  simpl.
  unfold map_drop in IHmap.
  rewrite <- IHmap; auto.
Qed.
  
Lemma map_drop_valid_size : forall (map:mapping index info) (i:index),
  map_valid_index_fix index_eq map i = true ->
  map_size (map_drop index_eq map i) < map_size map.
Proof.
  intros map i.
  induction map.
  simpl; intro H_abs; discriminate.
  intros H_i_valid.
  assert (exists b, b = map_valid_index_fix index_eq map i) as [b H_i_valid_map].
  exists (map_valid_index_fix index_eq map i); auto.
  rewrite <- H_i_valid_map in IHmap.
  induction b.  
  unfold map_valid_index_fix in H_i_valid.
  unfold map_drop; simpl.
  induction (index_eq i (item_index a)).
  induction (index_eq (item_index a) i).
  elim IHmap; auto.
  absurd (i=item_index a); auto.
  simpl.
  apply Lt.lt_n_S; auto.
  
  simpl.
  unfold map_valid_index_fix in H_i_valid.
  induction (index_eq i (item_index a)).
  induction (index_eq (item_index a) i).
  rewrite map_drop_invalid_size; auto.
  absurd (i = item_index a); auto.
  simpl.
  induction (index_eq (item_index a) i).
  absurd (i <> item_index a); auto.
  unfold map_valid_index_fix in H_i_valid_map.
  rewrite H_i_valid in H_i_valid_map; discriminate H_i_valid_map.
Qed.

Lemma map_valid_fix_valid : forall (map:mapping index info) (i:index),
  map_valid_index_fix index_eq map i = true <-> map_valid_index index_eq map i.
Proof.
  intros map i.
  split.
  unfold map_valid_index_fix, map_valid_index, map_apply.
  induction map.
  intro H_abs; discriminate H_abs.
  induction (index_eq i (item_index a)).
  rewrite a0.
  induction (index_eq (item_index a) (item_index a)).
  intro H_triv.
  exists (item_info a); auto.
  contradiction b; auto.
  case (index_eq (item_index a) i); auto.
  
  unfold map_valid_index_fix, map_valid_index, map_apply.
  induction map.
  intros [it H_error]; discriminate H_error.
  induction (index_eq i (item_index a)).
  rewrite a0.
  induction (index_eq (item_index a) (item_index a)); auto.
  case (index_eq (item_index a) i); auto.  
Qed.

End Mapping_Properties.
