(***********************************************************************
 * File: Read_Isolation.v
 * $Revision: 399 $
 * $Date: 2010-12-26 17:56:13 -0200 (Sun, 26 Dec 2010) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * This files proves read isolation property on valid executions. 
 **********************************************************************)

Require Import State_valid.
Require Import Env.
Require Import Exec.
Require Import Exec_invariant.
Require Export Observers.

Section Read_Isolation.

Variable ctxt : context.

(** If a read action is executed in a valid state, 
    then the vadd being read has a madd ma associated 
    in the active memory mapping and ma belongs to the Active OS 
*)

Definition read_isolation (s s':state) : Prop :=
  forall (va:vadd), 
         valid_state ctxt s -> 
         valid_exec ctxt s (Read va) s' ->

         (* exists pg, Some pg = va_mapped_to_pg_cache s va /\  *)
 	       (exists2 ma:madd, va_mapped_to_ma s (st_active_os s) va ma &
		        exists2 pg, Value _ pg = map_apply madd_eq (st_memory s) ma &
            
		        Os_PO (st_active_os s) = page_owned_by pg).

Theorem read_isolation_proof : forall (s s':state), read_isolation s s'.
Proof.
intros s s' va.
Intros_Valid_Hyp.
intro valid_read.
inversion_clear valid_read.
inversion_clear H; clear H0.
unfold pre_all_read in H1. 
elim H1; clear H1; intros ma (va_ma, ma_ptype).
unfold valid_virtual_mapping in H_valid_virtmap.
exists ma; trivial.
unfold va_mapped_to_ma in va_ma.
elim va_ma; clear va_ma; intros ma' current_vadd_ma' mp_va_ma.
elim mp_va_ma; clear mp_va_ma; intros mp_va_ma vadd_map_ma'_mp_va_ma ma_mp_va_ma.
unfold vadd_mapping in vadd_map_ma'_mp_va_ma.
destruct vadd_map_ma'_mp_va_ma as (pg, H_pg_map, H_pg_content).
elim H_valid_virtmap with (osi:=(st_active_os s)) (pt_addr:=ma') (mp:=mp_va_ma) (pt:=pg) (va:=va) (ma:=ma);
  clear H_valid_virtmap; intros; trivial. 
exists x; trivial.
symmetry; elim H0; intros.
elim H2; auto.
unfold os_current_vadd_mapping_madd in current_vadd_ma'.
destruct current_vadd_ma' as (aos, H_aos_map, (p2m, H_p2m_map, H_ma'_map)).
unfold valid_hypervisor in H_valid_hyper.
destruct H_valid_hyper with (st_active_os s) (os_current_page aos) p2m.
unfold hypervisor_mapping in H_p2m_map; assumption.

destruct H as (ma'', H_ma''_map, H_ma''_owner).
assert (ma' = ma'').
rewrite <- H_ma''_map in H_ma'_map.
injection H_ma'_map; auto.
subst ma''.
unfold maddress_owner in H_ma''_owner.
rewrite <- H_pg_map in H_ma''_owner.
assumption.

cut False; [intro; contradiction | rewrite H in H_ma'_map].
discriminate H_ma'_map.
Qed.

Definition read_isolation_cache (s s':state) : Prop :=
  forall (va:vadd), 
    valid_state ctxt s -> 
    valid_exec ctxt s (Read va) s' ->
    exists pg, Some pg = va_mapped_to_pg_cache s va /\
		  Os_PO (st_active_os s) = page_owned_by pg.

Lemma lemma_va_mapped_cache : forall (s:state) (va:vadd) (ma:madd) (pg:page),
  valid_state ctxt s ->
  va_mapped_to_ma s (st_active_os s) va ma ->
  Value _ pg = map_apply madd_eq (st_memory s) ma ->
  Some pg = va_mapped_to_pg_cache s va.
Proof.
  intros s va ma pg H_s_valid H_va_map H_pg_map.
  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va); auto.
  exists (map_apply vadd_eq (st_cache s) va); auto.
  destruct H as [it H_it_map].
  induction it; rewrite <- H_it_map.

  (* va is in the cache *)
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  unfold valid_cache, cache_included in H_valid_cache.
  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  destruct H_cache_included with va v as [ma' [H_va_map' [H_ptype H_pg_map']]]; auto.
  assert (ma=ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'; clear H_va_map'.
  rewrite <- H_pg_map in H_pg_map'; injection H_pg_map'; intros;
    subst v; auto.

  (* va is not in the cache *)
  clear H_it_map e.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va) as [it H_it_map].
  exists (map_apply vadd_eq (st_tlb s) va); auto.
  unfold va_mapped_to_ma_cache.
  induction it; rewrite <- H_it_map.
  (* va is in the TLB *)
  rename v into ma'.
  assert (ma = ma').
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  unfold valid_tlb, tlb_included in H_valid_tlb.
  destruct H_valid_tlb as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  destruct H_tlb_included with va ma' as [H_va_map']; auto.
  apply va_mapped_same_ma with s va; auto.
  subst ma'.
  rewrite <- H_pg_map; auto.

  (* va is not in the TLB *)
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_va_map.
  destruct H_va_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  rewrite <- H_aos_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- H_ma_map.
  rewrite <- H_pg_map; auto.
Qed.
  
  
Theorem read_isolation_cache_proof : forall (s s':state), read_isolation_cache s s'.
Proof.
  unfold read_isolation_cache.
  intros s s' va H_s_valid H_valid_exec.
  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].
  unfold pre_action, pre_all_read, pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_va_map H_ptype]] [H_accessible H_osi_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as H_pg_map.
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma).
  exists v; auto.
  discriminate.
  destruct H_pg_map as [pg H_pg_map].
  exists pg; split.
  apply lemma_va_mapped_cache with ma; auto.
  generalize H_s_valid.
  Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold va_mapped_to_ma in H_va_map.
  destruct H_va_map as [pt_addr H_pt_addr_map [mp H_mp_map H_ma_map]].
  unfold valid_virtual_mapping in H_valid_virtmap.
  unfold valid_current_page in H_valid_currpg.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as [pt_addr' H_pt_addr'_map [H_pt_type H_pt_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map H_valid_currpg.
  unfold maddress_owner in H_pt_owner.
  assert (exists pt_page, Value _ pt_page = map_apply madd_eq (st_memory s) pt_addr) as H_pt_page_map.
  induction (map_apply madd_eq (st_memory s) pt_addr).
  exists v; auto.
  discriminate H_pt_owner.
  destruct H_pt_page_map as [pt_page H_pt_page_map].
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma
    as [pg' H_pg'_map [H_owner_acc H_owner_nacc]]; auto.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as [pt_page' H_pt_page'_map H_mp_map].
  rewrite <- H_pt_page_map in H_pt_page'_map; injection H_pt_page'_map;
    intros; subst pt_page'; clear H_pt_page'_map; auto.
  rewrite <- H_pt_page_map in H_pt_owner; auto.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; intros; subst pg'; clear H_pg'_map.
  symmetry; apply H_owner_acc; auto.
Qed.
  

End Read_Isolation.