(***********************************************************************
 * File: Env.v
 * $Revision: 364 $
 * $Date: 2010-11-04 14:16:55 -0200 (Thu, 04 Nov 2010) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This module provides a general mechanism for specifying the update
 * of states.
 ************************************************************************)

Require Import State.
Set Implicit Arguments.

Section State_Component_Update.

(************************************************************************)
(* The following definitions are used to have encapsulated modification 
   operations for the components of the state                           *)
(************************************************************************)

(********************************************************************)
(*                    The Universe of State components              *)
(********************************************************************)

(* Name of the components of the State. *)
Inductive state_component : Set :=
  | Active_OS : state_component
  | Guest_OSs : state_component
  | Exec_Mode  : state_component
  | OS_Activity : state_component
  | Hypervisor_SC : state_component
  | System_Memory : state_component
  | System_Cache : state_component
  | System_TLB : state_component.

(* The type of each component of the state. *)
Definition state_component_domain (cmp : state_component) : Set :=
  match cmp with
  | Active_OS => os_ident
  | Guest_OSs => oss_map
  | Exec_Mode => exec_mode
  | OS_Activity => os_activity 
  | Hypervisor_SC => hypervisor
  | System_Memory => system_memory
  | System_Cache => cache
  | System_TLB => tlb  
  end.

(* The projection of each component of the state. *)
Definition state_component_interp 
  (st : state)(cmp : state_component) : state_component_domain cmp :=
  match cmp as x return (state_component_domain x) with
  | Active_OS => st_active_os st
  | Guest_OSs => st_oss st
  | Exec_Mode => st_exec_mode st
  | OS_Activity => st_aos_activity st 
  | Hypervisor_SC => st_hypervisor st
  | System_Memory => st_memory st
  | System_Cache => st_cache st
  | System_TLB => st_tlb st
  end.
  
(***********************************************************************)
(*                      Modification of a State component              *)
(***********************************************************************)

Section One_Component.

(* The component to be modified. *)
Variable cmp : state_component.

Section Component_Update.
(* A relation specifying the transformation of component cmp. *)
Variable R : state_component_domain cmp -> state_component_domain cmp -> Prop.

(* All the other components remain unchanged except a given one. *)
Definition state_component_update (inps outs : state) : Prop :=
  forall cmp' : state_component,
  IF cmp' = cmp
  then R (state_component_interp inps cmp) 
         (state_component_interp outs cmp)
  else state_component_interp inps cmp' = state_component_interp outs cmp'
       :> state_component_domain cmp'.

End Component_Update.

Section Component_Assign.
(* A component of the state is reassigned. *)

(* The new value of the component. *)
Variable value : state_component_domain cmp.

(* All the other components remain unchanged except a given one, which
   takes a given value. *)
Definition state_component_assign (inps outs : state) : Prop :=
  state_component_update (fun _ nval => nval = value) inps outs.

End Component_Assign.

End One_Component.

(***********************************************************************)
(* Shorthands for some frequently used predicates. *)

Definition modify_active_os := state_component_update Active_OS.
Definition modify_guest_OSs := state_component_update Guest_OSs.
Definition modify_exec_mode := state_component_update Exec_Mode.
Definition modify_os_activity := state_component_update OS_Activity.
Definition modify_hypervisor := state_component_update Hypervisor_SC.
Definition modify_system_memory := state_component_update System_Memory.

Definition assign_system_memory := state_component_assign System_Memory.


Section Two_Components.

(* The components to be modified. *)
Variable cmp cmp' : state_component.

Section Two_Components_Update.
(* The relations specifying the transformation of componenst cmp and cmp'. *)
Variable R : state_component_domain cmp -> state_component_domain cmp -> Prop.
Variable R': state_component_domain cmp' -> state_component_domain cmp' -> Prop.

(* All the components remain unchanged except the two given ones. *)
Definition state_components_update (inps outs : state) : Prop :=
  forall cmp'' : state_component,
  IF cmp'' <> cmp /\ cmp'' <> cmp'
  then state_component_interp inps cmp'' = 
       state_component_interp outs cmp'' (*:>state_component_domain cmp''*)
  else 
    R  (state_component_interp inps cmp) 
       (state_component_interp outs cmp) /\
    R' (state_component_interp inps cmp') 
       (state_component_interp outs cmp').

End Two_Components_Update.

Section Two_Components_Assign.
(* The two components of the state are reassigned. *)

(* The new values of the components. *)
Variable value : state_component_domain cmp.
Variable value' : state_component_domain cmp'.

(* All the other components remain unchanged except a given one, which
   takes a given value. *)
Definition state_components_assign (inps outs: state) : Prop :=
  state_components_update
     (fun _ nval => nval = value) 
     (fun _ nval => nval = value') 
     inps outs.

End Two_Components_Assign.

End Two_Components.

End State_Component_Update.


