(***********************************************************************
 * File: Execution_isolation.v
 * $Revision: 404 $
 * $Date: 2010-12-30 23:08:20 -0200 (Thu, 30 Dec 2010) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * Definition of equivalence between states (for the AOS and a given
   osi). Proofs of equivalence properties and hypervisor behaviour
   related to equivalence.
 **********************************************************************) 

Section State_equivalence.

Require Import Actions.
Require Import Exec.
Require Import Exec_invariant.

Variable ctxt : context. 

(**
   Get the page mapped to a padd for a given osi
*)
Definition padd_page (s:state) (osi:os_ident) (pa:padd) (pg:page) : Prop :=
    exists2 ma, hypervisor_mapping_madd (st_hypervisor s) osi pa ma &
      Value _ pg = map_apply madd_eq (st_memory s) ma.

(**
   Get the page mapped to a vadd for a given PT page.
*)
Definition vadd_page (s:state) (pt_page:page) (va:vadd) (pg:page) : Prop :=
    exists2 mp, PT_c mp = (page_content pt_page) &
      exists2 ma, Value _ ma = map_apply vadd_eq mp va &
        Value _ pg = map_apply madd_eq (st_memory s) ma.


(* osi equivalence *)

(* Two states s1 and s2 have equivalent hypervisor mappings for osi,
   if for every pa that is mapped to a RW page, that page is the same
   in both states. *)
Definition osi_hyper_mapping_eq (s1 s2:state) (osi:os_ident): Prop :=
  forall pa pg, padd_page s1 osi pa pg ->
      (exists val, (page_content pg) = RW_c val) ->
      padd_page s2 osi pa pg.

Definition osi_equivalent_hyper_mapping (s1 s2:state) (osi:os_ident): Prop :=
  (osi_hyper_mapping_eq s1 s2 osi) /\ (osi_hyper_mapping_eq s2 s1 osi).

(* Two states s1 and s2 have equivalent PT pages for osi, if for
   all PT of osi, if it maps a va to an arbitrary ma with a RW
   page, the content is the same in the other state. Also, there is a
   pa that maps to that ma. *)
Definition osi_pt_pages_eq (s1 s2:state) (osi:os_ident) : Prop :=
  forall pt_padd pt_page1 mp1,
    (* pt_padd maps to pt_page1, which is a PT page with map mp1. *)
    padd_page s1 osi pt_padd pt_page1 ->
    PT_c mp1 = (page_content pt_page1) ->
      (* Then, for s1 and s2 to be active_equivalent, pt_padd must map
         to a pt_page2, of type PT *)
      (exists2 pt_page2, padd_page s2 osi pt_padd pt_page2 &
        exists2 mp2, PT_c mp2 = (page_content pt_page2) &
        (forall va pa ma1 pg,
          (* If in mp1, va maps to ma1 *)
          (Value _ ma1 = map_apply vadd_eq mp1 va ->
          (* ma maps to pg *)
          Value _ pg = map_apply madd_eq (st_memory s1) ma1 ->
          (* va is os_accessible and maps to a RW page *)
          os_accessible ctxt va -> page_type (st_memory s1) ma1 = Some RW ->
          (* and pa is _the_ physical address that maps to ma1 *)
          hypervisor_mapping_madd (st_hypervisor s1) osi pa ma1 ->
            (* Then, in mp2, va must map to some ma2 *)
            exists2 ma2, Value _ ma2 = map_apply vadd_eq mp2 va &
              (* pa must map to ma2 *)
              (hypervisor_mapping_madd (st_hypervisor s2) osi pa ma2) /\
              (* And the page in ma2 must be the same as that mapped to ma1 in
                 s1 *)
              (Value _ pg = map_apply madd_eq (st_memory s2) ma2)))).

Definition osi_equivalent_pt_pages (s1 s2:state) (osi:os_ident): Prop :=
  (osi_pt_pages_eq s1 s2 osi) /\ (osi_pt_pages_eq s2 s1 osi).

(** Two states are osi equivalent, 
    if they are equal regarding only 
    the part of the states that corresponds 
    to osi.
*)
Definition osi_equivalent (s1 s2:state) (osi:os_ident):= 
  ((st_active_os s1 = osi /\ st_active_os s2 = osi /\ st_aos_activity s1 = st_aos_activity s2) \/
  (st_active_os s1 <> osi /\ st_active_os s2 <> osi)) /\

  (* Acá está la hypercall. Ver si se puede simplificar *)
  map_apply os_ident_eq (st_oss s1) osi =
  map_apply os_ident_eq (st_oss s2) osi /\
  
  (osi_equivalent_hyper_mapping s1 s2 osi) /\
  (osi_equivalent_pt_pages s1 s2 osi).

Lemma osi_equivalent_refl : forall (s:state) (osi:os_ident),
  osi_equivalent s s osi.
Proof.
  intros s osi.
  unfold osi_equivalent, osi_equivalent_hyper_mapping,
    osi_equivalent_pt_pages, osi_hyper_mapping_eq, osi_pt_pages_eq.
  repeat split; auto.
  case (os_ident_eq (st_active_os s) osi).
  left; repeat split; auto.
  right; repeat split; auto.

  intros pt_padd pt_page mp H_pt_page_map H_pt_page_content.
  exists pt_page; auto.
  exists mp; auto.
  intros va pa ma pg H_ma_map H_pg_map H_va_accessible H_pg_type H_ma_hyp_map.
  exists ma; auto.

  intros pt_padd pt_page mp H_pt_page_map H_pt_page_content.
  exists pt_page; auto.
  exists mp; auto.
  intros va pa ma pg H_ma_map H_pg_map H_va_accessible H_pg_type H_ma_hyp_map.
  exists ma; auto.
Qed.

Lemma osi_equivalent_symm : forall (s1 s2:state) (osi:os_ident),
  osi_equivalent s1 s2 osi -> osi_equivalent s2 s1 osi.
Proof.
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages.
  intros s1 s2 osi H_s1_eqosi_s2.
  destruct H_s1_eqosi_s2 as [H_active [H_osi [[H_hyp1 H_hyp2] [H_pt1 H_pt2]]]].
  repeat split; auto.
  destruct H_active as [[H_osi_active1 [H_osi_active2 H_activity]] | [H_osi_no_active1 H_osi_no_active2]];
    [left | right]; auto.
Qed.  

Lemma osi_eq_hyp_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_hyper_mapping_eq s1 s2 osi -> osi_hyper_mapping_eq s2 s3 osi ->
  osi_hyper_mapping_eq s1 s3 osi.
Proof.
  intros s1 s2 s3 osi.
  unfold osi_hyper_mapping_eq, padd_page.
  intros H_s1_eqhyp_s2 H_s2_eqhyp_s3
    pa pg [ma1 H_ma1_map H_pg_map] [val H_pg_content].
  destruct H_s2_eqhyp_s3 with pa pg as [ma3 H_ma3_map H_pg_map3].
  destruct H_s1_eqhyp_s2 with pa pg as [ma2 H_ma2_map H_pg_map2].
  exists ma1; auto.
  exists val; auto.
  exists ma2; auto.
  exists val; auto.
  exists ma3; auto.
Qed.

Lemma osi_eq_pt_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_pt_pages_eq s1 s2 osi -> osi_pt_pages_eq s2 s3 osi ->
  osi_pt_pages_eq s1 s3 osi.
Proof.
  intros s1 s2 s3 osi.
  unfold osi_pt_pages_eq, padd_page.
  intros H_s1_eqpt_s2 H_s2_eqpt_s3 pt_padd pt_page1 mp1
    [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.
  elim H_s1_eqpt_s2 with pt_padd pt_page1 mp1; auto.
  intros pt_page2 [pt_addr2 H_pt_addr2_map H_pt_page2_map]
    [mp2 H_pt_page2_content H_s1_eqpt_s2'].
  elim H_s2_eqpt_s3 with pt_padd pt_page2 mp2; auto.
  intros pt_page3 [pt_addr3 H_pt_addr3_map H_pt_page3_map]
    [mp3 H_pt_page3_content H_s2_eqpt_s3'].
  exists pt_page3.
  exists pt_addr3; auto.
  exists mp3; auto.
  intros va pa ma1 pg H_ma1_map H_pg_map H_va_accessible H_ma1_type H_ma1_hyp_map.
  elim H_s1_eqpt_s2' with va pa ma1 pg; auto.
  intros ma2 H_ma2_map [H_ma2_hyp_map H_pg_map'].
  elim H_s2_eqpt_s3' with va pa ma2 pg; auto.
  intros ma3 H_ma3_map [H_ma3_hyp_map H_pg_map''].
  exists ma3; auto.
  unfold page_type.
  rewrite <- H_pg_map'.
  unfold page_type in H_ma1_type.
  rewrite <- H_pg_map in H_ma1_type.
  assumption.
  exists pt_addr2; auto.
  exists pt_addr1; auto.
Qed.

Lemma osi_equivalent_trans : forall (s1 s2 s3:state) (osi:os_ident),
  osi_equivalent s1 s2 osi -> osi_equivalent s2 s3 osi -> osi_equivalent s1 s3 osi.
Proof.
  unfold osi_equivalent.
  intros s1 s2 s3 osi H_s1_eq_s2 H_s2_eq_s3.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  destruct H_s2_eq_s3 as [H_s2_active_s3 [H_s2_osi_s3 [[H_s2_hyp_s3 H_s3_hyp_s2] [H_s2_pt_s3 H_s3_pt_s2]]]].
  repeat split.
  
  destruct H_s1_active_s2 as [[H_s1_active [H_s2_active H_s1_activity_s2]] | [H_s1_not_active H_s2_not_active]]; 
    destruct H_s2_active_s3 as [[H_s2_active' [H_s3_active H_s2_activity_s3]] | [H_s2_not_active' H_s3_not_active]].

  left; repeat split; auto.
  rewrite <- H_s2_activity_s3.
  assumption.
  cut False; contradiction.
  cut False; contradiction.
  right; split; auto.

  rewrite H_s1_osi_s2; assumption.

  apply osi_eq_hyp_trans with s2; auto.
  apply osi_eq_hyp_trans with s2; auto.
  apply osi_eq_pt_trans with s2; auto.
  apply osi_eq_pt_trans with s2; auto.
Qed.


(* Active OS equivalence *)
Definition aos_equivalent_hyper_mapping (s1 s2:state) :=
  (osi_equivalent_hyper_mapping s1 s2 (st_active_os s1)).

Definition aos_equivalent_pt_pages (s1 s2:state) :=
  (osi_equivalent_pt_pages s1 s2 (st_active_os s1)).

(** Two states are active equivalent, 
    if they are equal regarding only 
    the part of the states that corresponds 
    to the AOS.
*)

Definition active_equivalent (s1 s2:state) := 
  st_active_os s1 = st_active_os s2 /\
  st_aos_activity s1 = st_aos_activity s2 /\

  (* Esto se deduce de la condición de estado válido, a partir de las
     2 previas. *)
  (* st_exec_mode s1 = st_exec_mode s2 /\ *)

  map_apply os_ident_eq (st_oss s1) (st_active_os s1) =
  map_apply os_ident_eq (st_oss s2) (st_active_os s2) /\
  
  (aos_equivalent_hyper_mapping s1 s2) /\
  (aos_equivalent_pt_pages s1 s2).


(* If osi is the AOS, osi equivalence and active equivalence are the same *)
Lemma aos_osi_equivalent : forall (s1 s2:state) (osi:os_ident),
  osi = st_active_os s1 -> active_equivalent s1 s2 ->
  osi_equivalent s1 s2 osi.
Proof.
  intros s1 s2 osi H_osi_active H_active_eq.
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages.
  subst osi.
  unfold active_equivalent, aos_equivalent_hyper_mapping, aos_equivalent_pt_pages in H_active_eq.
  generalize H_active_eq.
  unfold active_equivalent, aos_equivalent_hyper_mapping,
    aos_equivalent_pt_pages, padd_page, vadd_page;
    intros (H_active_os, (H_activity_aos, (H_oss,
      ((H_hyp1_eq, H_hyp2_eq), (H_pt1_eq, H_pt2_eq))))).
  rewrite <- H_active_os in H_oss.
  repeat split; auto.
Qed.

Lemma osi_aos_equivalent : forall (s1 s2:state) (osi:os_ident),
  osi = st_active_os s1 -> osi_equivalent s1 s2 osi ->
  active_equivalent s1 s2.
Proof.
  intros s1 s2 osi H_osi_active H_osi_eq.
  unfold active_equivalent, aos_equivalent_hyper_mapping, aos_equivalent_pt_pages.
  rewrite <- H_osi_active.
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages in H_osi_eq.
  destruct H_osi_eq as [H_active [H_osi [[H_hyp1 H_hyp2] [H_pt1 H_pt2]]]].
  destruct H_active as [[H_osi_active1 [H_osi_active2 H_activity]] | [H_osi_no_active1 H_osi_no_active2]].
  rewrite H_osi_active2.
  repeat split; auto.
  cut False; try contradiction; apply H_osi_no_active1; auto.
Qed.


Lemma active_eq_symm : forall (s1 s2:state),
  active_equivalent s1 s2 -> active_equivalent s2 s1.
Proof.
  intros s1 s2 H_active_eq.
  apply osi_aos_equivalent with (st_active_os s1).
  elim H_active_eq; auto.
  apply osi_equivalent_symm.
  apply aos_osi_equivalent; auto.
Qed.

Lemma active_eq_trans : forall (s1 s2 s3:state),
  active_equivalent s1 s2 -> active_equivalent s2 s3 ->
  active_equivalent s1 s3.
Proof.
  intros s1 s2 s3 H_s1_eq_s2 H_s2_eq_s3.
  apply osi_aos_equivalent with (st_active_os s1); auto.
  apply osi_equivalent_trans with s2.
  apply aos_osi_equivalent; auto.
  apply aos_osi_equivalent; auto.
  elim H_s1_eq_s2; auto.
Qed.

End State_equivalence.

Section Untrusted_actions_equivalence.
(* Hypervisor untrusted actions preserve state equivalence. *)

Variable ctxt:context.  
  
Lemma new_os_isolation_hyper : forall (s1 s2 s1' s2':state)
  (osi o:os_ident) (va:vadd) (pa:padd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_all_new o va pa None s1 s1' -> post_all_new o va pa None s2 s2' ->
  osi_hyper_mapping_eq s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va pa H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system1 H_system1_map H_s1']]]].
  destruct H_post2' as [ma2 H_ma2_hyp_map [pt_addr2
    H_pt_addr2_curr_map [newmem2 H_update2 [system2 H_system2_map H_s2']]]].
  unfold osi_equivalent_hyper_mapping, osi_hyper_mapping_eq, padd_page.

  rewrite H_s1'; simpl.
  rewrite H_s2'; simpl.
  unfold update_pt_page in H_update1, H_update2.
  destruct H_update1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_update2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pa' pg' [ma' H_ma'_hyp_map H_pg'_map] [val H_pg'_content].
  (* ma' cannot be the changed pt_addr1, as it contains a RW page *)
  assert (ma'<>pt_addr1) as H_ma'_ne_pt_addr1.
  intro H_ma'_eq_pt_addr1.
  rewrite H_newmem1 in H_pg'_map.
  rewrite H_ma'_eq_pt_addr1 in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value;
    rewrite H_pg'_value in H_pg'_content; discriminate H_pg'_content.
  
  elim H_s1_hyp_s2 with pa' pg'; auto.
  intros ma H_ma_map H_pg'_map2.
  exists ma; auto.

  assert (ma<>pt_addr2) as H_ma_ne_pt_addr2.
  intro H_ma_eq_pt_addr2.
  unfold os_current_vadd_mapping_madd in H_pt_addr2_curr_map.
  destruct H_pt_addr2_curr_map as [system2' H_system2'_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system2 p2m2; auto.
  intros pt_addr2' H_pt_addr2'_map [H_pt_addr2'_type H_pt_addr2'_owner].
  rewrite <- H_system2_map in H_system2'_map.
  injection H_system2'_map; intro; subst system2'; clear H_system2'_map.
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.
  rewrite <- H_ma_eq_pt_addr2 in H_pt_addr2'_type.
  unfold page_type in H_pt_addr2'_type.
  rewrite <- H_pg'_map2 in H_pt_addr2'_type.
  rewrite H_pg'_content in H_pt_addr2'_type.
  discriminate H_pt_addr2'_type.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold padd_page.
  exists ma'; auto.
  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  exists val; auto.
Qed.

Lemma new_os_isolation_pt_pages : forall (s1 s2 s1' s2':state)
  (osi o:os_ident) (va:vadd) (pa:padd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_all_new o va pa None s1 s1' -> post_all_new o va pa None s2 s2' ->
  osi_pt_pages_eq ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va pa H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system H_system_map H_s1']]]].
  destruct H_post2' as [ma2 H_ma2_hyp_map [pt_addr2
    H_pt_addr2_curr_map [newmem2 H_update2 [system2 H_system2_map H_s2']]]].

  rewrite H_s1'; simpl.
  rewrite H_s2'; simpl.

  unfold update_pt_page in H_update1, H_update2.
  destruct H_update1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_update2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pt_padd pt_page1' mp1' [pt_addr1'
    H_pt_addr1'_map H_pt_page1'_map] H_pt_page1'_content.

  unfold hypervisor_mapping_madd in H_pt_addr1'_map.
  destruct H_pt_addr1'_map as [p2m1 H_p2m1_map H_pt_addr1'_map].
  unfold hypervisor_mapping in H_p2m1_map.

  unfold os_current_vadd_mapping_madd in H_pt_addr1_curr_map.
  destruct H_pt_addr1_curr_map as [system1 H_system1_map [p2m1' H_p2m1'_map H_pt_addr1_curr_map]].
  rewrite <- H_system_map in H_system1_map.
  injection H_system1_map; intro; subst system1; clear H_system1_map.

  case (madd_eq pt_addr1 pt_addr1').
  intro H_pt_addr1'_eq_pt_addr1.

  assert (o=osi).
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m1.
  intros [ma_1 H_ma_1_map H_pt_addr1'_owner].
  rewrite <- H_pt_addr1'_map in H_ma_1_map.
  injection H_ma_1_map; intro; subst ma_1; clear H_ma_1_map.
  clear H_valid_hyper.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o (os_current_page system) p2m1'.
  intros [ma_1 H_ma_1_map H_pt_addr1'_owner'].
  rewrite <- H_pt_addr1_curr_map in H_ma_1_map.
  injection H_ma_1_map; intro; subst ma_1; clear H_ma_1_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_addr1'_owner'.
  rewrite H_pt_addr1'_owner in H_pt_addr1'_owner'.
  injection H_pt_addr1'_owner'; auto.
  intro H_error.
  rewrite <- H_pt_addr1_curr_map in H_error.
  discriminate H_error.
  assumption.
  intro H_error.
  rewrite <- H_pt_addr1'_map in H_error.
  discriminate H_error.
  assumption.
  subst o.
  simpl in H_p2m1_map.
  assert (p2m1'=p2m1).
  unfold hypervisor_mapping in H_p2m1'_map.
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; auto.
  subst p2m1'; clear H_p2m1'_map.

  assert (pt_padd=(os_current_page system)).
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  subst pt_padd.

  assert (system2=system).
  rewrite H_s1_osi_s2 in H_system_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; auto.
  subst system2.

  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_page1'_map.
  simpl in H_pt_page1'_map.
  rewrite lemma_map_add_apply in H_pt_page1'_map.
  injection H_pt_page1'_map; intro H_pt_page1'_value.
  
  unfold os_current_vadd_mapping_madd in H_pt_addr2_curr_map.

  destruct H_pt_addr2_curr_map as [system' H_system'_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  unfold hypervisor_mapping in H_p2m2_map.
  rewrite <- H_s1_osi_s2 in H_system'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  
  exists (Page (PT_c (map_add vadd_eq old_mp2 va ma2)) (page_owned_by old_pt2)).
  exists pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_newmem2.
  simpl.
  rewrite lemma_map_add_apply.
  reflexivity.

  exists (map_add vadd_eq old_mp2 va ma2); auto.
  simpl.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  case (vadd_eq va va').
  intro H_va_eq_va'.

  assert (ma1=ma') as H_ma1_eq_ma'.
  assert (map_apply vadd_eq mp1' va = Value vadd ma1) as H_ma1_map.
  rewrite H_pt_page1'_value in H_pt_page1'_content.
  unfold page_content in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite H_va_eq_va' in H_ma1_map.
  rewrite <- H_ma'_map in H_ma1_map.
  injection H_ma1_map; auto.

  assert (pa=pa') as H_pa_eq_pa'.
  apply hyp_mapped_same_paddr with ctxt s1 osi ma1; auto.
  rewrite H_ma1_eq_ma'.
  assumption.

  exists ma2.
  rewrite H_va_eq_va', lemma_map_add_apply.
  reflexivity.
  split.
  rewrite <- H_pa_eq_pa'.
  assumption.
  rewrite H_newmem2.

  assert ((os_current_page system) <> pa') as H_pg'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (pt_addr1=ma1) as H_pt_addr1_eq_ma1.
  apply hyp_mapped_same_maddr with s1 osi pa; auto.
  rewrite H_pa_eq_pa'.
  rewrite <- H_pa'_curr_pt.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_pt_addr1_eq_ma1 in H_pg'_type.
  rewrite H_ma1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.

  assert (pt_addr1<>ma1) as H_pt_addr1_ne_ma1.
  intro H_pt_addr1_eq_ma1.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma1.
  rewrite H_ma1_eq_ma'.
  assumption.

  assert (pt_addr2<>ma2) as H_pt_addr2_ne_ma2.
  intro H_pt_addr2_eq_ma2.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_pt_addr2_eq_ma2.
  rewrite <- H_pa_eq_pa'.
  assumption.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  
  rewrite H_newmem1 in H_pg'_map.
  rewrite H_ma1_eq_ma' in H_pt_addr1_ne_ma1.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite lemma_map_add_apply_distinct; auto.

  elim H_s1_hyp_s2 with pa pg'; auto.
  intros ma H_ma_hyp_map H_pg'_map2.
  assert (ma2=ma) as H_ma2_eq_ma.
  apply hyp_mapped_same_maddr with s2 osi pa; auto.
  rewrite H_ma2_eq_ma.
  assumption.
  unfold padd_page.
  exists ma'; auto.
  rewrite H_pa_eq_pa'.
  assumption.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  induction (page_content pg').
  exists o; auto.
  discriminate H_pg'_type.
  discriminate H_pg'_type.

  (* If va' is not the new address *)
  intro H_va_ne_va'.
  case (padd_eq pa' pa).
  intro H_pa'_eq_pa.
  assert (ma1=ma') as H_ma1_eq_ma'.
  apply hyp_mapped_same_maddr with s1 osi pa; auto.
  rewrite <- H_pa'_eq_pa.
  assumption.

  assert ((os_current_page system) <> pa') as H_pg'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (pt_addr1=ma1) as H_pt_addr1_eq_ma1.
  apply hyp_mapped_same_maddr with s1 osi pa; auto.
  rewrite <- H_pa'_eq_pa.
  rewrite <- H_pa'_curr_pt.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_pt_addr1_eq_ma1 in H_pg'_type.
  rewrite H_ma1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  discriminate H_pg'_type.
  rewrite H_pa'_eq_pa in H_pg'_not_curr_pt.

  rewrite lemma_map_add_apply_distinct; auto.
  rewrite H_pt_page1'_value in H_pt_page1'_content.
  unfold page_content in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.

  elim H_s1_pt_s2 with (os_current_page system) old_pt1 old_mp1; auto.
  intros pt_page2' H_pt_page2'_map [mp2' H_pt_page2'_content H_virt_map_eq].
  unfold padd_page in H_pt_page2'_map.
  destruct H_pt_page2'_map as [pt_addr2' H_pt_addr2'_map H_pt_page2'_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.

  assert (pt_addr1<>ma1) as H_pt_addr1_ne_ma1.
  intro H_pt_addr1_eq_ma1.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma1.
  rewrite H_ma1_eq_ma'.
  rewrite <- H_pa'_eq_pa.
  assumption.
  rewrite H_ma1_eq_ma' in H_pt_addr1_ne_ma1.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma.
  rewrite <- H_old_pt2_map in H_pt_page2'_map.
  injection H_pt_page2'_map; intro; subst pt_page2'; clear H_pt_page2'_map.
  rewrite <- H_old_pt2_content in H_pt_page2'_content.
  injection H_pt_page2'_content; intro; subst mp2'; clear H_pt_page2'_content.
  assumption.

  assert (pt_addr2<>ma) as H_pt_addr2_ne_ma.
  intro H_pt_addr2_eq_ma.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_pt_addr2_eq_ma.
  rewrite <- H_pa'_eq_pa.
  assumption.

  split; auto.
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.

  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.

  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite <- H_pt_addr1'_eq_pt_addr1.
  assumption.

  (* pa' is not the new page *)
  intro H_pa'_ne_pa.
  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  apply H_pa'_ne_pa.
  apply hyp_mapped_same_paddr with ctxt s1 osi ma1; auto.
  rewrite H_ma1_eq_ma'.
  assumption.

  assert ((os_current_page system) <> pa') as H_pa'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (ma'=pt_addr1') as H_ma'_eq_pt_addr1'.
  apply hyp_mapped_same_maddr with s1 osi pa'; auto.
  rewrite <- H_pa'_curr_pt.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite H_ma'_eq_pt_addr1' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_newmem1 in H_pg'_type.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_ne_ma'.
  intro H_pt_addr1_eq_ma'.
  apply H_pa'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma'.
  assumption.

  elim H_s1_pt_s2 with (os_current_page system) old_pt1 old_mp1; auto.
  intros pt_page2' H_pt_page2'_map [mp2' H_pt_page2'_content H_virt_map_eq].
  unfold padd_page in H_pt_page2'_map.
  destruct H_pt_page2'_map as [pt_addr2' H_pt_addr2'_map H_pt_page2'_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma.
  rewrite <- H_old_pt2_map in H_pt_page2'_map.
  injection H_pt_page2'_map; intro; subst pt_page2'; clear H_pt_page2'_map.
  rewrite <- H_old_pt2_content in H_pt_page2'_content.
  injection H_pt_page2'_content; intro; subst mp2'; clear H_pt_page2'_content.
  
  rewrite lemma_map_add_apply_distinct; auto.
  split; auto.
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.

  intro H_pt_addr2_eq_ma.
  apply H_pa'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi ma; auto.
  rewrite <- H_pt_addr2_eq_ma.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  
  rewrite H_pt_page1'_value in H_pt_page1'_content.
  simpl in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.

  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  unfold padd_page.
  exists pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.


  (* If pt_addr1' is not the current PT page *)
  intro H_pt_addr1'_ne_pt_addr1.
  destruct H_pt_addr2_curr_map as
    [system2' H_system2'_map [p2m2 H_p2m2_map H_pt_addr2_map]].
  unfold hypervisor_mapping in H_p2m2_map.
  rewrite <- H_system2_map in H_system2'_map.
  injection H_system2'_map; intro; subst system2'; clear H_system2'_map.
  case (os_ident_eq o osi).
  intro; subst o.
  rewrite <- H_s1_osi_s2 in H_system2_map.
  rewrite <- H_system_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.
  unfold hypervisor_mapping in H_p2m1'_map.
  simpl in H_p2m1_map.
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  simpl.
  elim H_s1_pt_s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  
  assert (pt_padd<>(os_current_page system)) as H_pt_padd_ne_curr_pt.
  intro H_pt_padd_eq_curr_pt.
  apply H_pt_addr1'_ne_pt_addr1.
  apply hyp_mapped_same_maddr with s1 osi pt_padd.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_padd_eq_curr_pt.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  
  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  apply H_pt_padd_ne_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_pt_addr2'_eq_pt_addr2.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_newmem2.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map
    H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi system p2m1; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_curr_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.
  
  rewrite H_newmem2.
  assert (pt_addr2<>ma) as H_pt_addr2_neq_ma.
  intro H_pt_addr2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi system p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  rewrite <- H_s1_osi_s2; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.
  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  simpl in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.

  (* If o is not osi *)
  intro H_o_ne_osi.
  elim H_s1_pt_s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2_osi H_p2m2_osi_map H_pt_addr2'_map].

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  exists p2m2_osi; auto.
  rewrite H_newmem2.

  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m2_osi; auto.
  intros [ma_2 H_ma_2_map H_pt_addr2_owner].
  rewrite <- H_pt_addr2'_map in H_ma_2_map.
  injection H_ma_2_map; intro; subst ma_2; clear H_ma_2_map.
  clear H_valid_hyper.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o (os_current_page system2) p2m2; auto.
  intros [ma_2 H_ma_2_map H_pt_addr2'_owner].
  rewrite <- H_pt_addr2_map in H_ma_2_map.
  injection H_ma_2_map; intro; subst ma_2; clear H_ma_2_map.
  rewrite <- H_pt_addr2'_eq_pt_addr2 in H_pt_addr2'_owner.
  rewrite H_pt_addr2_owner in H_pt_addr2'_owner.
  injection H_pt_addr2'_owner; intro H_abs;
    apply H_o_ne_osi; auto.

  intro H_error.
  rewrite <- H_pt_addr2_map in H_error.
  discriminate H_error.
  intro H_error.
  rewrite <- H_pt_addr2'_map in H_error.
  discriminate H_error.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.

  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  simpl.
  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system p2m1'; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_curr_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  simpl in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  simpl in H_pg'_map, H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.

  intro H_pt_addr2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system2 p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.
  assumption.
  assumption.

  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  simpl in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.
Qed.

Lemma new_os_isolation : forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd) (pa:padd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_all_new o va pa None s1 s1' -> post_all_new o va pa None s2 s2' ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va pa H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [ma1 H_ma1_hyp_map [pt_addr1 H_pt_addr1_curr_map
    [newmem1 H_update1 [system1 H_system1_map H_s1']]]].
  destruct H_post2' as [ma2 H_ma2_hyp_map [pt_addr2 H_pt_addr2_curr_map
    [newmem2 H_update2 [system2 H_system2_map H_s2']]]].

  unfold osi_equivalent.

  generalize (osi_equivalent_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  repeat split; try (rewrite H_s1', H_s2'; auto; fail).
  (* Prove osi data does not change *)
  rewrite H_s1', H_s2'; simpl.
  case (os_ident_eq o osi).
  intros H_o_eq_osi.
  rewrite H_o_eq_osi.
  repeat rewrite lemma_map_add_apply.
  assert (system1 = system2).
  subst osi.
  rewrite  <- H_s1_osi_s2  in H_system2_map.
  rewrite <- H_system2_map in H_system1_map.
  injection H_system1_map; auto.
  subst system2; reflexivity.

  intros H_o_ne_osi.
  repeat rewrite lemma_map_add_apply_distinct; auto.

  apply new_os_isolation_hyper with s1 s2 o va pa; auto.
  apply new_os_isolation_hyper with s2 s1 o va pa; auto.
  apply new_os_isolation_pt_pages with s1 s2 o va pa; auto.
  apply new_os_isolation_pt_pages with s2 s1 o va pa; auto.
Qed.

Lemma del_untrusted_os_isolation_hyper : forall (s1 s2 s1' s2':state)
  (osi o:os_ident) (va:vadd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_del_untrusted o va s1 s1' -> post_del_untrusted o va s2 s2' ->
  osi_hyper_mapping_eq s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_del_untrusted in H_post1', H_post2'.
  destruct H_post1' as [system1 H_system1_map [p2m1 H_p2m1_map [pt_addr1 H_pt_addr1_map [newmem1 H_drop1 H_s1']]]].
  destruct H_post2' as [system2 H_system2_map [p2m2 H_p2m2_map [pt_addr2 H_pt_addr2_map [newmem2 H_drop2 H_s2']]]].

  unfold osi_hyper_mapping_eq, padd_page.

  rewrite H_s1', H_s2'; simpl.

  unfold drop_va_pt_page in H_drop1, H_drop2.
  destruct H_drop1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_drop2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pa' pg' [ma' H_ma'_hyp_map H_pg'_map] [val H_pg'_content].
  (* ma' cannot be the changed pt_addr1, as it contains a RW page *)
  assert (ma'<>pt_addr1) as H_ma'_ne_pt_addr1.
  intro H_ma'_eq_pt_addr1.
  rewrite H_newmem1 in H_pg'_map.
  rewrite H_ma'_eq_pt_addr1 in H_pg'_map.
  rewrite lemma_map_add_apply in H_pg'_map.
  injection H_pg'_map; intro H_pg'_value;
    rewrite H_pg'_value in H_pg'_content; discriminate H_pg'_content.
  
  elim H_s1_hyp_s2 with pa' pg'; auto.
  intros ma H_ma_map H_pg'_map2.
  exists ma; auto.

  assert (ma<>pt_addr2) as H_ma_ne_pt_addr2.
  intro H_ma_eq_pt_addr2.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system2 p2m2; auto.
  intros pt_addr2' H_pt_addr2'_map [H_pt_addr2'_type H_pt_addr2'_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.
  rewrite <- H_ma_eq_pt_addr2 in H_pt_addr2'_type.
  unfold page_type in H_pt_addr2'_type.
  rewrite <- H_pg'_map2 in H_pt_addr2'_type.
  rewrite H_pg'_content in H_pt_addr2'_type.
  discriminate H_pt_addr2'_type.
  
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold padd_page.
  exists ma'; auto.
  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  exists val; auto.
Qed.

Lemma del_untrusted_os_isolation_pt_pages : forall (s1 s2 s1' s2':state)
  (osi o:os_ident) (va:vadd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_del_untrusted o va s1 s1' -> post_del_untrusted o va s2 s2' ->
  osi_pt_pages_eq ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_del_untrusted in H_post1', H_post2'.
  destruct H_post1' as [system1 H_system1_map [p2m1 H_p2m1_map [pt_addr1 H_pt_addr1_map [newmem1 H_drop1 H_s1']]]].
  destruct H_post2' as [system2 H_system2_map [p2m2 H_p2m2_map [pt_addr2 H_pt_addr2_map [newmem2 H_drop2 H_s2']]]].
  
  unfold osi_pt_pages_eq, padd_page.

  rewrite H_s1', H_s2'; simpl.
  unfold drop_va_pt_page in H_drop1, H_drop2.
  destruct H_drop1 as [old_pt1 H_old_pt1_map [old_mp1 H_old_pt1_content H_newmem1]].
  destruct H_drop2 as [old_pt2 H_old_pt2_map [old_mp2 H_old_pt2_content H_newmem2]].

  intros pt_padd pt_page1' mp1' [pt_addr1'
    H_pt_addr1'_map H_pt_page1'_map] H_pt_page1'_content.

  unfold hypervisor_mapping_madd in H_pt_addr1'_map.
  destruct H_pt_addr1'_map as [p2m1_aos H_p2m1_aos_map H_pt_addr1'_map].
  unfold hypervisor_mapping in H_p2m1_map.

  case (madd_eq pt_addr1 pt_addr1').
  intro H_pt_addr1'_eq_pt_addr1.

  assert (o=osi).
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m1_aos.
  intros [ma_1 H_ma_1_map H_pt_addr1'_owner].
  rewrite <- H_pt_addr1'_map in H_ma_1_map.
  injection H_ma_1_map; intro; subst ma_1; clear H_ma_1_map.
  clear H_valid_hyper.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o (os_current_page system1) p2m1.
  intros [ma_1 H_ma_1_map H_pt_addr1'_owner'].
  rewrite <- H_pt_addr1_map in H_ma_1_map.
  injection H_ma_1_map; intro; subst ma_1; clear H_ma_1_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_addr1'_owner'.
  rewrite H_pt_addr1'_owner in H_pt_addr1'_owner'.
  injection H_pt_addr1'_owner'; auto.
  intro H_error.
  rewrite <- H_pt_addr1_map in H_error.
  discriminate H_error.
  assumption.
  intro H_error.
  rewrite <- H_pt_addr1'_map in H_error.
  discriminate H_error.
  assumption.
  subst o.

  rewrite <- H_p2m1_map in H_p2m1_aos_map.
  injection H_p2m1_aos_map; intro; subst p2m1_aos; clear H_p2m1_aos_map.

  assert (pt_padd=(os_current_page system1)).
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  rewrite H_pt_addr1'_eq_pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  subst pt_padd.

  assert (system2=system1).
  rewrite <- H_s1_osi_s2 in H_system2_map.
  rewrite <- H_system1_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.
  reflexivity.
  subst system2.
  
  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite H_pt_addr1'_eq_pt_addr1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply in H_pt_page1'_map.
  injection H_pt_page1'_map; intro H_pt_page1'_value.
  
  unfold hypervisor_mapping in H_p2m2_map.
  
  exists (Page (PT_c (map_drop vadd_eq old_mp2 va)) (page_owned_by old_pt2)).
  exists pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply.
  reflexivity.

  exists (map_drop vadd_eq old_mp2 va); auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_pt_page1'_value in H_pt_page1'_content.
  simpl in H_pt_page1'_content.
  injection H_pt_page1'_content; intro H_mp1'_value.
  rewrite H_mp1'_value in H_ma'_map.

  assert (va<>va') as H_va_ne_va'.
  intro H_va_eq_va'.
  rewrite H_va_eq_va' in H_ma'_map.
  rewrite map_drop_apply in H_ma'_map.
  discriminate H_ma'_map.

  rewrite map_drop_apply_distinct in H_ma'_map; auto.
  rewrite map_drop_apply_distinct; auto.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.

  assert ((os_current_page system1) <> pa') as H_pg'_not_curr_pt.
  intro H_pa'_curr_pt.
  assert (pt_addr1=ma') as H_pt_addr1_eq_ma'.
  apply hyp_mapped_same_maddr with s1 osi (os_current_page system1).
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pa'_curr_pt.
  assumption.
  rewrite <- H_pt_addr1_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite H_newmem1 in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_ne_ma1.
  intro H_pt_addr1_eq_ma'.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s1 osi pt_addr1; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_addr1_eq_ma'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.


  elim H_s1_pt_s2 with (os_current_page system1) old_pt1 old_mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  
  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma2 H_ma2_map [H_ma2_hyp_map H_pg'_map2].
  exists ma2; auto.

  rewrite <- H_pt_addr2_map in H_pt_addr2'_map.
  injection H_pt_addr2'_map; intro; subst pt_addr2'; clear H_pt_addr2'_map.
  rewrite <- H_old_pt2_map in H_pt_page2_map.
  injection H_pt_page2_map; intro; subst pt_page2; clear H_pt_page2_map.
  rewrite <- H_old_pt2_content in H_pt_page2_content.
  injection H_pt_page2_content; intro; subst mp2.
  assumption.

  split; auto.
  rewrite H_newmem2.

  assert (pt_addr2<>ma2) as H_pt_addr2_ne_ma2.
  intro H_pt_addr2_eq_ma2.
  apply H_pg'_not_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_pt_addr2_eq_ma2.
  assumption.

  rewrite lemma_map_add_apply_distinct; auto.

  rewrite H_newmem1 in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite H_newmem1 in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.

  unfold hypervisor_mapping_madd; exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1; auto.
  unfold hypervisor_mapping_madd; exists p2m1; auto.


  (* If pt_addr1' is not the current PT page *)
  intro H_pt_addr1'_ne_pt_addr1.
  case (os_ident_eq o osi).
  intro; subst o.
  rewrite <- H_s1_osi_s2 in H_system2_map.
  rewrite <- H_system1_map in H_system2_map.
  injection H_system2_map; intro; subst system2; clear H_system2_map.

  rewrite <- H_p2m1_map in H_p2m1_aos_map.
  injection H_p2m1_aos_map; intro; subst p2m1_aos; clear H_p2m1_aos_map.
  
  elim H_s1_pt_s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2' H_p2m2'_map H_pt_addr2'_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  
  assert (pt_padd<>(os_current_page system1)) as H_pt_padd_ne_curr_pt.
  intro H_pt_padd_eq_curr_pt.
  apply H_pt_addr1'_ne_pt_addr1.
  apply hyp_mapped_same_maddr with s1 osi pt_padd.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_pt_padd_eq_curr_pt.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  
  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  apply H_pt_padd_ne_curr_pt.
  apply hyp_mapped_same_paddr with ctxt s2 osi pt_addr2; auto.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite <- H_pt_addr2'_eq_pt_addr2.
  assumption.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  intros va' pa' ma' pg' H_ma'_map H_pg'_map
    H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi system1 p2m1; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.
  
  rewrite H_newmem2.
  assert (pt_addr2<>ma) as H_pt_addr2_neq_ma.
  intro H_pt_addr2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with osi system1 p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  rewrite <- H_s1_osi_s2; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.
  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.

  (* If o is not osi *)
  intro H_o_ne_osi.
  elim H_s1_pt_s2 with pt_padd pt_page1' mp1'; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2' H_pt_addr2'_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2'_map.
  destruct H_pt_addr2'_map as [p2m2_osi H_p2m2_osi_map H_pt_addr2'_map].

  exists pt_page2.
  exists pt_addr2'.
  unfold hypervisor_mapping_madd.
  exists p2m2_osi; auto.
  rewrite H_newmem2.

  assert (pt_addr2'<>pt_addr2) as H_pt_addr2'_ne_pt_addr2.
  intro H_pt_addr2'_eq_pt_addr2.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o (os_current_page system2) p2m2; auto.
  intros [ma_2 H_ma_2_map H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_ma_2_map.
  injection H_ma_2_map; intro; subst ma_2; clear H_ma_2_map.
  clear H_valid_hyper.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m2_osi; auto.
  intros [ma_2 H_ma_2_map H_pt_addr2'_owner].
  rewrite <- H_pt_addr2'_map in H_ma_2_map.
  injection H_ma_2_map; intro; subst ma_2; clear H_ma_2_map.
  rewrite H_pt_addr2'_eq_pt_addr2 in H_pt_addr2'_owner.
  rewrite H_pt_addr2_owner in H_pt_addr2'_owner.
  injection H_pt_addr2'_owner; intro H_abs;
    apply H_o_ne_osi; assumption.

  intro H_error.
  rewrite <- H_pt_addr2'_map in H_error.
  discriminate H_error.
  intro H_error.
  rewrite <- H_pt_addr2_map in H_error.
  discriminate H_error.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.

  intros va' pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  rewrite H_newmem1 in H_pg'_map, H_pg'_type.
  unfold page_type in H_pg'_type.

  assert (pt_addr1<>ma') as H_pt_addr1_neq_ma'.
  intro H_pt_addr1_eq_ma'.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system1 p2m1; auto.
  intros pt_addr1'' H_pt_addr1''_map [H_pt_addr1_type H_pt_addr1_owner].
  rewrite <- H_pt_addr1_map in H_pt_addr1''_map.
  injection H_pt_addr1''_map; intro; subst pt_addr1''; clear H_pt_addr1''_map.
  rewrite H_pt_addr1_eq_ma' in H_pt_addr1_type.
  rewrite H_pt_addr1_eq_ma' in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type.
  simpl in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.

  elim H_virt_map_eq with va' pa' ma' pg'; auto.
  intros ma H_ma_map [H_ma_hyp_map H_pg'_map2].
  exists ma; auto.
  split; auto.

  rewrite H_newmem2.
  rewrite lemma_map_add_apply_distinct; auto.

  intro H_pt_addr2_eq_ma.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_injective_hyper H_va_has_pa.
  unfold valid_current_page in H_valid_currpg.
  elim H_valid_currpg with o system2 p2m2; auto.
  intros pt_addr2'' H_pt_addr2''_map [H_pt_addr2_type H_pt_addr2_owner].
  rewrite <- H_pt_addr2_map in H_pt_addr2''_map.
  injection H_pt_addr2''_map; intro; subst pt_addr2''; clear H_pt_addr2''_map.
  rewrite H_pt_addr2_eq_ma in H_pt_addr2_type.
  unfold page_type in H_pt_addr2_type.
  rewrite <- H_pg'_map2 in H_pt_addr2_type.
  rewrite H_pg'_type in H_pt_addr2_type.
  discriminate H_pt_addr2_type.

  unfold page_type.
  rewrite <- H_pg'_map.
  rewrite H_pg'_type; reflexivity.

  unfold padd_page.
  exists pt_addr1'.
  unfold hypervisor_mapping_madd.
  exists p2m1_aos; auto.
  rewrite H_newmem1 in H_pt_page1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_page1'_map; auto.
Qed.

Lemma del_untrusted_os_isolation : forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd),
  osi_equivalent ctxt s1 s2 osi ->
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  post_del_untrusted o va s1 s1' -> post_del_untrusted o va s2 s2' ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o va H_s1_eq_s2 H_s1_valid H_s2_valid H_post1 H_post2.
  generalize H_post1 H_post2.
  intros H_post1' H_post2'.
  unfold post_all_new in H_post1', H_post2'.
  destruct H_post1' as [system1 H_system1_map [p2m1 H_p2m1_map [pt_addr1 H_pt_addr1_map [newmem1 H_drop1 H_s1']]]].
  destruct H_post2' as [system2 H_system2_map [p2m2 H_p2m2_map [pt_addr2 H_pt_addr2_map [newmem2 H_drop2 H_s2']]]].

  unfold osi_equivalent.

  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize (osi_equivalent_symm ctxt s1 s2 osi); intro H_s2_eq_s1.
  
  repeat split; try (rewrite H_s1', H_s2'; auto; fail).
  
  (* Prove AOS data does not change *)
  rewrite H_s1', H_s2'; simpl.
  case (os_ident_eq o osi).
  intros H_o_eq_osi.
  rewrite H_o_eq_osi.
  repeat rewrite lemma_map_add_apply.
  assert (system1 = system2).
  subst osi.
  rewrite <- H_s1_osi_s2  in H_system2_map.
  rewrite <- H_system2_map in H_system1_map.
  injection H_system1_map; auto.
  subst system2; reflexivity.

  intros H_o_ne_osi.
  repeat rewrite lemma_map_add_apply_distinct; auto.

  apply del_untrusted_os_isolation_hyper with s1 s2 o va; auto.
  apply del_untrusted_os_isolation_hyper with s2 s1 o va; auto.
  apply del_untrusted_os_isolation_pt_pages with s1 s2 o va; auto.
  apply del_untrusted_os_isolation_pt_pages with s2 s1 o va; auto.
Qed.

Lemma pin_untrusted_os_isolation_hyper : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_pin_untrusted o pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_untrusted o pa type) s2' ->
  osi_hyper_mapping_eq s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa type H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_hyper_mapping_eq in *.
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_untrusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].

  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pa' pg' H_pg'_map [val H_pg'_content].
  case (os_ident_eq o osi).
  (* If o is osi, the hyp mapping that changes is relevant *)
  intros H_o_eq_osi; subst o.
  case (padd_eq pa pa').
  (* If pa' is the new pa, it changes to the same page *)
  intros; subst pa'.
  unfold hypervisor_mapping_madd.
  exists ma2.
  exists (map_add padd_eq p2m2 pa ma2).
  rewrite lemma_map_add_apply; auto.
  rewrite lemma_map_add_apply; auto.  

  rewrite lemma_map_add_apply.
  destruct H_pg'_map as [ma H_hyp_pa_ma1 H_pg'_value].
  assert (ma = ma1).
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_map'.
  rewrite H_p2m1'_map' in H_ma_map.
  rewrite lemma_map_add_apply in H_ma_map; injection H_ma_map; auto.
  subst ma.

  induction (type).
  (* The pinned page is a PT. This case is absurd. *)
  rewrite lemma_map_add_apply in H_pg'_value.
  injection H_pg'_value; intro H_pg'.
  rewrite H_pg' in H_pg'_content.
  simpl in H_pg'_content.
  discriminate H_pg'_content.
  (* The pinned page has RW type *)
  rewrite lemma_map_add_apply in H_pg'_value; auto.

  (* If pa' is not the new pa, it does not change. *)
  intro H_pa_ne_pa'.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_untrusted osi pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  rename H_injective_hyper into H_inj_hyper_s1'.

  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  unfold hypervisor_mapping in H_post_hyp2.

  destruct H_pg'_map as [ma' H_hyp_pa_ma1 H_pg'_value].
  assert (ma1 <> ma').
  intros H_ma1_eq_ma'.
  apply H_pa_ne_pa'.

  unfold injective_hyper_mappings in H_inj_hyper_s1'.
  rewrite H_inj_hyper_s1' with osi (map_add padd_eq p2m1 pa ma1) pa pa'; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply.
  reflexivity.

  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map; intros H_p2m1'_map.
  rewrite lemma_map_add_apply.
  rewrite <- H_p2m1'_map.
  rewrite H_ma1_eq_ma'.
  assumption.

  rewrite lemma_map_add_apply_distinct in H_pg'_value; auto.

  destruct (H_valid_hyper osi pa' p2m2 H_post_hyp2)
    as [[ma H_ma_map H_ma_owner]|H_error].
  exists ma.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  assert (ma <> ma2).
  intros H_ma_eq_ma2.
  apply H_pa_ne_pa'.

  clear H_valid_hyper.
  assert (valid_state ctxt s2') as H_s2'_valid.
  apply valid_state_invariant with s2 (Page_pin_untrusted osi pa type); auto.
  generalize H_s2'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi (map_add padd_eq p2m2 pa ma2) pa pa'; auto.

  rewrite H_s2'; simpl.
  rewrite lemma_map_add_apply.
  reflexivity.
  cut (Value padd ma = map_apply padd_eq (map_add padd_eq p2m2 pa ma2) pa').
  intros H_ma_map'.
  rewrite <- H_ma_map'.
  rewrite lemma_map_add_apply.
  subst ma; reflexivity.
  rewrite lemma_map_add_apply_distinct; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  elim H_s1_hyp_s2 with pa' pg'.
  intros max H_max_map.
  unfold hypervisor_mapping_madd in H_max_map.
  assert (max=ma).
  destruct H_max_map as [p2m2' H_p2m2'_map H_max_map].
  rewrite <- H_p2m2'_map in H_post_hyp2.
  injection H_post_hyp2.
  intros H_p2m2_eq_p2m2'; subst p2m2'.
  rewrite <- H_max_map in H_ma_map.
  injection H_ma_map; auto.
  subst max.
  auto.

  unfold padd_page.
  exists ma'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map;intros H_p2m1'_map.
  rewrite H_p2m1'_map in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  exists val; auto.
  
  elim H_s1_hyp_s2 with pa' pg'.
  intros ma H_ma_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m2' H_p2m2'_map H_ma_map].
  rewrite <- H_p2m2'_map in H_post_hyp2.
  injection H_post_hyp2.
  intros H_p2m2_eq_p2m2'; subst p2m2'.
  rewrite H_error in H_ma_map.
  discriminate H_ma_map.

  unfold padd_page, hypervisor_mapping_madd.
  exists ma'; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1' H_p2m1'_map H_ma'_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; clear H_p2m1'_map;intros H_p2m1'_map.
  rewrite H_p2m1'_map in H_ma'_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_map; auto.
  exists val; auto.
  
  (* Now prove the case when osi does not change*)
  intros H_o_ne_osi.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct; auto.
  destruct H_pg'_map as [ma' H_hyp_pa_ma1 H_pg'_value].
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1_aos H_p2m1_aos_map H_ma'_map].
  rewrite lemma_map_add_apply_distinct in H_p2m1_aos_map; auto.

  (* The new page in s1' is in ma1, it cannot be ma' *)
  cut (ma1 <> ma').
  intros H_ma1_ne_ma'.
  rewrite lemma_map_add_apply_distinct in H_pg'_value; auto.
  elim H_s1_hyp_s2 with pa' pg'.
  intros ma H_ma_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m2_aos H_p2m2_aos_map H_ma_map].
  exists ma.
  exists p2m2_aos; auto.
  cut (ma <> ma2).
  intros H_ma_ne_ma2.
  rewrite lemma_map_add_apply_distinct; auto.

  intros H_ma_eq_ma2.
  (* Prove ma owner is AOS *)
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct (H_valid_hyper osi pa' p2m2_aos H_p2m2_aos_map)
    as [[ma_2 H_ma_2_map H_ma_2_owner]|H_error].
  assert (ma_2=ma).
  rewrite <- H_ma_map in H_ma_2_map.
  injection H_ma_2_map; auto.
  subst ma_2.
  (* As ma2 has no owner, it is not the case that ma=ma2 *)
  unfold maddress_owner in H_ma_2_owner.
  rewrite <- H_pg'_map2 in H_ma_2_owner.
  rewrite <- H_ma_eq_ma2 in H_pg2_map.
  rewrite <- H_pg2_map in H_pg'_map2.
  injection H_pg'_map2; intro H_pg'_eq_pg2.
  rewrite H_pg'_eq_pg2 in H_ma_2_owner.
  rewrite H_pg2_owner in H_ma_2_owner.
  discriminate H_ma_2_owner.
  rewrite <- H_ma_map in H_error.
  discriminate H_error.

  (* Prove pg' is in pa' in s1 *)
  unfold padd_page, hypervisor_mapping_madd.
  exists ma'; auto.
  exists p2m1_aos; auto.

  (* Prove val is pg' content *)
  exists val; auto.

  (* Finally, prove ma1 <> ma' *)
  intros H_ma1_eq_ma'.
  (* Prove ma' owner is osi *)
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct (H_valid_hyper osi pa' p2m1_aos H_p2m1_aos_map) as [[ma_2 H_ma_2_map H_ma_2_owner]|H_error].
  assert (ma_2=ma').
  rewrite <- H_ma'_map in H_ma_2_map.
  injection H_ma_2_map; auto.
  subst ma_2.
  (* As ma1 has no owner, it is not the case that ma1=ma' *)
  unfold maddress_owner in H_ma_2_owner.
  rewrite H_ma1_eq_ma' in H_pg1_map.
  rewrite <- H_pg1_map in H_ma_2_owner.
  rewrite H_pg1_owner in H_ma_2_owner.
  discriminate H_ma_2_owner.
  rewrite <- H_ma'_map in H_error.
  discriminate H_error.
Qed.

Lemma pin_untrusted_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_pin_untrusted o pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_untrusted o pa type) s2' ->
  osi_pt_pages_eq ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa type H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  unfold osi_equivalent, osi_equivalent_pt_pages, osi_pt_pages_eq in *.
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_untrusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].
  unfold hypervisor_mapping in H_post_hyp1, H_post_hyp2.

  unfold pre_action, pre_page_pin_untrusted, pre_all_pin in H_pre2.
  destruct H_pre1 as [H_untrusted_aos [H_aos_hcall [H_hyper_running [[p2m1' H_p2m1'_map H_pa_err1] H_pre1]]]].
  rewrite <- H_post_hyp1 in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_eq_p2m1; subst p2m1'; clear H_p2m1'_map.
  destruct H_pre2 as [H_untrusted_aos2 [H_aos_hcall2 [H_hyper_running2 [[p2m2' H_p2m2'_map H_pa_err2] H_pre2]]]].
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intro H_p2m2'_eq_p2m2; subst p2m2'; clear H_p2m2'_map.

  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pt_padd pt_page1 mp1 [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.

  case (os_ident_eq o osi).
  (* If o is osi, no osi owned PT page in memory changes *)
  intros H_o_eq_osi; subst o.

  case (padd_eq pt_padd pa).
  intros H_pt_padd_eq_pa.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite H_pt_padd_eq_pa in H_pt_addr1_map.
  rewrite lemma_map_add_apply in H_pt_addr1_map.
  injection H_pt_addr1_map; intro H_pt_addr1_eq_ma1; clear H_pt_addr1_map.
  rewrite H_pt_addr1_eq_ma1, lemma_map_add_apply in H_pt_page1_map.
  injection H_pt_page1_map; intro H_pt_page1_value.
  induction type; try (rewrite H_pt_page1_value in H_pt_page1_content; discriminate H_pt_page1_content).

  exists (Page (PT_c (map_empty vadd madd)) (Os_PO osi)).
  exists ma2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite H_pt_padd_eq_pa.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite lemma_map_add_apply; reflexivity.
  exists (map_empty vadd madd); auto.

  rewrite H_pt_page1_value in H_pt_page1_content.
  unfold page_content in H_pt_page1_content.
  injection H_pt_page1_content; intro H_mp1_empty.
  subst mp1.
  intros va pa' ma' pg' H_abs.
  simpl in H_abs; discriminate H_abs.
  
  (* If pt_padd is not the pinned page *)
  intro H_pt_padd_ne_pa.
  assert (ma1<>pt_addr1) as H_ma1_ne_pt_addr1.
  intros H_ma1_eq_pt_addr1.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite <- H_ma1_eq_pt_addr1 in H_pt_addr1_map.
  apply H_pt_padd_ne_pa.
  assert (Value padd ma1 = map_apply padd_eq (map_add padd_eq p2m1 pa ma1) pa) as H_ma1_hyp_map.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite H_pt_addr1_map in H_ma1_hyp_map.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_untrusted osi pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi (map_add padd_eq p2m1 pa ma1) pt_padd pa; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply; reflexivity.

  elim H_s1_pt_s2 with pt_padd pt_page1 mp1.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  
  assert (ma2<>pt_addr2) as H_ma2_ne_pt_addr2.
  intro H_ma2_eq_pt_addr2.
  rewrite H_ma2_eq_pt_addr2 in H_pg2_map.
  rewrite <- H_pt_page2_map in H_pg2_map.
  injection H_pg2_map; intros; subst pg2.
  rewrite <- H_pt_page2_content in H_pg2_content.
  discriminate H_pg2_content.

  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold hypervisor_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [p2m2' H_p2m2'_map H_pt_addr2_map].
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; auto.

  rewrite lemma_map_add_apply_distinct; auto.

  exists mp2; auto.
  
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.
  case (padd_eq pa pa').
  intro H_pa_eq_pa'.
  assert (ma1=ma') as H_ma1_eq_ma'.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  subst p2m1'.
  rewrite <- H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite lemma_map_add_apply in H_ma'_hyp_map.
  injection H_ma'_hyp_map; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite lemma_map_add_apply_distinct in H_pt_addr1_map; auto.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_hyper H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_virtual_mapping in H_valid_virtmap.
  elim H_valid_virtmap with osi pt_addr1 mp1 pt_page1 va ma'; auto.
  clear H_valid_virtmap.
  intros pg1' H_pg1'_map [H_acc H_not_acc].
  assert (page_owned_by pg1'=Os_PO osi) as H_pg1'_owner.
  apply H_acc; auto.
  rewrite H_ma1_eq_ma' in H_pg1_map.
  rewrite <- H_pg1_map in H_pg1'_map.
  injection H_pg1'_map; intros; subst pg1'.
  rewrite H_pg1_owner in H_pg1'_owner.
  discriminate H_pg1'_owner.
  clear H_valid_virtmap.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m1; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [pt_addr1' H_pt_addr1'_map H_pt_addr1'_owner].
  unfold maddress_owner in H_pt_addr1'_owner.
  rewrite <- H_pt_addr1_map in H_pt_addr1'_map.
  injection H_pt_addr1'_map; intros; subst pt_addr1'; clear H_pt_addr1'_map.
  rewrite <- H_pt_page1_map in H_pt_addr1'_owner.
  assumption.
  intro H_abs.
  rewrite <- H_pt_addr1_map in H_abs.
  discriminate H_abs.

  intros H_pa_ne_pa'.
  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  apply H_pa_ne_pa'.

  assert (valid_state ctxt s1') as H_s1'_valid.
  apply valid_state_invariant with s1 (Page_pin_untrusted osi pa type); auto.
  generalize H_s1'_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi (map_add padd_eq p2m1 pa ma1) pa pa'; auto.

  rewrite H_s1'; simpl.
  rewrite lemma_map_add_apply; reflexivity.
  rewrite lemma_map_add_apply.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite <- H_ma1_eq_ma' in H_ma'_hyp_map.
  assumption.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].
  exists ma2'; auto.
  destruct H_ma2'_hyp_map as [p2m2' H_p2m2'_map H_ma2'_hyp_map].
  rewrite <- H_post_hyp2 in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.
  split.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_add padd_eq p2m2 pa ma2); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold hypervisor_mapping_madd in H_ma2'_hyp_map.

  assert (ma2'<>ma2) as H_ma2'_ne_ma2.
  intro H_ma2'_eq_ma2.
  rewrite H_ma2'_eq_ma2 in H_pg'_map2.
  rewrite <- H_pg2_map in H_pg'_map2.
  injection H_pg'_map2; intro H_pg'_eq_pg2.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map, H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  rewrite H_pg'_eq_pg2 in H_pg'_type.
  induction (page_content pg2); discriminate.

  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold page_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg'_map.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  assumption.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  rewrite H_p2m1'_value in H_ma'_hyp_map.
  rewrite lemma_map_add_apply_distinct in H_ma'_hyp_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value.
  rewrite H_p2m1'_value in H_pt_addr1_map.
  rewrite lemma_map_add_apply_distinct in H_pt_addr1_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  assumption.

  (* Now prove the case when osi does not change*)
  intros H_o_ne_osi.
  elim H_s1_pt_s2 with pt_padd pt_page1 mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_pt_addr2.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_page2_map.
  rewrite <- H_pg2_map in H_pt_page2_map.
  injection H_pt_page2_map; intro H_pt_page2_eq_pg2.
  rewrite H_pt_page2_eq_pg2 in H_pt_page2_content.
  rewrite H_pg2_content in H_pt_page2_content.
  discriminate H_pt_page2_content.
  exists mp2; auto.
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1_aos H_p2m1_aos_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct in H_p2m1_aos_map; auto.

  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  assert (maddress_owner (st_memory s1) ma' = Os_PO osi) as H_ma'_owner.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pa' p2m1_aos; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma1' H_ma1'_map H_ma1'_owner].
  rewrite <- H_ma'_hyp_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.
  assumption.
  intro H_abs; rewrite <- H_ma'_hyp_map in H_abs; discriminate H_abs.
  rewrite <- H_ma1_eq_ma' in H_ma'_owner.
  unfold maddress_owner in H_ma'_owner.
  rewrite <- H_pg1_map in H_ma'_owner.
  rewrite H_pg1_owner in H_ma'_owner.
  discriminate H_ma'_owner.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].
  exists ma2'; auto.
  split.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_ma2'.
  subst ma2'.
  rewrite <- H_pg2_map in H_pg'_map2.
  injection H_pg'_map2; intro H_pg'_eq_pg2.
  unfold page_type in H_pg'_type.
  
  rewrite <- H_pg'_map in H_pg'_type.
  rewrite H_pg'_eq_pg2 in H_pg'_type.
  rewrite H_pg2_content in H_pg'_type.
  discriminate H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  unfold hypervisor_mapping_madd; exists p2m1_aos; auto.

  (* last goal *)
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1_aos H_p2m1_aos_map H_pt_addr1_map].
  rewrite lemma_map_add_apply_distinct in H_p2m1_aos_map; auto.
  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1_aos; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  
  intro H_ma1_eq_pt_addr1.

  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m1_aos; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [pt_addr1' H_pt_addr1'_map H_pt_addr1'_owner].
  rewrite <- H_pt_addr1_map in H_pt_addr1'_map.
  injection H_pt_addr1'_map; intro; subst pt_addr1'; clear H_pt_addr1'_map.
  unfold maddress_owner in H_pt_addr1'_owner.
  rewrite <- H_ma1_eq_pt_addr1 in H_pt_addr1'_owner.
  rewrite <- H_pg1_map in H_pt_addr1'_owner.
  rewrite H_pg1_owner in H_pt_addr1'_owner.
  discriminate H_pt_addr1'_owner.

  intro H_abs.
  rewrite <- H_pt_addr1_map in H_abs.
  discriminate H_abs.
Qed.

Lemma pin_untrusted_os_isolation : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd) (type:ptype),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_pin_untrusted o pa type) s1' ->
  valid_exec ctxt s2 (Page_pin_untrusted o pa type) s2' ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa type H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages in *.
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_pin_untrusted in H_post1, H_post2.
  unfold post_all_pin in H_post1, H_post2.
  destruct H_post1 as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  destruct H_post2 as [ma2 [pg2 H_pg2_map [H_pg2_owner
    [H_pg2_content [p2m2 H_post_hyp2 [os2 H_os2_map H_s2']]]]]].

  generalize (osi_equivalent_symm ctxt s1 s2 osi); intro H_s2_equiv_s1.
  
  repeat split; try (rewrite H_s1'; rewrite H_s2'; simpl; assumption).
  
  (* Prove osi does not change *)
  rewrite H_s1'; rewrite H_s2'; simpl.
  case (os_ident_eq o osi).
  intros H_o_eq_osi.
  rewrite H_o_eq_osi.
  repeat rewrite lemma_map_add_apply.
  assert (os1 = os2).
  subst osi.
  rewrite <- H_s1_osi_s2  in H_os2_map.
  rewrite <- H_os2_map in H_os1_map.
  injection H_os1_map; auto.
  subst os2; reflexivity.

  intros H_osi_ne_aos.
  repeat rewrite lemma_map_add_apply_distinct; auto.

  apply pin_untrusted_os_isolation_hyper with s1 s2 o pa type; auto.
  apply pin_untrusted_os_isolation_hyper with s2 s1 o pa type; auto.
  apply pin_untrusted_os_isolation_pt_pages with s1 s2 o pa type; auto.
  apply pin_untrusted_os_isolation_pt_pages with s2 s1 o pa type; auto.
Qed.

Lemma unpin_untrusted_os_isolation_hyper : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_unpin_untrusted o pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_untrusted o pa) s2' ->
  osi_hyper_mapping_eq s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted, post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pa' pg' H_pg'_map [val H_pg'_content].
  case (os_ident_eq o osi).
  (* If o is osi, the hyp mapping that changes is relevant *)
  intros H_o_eq_osi; subst o.

  destruct H_pg'_map as [ma' H_ma'_hyp_map H_pg'_map].
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  simpl in H_p2m1'_map.
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro; subst p2m1'; clear H_p2m1'_map.
  assert (pa<>pa') as H_pa_ne_pa'.
  intro H_pa_eq_pa'.
  rewrite H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite map_drop_apply in H_ma'_hyp_map.
  discriminate H_ma'_hyp_map.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.

  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  apply H_pa_ne_pa'.
  apply hyp_mapped_same_paddr with ctxt s1 osi ma'; auto.
  rewrite <- H_ma1_eq_ma'.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.
  unfold hypervisor_mapping_madd; exists p2m1; auto.

  simpl in H_pg'_map.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pa' p2m2; auto.
  intros [ma H_ma_hyp_map H_ma_owner].
  clear H_valid_hyper.
  
  exists ma.
  unfold hypervisor_mapping_madd.
  exists (map_drop padd_eq p2m2 pa); auto.
  simpl.
  rewrite lemma_map_add_apply; auto.
  rewrite map_drop_apply_distinct; auto.
  assert (ma <> ma2) as H_ma_ne_ma2.
  intro H_ma_eq_ma2.
  apply H_pa_ne_pa'.
  apply hyp_mapped_same_paddr with ctxt s2 osi ma2; auto.
  exists p2m2; auto.
  rewrite <- H_ma_eq_ma2.
  unfold hypervisor_mapping_madd.
  exists p2m2; auto.
  simpl.
  rewrite lemma_map_add_apply_distinct; auto.
  elim H_s1_hyp_s2 with pa' pg'.
  intros ma2' H_ma2'_hyp_map H_pg'_map2.
  assert (ma2'=ma).
  apply hyp_mapped_same_maddr with s2 osi pa'; auto.
  exists p2m2; auto.
  subst ma2'; clear H_ma2'_hyp_map.
  assumption.

  unfold padd_page.
  exists ma'; auto.
  exists p2m1; auto.
  exists val; auto.
  
  intro H_abs.
  clear H_valid_hyper.
  elim H_s1_hyp_s2 with pa' pg'.
  intros ma H_ma_hyp_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_hyp_map.
  destruct H_ma_hyp_map as [p2m2' H_p2m2'_map H_ma_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_ma_map in H_abs.
  discriminate H_abs.
  unfold padd_page.
  exists ma'; auto.
  exists p2m1; auto.
  exists val; auto.

  (* Now prove the case when osi does not change*)
  intros H_o_ne_osi.
  unfold padd_page, hypervisor_mapping_madd.
  simpl.
  rewrite lemma_map_add_apply_distinct; try (rewrite <- H_active_os; assumption).
  destruct H_pg'_map as [ma' H_hyp_pa_ma1 H_pg'_value].
  unfold hypervisor_mapping_madd in H_hyp_pa_ma1.
  destruct H_hyp_pa_ma1 as [p2m1_osi H_p2m1_osi_map H_ma'_map].
  simpl in H_p2m1_osi_map.
  rewrite lemma_map_add_apply_distinct in H_p2m1_osi_map; auto.

  (* The new page in s1' is in ma1, it cannot be ma' *)
  cut (ma1 <> ma').
  intros H_ma1_ne_ma'.
  simpl in H_pg'_value.
  rewrite lemma_map_add_apply_distinct in H_pg'_value; auto.
  elim H_s1_hyp_s2 with pa' pg'.
  intros ma H_ma_map H_pg'_map2.
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m2_osi H_p2m2_osi_map H_ma_map].
  exists ma.
  exists p2m2_osi; auto.
  cut (ma <> ma2).
  intros H_ma_ne_ma2.
  rewrite lemma_map_add_apply_distinct; auto.

  intros H_ma_eq_ma2.
  (* Prove ma owner is osi *)
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct (H_valid_hyper osi pa' p2m2_osi H_p2m2_osi_map)
    as [[ma_2 H_ma_2_map H_ma_2_owner]|H_error].
  assert (ma_2=ma).
  rewrite <- H_ma_map in H_ma_2_map.
  injection H_ma_2_map; auto.
  subst ma_2.
  (* As ma2 has o as owner, it is not the case that ma=ma2 *)
  assert (maddress_owner (st_memory s2) ma2 = Os_PO o) as H_ma2_owner.
  unfold maddress_owner.
  elim H_valid_hyper with o pa p2m2; auto.
  intros [ma2' H_ma2'_map H_ma2'_owner].
  rewrite <- H_ma2_hyp_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.

  rewrite <- H_ma_eq_ma2 in H_ma2'_owner.
  rewrite H_ma_2_owner in H_ma2'_owner.
  injection H_ma2'_owner.
  intro H_abs.
  cut False; [intro; contradiction | apply H_o_ne_osi; auto].
  intro H_error.
  rewrite <- H_ma2_hyp_map in H_error; discriminate H_error.
  rewrite <- H_ma_eq_ma2 in H_ma2_owner.
  rewrite H_ma2_owner in H_ma_2_owner.
  injection H_ma_2_owner.
  intro H_abs.
  cut False; [intro; contradiction | apply H_o_ne_osi; auto].
  rewrite <- H_ma_map in H_error; discriminate H_error.
  
  (* Prove pg' is in pa' in s1 *)
  unfold padd_page, hypervisor_mapping_madd.
  exists ma'; auto.
  exists p2m1_osi; auto.

  (* Prove val is pg' content *)
  exists val; auto.

  (* Finally, prove ma1 <> ma' *)
  intros H_ma1_eq_ma'.
  (* Prove ma' owner is osi *)
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct (H_valid_hyper osi pa' p2m1_osi H_p2m1_osi_map) as [[ma_2 H_ma_2_map H_ma_2_owner]|H_error].
  assert (ma_2=ma').
  rewrite <- H_ma'_map in H_ma_2_map.
  injection H_ma_2_map; auto.
  subst ma_2.
  (* As ma1 owner is o, it is not the case that ma1=ma' *)
  assert (maddress_owner (st_memory s1) ma1 = Os_PO o) as H_ma1_owner.
  unfold maddress_owner.
  elim H_valid_hyper with o pa p2m1; auto.
  intros [ma1' H_ma1'_map H_ma1'_owner].
  rewrite <- H_ma1_hyp_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.

  rewrite H_ma1_eq_ma' in H_ma1'_owner.
  rewrite H_ma_2_owner in H_ma1'_owner.
  injection H_ma1'_owner.
  intro H_abs.
  cut False; [intro; contradiction | apply H_o_ne_osi; auto].
  intro H_error.
  rewrite <- H_ma1_hyp_map in H_error; discriminate H_error.
  rewrite H_ma1_eq_ma' in H_ma1_owner.
  rewrite H_ma1_owner in H_ma_2_owner.
  injection H_ma_2_owner.
  intro H_abs.
  cut False; [intro; contradiction | apply H_o_ne_osi; auto].
  rewrite <- H_ma'_map in H_error; discriminate H_error.
  assumption.
Qed.

Lemma unpin_untrusted_os_isolation_pt_pages : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_unpin_untrusted o pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_untrusted o pa) s2' ->
  osi_pt_pages_eq ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  unfold osi_equivalent, osi_equivalent_hyper_mapping,
    osi_equivalent_pt_pages, osi_pt_pages_eq in *.
  destruct H_s1_eq_s2 as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted in H_post1, H_post2.
  unfold post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin in H_pre1, H_pre2.
  elim H_pre1.
  intros H_untrusted_aos1 [H_aos1_hcall [H_hyper_running1
    [system1' H_sys1'_map [H_pa_ne_osi_curr_page1 [
      p2m1' H_p2m1'_map [ma1' H_ma1'_map [H_ma1'_ptype H_no_va_mapped1]]]]]]].
  unfold hypervisor_mapping in H_p2m1'_map.
  rewrite <- H_p2m1_map in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_eq_p2m1; subst p2m1'; clear H_p2m1'_map.
  elim H_pre2.
  intros H_untrusted_aos2 [H_aos2_hcall [H_hyper_running2
    [system2' H_sys2'_map [H_pa_ne_osi_curr_page2 [
      p2m2' H_p2m2'_map [ma2' H_ma2'_map [H_ma2'_ptype H_no_va_mapped2]]]]]]].
  unfold hypervisor_mapping in H_p2m2'_map.
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intro H_p2m2'_eq_p2m2; subst p2m2'; clear H_p2m2'_map.
  rewrite <- H_system1_map in H_sys1'_map.
  injection H_sys1'_map; intro; subst system1'; clear H_sys1'_map.
  rewrite <- H_system2_map in H_sys2'_map.
  injection H_sys2'_map; intro; subst system2'; clear H_sys2'_map.
  rewrite <- H_ma1_hyp_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.
  rewrite <- H_ma2_hyp_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.
  
  unfold padd_page.
  rewrite H_s1'; rewrite H_s2'; simpl.
  intros pt_padd pt_page1 mp1 [pt_addr1 H_pt_addr1_map H_pt_page1_map] H_pt_page1_content.

  case (os_ident_eq o osi).
  (* If o is osi, no osi owned PT page in memory changes *)
  intros H_osi_eq_aos; subst o.

  (* pt_padd cannot be the unpinned page *)
  assert (pt_padd<>pa) as H_pt_padd_ne_pa.
  intros H_pt_padd_eq_pa.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros H_p2m1'_value; subst p2m1'; clear H_p2m1'_map.
  rewrite H_pt_padd_eq_pa in H_pt_addr1_map.
  rewrite map_drop_apply in H_pt_addr1_map.
  discriminate H_pt_addr1_map.
  
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1' H_p2m1'_map H_pt_addr1_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intros; subst p2m1'; clear H_p2m1'_map.
  rewrite map_drop_apply_distinct in H_pt_addr1_map; auto.

  assert (ma1<>pt_addr1) as H_ma1_ne_pt_addr1.
  intros H_ma1_eq_pt_addr1.
  rewrite <- H_ma1_eq_pt_addr1 in H_pt_addr1_map.
  apply H_pt_padd_ne_pa.
  rewrite H_ma1_hyp_map in H_pt_addr1_map.

  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi p2m1 pt_padd pa; auto.

  elim H_s1_pt_s2 with pt_padd pt_page1 mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  unfold hypervisor_mapping_madd in H_pt_addr2_map.
  destruct H_pt_addr2_map as [p2m2' H_p2m2'_map H_pt_addr2_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.

  assert (ma2<>pt_addr2) as H_ma2_ne_pt_addr2.
  intro H_ma2_eq_pt_addr2.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_addr2_map.
  apply H_pt_padd_ne_pa.
  rewrite H_ma2_hyp_map in H_pt_addr2_map.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi p2m2 pt_padd pa; auto.

  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m2 pa); auto.
  rewrite map_drop_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists mp2; auto.
  
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.

  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1' H_p2m1'_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply in H_p2m1'_map.
  injection H_p2m1'_map; intro H_p2m1'_value.
  rewrite H_p2m1'_value in H_ma'_hyp_map.
  assert (pa<>pa') as H_pa_ne_pa'.
  intro H_pa_eq_pa'.
  rewrite H_pa_eq_pa' in H_ma'_hyp_map.
  rewrite map_drop_apply in H_ma'_hyp_map.
  discriminate H_ma'_hyp_map.

  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.
  apply H_pa_ne_pa'.

  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi p2m1 pa pa'; auto.

  rewrite <- H_ma'_hyp_map.
  rewrite <- H_ma1_eq_ma'; auto.
  
  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  rewrite map_drop_apply_distinct in H_ma'_hyp_map; auto.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].
  exists ma2'; auto.
  destruct H_ma2'_hyp_map as [p2m2' H_p2m2'_map H_ma2'_hyp_map].
  rewrite <- H_p2m2_map in H_p2m2'_map.
  injection H_p2m2'_map; intros; subst p2m2'; clear H_p2m2'_map.
  split.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply.
  exists (map_drop padd_eq p2m2 pa); auto.
  rewrite map_drop_apply_distinct; auto.

  assert (ma2'<>ma2) as H_ma2'_ne_ma2.
  intro H_ma2'_eq_ma2.
  rewrite H_ma2'_eq_ma2 in H_pg'_map2.
  apply H_pa_ne_pa'.

  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd H_valid_osexmd
      H_valid_untosexmd H_valid_virtmap H_valid_hyper H_valid_currpg H_va_has_pa.
  unfold injective_hyper_mappings in H_injective_hyper.
  rewrite H_injective_hyper with osi p2m2 pa pa'; auto.

  rewrite <- H_ma2'_hyp_map.
  rewrite H_ma2'_eq_ma2; auto.

  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.

  unfold page_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg'_map.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  rewrite <- H_pg'_map in H_pg'_type.
  assumption.

  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1; auto.

  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.

  (* Now prove the case when osi does not change*)
  intros H_o_ne_osi.
  unfold hypervisor_mapping_madd in H_pt_addr1_map.
  destruct H_pt_addr1_map as [p2m1_osi H_p2m1_osi_map H_pt_addr1_map].
  rewrite lemma_map_add_apply_distinct in H_p2m1_osi_map; auto.

  elim H_s1_pt_s2 with pt_padd pt_page1 mp1; auto.
  intros pt_page2 H_pt_page2_map [mp2 H_pt_page2_content H_virt_map_eq].
  unfold padd_page in H_pt_page2_map.
  destruct H_pt_page2_map as [pt_addr2 H_pt_addr2_map H_pt_page2_map].
  
  exists pt_page2.
  exists pt_addr2.
  unfold hypervisor_mapping_madd.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_pt_addr2.
  rewrite <- H_ma2_eq_pt_addr2 in H_pt_page2_map.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o pa p2m2; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma2' H_ma2'_map H_ma2_owner].
  rewrite <- H_ma2_hyp_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_pt_addr2_map as [p2m2_osi H_p2m2_osi_map H_pt_addr2_map].
  elim H_valid_hyper with osi pt_padd p2m2_osi; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma2' H_ma2'_map H_ma2_owner'].
  rewrite <- H_pt_addr2_map in H_ma2'_map.
  injection H_ma2'_map; intro; subst ma2'; clear H_ma2'_map.
  rewrite <- H_ma2_eq_pt_addr2 in H_ma2_owner'.
  rewrite H_ma2_owner' in H_ma2_owner.
  injection H_ma2_owner.
  intro H_abs; apply H_o_ne_osi; auto.

  intro H_error.
  rewrite <- H_pt_addr2_map in H_error.
  discriminate H_error.
  intro H_error.
  rewrite <- H_ma2_hyp_map in H_error.
  discriminate H_error.

  exists mp2; auto.
  intros va pa' ma' pg' H_ma'_map H_pg'_map H_va_accessible H_pg'_type H_ma'_hyp_map.
  unfold hypervisor_mapping_madd in H_ma'_hyp_map.
  destruct H_ma'_hyp_map as [p2m1'_osi H_p2m1'_osi_map H_ma'_hyp_map].
  rewrite lemma_map_add_apply_distinct in H_p2m1'_osi_map; auto.
  rewrite <- H_p2m1_osi_map in H_p2m1'_osi_map.
  injection H_p2m1'_osi_map; intro; subst p2m1'_osi; clear H_p2m1'_osi_map.
  
  assert (ma1<>ma') as H_ma1_ne_ma'.
  intro H_ma1_eq_ma'.
  rewrite H_ma1_eq_ma' in H_pg'_type.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply in H_pg'_type; auto.
  simpl in H_pg'_type; discriminate H_pg'_type.

  elim H_virt_map_eq with va pa' ma' pg'; auto.
  intros ma2' H_ma2'_map [H_ma2'_hyp_map H_pg'_map2].

  unfold hypervisor_mapping_madd in H_ma2'_map.
  destruct H_ma2'_hyp_map as [p2m2_osi H_p2m2_osi_map H_ma2'_hyp_map].

  exists ma2'; auto.
  split.
  unfold hypervisor_mapping_madd; rewrite lemma_map_add_apply_distinct; auto.
  exists p2m2_osi; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  intro H_ma2_eq_ma2'.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o pa p2m2; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma2'' H_ma2''_map H_ma2_owner].
  rewrite <- H_ma2_hyp_map in H_ma2''_map.
  injection H_ma2''_map; intro; subst ma2''; clear H_ma2''_map.
  generalize H_s2_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  destruct H_pt_addr2_map as [p2m2_osi' H_p2m2_osi'_map H_pt_addr2_map].
  rewrite <- H_p2m2_osi_map in H_p2m2_osi'_map.
  injection H_p2m2_osi'_map; intro; subst p2m2_osi'; clear H_p2m2_osi'_map.
  elim H_valid_hyper with osi pa' p2m2_osi; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma2'' H_ma2''_map H_ma2_owner'].
  rewrite <- H_ma2'_hyp_map in H_ma2''_map.
  injection H_ma2''_map; intro; subst ma2''; clear H_ma2''_map.
  rewrite <- H_ma2_eq_ma2' in H_ma2_owner'.
  rewrite H_ma2_owner' in H_ma2_owner.
  injection H_ma2_owner.
  intro H_abs; apply H_o_ne_osi; auto.

  intro H_error.
  rewrite <- H_ma2'_hyp_map in H_error.
  discriminate H_error.
  intro H_error.
  rewrite <- H_ma2_hyp_map in H_error.
  discriminate H_error.

  rewrite lemma_map_add_apply_distinct in H_pg'_map; auto.
  unfold page_type in H_pg'_type.
  rewrite lemma_map_add_apply_distinct in H_pg'_type; auto.
  unfold hypervisor_mapping_madd; exists p2m1_osi; auto.

  (* last goal *)
  unfold padd_page.
  exists pt_addr1.
  unfold hypervisor_mapping_madd.
  exists p2m1_osi; auto.
  rewrite lemma_map_add_apply_distinct in H_pt_page1_map; auto.
  
  intro H_ma1_eq_pt_addr1.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with o pa p2m1; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma1' H_ma1'_map H_ma1_owner].
  rewrite <- H_ma1_hyp_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.
  generalize H_s1_valid; Intros_Valid_Hyp;
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap
      H_valid_currpg H_injective_hyper H_va_has_pa.
  unfold valid_hypervisor in H_valid_hyper.
  elim H_valid_hyper with osi pt_padd p2m1_osi; auto.
  clear H_valid_hyper; intros H_valid_hyper.
  destruct H_valid_hyper as [ma1' H_ma1'_map H_ma1_owner'].
  rewrite <- H_pt_addr1_map in H_ma1'_map.
  injection H_ma1'_map; intro; subst ma1'; clear H_ma1'_map.
  rewrite <- H_ma1_eq_pt_addr1 in H_ma1_owner'.
  rewrite H_ma1_owner' in H_ma1_owner.
  injection H_ma1_owner.
  intro H_abs; apply H_o_ne_osi; auto.
  intro H_error.
  rewrite <- H_pt_addr1_map in H_error.
  discriminate H_error.
  intro H_error.
  rewrite <- H_ma1_hyp_map in H_error.
  discriminate H_error.
Qed.

Lemma unpin_untrusted_os_isolation : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd),
  valid_state ctxt s1 -> valid_state ctxt s2 ->
  osi_equivalent ctxt s1 s2 osi ->
  valid_exec ctxt s1 (Page_unpin_untrusted o pa) s1' ->
  valid_exec ctxt s2 (Page_unpin_untrusted o pa) s2' ->
  osi_equivalent ctxt s1' s2' osi.
Proof.
  intros s1 s2 s1' s2' osi o pa H_s1_valid H_s2_valid H_s1_eq_s2 H_exec1 H_exec2.
  generalize H_s1_eq_s2; intro H_s1_eq_s2'.
  destruct H_s1_eq_s2' as [H_s1_active_s2 [H_s1_osi_s2 [[H_s1_hyp_s2 H_s2_hyp_s1] [H_s1_pt_s2 H_s2_pt_s1]]]].
  unfold osi_equivalent, osi_equivalent_hyper_mapping, osi_equivalent_pt_pages in *.
  generalize H_exec1; generalize H_exec2.
  intros H_exec2' H_exec1'.
  inversion_clear H_exec1'.
  inversion_clear H_exec2'.
  rename H into H_pre1.
  rename H0 into H_post1.
  rename H1 into H_pre2.
  rename H2 into H_post2.
  unfold post_action, post_page_unpin_untrusted in H_post1, H_post2.
  unfold post_all_unpin in H_post1, H_post2.
  destruct H_post1 as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  destruct H_post2 as [p2m2 H_p2m2_map [ma2 H_ma2_hyp_map [system2 H_system2_map H_s2']]].

  unfold hypervisor_mapping in H_p2m1_map, H_p2m2_map.

  generalize (osi_equivalent_symm ctxt s1 s2); intro H_s2_equiv_s1.
  
  repeat split; try (rewrite H_s1'; rewrite H_s2'; simpl; assumption).
  
  (* Prove osi does not change *)
  rewrite H_s1'; rewrite H_s2'; simpl.
  case (os_ident_eq o osi).
  intros H_o_eq_osi.
  rewrite H_o_eq_osi.
  repeat rewrite lemma_map_add_apply.
  assert (system1 = system2).
  subst o.
  rewrite <- H_s1_osi_s2  in H_system2_map.
  rewrite <- H_system2_map in H_system1_map.
  injection H_system1_map; auto.
  subst system2; reflexivity.

  intros H_o_ne_osi.
  repeat rewrite lemma_map_add_apply_distinct; auto.

  apply unpin_untrusted_os_isolation_hyper with s1 s2 o pa; auto.
  apply unpin_untrusted_os_isolation_hyper with s2 s1 o pa; auto.
  apply unpin_untrusted_os_isolation_pt_pages with s1 s2 o pa; auto.
  apply unpin_untrusted_os_isolation_pt_pages with s2 s1 o pa; auto.
Qed.


End Untrusted_actions_equivalence.

