(***********************************************************************
 * File: State_valid.v
 * $Revision: 455 $
 * $Date: 2011-01-25 20:29:37 -0200 (Tue, 25 Jan 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This file specifies properties that have to be satisfied by the state
 * of the system
 * 
 **********************************************************************)
Require Import List.
Require Export Maps_proof.
Require Export Env.
Require Import Cache.

(** A machine address belongs at most to one OS (memory partitioning) *)
(** st_mem_owner : madd -> option os_ident *)

Definition maddress_owner (mem:system_memory)(ma:madd): page_owner :=
  let epage := map_apply madd_eq mem ma in
  match epage with
  | Error _ => No_Owner
  | Value pg => page_owned_by pg
  end.

Definition oss_map_apply := map_apply os_ident_eq.
Definition active_os_info (ois:os_ident)(s:state) := 
  let ossmap := st_oss s in
  oss_map_apply os ossmap ois.

(** Os Active Memory Mapping *)
Definition os_active_mapping (ma:madd) (s:state) : Prop :=
  let aos := st_active_os s in	
  exists2 mapphyma, 
    Value _ mapphyma = map_apply os_ident_eq (st_hypervisor s) aos &
    exists2 os, Value _ os = active_os_info aos s &
                let ospgpa := os_current_page os in
                Value _ ma = map_apply padd_eq mapphyma ospgpa.

Definition hypervisor_mapping
  (osi:os_ident) (hyp:hypervisor) (mp:mapping padd madd) : Prop :=
  Value _ mp = map_apply os_ident_eq hyp osi.
               
(* Gives the madd of a padd mapped in hypervisor mapping of the given OS. *)
Definition hypervisor_mapping_madd
  (hyp:hypervisor) (osi:os_ident) (pa:padd) (ma:madd) : Prop :=
  exists2 p2m, Value _ p2m = map_apply os_ident_eq hyp osi &
               Value _ ma = map_apply padd_eq p2m pa.

Definition vadd_mapping
    (ma:madd) (smem:system_memory) (mp:mapping vadd madd) : Prop :=
  exists2 pg, Value _ pg = map_apply madd_eq smem ma &
              PT_c mp = page_content pg.

(* The machine address of the current memory mapping of the given OS *)
Definition os_current_vadd_mapping_madd
  (s:state) (osi:os_ident) (ma:madd) : Prop :=
  exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
               exists2 p2m, hypervisor_mapping osi (st_hypervisor s) p2m &
                            Value _ ma = map_apply padd_eq p2m (os_current_page system).

(* The machine address of the current memory mapping of the active OS *)
Definition current_vadd_mapping_madd
  (s:state) (ma:madd) : Prop :=
  os_current_vadd_mapping_madd s (st_active_os s) ma.

(* The current memory mapping of the active OS *)
Definition current_vadd_mapping
  (s:state) (mp: mapping vadd madd) : Prop :=
  exists2 ma, current_vadd_mapping_madd s ma &
              vadd_mapping ma (st_memory s) mp. 

(* ma is the image of va in the current memmap of osi. *)
Definition va_mapped_to_ma
  (s:state) (osi:os_ident) (va:vadd) (ma:madd) : Prop :=
     exists2 pt_addr, os_current_vadd_mapping_madd s osi pt_addr &
        exists2 mp, vadd_mapping pt_addr (st_memory s) mp &
           Value vadd ma = map_apply vadd_eq mp va.

(** va is mapped in the Active Memory Mapping *)
Definition vadd_mapped (va:vadd) (ma:madd) (s:state) : Prop :=
  os_active_mapping ma s ->
  exists2 mp, vadd_mapping ma (st_memory s) mp & 
              is_Value (map_apply vadd_eq mp va).

(** va is mapped to pg for the AOS *)
Definition va_mapped_to_ma_cache (s:state) (va:vadd) : option madd :=
  match map_apply vadd_eq (st_tlb s) va with
    | Value ma => Some ma
    | _ =>
      match map_apply os_ident_eq (st_oss s) (st_active_os s) with
        | Value aos =>
          match map_apply os_ident_eq (st_hypervisor s) (st_active_os s) with
            | Value p2m =>
              match map_apply padd_eq p2m (os_current_page aos) with
                | Value pt_addr =>
                  match map_apply madd_eq (st_memory s) pt_addr with
                    | Value pt_page =>
                      match page_content pt_page with
                        | PT_c mp =>
                          match map_apply vadd_eq mp va with
                            | Value ma => Some ma
                            | _ => None
                          end
                        | _ => None
                      end
                    | _ => None
                  end
                | _ => None
              end
            | _ => None
          end
        | _ => None
      end
  end.
  
Definition va_mapped_to_pg_cache (s:state) (va:vadd) : option page :=
  match map_apply vadd_eq (st_cache s) va with
    | Value pg => Some pg 
    | _ =>
      match va_mapped_to_ma_cache s va with
        | Some ma => 
          match map_apply madd_eq (st_memory s) ma with
            | Value pg => Some pg
            | _ => None
          end
        | _ => None
      end
  end.



  (*         match map_apply os_ident_eq (st_oss s) (st_active_os s) with *)
  (*           | Value aos => *)
  (*             match map_apply os_ident_eq (st_hypervisor s) (st_active_os s) with *)
  (*               | Value p2m => *)
  (*                 match map_apply padd_eq p2m (os_current_page aos) with *)
  (*                   | Value pt_addr => *)
  (*                     match map_apply madd_eq (st_memory s) pt_addr with *)
  (*                       | Value pt_page => *)
  (*                         match page_content pt_page with *)
  (*                           | PT_c mp => *)
  (*                             match map_apply vadd_eq mp va with *)
  (*                               | Value ma => *)
  (*                                 match map_apply madd_eq (st_memory s) ma with *)
  (*                                   | Value pg => Some pg *)
  (*                                   | _ => None *)
  (*                                 end *)
  (*                               | _ => None *)
  (*                             end *)
  (*                           | _ => None *)
  (*                         end *)
  (*                       | _ => None *)
  (*                     end *)
  (*                   | _ => None *)
  (*                 end *)
  (*               | _ => None *)
  (*             end *)
  (*           | _ => None *)
  (*         end *)
  (*     end *)
  (* end. *)
  

(** Current call (if any) of the given OS *)
Definition get_os_hypercall
  (s:state) (osi:os_ident) (hcall:option Hyperv_call) : Prop :=
  exists2 system, Value _ system = map_apply os_ident_eq (st_oss s) osi &
      os_hcall system = hcall.

(** Create an empty page of the given type with given owner. *)
Definition new_page (type:option ptype) (owner:page_owner) : page :=
  match type with
    | Some RW => Page (RW_c None) owner
    | Some PT => Page (PT_c (map_empty vadd madd)) owner
    | None    => Page Other_c owner
  end.

Definition os_accessible ctxt va := (ctxt_vadd_accessible ctxt) va = true.
Definition os_running s := st_aos_activity s = OSA_running.
Definition hyper_running s := st_aos_activity s = OSA_waiting.
Definition trusted_aos ctxt s := ctxt_oss ctxt (st_active_os s) = true.


(* *** Valid state conditions *** *)


(* Trusted OSs do not make hypercalls *)

Definition trusted_os_not_hypercall 
  (ctxtos:os_ident -> bool)(ossmap:oss_map) : Prop :=
  forall (osi:os_ident) (tos:os),
    ctxtos osi = true -> Value os_ident tos = map_apply os_ident_eq ossmap osi ->
    os_hcall tos = None.

(* If active OS is running, it should not have a hypercall. *)
Definition running_os_not_hypercall
  (aosi:os_ident) (activity:os_activity) (ossmap:oss_map) : Prop :=
  exists2 aos:os, 
    Value os_ident aos = map_apply os_ident_eq ossmap aosi &
    (activity = OSA_running -> os_hcall aos = None).

(* If hypervisor is running (aos is waiting) then execution mode is
   supervisor *)
Definition valid_hyper_exec_mode
  (activity:os_activity) (mode:exec_mode) : Prop :=
  activity = OSA_waiting -> mode = EXMD_svc.


(* If trusted OS is running, execution mode is supervisor *)
Definition valid_trusted_os_exec_mode
  (aosi:os_ident) (mode:exec_mode) (ctxtos:os_ident -> bool) : Prop :=
  ctxtos aosi = true -> mode = EXMD_svc.


(* If untrusted OS is running, execution mode is user *)
Definition valid_untrusted_os_exec_mode
  (aosi:os_ident) (activity:os_activity) (mode:exec_mode) (ctxtos:os_ident -> bool) : Prop :=
  ctxtos aosi = false -> activity = OSA_running -> 
  mode = EXMD_usr.

(* Hypervisor should map an OS padd to a madd owned by that OS. *)
Definition valid_hypervisor
  (hyper:hypervisor) (smem:system_memory) : Prop :=
  forall (osi:os_ident) (pa:padd),
    (* p2m is the hypervisor mapping from physical address to machine address *)
     forall p2m:map_physical_machine_addr,
       Value os_ident p2m = map_apply os_ident_eq hyper osi ->
      (* For every possible physical address, get the machine address
      from the mapping, and check that it is owned by the OS *)
       (exists2 ma:madd,
         Value padd ma = map_apply padd_eq p2m pa &
         maddress_owner smem ma = Os_PO osi)       (* Should we allow to map to an hypervisor owned page? *)
         \/ map_apply padd_eq p2m pa = Error madd pa. (* No mapping for that padd *)

Definition valid_virtual_mapping
  (smem:system_memory) (ctxt:context) : Prop :=
  forall osi:os_ident,
    forall pt_addr:madd, forall mp, 
      forall pt, Value _ pt = map_apply madd_eq smem pt_addr ->
        PT_c mp = page_content pt ->
        page_owned_by pt = Os_PO osi ->
        forall va,
          forall ma, Value _ ma = map_apply vadd_eq mp va ->
            exists2 pg, Value _ pg = map_apply madd_eq smem ma &
              (os_accessible ctxt va -> page_owned_by pg = Os_PO osi) /\
              (~os_accessible ctxt va -> page_owned_by pg = Hyp_PO).

(* Current page table of any OS is of type PT, and owned by that OS. *)
Definition valid_current_page
  (hyper:hypervisor) (smem:system_memory) (ossmap:oss_map) : Prop :=
  forall osi:os_ident,
    forall system:os, 
      Value os_ident system = map_apply os_ident_eq ossmap osi ->
    (* p2m is the hypervisor mapping from physical address to machine address *)
    forall p2m:map_physical_machine_addr,
      Value os_ident p2m = map_apply os_ident_eq hyper osi ->
    (* pt_madd is the madd of the current page table. Check its type and owner *)
    exists2 pt_madd:madd, 
      Value padd pt_madd = map_apply padd_eq p2m (os_current_page system) &
      page_type smem pt_madd = Some PT /\
      maddress_owner smem pt_madd = Os_PO osi.

(* Hypervisors mappings must be injective *)
Definition injective_hyper_mappings
  (hyper:hypervisor) : Prop :=
  forall (osi:os_ident) (p2m:map_physical_machine_addr),
    (* p2m is the hypervisor mapping from physical address to machine address *)
       Value os_ident p2m = map_apply os_ident_eq hyper osi ->
       injective_mapping padd_eq p2m.

(*madds in page tables, must be assigned to a pinned padd by hypervisor*)
Definition va_has_valid_pa (smem:system_memory) (hyper:hypervisor) 
  (ctxt:context) : Prop :=
  forall osi pt_padd pt_addr pt_page mp va ma,
    Value _ pt_page = map_apply madd_eq smem pt_addr ->
      hypervisor_mapping_madd hyper osi pt_padd pt_addr ->
      PT_c mp = page_content pt_page ->
      Value _ ma = map_apply vadd_eq mp va ->
      os_accessible ctxt va ->
    (exists pa, hypervisor_mapping_madd hyper osi pa ma).


Definition cache_included (s:state) : Prop :=
  let aos := st_active_os s in
  let mem := st_memory s in
  let cache := st_cache s in
    forall va pg, Value _ pg = map_apply vadd_eq cache va ->
      (exists ma, va_mapped_to_ma s aos va ma /\
        page_type mem ma = Some RW /\
        Value _ pg = map_apply madd_eq mem ma).

Definition valid_cache (s:state) : Prop :=
  map_size (st_cache s) <= max_cache /\
  cache_distinct (st_cache s) /\
  cache_included s.


Definition tlb_included (s:state) : Prop :=
  let aos := st_active_os s in
  let tlb := st_tlb s in
    forall va ma, Value _ ma = map_apply vadd_eq tlb va ->
      va_mapped_to_ma s aos va ma.

Definition valid_tlb (s:state) : Prop :=
  map_size (st_tlb s) <= max_tlb /\
  tlb_distinct (st_tlb s) /\
  tlb_included s.


(* System properties and state validity *)

Section Valid_State.

Variable ctxt : context.

(** Valid state *)
Definition valid_state (st:state) : Prop :=
  let vadd_accessible := ctxt_vadd_accessible ctxt in
  let oss := ctxt_oss ctxt in
  let mem := st_memory st in
  let ossmap := st_oss st in
  let aos := st_active_os st in
  let activity := st_aos_activity st in
  let mode := st_exec_mode st in
  let hyper := st_hypervisor st in

    (*******************************************************)
    (* It doesn't seem necessary to require this property. *)
    (*******************************************************)
    (* The mapping from a madd to a page should be injective *)
    (* injective_mapping madd_eq mem /\ *)

    (* Trusted OSs do not make hypercalls *)
    trusted_os_not_hypercall oss ossmap /\
    (* If active OS is running, it should not have a hypercall. *)
    running_os_not_hypercall aos activity ossmap /\
    (* Hypervisor always runs in supervisor mode *)
    valid_hyper_exec_mode activity mode /\
    (* Trusted OSs also run in supervisor mode *)
    valid_trusted_os_exec_mode aos mode oss /\ 
    (* Untrusted OSs must run in user mode (if they are not waiting) *)
    valid_untrusted_os_exec_mode aos activity mode oss /\
    (* Hypervisor should map an OS padd to a madd owned by that OS. *)
    valid_hypervisor hyper mem /\
    (* If a mapped virtual address is accessible by an OS, it is owned
       by that OS. If a mapped virtual address is not accessible by an
       OS, it is owned by the Hypervisor. *)
    valid_virtual_mapping mem ctxt /\
    (* Current page table of any OS is of type PT, and owned by that OS. *)
    valid_current_page hyper mem ossmap /\
    (*Hypervisor Mappings must be injective for all os*)
    injective_hyper_mappings hyper /\
    (*Virtual memory must be assigned pinned Phisical memory*)
    va_has_valid_pa mem hyper ctxt /\

    (* Cache and TLB validity *)
    valid_cache st /\
    valid_tlb st.


End Valid_State.