(***********************************************************************
 * File: Write_Isolation.v
 * $Revision: 399 $
 * $Date: 2010-12-26 17:56:13 -0200 (Sun, 26 Dec 2010) $
 * $LastChangedBy: cluna $
 ***********************************************************************
 * This files proves write isolation property on valid executions. 
 **********************************************************************)

Require Import State_valid.
Require Import Env.
Require Import Exec.
Require Import Exec_invariant.
Require Export Observers.

Section Write_Isolation.

Variable ctxt : context.

(** Lemmas for write_isolation *)

(* If states differ only in one page owned by the AOS,
   all pages owned by other OSs or by the hypervisor do not change. *) 
Lemma same_not_owned_page :
  forall (s s':state) (new_ma:madd) (old_pg new_pg:page),
    Value _ old_pg = map_apply madd_eq (st_memory s) new_ma -> 
    st_memory s' = map_add madd_eq (st_memory s) new_ma new_pg ->
    Os_PO (st_active_os s) = page_owned_by old_pg ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
      ~ owner_of_page pg (st_active_os s) ->
      Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' new_ma old_pg new_pg H_old_pg_map H_smem'_map H_old_pg_owner
    ma pg H_pg_map H_pg_owner.

  rewrite H_smem'_map.

  rewrite H_pg_map; symmetry.
  apply lemma_map_add_apply_distinct.

  intro H_ma_eq_new_ma.

  assert (pg = old_pg).
  
  subst new_ma.
  rewrite <- H_pg_map in H_old_pg_map.
  injection H_old_pg_map; auto.

  subst old_pg; eauto. 
Qed.

 
(* Executing a valid write does not change other memory. *)
Lemma lemma_write_action_isolation :
  forall (s s':state) (va_write:vadd) (val:value),
    valid_state ctxt s ->
    valid_exec ctxt s (Write va_write val) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
      ~ owner_of_page pg (st_active_os s) ->
      Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' va_write val H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_write, post_all_write in H_post.
  destruct H_post as (ma_write, H_ma_write_map, (smem', H_up_rw, H_assign_smem')).

  unfold update_rw_page in H_up_rw.
  destruct H_up_rw as (pg_write, H_pg_write_map, H_smem'_map). (* pg_write is the page that
                                                                  the action changes *)

  (* Since this is a valid write, pg_write must be owned by the AOS *)
  assert (Os_PO (st_active_os s) = page_owned_by pg_write).
  unfold pre_action, pre_write in H_pre.
  destruct H_pre as (H_os_running, (H_os_accessible, (ma_write', H_ma_write'_map, H_ptype))).

  assert (va_mapped_to_ma s (st_active_os s) va_write ma_write) as H_ma_write_map2.
  unfold va_mapped_to_ma_cache in H_ma_write_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) va_write) as [it H_ma_write_map2].
  induction (map_apply vadd_eq (st_tlb s) va_write).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  rewrite <- H_ma_write_map2 in H_ma_write_map.

  induction it.
  injection H_ma_write_map; intro; subst v.
  apply valid_state_included_tlb with ctxt; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  assert (exists pt_addr, Value _ pt_addr = map_apply padd_eq v0 (os_current_page v)) as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v0 (os_current_page v)); try discriminate.
  exists v1; auto.
  rewrite <-  H_pt_addr_map in H_ma_write_map.
  exists pt_addr.
  exists v; auto.
  exists v0; auto.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  assert (exists mp, PT_c mp = (page_content v1)) as [mp H_mp_map].
  induction (page_content v1); try discriminate.
  exists m; auto.
  rewrite <- H_mp_map in H_ma_write_map.
  exists mp.
  exists v1; auto.
  induction (map_apply vadd_eq mp va_write); try discriminate.
  injection H_ma_write_map; intro; subst v2; auto.

  destruct H_ma_write_map2 as (pt_addr, H_curr_pt, (mp, H_mp_map, H_ma_write_map2)).
  generalize H_s_valid; clear H_s_valid; Intros_Valid_Hyp.
  unfold valid_virtual_mapping in H_valid_virtmap.

  unfold os_current_vadd_mapping_madd in H_curr_pt.
  destruct H_curr_pt as (aos, H_aos_map, (p2m, H_hyp_mapping, H_pt_addr_map)).
  unfold hypervisor_mapping in H_hyp_mapping.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as (pt_page, H_pt_page_map, H_pt_page_content).
  
  destruct H_valid_virtmap with (osi := st_active_os s) (pt_addr:=pt_addr) (pt:=pt_page)
    (mp:=mp) (va:=va_write) (ma:=ma_write) as [pg' H_pg'_map (H_owner, H_hyp)]; auto.

  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (osi:=(st_active_os s)) (pa:=(os_current_page aos)) (p2m:=p2m); auto.
  destruct H as (pt_addr', H_pt_addr'_map, H_pt_addr'_owner).
  assert (pt_addr'=pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map; injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr'_owner.
  rewrite <- H_pt_page_map in H_pt_addr'_owner.
  assumption.
  cut False. contradiction.
  rewrite H in H_pt_addr_map.
  discriminate H_pt_addr_map.
  
  (* pg' and pg_write are the pages in smem at address ma_write *)
  assert (pg_write = pg').
  rewrite <- H_pg'_map in H_pg_write_map; injection H_pg_write_map; auto.
  
  subst pg'.
  symmetry; apply H_owner; auto.

  (* Now that we know that pg_write is owned by the AOS, use the lemma *)
  apply same_not_owned_page with
      s ma_write pg_write (Page (RW_c (Some val)) (page_owned_by pg_write));
    auto.

  destruct H_assign_smem' as [pg' H_pg'_ma
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  rewrite H_s'; rewrite H_smem'_map; simpl; auto.
Qed.

(* Executing a valid new_trusted does not change other memory. *)
Lemma lemma_new_trusted_isolation :
  forall (s s':state) (new_va:vadd) (new_pa:padd),
    valid_state ctxt s ->
    valid_exec ctxt s (New_trusted new_va new_pa) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->      
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof. 
  intros s s' new_va new_pa H_s_valid H_valid_exec
    ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_new_trusted, post_all_new in H_post.
  destruct H_post as (new_ma, H_hyp_map_madd, (pt_addr, H_curr_map,
    (smem', H_up_pt_page, (aos, H_aos_map, H_s')))).

  unfold update_pt_page in H_up_pt_page.

  (* pt_page is the current PT of the AOS. *)
  destruct H_up_pt_page as (pt_page, H_pt_page_map, (oldmap, H_pt_page_content, H_smem')).

  
  apply same_not_owned_page with s pt_addr pt_page
    (Page (PT_c (map_add vadd_eq oldmap new_va new_ma)) (page_owned_by pt_page)); auto.

  (* We now need to prove that pt_page is the only page that changes in s'. *)
  rewrite H_s'; rewrite H_smem'; simpl; auto.

  (* Now, prove that AOS is the owner of pt_page, using valid_current_page *)
  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
    H_valid_untosexmd H_valid_hyper H_valid_virtmap.

  unfold os_current_vadd_mapping_madd in H_curr_map.
  destruct H_curr_map as (aos', H_aos'_map, (p2m, H_hyp_map, H_pt_addr_map)).

  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as (pt_addr', H_pt_addr'_map, (H_pt_addr_type, H_pt_addr_owner)); auto.

  (* Prove that objects in valid_current_page are the same we had *)
  assert (aos = aos').
  rewrite <- H_aos'_map in H_aos_map.
  injection H_aos_map; auto.
  subst aos'; clear H_aos'_map.

  assert (pt_addr' = pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map.
  injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.

  (* Now we know the owner of pt_page *)
  rewrite <- H_pt_addr_owner.

  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
Qed.
  
(* Executing a valid del_trusted does not change other memory. *)
Lemma lemma_del_trusted_isolation :
  forall (s s':state) (va:vadd),
    valid_state ctxt s ->
    valid_exec ctxt s (Del_trusted va) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
     Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' del_va H_s_valid H_valid_exec
    ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_del_trusted, post_all_del in H_post.
  destruct H_post as (pt_addr, H_curr_map_madd, (smem', H_drop_pt_page, H_s')).

  unfold drop_va_pt_page in H_drop_pt_page.
  
  (* pt_page is the current PT of the AOS. *)
  destruct H_drop_pt_page as (pt_page, H_pt_page_map, (oldmap, H_pt_page_content, H_smem')).

  apply same_not_owned_page with s pt_addr pt_page
    (Page (PT_c (map_drop vadd_eq oldmap del_va)) (page_owned_by pt_page)); auto.

  (* We now need to prove that pt_page is the only page that changes in s'. *)
  rewrite H_s'; rewrite H_smem'; simpl; auto.

  (* Now, prove that AOS is the owner of pt_page, using valid_current_page *)
  generalize H_s_valid; Intros_Valid_Hyp.
  clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
    H_valid_untosexmd H_valid_hyper H_valid_virtmap.

  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd in H_curr_map_madd.

  destruct H_curr_map_madd as (aos, H_aos_map, (p2m, H_hyp_map, H_pt_addr_map)).

  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as (pt_addr', H_pt_addr'_map, (H_pt_addr_type, H_pt_addr_owner)); auto.

  (* Prove that objects in valid_current_page are the same we had *)
  assert (pt_addr' = pt_addr).
  rewrite <- H_pt_addr'_map in H_pt_addr_map.
  injection H_pt_addr_map; auto.
  subst pt_addr'; clear H_pt_addr'_map.

  (* Now we know the owner of pt_page *)
  rewrite <- H_pt_addr_owner.

  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
Qed.

(* Executing a valid lswitch (trusted or untrusted) does not change
   other memory. *)
Lemma lemma_lswitch_isolation :
  forall (s s':state) (osi:os_ident) (pa:padd) (a:action),
    valid_state ctxt s ->
    ((a = (Lswitch_trusted pa)) \/ (a = (Lswitch_untrusted osi pa))) -> 
    valid_exec ctxt s a s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' osi pa a H_s_valid H_action H_valid_exec ma pg
    H_pg_map H_pg_owner.

  (* Memory doesn't change, so we only need to show
       (st_memory s' = st_memory s)
     for both actions. *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  destruct H_action as [H_lswitch_trusted | H_lswitch_untrusted]; subst a;
    unfold post_action, post_lswitch_trusted, post_lswitch_untrusted, post_all_lswitch in H_post;
      destruct H_post as (aos, H_aos_map, H_s'); rewrite H_s'; auto.
Qed.

(* Executing a valid page_pin_trusted does not change other memory used (reserved). *)
Lemma lemma_page_pin_trusted_isolation :
  forall (s s':state) (new_pa:padd) (page_type:ptype),
    valid_state ctxt s ->
    valid_exec ctxt s (Page_pin_trusted new_pa page_type) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->
    page_owned_by pg <> No_Owner ->
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' new_pa page_type H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner H_pg_no_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as (new_ma, (new_pg, H_new_pg_map, (H_new_pg_no_own,
      (H_new_pg_no_content, (p2m, H_hyp_map, (aos, H_aos_map, H_s')))))).

  (* Here we cannot use lemma same_not_owned_page, since the page that
     changes is not owned by AOS. We prove that new_ma (the madd of
     the page that is pinned), is different from any ma owned by other
     os, because it is owned by no-one. *)

  assert (new_ma <> ma).
  intros H_new_ma_eq_ma.
  subst new_ma.
  assert (new_pg = pg).
  rewrite <- H_new_pg_map in H_pg_map; injection H_pg_map; auto.

  subst new_pg.
  unfold owner_of_page in H_pg_owner.
  rewrite H_new_pg_no_own in H_pg_owner.
  rewrite H_new_pg_no_own in H_pg_no_owner.
  elim H_pg_no_owner; auto.

  (* Now we know that pg does not change, since the only memory change
     is in new_ma. *)
  replace (st_memory s') with (map_add madd_eq (st_memory s) new_ma
    (new_page (Some page_type) (Os_PO (st_active_os s)))).

  rewrite H_pg_map.
  symmetry; apply lemma_map_add_apply_distinct; auto.
  rewrite H_s'.
  simpl.
  reflexivity.
Qed.

(* Executing a valid page_unpin_trusted does not change other memory. *)
Lemma lemma_page_unpin_trusted_isolation :
  forall (s s':state) (pa:padd),
    valid_state ctxt s ->
    valid_exec ctxt s (Page_unpin_trusted pa) s' ->
    (forall (ma:madd) (pg:page), Value _ pg = map_apply madd_eq (st_memory s) ma ->
    ~ owner_of_page pg (st_active_os s) ->      
    Value _ pg = map_apply madd_eq (st_memory s') ma).
Proof.
  intros s s' unpin_pa H_s_valid H_valid_exec ma pg H_pg_map H_pg_owner.

  inversion_clear H_valid_exec as [s_vexec s'_vexec act_vexec H_pre H_post].

  unfold post_action, post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as (p2m, H_hyp_map, (unpin_ma, H_unpin_ma_map, (aos, H_aos_map, H_s'))).

  unfold pre_action, pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as (H_os_running, (H_trusted_aos, (aos', H_aos'_map,
    (H_pa_not_curr_pt, (p2m', H_hyp_map', (unpin_ma', H_unpin_ma'_map,
      (H_page_type, H_no_va_ma))))))).

  (* Prove that objects in pre and post are the same *)
  assert (p2m = p2m').
  apply lemma_same_hypervisor_mapping with (st_active_os s) (st_hypervisor s); auto.
  subst p2m'; clear H_hyp_map'.
  assert (unpin_ma = unpin_ma').
  rewrite <- H_unpin_ma'_map in H_unpin_ma_map.
  injection H_unpin_ma_map; auto.
  subst unpin_ma'; clear H_unpin_ma'_map.

  (* Obtain the changed page using lemma_page_type_exists_page *)
  assert (exists unpin_pg:page, Value madd unpin_pg = map_apply madd_eq (st_memory s) unpin_ma).
  apply lemma_page_type_exists_page; auto.
  destruct H as (unpin_pg, H_unpin_pg_map).

  (* Now we can apply lemma same_not_owned_page, using the changed page. *)
  apply same_not_owned_page with s unpin_ma unpin_pg (new_page None No_Owner); auto.

  (* Prove that page in unpin_ma is the only page that changes in s'. *)
  rewrite H_s'; simpl; reflexivity.

  (* Now, prove that AOS is the owner of unpin_pg, using valid_hypervisor. *)
  generalize H_s_valid; Intros_Valid_Hyp; clear H_trstos_nothypcall
    H_running_nothypcall H_valid_exmd H_valid_osexmd H_valid_untosexmd
    H_valid_virtmap H_valid_currpg.

  unfold valid_hypervisor in H_valid_hyper.
  destruct H_valid_hyper with (st_active_os s) unpin_pa p2m as
    [(unpin_ma', H_unpin_ma'_map, H_unpin_ma'_owner) | H_unpin_pa_error]; auto.

  (* valid_hypervisor is an or. Either the page is owned by AOS, or
     the page is not mapped. *)
  clear H_valid_hyper.
  
  assert (unpin_ma' = unpin_ma);
    [rewrite <- H_unpin_ma'_map in H_unpin_ma_map; injection H_unpin_ma_map; auto |
      subst unpin_ma'; clear H_unpin_ma'_map].

  unfold maddress_owner in H_unpin_ma'_owner.
  rewrite <- H_unpin_ma'_owner; rewrite <- H_unpin_pg_map; reflexivity.

  (* Absurd case where the page is not mapped in p2m.
     We know by H_unpin_ma_map that it is. *)
  cut False; [contradiction |
    rewrite H_unpin_pa_error in H_unpin_ma_map; discriminate].
Qed.

(** No action executed by an OS in a valid state changes memory that 
    does not belong to the AOS.
*)


(* Property write_isolation *)
Definition write_isolation' (s s':state) : Prop :=
 forall (a:action),
       valid_state ctxt s ->
       ~(hyper_running s) ->
       valid_exec ctxt s a s' ->     
       forall (ma:madd) (pg:page), 
          content_of_address s ma pg ->
          ~ owner_of_page pg (st_active_os s) ->
          page_owned_by pg <> No_Owner ->
          content_of_address s' ma pg.


Theorem write_isolation_proof' : forall (s s':state), write_isolation' s s'.
Proof.
  intros s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg H_pg_map H_pg_owner H_pg_no_owner.

  (* Prove by cases on the action. Actions where the state does not
     change, or the hyper is running are proved directly. Others have
     their own lemma. *)
  induction a.

  (* Silent action. *)
  inversion_clear H_valid_exec.
  unfold post_action, post_silent, same_state in H0.
  destruct H0 as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_mem System_Cache System_Memory.
  rewrite <- H_mem; try (intro; discriminate); auto.

  (* Read action (memory does not change) *)
  inversion_clear H_valid_exec.
  unfold post_action, post_read, post_read_hyper,
    post_all_read in H0.
  unfold pre_action, pre_read, pre_all_read in H.
  destruct H as [[ma' [H_ma'_map H_ptype]] [H_accessible H_running]].
  assert (exists pg', Value _ pg' = map_apply madd_eq (st_memory s) ma') as [pg' H_pg'_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma').
  exists v0; auto.
  discriminate H_ptype.
  assert (Some ma' = va_mapped_to_ma_cache s v) as H_ma'_cache_map.

  unfold va_mapped_to_ma_cache.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v0); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma'_map.
  destruct H_ma'_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]] [mp [pt_page H_pt_page_map H_mp_map] H_ma'_map]].
  induction it.
  elim valid_state_included_tlb with ctxt s v v0; auto.
  intros pt_addr' H_pt_addr'_map [mp' H_mp'_map H_v0_map].
  unfold os_current_vadd_mapping_madd, hypervisor_mapping, vadd_mapping in H_pt_addr'_map.
  destruct H_pt_addr'_map as [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]].
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map; intro;
    subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro;
    subst p2m'; clear H_p2m'_map.
  unfold vadd_mapping in H_mp'_map.
  destruct H_mp'_map as [pt_page' H_pt_page'_map H_mp'_map].
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map; intro;
    subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map; injection H_pt_page'_map; intro;
    subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_mp'_map; injection H_mp'_map; intro;
    subst mp'; clear H_mp'_map.
  rewrite <- H_ma'_map in H_v0_map; injection H_v0_map; intro; subst v0; auto.

  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  injection H_aos_map; intro; subst v0.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  injection H_p2m_map; intro; subst v0.
  induction (map_apply padd_eq p2m (os_current_page aos)); try discriminate.
  injection H_pt_addr_map; intro; subst v0.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  injection H_pt_page_map; intro; subst v0.
  induction (page_content pt_page); try discriminate.
  injection H_mp_map; intro; subst m.
  induction (map_apply vadd_eq mp v); try discriminate.
  injection H_ma'_map; intro; subst v0; auto.

  destruct H0 with ma' pg' as [new_cache H_new_cache
    [new_tlb H_new_tlb H_assign]]; auto.

  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v0); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with ctxt s v v0; auto.
  intros ma'' H_ma''_map H_v0_map.
  assert (ma' = ma'').
  elim va_mapped_same_ma with s v ma' ma''; auto.
  subst ma''; clear H_ma''_map.
  rewrite <- H_pg'_map in H_v0_map.
  injection H_v0_map; intro; subst v0; auto.

  rewrite <- H_ma'_cache_map.
  rewrite <- H_pg'_map; auto.

  elim H_assign with System_Memory; auto.
  unfold state_component_interp.
  intros [H_triv H_new_mem].
  unfold content_of_address.
  rewrite <- H_new_mem; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; intro H_abs'; discriminate H_abs'.

  (* Read hyper action (memory does not change) *)
  inversion_clear H_valid_exec.
  unfold post_action, post_read, post_read_hyper,
    post_all_read in H0.
  unfold pre_action, pre_read_hyper, pre_all_read in H.
  destruct H as [[ma' [H_ma'_map H_ptype]] H_hyper_running].
  absurd (hyper_running s); auto.

  (* Write action *)
  apply lemma_write_action_isolation with s v v0; auto.
  
  (* Write Hyper (hyper is running). *)
  inversion_clear H_valid_exec.
  unfold pre_action, pre_write_hyper, hyper_running in H.
  unfold hyper_running in H_not_hyp_run.
  destruct H as (H_act, H').
  rewrite H_act in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* New untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_new_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrusted, (ma', H_hyp_map, (H_ptype, (H_ma'_owner, (H_os_acc, (H_hcall, H_activity)))))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* New trusted *)
  apply lemma_new_trusted_isolation with s v p; auto.

  (* New hyper (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_new_hyper, hyper_running in H_pre.
  destruct H_pre as (H_waitning, H').
  unfold hyper_running in H_not_hyp_run.
  rewrite H_waitning in H_not_hyp_run; elim H_not_hyp_run; auto.


  (* Del untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_del_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_hyp_mapping,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Del trusted *)
  apply lemma_del_trusted_isolation with s v; auto.

  (* Del hyper (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_del_hyper, pre_all_del, hyper_running in H_pre.
  destruct H_pre as (pt_addr, H_curr_map, (H_pre', (H_activity, H_not_accessible))). 
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.
  
  (* Switch (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_switch, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_pre').
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* LSwitch trusted *)
  apply lemma_lswitch_isolation with s (st_active_os s) p (Lswitch_trusted p); auto.

  (* LSwitch untrusted *)
  apply lemma_lswitch_isolation with s o p (Lswitch_untrusted o p); auto.

  (* Hcall (memory does not change). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_hcall in H_post.
  destruct H_post as (aos, H_aos_map, H_s').
  rewrite H_s'; auto.

  (* Ret ctrl (memory does not change). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold post_action, post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update in H_post.
  destruct H_post with System_Memory as [((H_triv1, H_triv2), H_eq_comp) | H_abs].
  unfold state_component_interp in H_eq_comp.
  unfold content_of_address.
  rewrite <- H_eq_comp.
  assumption.
  destruct H_abs as (H_abs, (H_exec_mode, H_activity)).
  cut False; [contradiction | (apply H_abs; split; discriminate)].

  (* Chmod (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_chmod, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_hcall).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Page pin trusted *)
  apply lemma_page_pin_trusted_isolation with s p p0; auto.

  (* Page unpin trusted *)
  apply lemma_page_unpin_trusted_isolation with s p; auto.

  (* Page pin untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_page_pin_untrusted, pre_all_pin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.

  (* Page unpin untrusted (hyper is running). *)
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyp_run.
  rewrite H_activity in H_not_hyp_run; elim H_not_hyp_run; auto.
Qed.
  

(* Write isolation - other formalization
   Write isolation captures the intuition that an OS cannot modify memory that
   it does not own.
*)

Definition write_isolation (s s':state) : Prop :=
 forall (a:action),
       valid_state ctxt s ->
       ~(hyper_running s) ->
       valid_exec ctxt s a s' ->     
       forall (ma:madd) (pg:page) (osi:os_ident), 
          content_of_address s ma pg ->
          ((owner_of_page pg osi /\ osi <> (st_active_os s)) \/
          page_owned_by pg = Hyp_PO) ->
          content_of_address s' ma pg.


Theorem write_isolation_proof : forall (s s':state), write_isolation s s'.
Proof.
  intros s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg osi H_pg_map H_pg_owner.
  generalize (write_isolation_proof' s s' a H_s_valid H_not_hyp_run H_valid_exec ma pg H_pg_map); intro H_write_isolation.
  apply H_write_isolation.
  unfold owner_of_page; unfold owner_of_page in H_pg_owner; unfold not; intro H_aos.
  rewrite <- H_aos in H_pg_owner.
  destruct H_pg_owner as [(H_aos1,H_aos2) | H_hyp]; try discriminate.
  inversion H_aos1.
  rewrite H0 in H_aos2; elim H_aos; auto.
  unfold owner_of_page; unfold owner_of_page in H_pg_owner; unfold not; intro H_aos.
  rewrite H_aos in H_pg_owner.
  destruct H_pg_owner as [(H_aos1,H_aos2) | H_hyp]; try discriminate.
Qed.



Lemma cache_owner : forall (s:state) (va:vadd) (pg:page),
  valid_state ctxt s ->
  Value _ pg = cache_value (st_cache s) va ->
  (page_owned_by pg = Hyp_PO \/
    page_owned_by pg = Os_PO (st_active_os s)).
Proof.
  intros s va pg.
  Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  intro H_pg_map.
  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_pg_content H_pg_map_mem]]]; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg' H_pg'_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  rewrite <- H_pg_map_mem in H_pg'_map.
  injection H_pg'_map; intro; subst pg'; clear H_pg'_map.
  unfold os_accessible in *.
  induction (ctxt_vadd_accessible ctxt va); [right | left]; auto.
Qed.


Lemma not_accessible_va_unchanged : forall a s s' va pg pg',
  valid_state ctxt s ->
  ~(hyper_running s) ->
  valid_exec ctxt s a s' ->
  ~(os_accessible ctxt va) ->
  Value _ pg = cache_value (st_cache s) va ->
  Value _ pg' = cache_value (st_cache s') va ->
  pg=pg'.
Proof.
  intros a s s' va pg pg' H_s_valid H_not_hyper_running H_valid_exec
    H_va_not_accessible H_pg_map H_pg'_map.
  generalize H_valid_exec; intro H_valid_exec'.
  induction a;
    inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post);
      unfold pre_action in H_pre; unfold post_action in H_post.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold cache_value in H_pg_map.
  elim valid_state_included_cache with ctxt s va pg; auto.
  intros ma H_ma_map H_pg_map_mem.
  elim valid_state_included_cache with ctxt s' va pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_mem System_Cache System_Memory.
  rewrite <- H_mem in H_pg'_map_mem; try (intro; discriminate).
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma'_map.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_hyp System_Cache
    Active_OS Guest_OSs Hypervisor_SC.
  rewrite <- H_aos in H_ma'_map; try (intro; discriminate).
  rewrite <- H_oss in H_ma'_map; try (intro; discriminate).
  rewrite <- H_hyp in H_ma'_map; try (intro; discriminate).
  rewrite <- H_mem in H_ma'_map; try (intro; discriminate).
  assert (va_mapped_to_ma s (st_active_os s) va ma') as H_ma'_map_s.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping; auto.

  assert (ma=ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'; clear H_ma'_map_s.
  rewrite <- H_pg_map_mem in H_pg'_map_mem.
  injection H_pg'_map_mem; auto.

  apply valid_state_invariant with s Silent; auto.
  
  (* Read *)
  unfold pre_read, pre_all_read in H_pre.
  unfold cache_value in H_pg_map.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  induction (vadd_eq v va).
  subst v; contradiction.
  unfold post_read, post_all_read in H_post.
  assert (exists pg_read, Value _ pg_read = map_apply madd_eq (st_memory s) ma) as [pg_read H_pg_read_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma).
  exists v0; auto.
  discriminate H_ptype.

  assert (Some ma = va_mapped_to_ma_cache s v) as H_ma_cache_map.

  unfold va_mapped_to_ma_cache.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v0); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]] [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  induction it.
  elim valid_state_included_tlb with ctxt s v v0; auto.
  intros pt_addr' H_pt_addr'_map [mp' H_mp'_map H_v0_map].
  unfold os_current_vadd_mapping_madd, hypervisor_mapping, vadd_mapping in H_pt_addr'_map.
  destruct H_pt_addr'_map as [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]].
  rewrite <- H_aos_map in H_aos'_map; injection H_aos'_map; intro;
    subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map; injection H_p2m'_map; intro;
    subst p2m'; clear H_p2m'_map.
  unfold vadd_mapping in H_mp'_map.
  destruct H_mp'_map as [pt_page' H_pt_page'_map H_mp'_map].
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map; intro;
    subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page'_map; injection H_pt_page'_map; intro;
    subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_mp_map in H_mp'_map; injection H_mp'_map; intro;
    subst mp'; clear H_mp'_map.
  rewrite <- H_ma_map in H_v0_map; injection H_v0_map; intro; subst v0; auto.

  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  injection H_aos_map; intro; subst v0.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  injection H_p2m_map; intro; subst v0.
  induction (map_apply padd_eq p2m (os_current_page aos)); try discriminate.
  injection H_pt_addr_map; intro; subst v0.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  injection H_pt_page_map; intro; subst v0.
  induction (page_content pt_page); try discriminate.
  injection H_mp_map; intro; subst m.
  induction (map_apply vadd_eq mp v); try discriminate.
  injection H_ma_map; intro; subst v0; auto.

  destruct H_post with ma pg_read as [new_cache H_new_cache [new_tlb H_new_tlb H_assign]]; auto.

  unfold va_mapped_to_pg_cache.
  assert (exists it, it = map_apply vadd_eq (st_cache s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) v).
  exists (Value vadd v0); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map.
  induction it.
  elim valid_state_included_cache with ctxt s v v0; auto.
  intros ma' H_ma'_map H_v0_map.
  assert (ma = ma').
  elim va_mapped_same_ma with s v ma ma'; auto.
  subst ma'; clear H_ma'_map.
  rewrite <- H_pg_read_map in H_v0_map.
  injection H_v0_map; intro; subst v0; auto.
  rewrite <- H_ma_cache_map.
  rewrite <- H_pg_read_map; auto.

  elim H_assign with System_Cache.
  intros [[H_abs H_cmp] H_interp]; contradiction H_abs; auto.
  unfold state_component_interp.
  intros [H_triv [H_cache H_tlb]].
  rewrite H_cache in H_pg'_map.
  unfold cache_value in H_pg'_map.
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
  rewrite <- H_old_entries with va pg' in H_pg_map; auto.
  injection H_pg_map; auto.

  (* Read hyper *)
  unfold pre_read_hyper in H_pre.
  destruct H_pre as [H_pre H_hyper_running].
  contradiction.

  (* Write *)
  unfold pre_write in H_pre.
  unfold cache_value in H_pg_map, H_pg'_map.
  destruct H_pre as [H_os_running [H_v_accessible [ma H_ma_map H_ptype]]].
  induction (vadd_eq v va).
  subst v; contradiction.
  unfold post_write, post_all_write in H_post.
  destruct H_post as [ma' H_ma'_map [new_mem H_new_mem [pg_write H_pg_write_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].

  assert (ma = ma').
  apply va_mapped_same_ma with s v; auto.
  unfold va_mapped_to_ma_cache in H_ma'_map.
  assert (exists it, it = map_apply vadd_eq (st_tlb s) v) as [it H_it_map].
  induction (map_apply vadd_eq (st_tlb s) v).
  exists (Value vadd v1); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map in H_ma'_map.
  induction it.
  injection H_ma'_map; intro; subst v1.
  apply valid_state_included_tlb with ctxt; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  induction (map_apply os_ident_eq (st_oss s) (st_active_os s)); try discriminate.
  induction (map_apply os_ident_eq (st_hypervisor s) (st_active_os s)); try discriminate.
  assert (exists pt_addr, Value _ pt_addr = map_apply padd_eq v2 (os_current_page v1)) as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v2 (os_current_page v1)); try discriminate.
  exists v3; auto.
  rewrite <-  H_pt_addr_map in H_ma'_map.
  exists pt_addr.
  exists v1; auto.
  exists v2; auto.
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate.
  assert (exists mp, PT_c mp = (page_content v3)) as [mp H_mp_map].
  induction (page_content v3); try discriminate.
  exists m; auto.
  rewrite <- H_mp_map in H_ma'_map.
  exists mp.
  exists v3; auto.
  induction (map_apply vadd_eq mp v); try discriminate.
  injection H_ma'_map; intro; subst v4; auto.
  
  subst ma'.
  subst s'; simpl in H_pg'_map.
  unfold cache_add in H_pg'_map.
  
  inversion_clear H_new_cache as [H_pg_new_map [H_old_entries
    [H_new_size H_new_distinct]]]; auto.
  assert (Value _ pg' = map_apply vadd_eq (fix_cache_synonym s (st_cache s) ma) va)
    as H_pg'_map_fix.
  apply H_old_entries; auto.

  generalize map_apply_fix_distinct.
  intro H_fix_distinct.
  unfold cache_value in H_fix_distinct.
  rewrite <- H_fix_distinct with s (st_cache s) va ma pg' in H_pg_map; auto.
  injection H_pg_map; auto.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_distinct H_cache_included]] H_tlb_valid]]; auto.

  (* Write hyper *)
  unfold pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  contradiction.

  (* New untrusted *)
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as [H_untrusted [ma H_ma_map [H_ptype [H_owner [H_accessible [H_hcall H_hyper_running]]]]]].
  contradiction.

  (* New trusted *)
  unfold post_new_trusted, post_all_new in H_post.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try contradiction b; auto.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map
    [new_mem H_new_mem [aos H_aos_map H_s']]]].
  rewrite H_s' in H_pg'_map; simpl in H_pg'_map.
  unfold cache_value in H_pg_map, H_pg'_map.
  induction (vadd_eq v va).
  subst v; rewrite map_drop_apply in H_pg'_map; discriminate.
  rewrite map_drop_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.

  (* New hyper *)
  unfold pre_new_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  contradiction.

  (* Del untrusted *)
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as [H_untrusted [H_hcall [aos H_aos_map [p2m H_p2m_map
    [pt_addr H_pt_addr_map [H_pre [H_hyper_running H_v_accessible]]]]]]].
  contradiction.
  
  (* Del trusted *)
  unfold post_del_trusted, post_all_del, current_vadd_mapping_madd,
    os_current_vadd_mapping_madd, hypervisor_mapping in H_post.
  destruct H_post as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]] [new_mem H_new_mem H_s']].
  rewrite H_s' in H_pg'_map; simpl in H_pg'_map.
  unfold cache_value in H_pg_map, H_pg'_map.
  induction (vadd_eq v va).
  subst v; rewrite map_drop_apply in H_pg'_map; discriminate.
  rewrite map_drop_apply_distinct in H_pg'_map; auto.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.

  (* Del hyper *)
  unfold pre_del_hyper in H_pre.
  destruct H_pre as [pt_addr H_pt_addr_map [H_pre [H_hyper_running H_v_acc]]].
  contradiction.

  (* Switch *)
  unfold pre_switch in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  contradiction.

  (* Lswitch trusted *)
  unfold post_lswitch_trusted, post_all_lswitch in H_post.
  destruct H_post as [aos H_aos_map H_s'].
  rewrite H_s' in H_pg'_map; simpl in H_pg'_map.
  induction (os_ident_eq (st_active_os s) (st_active_os s)); try (contradiction b; auto; fail).
  discriminate H_pg'_map.

  (* Lswitch untrusted *)
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as [H_untrst [aos H_aos [H_hcall [H_hyper_running H_pre]]]].
  contradiction.

  (* Hcall *)
  unfold post_hcall in H_post.
  destruct H_post as [aos H_aos_map H_s'].
  rewrite H_s' in H_pg'_map.
  simpl in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.

  (* ret ctrl *)
  unfold post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update in H_post.
  elim H_post with (System_Cache).
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.
  intros [H_abs H_interp].
  contradiction (H_abs).
  split; intro H_disc; discriminate H_disc.

  (* Chmod *)
  unfold pre_chmod in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  contradiction.

  (* Pin trusted *)
  unfold post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as [ma [pg_free H_pg_free_map [H_pg_free_owner [H_pg_free_content
    [p2m H_p2m_map [aos H_aos_map H_s']]]]]].
  rewrite H_s' in H_pg'_map.
  simpl in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.

  (* Unpin trusted *)
  unfold post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [aos H_aos_map H_s']]].
  rewrite H_s' in H_pg'_map.
  simpl in H_pg'_map.
  rewrite <- H_pg_map in H_pg'_map; injection H_pg'_map; auto.

  (* Pin untrusted *)
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as [H_untrst [H_hcall [H_hyper_running H_pre]]].
  contradiction.

  (* Unpin untrusted *)
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as [H_untrst [H_hcall [H_hyper_running H_pre]]].
  contradiction.
Qed.


Definition write_isolation_cache (s s':state) : Prop :=
 forall (a:action) (va:vadd) (pg pg':page) (osi:os_ident),
       valid_state ctxt s ->
       ~(hyper_running s) ->
       valid_exec ctxt s a s' ->     

       ((owner_of_page pg osi /\ osi <> (st_active_os s)) \/
         page_owned_by pg = Hyp_PO) ->
       
       Value _ pg = cache_value (st_cache s) va ->
       (* pg could have been deleted from (st_cache s') by the replacemente policy. *)
       Value _ pg' = cache_value (st_cache s') va ->
       pg = pg'.

Theorem write_isolation_cache_proof : forall s s', write_isolation_cache s s'.
Proof.
  unfold write_isolation_cache.
  intros s s' a va pg pg' osi H_s_valid H_not_hyper_running H_valid_exec
    [[H_pg_owner H_osi_not_active] | H_pg_owner] H_pg_map H_pg'_map.
  unfold owner_of_page in H_pg_owner.
  elim cache_owner with s va pg; auto.
  intro H_hyp_owner.
  rewrite <- H_pg_owner in H_hyp_owner.
  discriminate H_hyp_owner.
  intros H_aos_owner.
  rewrite <- H_pg_owner in H_aos_owner.
  injection H_aos_owner; intro H_aos_eq_osi.
  absurd (osi = (st_active_os s)); auto.

  (* Hypervisor page *)
  apply not_accessible_va_unchanged with a s s' va; auto.
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_pg_content H_pg_map_mem]]]; auto.
  intro H_va_accessible.

  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg'' H_pg''_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  rewrite <- H_pg_map_mem in H_pg''_map.
  injection H_pg''_map; intro; subst pg''; clear H_pg''_map.
  rewrite H_va_accessible_owner in H_pg_owner; auto.
  discriminate H_pg_owner.
Qed.


End Write_Isolation.