(***********************************************************************
 * File: Erasure.v
 * $Revision: 476 $
 * $Date: 2011-02-23 18:19:58 -0200 (Wed, 23 Feb 2011) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * Erasure over states
 **********************************************************************)
Require Import List.
Require Import Actions.
Require Import Exec_invariant.
Require Import Write_Isolation.
Require Import State_equivalence.
Require Import Execution_isolation.

Set Undo 20000.

Section Erasure.

Variable ctxt : context. 
Variable osi : os_ident.

Definition activity_erase (s:state) : os_activity :=
  if (os_ident_eq (st_active_os s) osi) then
    (st_aos_activity s)
  else
    OSA_waiting.

Definition exec_erase (s:state) : exec_mode :=
  if (ctxt_oss ctxt osi) then
    (* osi is trusted. Go to svc mode *)
    EXMD_svc
  else
    match (activity_erase s) with
      | OSA_waiting => EXMD_svc
      | OSA_running => EXMD_usr
    end.


Fixpoint oss_erase (oss:oss_map) : oss_map :=
  match oss with
    | nil => nil
    | it::oss' =>
      if (os_ident_eq (item_index it) osi) then
        it::nil
      else
        oss_erase oss'
  end.        

Fixpoint hyper_erase (hyp:hypervisor) : hypervisor :=
  match hyp with
    | nil => nil
    | it::hyp' =>
      if (os_ident_eq (item_index it) osi) then
        it::nil
      else
        hyper_erase hyp'
  end.

(* Erase non accessible va from mp map *)
Fixpoint mp_erase (mp:mapping vadd madd) : mapping vadd madd :=
  match mp with
    | nil => nil
    | it::mp' =>
      let va := (item_index it) in
        if ((ctxt_vadd_accessible ctxt) va) then
          it::(mp_erase mp')
        else
          (mp_erase mp')
  end.

Fixpoint mem_erase (mem:system_memory) : system_memory :=
  match mem with
    | nil => nil
    | it::mem' => 
      match (page_owned_by (item_info it)) with
        | No_Owner => (Item (item_index it) (Page Other_c No_Owner))::(mem_erase mem')
        | Hyp_PO => (Item (item_index it) (Page Other_c No_Owner))::(mem_erase mem')
        | Os_PO o =>
          if (os_ident_eq o osi) then
            match (page_content (item_info it)) with
              | RW_c _ => it::(mem_erase mem')
              | PT_c mp => (Item (item_index it) (Page (PT_c (mp_erase mp)) (Os_PO osi)))::(mem_erase mem')
              | Other_c => it::(mem_erase mem')
            end
          else
            (Item (item_index it) (Page Other_c No_Owner))::(mem_erase mem')
      end
  end.

Fixpoint cache_erase_fix (c:cache) : cache :=
  match c with
    | nil => nil
    | it::c' => 
      if (ctxt_vadd_accessible ctxt (item_index it)) then
        it::(cache_erase_fix c')
      else
        cache_erase_fix c'
  end.

Definition cache_erase (s:state) : cache :=
  if (os_ident_eq (st_active_os s) osi) then
    cache_erase_fix (st_cache s)
  else
    map_empty vadd page.

Fixpoint tlb_erase_fix (t:tlb) : tlb :=
  match t with
    | nil => nil
    | it::t' =>
      if (ctxt_vadd_accessible ctxt (item_index it)) then
        it::(tlb_erase_fix t')
      else
        tlb_erase_fix t'
  end.

Definition tlb_erase (s:state) : tlb :=
  if (os_ident_eq (st_active_os s) osi) then
    tlb_erase_fix (st_tlb s)
  else
    map_empty vadd madd.

(* Function to remove all non-osi components from the state *)
Definition state_erase (s:state) : state :=
  State osi (oss_erase (st_oss s)) (exec_erase s)  (activity_erase s)
    (hyper_erase (st_hypervisor s)) (mem_erase (st_memory s))
    (cache_erase s) (tlb_erase s).


Definition action_erase (s:state) (a:action) : action :=
  match st_aos_activity s with
    | OSA_running => 
      if (os_ident_eq (st_active_os s) osi) then
        a
      else
        Silent
    | OSA_waiting =>
      match a with
        | New_untrusted o _ _ => if (os_ident_eq o osi) then a else Silent
        | Del_untrusted o _ => if (os_ident_eq o osi) then a else Silent
        | Lswitch_untrusted o _ => if (os_ident_eq o osi) then a else Silent
        | Page_pin_untrusted o _ _ => if (os_ident_eq o osi) then a else Silent
        | Page_unpin_untrusted o _ => if (os_ident_eq o osi) then a else Silent
        | Chmod => if (os_ident_eq (st_active_os s) osi) then a else Silent
        | _ => Silent
      end
  end.


Definition osi_equivalent_upto_aos (s1 s2:state) : Prop :=
  map_apply os_ident_eq (st_oss s1) osi =
  map_apply os_ident_eq (st_oss s2) osi /\
  
  (osi_equivalent_hyper_mapping s1 s2 osi) /\
  (osi_equivalent_pt_pages ctxt s1 s2 osi).

(* Auxiliary lemmas *)
Lemma oss_erase_osi : forall (oss:oss_map),
  map_apply os_ident_eq (oss_erase oss) osi = map_apply os_ident_eq oss osi.
Proof.
  intros oss.
  induction oss; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi);
    induction (os_ident_eq osi (item_index a));
      try (contradiction b; auto; fail); auto.
  simpl.
  induction (os_ident_eq osi (item_index a));
    try (contradiction b; auto; fail); auto.
Qed.  

Lemma oss_erase_not_osi : forall (oss:oss_map) (o:os_ident),
  o<>osi -> 
  map_apply os_ident_eq (oss_erase oss) o = Error _ o.
Proof.
  intros oss o H_o_not_osi.
  induction oss; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi); auto.
  simpl.
  induction (os_ident_eq o (item_index a)); auto.
  rewrite <- a1 in a0; contradiction a0.
Qed.

Lemma oss_no_osi_nil_erase : forall (oss:oss_map),
  map_apply os_ident_eq oss osi = Error _ osi ->
  oss_erase oss = nil.
Proof.
  intros oss H_error.
  induction oss; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi).
  simpl in H_error.
  induction (os_ident_eq osi (item_index a)); [discriminate H_error | contradiction b; auto].
  apply IHoss.
  rewrite <- H_error.
  simpl.
  induction (os_ident_eq osi (item_index a)); try (contradiction b; auto; fail); auto.
Qed.

Lemma same_oss_erase : forall (oss1 oss2:oss_map),
  map_apply os_ident_eq oss1 osi = map_apply os_ident_eq oss2 osi ->
  oss_erase oss1 = oss_erase oss2.
Proof.
  induction oss1.
  simpl.
  intros oss2 H_oss2_error.
  rewrite oss_no_osi_nil_erase; auto.

  intros oss2 H_eq_map.
  induction oss2.
  replace (oss_erase nil) with
    (nil:oss_map); auto.
  rewrite oss_no_osi_nil_erase; auto.
  
  simpl.
  induction (os_ident_eq (item_index a) osi);
      induction (os_ident_eq (item_index a0) osi).
  assert (a=a0) as H_a_eq_a0.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
      induction (os_ident_eq osi (item_index a0));
        try (contradiction b; auto; fail).
  injection H_eq_map; intro H_info.
  rewrite <- a1 in a2.
  induction a; induction a0; simpl in a2, H_info.
  rewrite a2; rewrite H_info; auto.
  rewrite H_a_eq_a0; auto.
  
  rewrite <- IHoss2.
  simpl.
  induction (os_ident_eq (item_index a) osi); [auto | contradiction b0; auto].
  rewrite H_eq_map.
  simpl.
  induction (os_ident_eq osi (item_index a0)); [contradiction b; auto | auto ].

  rewrite IHoss1 with (a0::oss2).
  simpl.
  induction (os_ident_eq (item_index a0) osi); [auto | contradiction b0; auto].
  simpl.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
    induction (os_ident_eq osi (item_index a0));
      try (contradiction b; auto; fail); auto.
  
  apply IHoss1.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
    induction (os_ident_eq osi (item_index a0));
      try (contradiction b0; auto; fail); auto.
  contradiction b; auto.
Qed.

Lemma oss_erase_add_osi: forall (oss:oss_map) (system:os),
  oss_erase (map_add os_ident_eq oss osi system) = map_add os_ident_eq (oss_erase oss) osi system.
Proof.
  intros oss system.
  induction oss.
  simpl.
  induction (os_ident_eq osi osi); simpl; auto.
  contradiction b; auto.

  simpl.
  induction (os_ident_eq (item_index a) osi).
  simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  induction (os_ident_eq (item_index a) osi); try contradiction b; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi); try (contradiction b; auto; fail); auto.
Qed.

Lemma oss_erase_add_not_osi: forall (oss:oss_map) (o:os_ident) (system:os),
  o<>osi -> oss_erase oss = oss_erase (map_add os_ident_eq oss o system).
Proof.
  intros oss o system H_o_ne_osi.
  induction oss.
  simpl.
  induction (os_ident_eq o osi); auto; contradiction H_o_ne_osi; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi).
  induction (os_ident_eq (item_index a) o).
  rewrite a0 in a1.
  contradiction H_o_ne_osi; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi); try contradiction b0; auto.
  induction (os_ident_eq (item_index a) o).
  simpl.
  induction (os_ident_eq o osi); auto; try (contradiction H_o_ne_osi; auto).
  simpl.
  induction (os_ident_eq (item_index a) osi); auto; try (contradiction b; auto).
Qed.

Lemma hyper_erase_osi : forall (hyp:hypervisor),
  map_apply os_ident_eq (hyper_erase hyp) osi = map_apply os_ident_eq hyp osi.
Proof.
  intros hyp.
  induction hyp; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi).
  induction (os_ident_eq osi (item_index a)); try contradiction b; auto.
  simpl.
  induction (os_ident_eq osi (item_index a)); try contradiction b; auto.
  induction (os_ident_eq osi (item_index a)); try (contradiction b; auto; fail); auto.
Qed.

Lemma hyper_erase_not_osi : forall (hyp:hypervisor) (o:os_ident),
  o<>osi -> 
  map_apply os_ident_eq (hyper_erase hyp) o = Error _ o.
Proof.
  intros hyp o H_o_not_osi.
  induction hyp; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi); auto.
  simpl.
  induction (os_ident_eq o (item_index a)); auto.
  subst o.
  contradiction H_o_not_osi.
Qed.

Lemma hyper_no_osi_nil_erase : forall (hyp:hypervisor),
  map_apply os_ident_eq hyp osi = Error _ osi ->
  hyper_erase hyp = nil.
Proof.
  intros hyp H_error.
  induction hyp; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi).
  simpl in H_error.
  induction (os_ident_eq osi (item_index a)); [discriminate H_error | contradiction b; auto].
  apply IHhyp.
  rewrite <- H_error.
  simpl.
  induction (os_ident_eq osi (item_index a)); try (contradiction b; auto; fail); auto.
Qed.
  
Lemma same_hyper_erase : forall (hyp1 hyp2:hypervisor),
  map_apply os_ident_eq hyp1 osi = map_apply os_ident_eq hyp2 osi ->
  hyper_erase hyp1 = hyper_erase hyp2.
Proof.
  induction hyp1.
  simpl.
  intros hyp2 H_hyp2_error.
  rewrite hyper_no_osi_nil_erase; auto.

  intros hyp2 H_eq_map.
  induction hyp2.
  replace (hyper_erase nil) with
    (nil:hypervisor); auto.
  rewrite hyper_no_osi_nil_erase; auto.
  
  simpl.
  induction (os_ident_eq (item_index a) osi);
      induction (os_ident_eq (item_index a0) osi).
  assert (a=a0) as H_a_eq_a0.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
      induction (os_ident_eq osi (item_index a0));
        try (contradiction b; auto; fail).
  injection H_eq_map; intro H_info.
  rewrite <- a1 in a2.
  induction a; induction a0; simpl in a2, H_info.
  rewrite a2; rewrite H_info; auto.
  rewrite H_a_eq_a0; auto.
  
  rewrite <- IHhyp2.
  simpl.
  induction (os_ident_eq (item_index a) osi); [auto | contradiction b0; auto].
  rewrite H_eq_map.
  simpl.
  induction (os_ident_eq osi (item_index a0)); [contradiction b; auto | auto ].

  rewrite IHhyp1 with (a0::hyp2).
  simpl.
  induction (os_ident_eq (item_index a0) osi); [auto | contradiction b0; auto].
  simpl.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
    induction (os_ident_eq osi (item_index a0));
      try (contradiction b; auto; fail); auto.
  
  apply IHhyp1.
  simpl in H_eq_map.
  induction (os_ident_eq osi (item_index a));
    induction (os_ident_eq osi (item_index a0));
      try (contradiction b0; auto; fail); auto.
  contradiction b; auto.
Qed.

Lemma hyper_erase_add_osi : forall (hyp:hypervisor) (p2m:map_physical_machine_addr),
  hyper_erase (map_add os_ident_eq hyp osi p2m) =
  map_add os_ident_eq (hyper_erase hyp) osi p2m.
Proof.
  intros hyp p2m.
  induction hyp; simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  induction (os_ident_eq (item_index a) osi).
  simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  induction (os_ident_eq (item_index a) osi); try contradiction b; auto.
  simpl.
  induction (os_ident_eq (item_index a) osi); try (contradiction b; auto; fail); auto.
Qed.
  
Lemma mp_erase_accessible : forall (mp:mapping vadd madd) (va:vadd),
  os_accessible ctxt va ->
  map_apply vadd_eq mp va = map_apply vadd_eq (mp_erase mp) va.
Proof.
  intros mp va H_va_accessible.
  induction mp; auto.
  simpl.
  unfold os_accessible in H_va_accessible.
  induction (vadd_eq va (item_index a)).
  rewrite <- a0.
  rewrite H_va_accessible.
  simpl.
  induction (vadd_eq va (item_index a)); auto; contradiction b.

  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); auto; contradiction b.
Qed.
  
Lemma mp_erase_not_accessible : forall (mp:mapping vadd madd) (va:vadd),
  ~ os_accessible ctxt va ->
  map_apply vadd_eq (mp_erase mp) va = Error _ va.
Proof.
  intros mp va H_va_not_accessible.
  induction mp; auto.
  simpl.
  unfold os_accessible in H_va_not_accessible.
  assert ({ctxt_vadd_accessible ctxt (item_index a)=true}+
          {ctxt_vadd_accessible ctxt (item_index a)=false}) as H;
    [induction (ctxt_vadd_accessible ctxt (item_index a)); auto | idtac].
  elim H; [intro H_a_accessible | intro H_a_not_accessible]; clear H.
  rewrite H_a_accessible.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  rewrite <- a0 in H_a_accessible; contradiction H_va_not_accessible.

  rewrite H_a_not_accessible; auto.
Qed.

Lemma mp_erase_add_osi : forall (mp:mapping vadd madd) (va:vadd) (ma:madd),
  ctxt_vadd_accessible ctxt va = true ->
  mp_erase (map_add vadd_eq mp va ma) = map_add vadd_eq (mp_erase mp) va ma.
Proof.
  intros mp va ma H_va_accessible.
  induction mp.
  simpl.
  rewrite H_va_accessible; auto.

  simpl.
  induction (vadd_eq (item_index a) va).
  rewrite a0; rewrite H_va_accessible.
  
  simpl.
  rewrite H_va_accessible.
  induction (vadd_eq (item_index a) va); try contradiction b; auto.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq (item_index a) va); try (contradiction b; auto; fail).
  rewrite IHmp; auto.
Qed.

Lemma mp_erase_drop_osi_acc : forall (mp:mapping vadd madd) (va:vadd),
  ctxt_vadd_accessible ctxt va = true ->
  mp_erase (map_drop vadd_eq mp va) = map_drop vadd_eq (mp_erase mp) va.
Proof.
  intros mp va H_va_accessible.
  induction mp; auto.

  simpl.
  induction (vadd_eq va (item_index a)).
  rewrite <- a0; rewrite H_va_accessible.
  simpl.
  induction (vadd_eq va (item_index a)); try contradiction b; auto.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); try (contradiction b; auto; fail).
  rewrite IHmp; auto.
Qed.

Lemma mp_erase_add_not_acc_osi : forall (mp: mapping vadd madd) (va:vadd) (ma:madd),
  ctxt_vadd_accessible ctxt va = false ->
  mp_erase (map_add vadd_eq mp va ma) = mp_erase mp.
Proof.
  intros mp va ma H_va_not_accessible.
  induction mp.
  simpl; rewrite H_va_not_accessible; auto.

  simpl.
  induction (vadd_eq (item_index a) va).
  rewrite a0.
  rewrite H_va_not_accessible.
  simpl.
  rewrite H_va_not_accessible; auto.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)).
  rewrite IHmp; auto.
  apply IHmp.
Qed.

Lemma mp_erase_drop_osi_not_acc : forall (mp: mapping vadd madd) (va:vadd),
  ctxt_vadd_accessible ctxt va = false ->
  mp_erase (map_drop vadd_eq mp va) = mp_erase mp.
Proof.
  intros mp va H_va_not_accessible.
  induction mp; auto.

  simpl.
  induction (vadd_eq va (item_index a)).
  rewrite <- a0.
  rewrite H_va_not_accessible.
  apply IHmp.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)).
  rewrite IHmp; auto.
  apply IHmp.
Qed.

Lemma mp_erase_old : forall (mp:mapping vadd madd) (va:vadd) (ma:madd),
  Value _ ma =  map_apply vadd_eq (mp_erase mp) va ->
  Value _ ma =  map_apply vadd_eq mp va.
Proof.
  intros mp va ma H_ma_map.

  assert (os_accessible ctxt va \/ ~ os_accessible ctxt va) as [H_va_accessible | H_va_not_accessible].
  unfold os_accessible; induction (ctxt_vadd_accessible ctxt va); [left | right]; auto.
  rewrite <- mp_erase_accessible in H_ma_map; auto.
  rewrite mp_erase_not_accessible in H_ma_map; auto; discriminate H_ma_map.
Qed.

Lemma mem_erase_owner : forall (mem:system_memory) (ma:madd),
  maddress_owner mem ma = Os_PO osi -> maddress_owner (mem_erase mem) ma = Os_PO osi.
Proof.
  intros mem ma H_ma_owner.
  unfold maddress_owner in *.
  induction mem; auto.

  simpl in H_ma_owner.
  simpl.
  induction (madd_eq ma (item_index a)).
  rewrite H_ma_owner.
  induction (os_ident_eq osi osi); try (contradiction b; auto).
  induction (page_content (item_info a)); simpl;
    induction (madd_eq ma (item_index a)); try contradiction b; auto.

  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
  induction (os_ident_eq o osi).
  subst o.
  induction (page_content (item_info a)); simpl;
    induction (madd_eq ma (item_index a)); try (contradiction b; fail);
      apply IHmem; auto.
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
Qed.

Lemma mem_erase_not_os_owner : forall (mem:system_memory) (ma:madd) (pg:page) (o:os_ident),
  o<>osi -> Value _ pg = map_apply madd_eq (mem_erase mem) ma ->
  page_owned_by pg <> Os_PO o.
Proof.
  intros mem ma pg o H_o_ne_osi H_pg_map H_pg_owner.
  induction mem.
  simpl in H_pg_map; discriminate H_pg_map.

  simpl in H_pg_map.
  assert (exists owner, page_owned_by (item_info a) = owner) as [owner H_a_owner].
  induction (page_owned_by (item_info a)); [exists Hyp_PO | exists (Os_PO o0) | exists No_Owner]; auto.
  induction owner; rewrite H_a_owner in H_pg_map.

  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try (apply IHmem; auto).
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_pg_owner; discriminate H_pg_owner.
  
  induction (os_ident_eq o0 osi).
  subst o0.
  induction (page_content (item_info a)).
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try (apply IHmem; auto; fail).
  apply H_o_ne_osi.
  injection H_pg_map; intro H_pg_value; rewrite <- H_pg_value in H_a_owner.
  rewrite H_pg_owner in H_a_owner; injection H_a_owner; auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  apply H_o_ne_osi.
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_pg_owner; simpl in H_pg_owner.
  injection H_pg_owner; auto.
  apply IHmem; auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try (apply IHmem; auto; fail).
  injection H_pg_map; intro H_pg_value; rewrite <- H_pg_value in H_a_owner.
  apply H_o_ne_osi; rewrite H_pg_owner in H_a_owner; injection H_a_owner; auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try (apply IHmem; auto; fail).
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_pg_owner; discriminate H_pg_owner.

  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try (apply IHmem; auto; fail).
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_pg_owner.
  simpl in H_pg_owner; discriminate H_pg_owner.
Qed.

Lemma mem_erase_old_owner : forall (mem:system_memory) (ma:madd),
  maddress_owner (mem_erase mem) ma = Os_PO osi ->
  maddress_owner mem ma = Os_PO osi.
Proof.
  intros mem ma H_ma_owner.
  unfold maddress_owner in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (mem_erase mem) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (mem_erase mem) ma); try discriminate H_ma_owner.
  exists v; auto.
  rewrite <- H_pg_map in H_ma_owner.
  
  induction mem.
  simpl in H_pg_map; discriminate H_pg_map.
  
  simpl.
  simpl in H_pg_map.
  assert (exists owner, page_owned_by (item_info a) = owner) as [owner H_a_owner].
  induction (page_owned_by (item_info a)); [exists Hyp_PO | exists (Os_PO o) | exists No_Owner]; auto.
  induction owner.
  rewrite H_a_owner in H_pg_map.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_ma_owner; discriminate H_ma_owner.
  apply IHmem; auto.
  rewrite H_a_owner in H_pg_map.
  induction (os_ident_eq o osi).
  subst o.
  induction (page_content (item_info a)); simpl in H_pg_map;
    induction (madd_eq ma (item_index a)); auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro H_pg_value; rewrite H_pg_value in H_ma_owner; discriminate H_ma_owner.
  apply IHmem; auto.
  rewrite H_a_owner in H_pg_map.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); auto.
  injection H_pg_map; intro H_pg_value; subst pg; discriminate H_ma_owner.
Qed.

Lemma other_mem_erase_add_osi : forall (mem:system_memory) (ma:madd),
   mem_erase (map_add madd_eq mem ma (Page Other_c No_Owner)) = 
   map_add madd_eq (mem_erase mem) ma (Page Other_c No_Owner).
Proof.
  intros mem ma.
  induction mem; auto.
  simpl.
  induction (madd_eq (item_index a) ma).
  simpl.
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try contradiction b; auto.
  induction (os_ident_eq o osi).
  subst o.
  induction (page_content (item_info a)); simpl;
    induction (madd_eq (item_index a) ma); try contradiction b; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try contradiction b0; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try contradiction b; auto.
  simpl.
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  induction (os_ident_eq o osi); simpl; 
    try induction (page_content (item_info a));
      simpl; induction (madd_eq (item_index a) ma);
        try (contradiction b; fail); rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
Qed.

Lemma rw_mem_erase_osi : forall (mem:system_memory) (ma:madd) (pg:page),
  Value _ pg = map_apply madd_eq mem ma ->
  page_owned_by pg = Os_PO osi -> (exists val, page_content pg = RW_c val) ->
  Value _ pg = map_apply madd_eq (mem_erase mem) ma.
Proof.
  intros mem ma pg H_pg_map H_pg_owner [val H_pg_content].
  induction mem; auto.

  simpl in H_pg_map.
  simpl.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro; subst pg.
  rewrite H_pg_owner.
  induction (os_ident_eq osi osi);
    try (contradiction b; auto).
  rewrite H_pg_content.
  simpl.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.

  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; auto; fail); auto.
  induction (os_ident_eq o osi).
  induction (page_content (item_info a));
    simpl; induction (madd_eq ma (item_index a)); auto;
      contradiction b.
  simpl.
  induction (madd_eq ma (item_index a));
    try (contradiction b; fail); auto.
  simpl.
  induction (madd_eq ma (item_index a)); auto.
  contradiction b.
Qed.

Lemma rw_mem_erase_old_mem : forall (mem:system_memory) (ma:madd) (pg:page),
  Value _ pg = map_apply madd_eq (mem_erase mem) ma ->
  page_owned_by pg = Os_PO osi -> (exists val, page_content pg = RW_c val) -> 
  Value _ pg = map_apply madd_eq mem ma.
Proof.
  intros mem ma pg H_pg_map H_pg_owner [val H_pg_content].
  induction mem; auto.

  simpl in H_pg_map.
  simpl.
  induction (madd_eq ma (item_index a)).
  induction (page_owned_by (item_info a)).
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try contradiction b.
  injection H_pg_map; intro H_pg_value;
    rewrite H_pg_value in H_pg_owner; discriminate H_pg_owner.
  induction (os_ident_eq o osi).
  induction (page_content (item_info a)).
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try contradiction b.
  injection H_pg_map; intro H_pg_value;
    rewrite H_pg_value in H_pg_content; discriminate H_pg_content.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro H_pg_value;
    rewrite H_pg_value in H_pg_content; discriminate H_pg_content.
  contradiction b0.
  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.

  injection H_pg_map; intro H_pg_value;
    rewrite H_pg_value in H_pg_owner; discriminate H_pg_owner.
  induction (page_owned_by (item_info a)); simpl in H_pg_map;
    induction (madd_eq ma (item_index a)); try (contradiction b; fail); auto.
  induction (os_ident_eq o osi).
  induction (page_content (item_info a)); simpl in H_pg_map;
    induction (madd_eq ma (item_index a)); try (contradiction b; auto; fail); auto.
  induction (page_content (item_info a)); simpl in H_pg_map;
    induction (madd_eq ma (item_index a)); try (contradiction b; auto; fail); auto.
Qed.

Lemma pt_mem_erase_osi : forall (mem:system_memory) (pt_addr:madd)
  (pt_page:page) (mp:mapping vadd madd),
    Value _ pt_page = map_apply madd_eq mem pt_addr ->
    page_owned_by pt_page = Os_PO osi -> page_content pt_page = PT_c mp ->
    Value _ (Page (PT_c (mp_erase mp)) (Os_PO osi)) = map_apply madd_eq (mem_erase mem) pt_addr.
Proof.
  intros mem pt_addr pt_page mp H_pt_page_map
    H_pt_page_owner H_pt_page_content.
  induction mem.
  simpl in H_pt_page_map; discriminate H_pt_page_map.
  simpl.
  induction (madd_eq pt_addr (item_index a)).
  replace (item_info a) with pt_page.
  rewrite H_pt_page_owner.
  rewrite H_pt_page_content.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  simpl.
  induction (madd_eq pt_addr (item_index a)); auto; contradiction b.

  rewrite a0 in H_pt_page_map.
  simpl in H_pt_page_map.
  induction (madd_eq (item_index a) (item_index a)); try contradiction b; auto.
  injection H_pt_page_map; auto.

  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)); try (contradiction b; fail).

  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq pt_addr (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
  induction (os_ident_eq o osi); 
    [induction (page_content (item_info a)) | idtac]; simpl; 
      induction (madd_eq pt_addr (item_index a));
        try (contradiction b; fail); apply IHmem; auto.
  simpl; induction (madd_eq pt_addr (item_index a));
    try (contradiction b; fail); apply IHmem; auto.
Qed.

Lemma pt_mem_erase_drop_osi : forall (mem:system_memory) (va:vadd) (pt_addr:madd)
  (pt_page:page) (oldmap:mapping vadd madd), 
  ~ (os_accessible ctxt va) -> Value _ pt_page = map_apply madd_eq mem pt_addr ->
  PT_c oldmap = page_content pt_page -> page_owned_by pt_page = Os_PO osi ->
  mem_erase mem =
  mem_erase (map_add madd_eq mem pt_addr (Page (PT_c (map_drop vadd_eq oldmap va)) (page_owned_by pt_page))).
Proof.
  intros mem va pt_addr pt_page oldmap H_va_not_accessible
    H_pt_page_map H_pt_page_content H_pt_page_owner.
  induction mem.
  simpl in H_pt_page_map; discriminate H_pt_page_map.

  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  simpl.
  injection H_pt_page_map; intro H_pt_page_value; rewrite <- H_pt_page_value.
  rewrite H_pt_page_owner.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  rewrite <- H_pt_page_content.
  induction (madd_eq (item_index a) pt_addr); try contradiction b; auto.
  simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  rewrite a2.
  replace (mp_erase oldmap) with (mp_erase (map_drop vadd_eq oldmap va)); auto.
  unfold os_accessible in H_va_not_accessible.
  apply mp_erase_drop_osi_not_acc.  
  induction (ctxt_vadd_accessible ctxt va);
    try (contradiction H_va_not_accessible; auto; fail); auto.
  simpl.
  induction (madd_eq (item_index a) pt_addr); try (contradiction b; auto; fail).
  simpl.
  induction (page_owned_by (item_info a)).
  rewrite IHmem; auto.
  induction (os_ident_eq o osi); induction (page_content (item_info a));
    rewrite IHmem; auto.
  rewrite IHmem; auto.
Qed.

Lemma pt_mem_erase_add_not_acc_osi : forall (mem:system_memory) (va:vadd) (pt_addr ma:madd)
  (pt_page:page) (oldmap:mapping vadd madd), 
  ~ (os_accessible ctxt va) -> Value _ pt_page = map_apply madd_eq mem pt_addr ->
  PT_c oldmap = page_content pt_page -> page_owned_by pt_page = Os_PO osi ->
  mem_erase mem =
  mem_erase (map_add madd_eq mem pt_addr (Page (PT_c (map_add vadd_eq oldmap va ma)) (page_owned_by pt_page))).
Proof.
  intros mem va pt_addr ma pt_page oldmap H_va_not_accessible
    H_pt_page_map H_pt_page_content H_pt_page_owner.
  induction mem.
  simpl in H_pt_page_map; discriminate H_pt_page_map.

  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  simpl.
  injection H_pt_page_map; intro H_pt_page_value; rewrite <- H_pt_page_value.
  rewrite H_pt_page_owner.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  rewrite <- H_pt_page_content.
  induction (madd_eq (item_index a) pt_addr); try contradiction b; auto.
  simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  rewrite a2.
  replace (mp_erase oldmap) with (mp_erase (map_add vadd_eq oldmap va ma)); auto.
  unfold os_accessible in H_va_not_accessible.
  apply mp_erase_add_not_acc_osi.  
  induction (ctxt_vadd_accessible ctxt va);
    try (contradiction H_va_not_accessible; auto; fail); auto.
  simpl.
  induction (madd_eq (item_index a) pt_addr); try (contradiction b; auto; fail).
  simpl.
  induction (page_owned_by (item_info a)).
  rewrite IHmem; auto.
  induction (os_ident_eq o osi); induction (page_content (item_info a));
    rewrite IHmem; auto.
  rewrite IHmem; auto.
Qed.


Lemma pt_mem_erase_old_mem : forall (mem:system_memory) (pt_addr:madd)
  (pt_page:page) (mp:mapping vadd madd),
  Value _ pt_page = map_apply madd_eq (mem_erase mem) pt_addr ->
  page_owned_by pt_page = Os_PO osi -> page_content pt_page = PT_c mp -> 
  exists2 old_mp, mp = mp_erase old_mp &
  exists2 old_pt_page, Value _ old_pt_page = map_apply madd_eq mem pt_addr &
    page_content old_pt_page = PT_c old_mp.
Proof.
  intros mem pt_addr pt_page mp H_pt_page_map H_pt_page_owner
    H_pt_page_content.
  induction mem; simpl in H_pt_page_map;
    try discriminate H_pt_page_map.
  simpl.

  assert (exists owner, page_owned_by (item_info a) = owner) as [owner H_a_owner].
  induction (page_owned_by (item_info a));
    [exists Hyp_PO | exists (Os_PO o) | exists No_Owner]; auto.
  induction owner.
  rewrite H_a_owner in H_pt_page_map.
  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value;
    rewrite H_pt_page_value in H_pt_page_content;
      discriminate H_pt_page_content.
  apply IHmem; auto.

  rewrite H_a_owner in H_pt_page_map.
  induction (os_ident_eq o osi).
  subst o.
  assert (exists content, page_content (item_info a) = content) as [content H_a_content].
  induction (page_content (item_info a)); [exists (RW_c o)| exists (PT_c m) | exists Other_c]; auto.
  induction (content); rewrite H_a_content in H_pt_page_map.
  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite <- H_pt_page_value in H_a_content.
  rewrite H_a_content in H_pt_page_content; discriminate H_pt_page_content.
  apply IHmem; auto.

  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite H_pt_page_value in H_pt_page_content.
  simpl in H_pt_page_content.
  injection H_pt_page_content; auto.
  intro H_em_eq_mp.
  exists m; auto.
  exists (item_info a); auto.
  apply IHmem; auto.

  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite <- H_pt_page_value in H_a_content.
  rewrite H_a_content in H_pt_page_content.
  discriminate H_pt_page_content.
  apply IHmem; auto.
  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite H_pt_page_value in H_pt_page_owner.
  simpl in H_pt_page_owner; discriminate H_pt_page_owner.
  apply IHmem; auto.
  
  rewrite H_a_owner in H_pt_page_map.
  simpl in H_pt_page_map.
  induction (madd_eq pt_addr (item_index a)).
  injection H_pt_page_map; intro H_pt_page_value.
  rewrite H_pt_page_value in H_pt_page_owner; simpl in H_pt_page_owner;
    discriminate H_pt_page_owner.
  apply IHmem; auto.
Qed.
  
Lemma rw_mem_erase_add_osi : forall (mem:system_memory) (ma:madd) (new_pg:page),
  page_owned_by new_pg = Os_PO osi -> (exists val, page_content new_pg = RW_c val) ->
  mem_erase (map_add madd_eq mem ma new_pg) = map_add madd_eq (mem_erase mem) ma new_pg.
Proof.
  intros mem ma new_pg H_new_pg_owner [val H_new_pg_content].
  
  induction mem.
  simpl.
  rewrite H_new_pg_owner.
  rewrite H_new_pg_content.
  induction (os_ident_eq osi osi); auto.
  contradiction b; auto.

  simpl.
  induction (madd_eq (item_index a) ma).
  simpl.
  rewrite H_new_pg_owner.
  rewrite H_new_pg_content.
  induction (os_ident_eq osi osi); try (contradiction b; auto; fail).
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b.
  induction (os_ident_eq o osi).
  subst o.
  induction (page_content (item_info a)); simpl;
    induction (madd_eq (item_index a) ma); auto; contradiction b.
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b0.
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b.

  simpl.
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  induction (os_ident_eq o osi).
  induction (page_content (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
Qed.

Lemma pt_mem_erase_add_osi : forall (mem:system_memory) (ma:madd) (mp:mapping vadd madd),
  mem_erase (map_add madd_eq mem ma (Page (PT_c mp) (Os_PO osi))) =
  map_add madd_eq (mem_erase mem) ma (Page (PT_c (mp_erase mp)) (Os_PO osi)).
Proof.
  intros mem ma mp.
  
  induction mem.
  simpl.
  induction (os_ident_eq osi osi); auto.
  contradiction b; auto.

  simpl.
  induction (madd_eq (item_index a) ma).
  simpl.
  induction (os_ident_eq osi osi); try (contradiction b; auto; fail).
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b.
  induction (os_ident_eq o osi).
  subst o.
  induction (page_content (item_info a)); simpl;
    induction (madd_eq (item_index a) ma); auto; contradiction b.
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b0.
  simpl.
  induction (madd_eq (item_index a) ma); auto; contradiction b.

  simpl.
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  induction (os_ident_eq o osi).
  induction (page_content (item_info a)).
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; fail).
  rewrite IHmem; auto.
Qed.


Lemma mem_erase_add_not_osi : forall (mem:system_memory) (ma:madd) (old_pg new_pg:page),
  Value _ old_pg = map_apply madd_eq mem ma ->
  page_owned_by old_pg <> Os_PO osi -> page_owned_by new_pg <> Os_PO osi ->
  mem_erase (map_add madd_eq mem ma new_pg) = mem_erase mem.
Proof.
  intros mem ma old_pg new_pg H_old_pg_exists H_old_pg_owner H_new_pg_owner.
  
  induction mem.
  simpl in H_old_pg_exists; discriminate H_old_pg_exists.
  
  induction (madd_eq (item_index a) ma).
  simpl in H_old_pg_exists.  
  simpl.
  induction (madd_eq (item_index a) ma); try contradiction b; auto.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.
  injection H_old_pg_exists; intro H_old_pg_value.
  rewrite H_old_pg_value in H_old_pg_owner.
  simpl.
  induction (page_owned_by new_pg);
    induction (page_owned_by (item_info a)); try rewrite a0; auto.
  induction (os_ident_eq o osi); auto.
  subst o; contradiction H_old_pg_owner; auto.
  induction (page_content new_pg);
    induction (os_ident_eq o osi); auto;
      subst o; contradiction H_new_pg_owner; auto.
  induction (os_ident_eq o osi).
  subst o; contradiction H_new_pg_owner; auto.
  induction (os_ident_eq o0 osi); auto.
  subst o0; contradiction H_old_pg_owner; auto.
  induction (os_ident_eq o osi); auto.
  subst o; contradiction H_new_pg_owner; auto.
  induction (os_ident_eq o osi); auto.
  subst o; contradiction H_old_pg_owner; auto.

  simpl in H_old_pg_exists.
  simpl.
  induction (madd_eq (item_index a) ma); try (contradiction b; auto; fail).
  induction (madd_eq ma (item_index a)); try (contradiction b; auto; fail).
  simpl.
  induction (page_owned_by (item_info a)).
  rewrite IHmem; auto.

  induction (os_ident_eq o osi);
    induction (page_content (item_info a));
      rewrite IHmem; auto.
  rewrite IHmem; auto.
Qed.

Lemma other_mem_erase_osi : forall (mem:system_memory) (ma:madd) (pg:page),
  Value _ pg = map_apply madd_eq mem ma ->
  page_owned_by pg = Os_PO osi -> page_content pg = Other_c ->
  Value _ pg = map_apply madd_eq (mem_erase mem) ma.
Proof.
  intros mem ma pg H_pg_map H_pg_owner H_pg_content.
  induction mem.
  simpl in H_pg_map; discriminate H_pg_map.

  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro H_pg_value.
  simpl.
  rewrite <- H_pg_value.
  rewrite H_pg_owner.
  induction (os_ident_eq osi osi); try (contradiction b; auto).
  rewrite H_pg_content.
  simpl.
  induction (madd_eq ma (item_index a));  try contradiction b.
  rewrite H_pg_value; auto.
  simpl.

  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
  induction (os_ident_eq o osi);
    induction (page_content (item_info a)); simpl;
      induction (madd_eq ma (item_index a)); try (contradiction b; fail);
        apply IHmem; auto.

  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
Qed.

Lemma other_mem_erase_not_osi : forall (mem:system_memory) (ma:madd) (pg:page),
  Value _ pg = map_apply madd_eq mem ma ->
  page_owned_by pg <> Os_PO osi -> page_content pg = Other_c ->
  Value _ (Page Other_c No_Owner) = map_apply madd_eq (mem_erase mem) ma.
Proof.
  intros mem ma pg H_pg_map H_pg_owner H_pg_content.
  induction mem.
  simpl in H_pg_map; discriminate H_pg_map.

  simpl in H_pg_map.
  induction (madd_eq ma (item_index a)).
  injection H_pg_map; intro H_pg_value.
  simpl.
  rewrite <- H_pg_value.
  rewrite H_pg_content.
  assert (exists owner, owner = page_owned_by pg) as [owner H_pg_owner'].
  induction (page_owned_by pg); [exists Hyp_PO | exists (Os_PO o) | exists No_Owner]; auto.
  rewrite <- H_pg_owner'.
  induction owner.
  simpl.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.
  induction (os_ident_eq o osi).
  subst o.
  contradiction H_pg_owner; auto.
  simpl.
  induction (madd_eq ma (item_index a)); try contradiction b0; auto.
  simpl.
  induction (madd_eq ma (item_index a)); try contradiction b; auto.
  simpl.
  induction (page_owned_by (item_info a)).
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
  induction (os_ident_eq o osi);
    induction (page_content (item_info a)); simpl;
      induction (madd_eq ma (item_index a)); try (contradiction b; fail);
        apply IHmem; auto.
  simpl.
  induction (madd_eq ma (item_index a)); try (contradiction b; fail).
  apply IHmem; auto.
Qed.

Lemma cache_erase_less_elements : forall (c:cache),
  map_size (cache_erase_fix c) <= map_size c.
Proof.
  intros c.
  induction c; auto.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); simpl; auto.
  induction (page_content (item_info a)); simpl; omega.
Qed.

Lemma cache_erase_apply : forall (c:cache) (va:vadd) (pg:page),
  cache_distinct c ->
  Value _ pg = cache_value (cache_erase_fix c) va ->
  Value _ pg = cache_value c va.
Proof.
  intros c va pg H_cache_distinct H_pg_map.
  unfold cache_value in *.
  induction H_cache_distinct; auto.

  simpl; simpl in H_pg_map.
  induction (vadd_eq va va0).
  subst va0.
  induction (ctxt_vadd_accessible ctxt va).
  simpl in H_pg_map.
  induction (vadd_eq va va); auto.
  contradiction b; auto.
  rewrite IHH_cache_distinct; auto.
  elim H.
  unfold cache_value.
  exists pg; auto.

  induction (ctxt_vadd_accessible ctxt va0).
  simpl in H_pg_map.
  induction (vadd_eq va va0); try (contradiction b; assumption).
  apply IHH_cache_distinct; auto.
  apply IHH_cache_distinct; auto.
Qed.

Lemma cache_erase_accessible : forall (c:cache) (va:vadd),
  os_accessible ctxt va ->
  map_apply vadd_eq c va = map_apply vadd_eq (cache_erase_fix c) va.
Proof.
  intros c va H_va_accessible.
  induction c; auto.

  simpl.
  unfold os_accessible in H_va_accessible.
  induction (vadd_eq va (item_index a)).
  rewrite <- a0.
  rewrite H_va_accessible.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  contradiction b; auto.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  contradiction b; auto.
Qed.


Lemma cache_erase_not_accessible : forall (c:cache) (va:vadd),
  ~(os_accessible ctxt va) ->
  map_apply vadd_eq(cache_erase_fix c) va = Error _ va.
Proof.
  intros c va H_va_not_accessible.
  induction c; auto.
  simpl.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt (item_index a)) as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt (item_index a)); [exists true | exists false]; auto.
  rewrite <- H_acc.
  induction acc; auto.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  unfold os_accessible in H_va_not_accessible.
  rewrite <- a0 in H_acc.
  elim H_va_not_accessible; auto.
Qed.

Lemma cache_erase_owner : forall (s:state) (va:vadd) (pg:page),
  valid_state ctxt s ->
  Value _ pg = map_apply vadd_eq (cache_erase s) va ->
  page_owned_by pg = Os_PO osi.
Proof.
  intros s va pg H_s_valid H_pg_map.
  unfold cache_erase in H_pg_map.
  induction (os_ident_eq (st_active_os s) osi); try discriminate.

  generalize H_pg_map; rewrite cache_erase_apply with (st_cache s) va pg; auto.
  2: generalize H_s_valid; Intros_Valid_Hyp;
      destruct H_va_has_pa as [H_va_has_pa [[H_cache_size
        [H_cache_distinct H_cache_included]] H_valid_tlb]]; auto.
  intro H_pg_map_s.
  rewrite <- H_pg_map in H_pg_map_s.

  assert (os_accessible ctxt va) as H_va_accessible.
  unfold os_accessible.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va) as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  rewrite <- H_acc.
  induction acc; auto.
  rewrite cache_erase_not_accessible in H_pg_map.
  discriminate H_pg_map.
  unfold os_accessible; intro; rewrite <- H_acc in H; discriminate.

  generalize H_s_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma [H_ma_map [H_pg_content H_pg_map_mem]]]; auto.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  unfold valid_virtual_mapping in H_valid_virtmap.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg' H_pg'_map [H_va_accessible_owner H_va_not_accessible_owner]]; auto.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_ptype H_pt_page_owner]]; auto.
  unfold maddress_owner in H_pt_page_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_page_owner; auto.
  rewrite <- H_pg_map_mem in H_pg'_map.
  injection H_pg'_map; intro; subst pg'; clear H_pg'_map.
  unfold os_accessible in *.
  rewrite <- a.
  apply H_va_accessible_owner; auto.
Qed.
    

Lemma cache_erase_distinct : forall (c:cache),
  cache_distinct c -> cache_distinct (cache_erase_fix c).
Proof.
  intros c H_c_distinct.
  induction H_c_distinct; try constructor.
  simpl.
  induction (ctxt_vadd_accessible ctxt va); auto.
  constructor; auto.
  intros [pg' H_pg'_map].
  apply H.
  exists pg'.
  apply cache_erase_apply; auto.
Qed.

Lemma cache_erase_included : forall (s:state),
  valid_state ctxt s -> cache_included (state_erase s).
Proof.
  intros s H_s_valid.
  unfold cache_included in *.
  intros va pg H_pg_map'.
  unfold state_erase in H_pg_map'; simpl in H_pg_map'.
  unfold cache_erase in H_pg_map'.
  induction (os_ident_eq (st_active_os s) osi); try discriminate.
  assert (Value vadd pg = map_apply vadd_eq (st_cache s) va) as H_pg_map.
  apply cache_erase_apply; auto.
  generalize H_s_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [[H_cache_size
      [H_cache_distinct H_cache_included]] H_tlb_valid]]; auto.

  assert (os_accessible ctxt va) as H_va_accessible.
  unfold os_accessible.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va) as [acc H_acc_map].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  rewrite <- H_acc_map.
  induction acc; auto.
  rewrite cache_erase_not_accessible in H_pg_map'.
  discriminate H_pg_map'.
  unfold os_accessible; intro H_accessible;
    rewrite <- H_acc_map in H_accessible; discriminate H_accessible.
  assert (page_owned_by pg = Os_PO osi) as H_pg_owner.
  apply cache_erase_owner with s va; auto.
  unfold cache_erase.
  induction (os_ident_eq (st_active_os s) osi); auto.
  contradiction b; auto.

  elim valid_state_included_cache with ctxt s va pg; auto.
  intros ma H_va_mapped H_pg_map_mem.
  exists ma.
  generalize H_va_mapped; intro H_ma_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping in H_ma_map.
  destruct H_ma_map as [pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map] H_ma_map]].
  rewrite a in H_aos_map, H_p2m_map.
  unfold state_erase; simpl.
  split.
  exists pt_addr.
  exists aos.
  rewrite oss_erase_osi; auto.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  exists (mp_erase mp).
  exists (Page (PT_c (mp_erase mp)) (Os_PO osi)); auto.
  elim pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp; auto.

  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as [pt_addr' H_pt_addr'_map [H_pt_addr_type H_pt_addr_owner]]; auto.
  rewrite a; auto.
  rewrite a; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite <- a; auto.

  rewrite <- mp_erase_accessible; auto.

  assert (page_type (st_memory s) ma = Some RW) as H_ma_type.
  generalize H_s_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [[H_cache_size [H_cache_distinct H_cache_included]] H_tlb_valid]].
  unfold cache_included in H_cache_included.
  destruct H_cache_included with va pg as [ma' [H_ma'_map [H_ptype H_pg_map_mem']]]; auto.
  assert (ma = ma').
  apply va_mapped_same_ma with s va; auto.
  subst ma'; auto.
  split.
  unfold page_type in *.
  rewrite <- H_pg_map_mem in H_ma_type.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  induction (page_content pg); try discriminate.
  exists o; auto.

  rewrite rw_mem_erase_osi with (st_memory s) ma pg; auto.
  unfold page_type in H_ma_type.
  rewrite <- H_pg_map_mem in H_ma_type.
  induction (page_content pg); try discriminate.
  exists o; auto.
Qed.


Lemma tlb_erase_less_elements : forall (t:tlb),
  map_size (tlb_erase_fix t) <= map_size t.
Proof.
  intros t.
  induction t; auto.
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); simpl; omega.
Qed.

Lemma tlb_erase_apply : forall (t:tlb) (va:vadd) (ma:madd),
  tlb_distinct t ->
  Value _ ma = tlb_value (tlb_erase_fix t) va ->
  Value _ ma = tlb_value t va.
Proof.
  intros t va ma H_tlb_distinct H_ma_map.
  unfold tlb_value in *.
  induction H_tlb_distinct; auto.

  simpl; simpl in H_ma_map.
  induction (vadd_eq va va0).
  subst va0.
  induction (ctxt_vadd_accessible ctxt va).
  simpl in H_ma_map.
  induction (vadd_eq va va); auto.
  contradiction b; auto.
  rewrite IHH_tlb_distinct; auto.
  elim H.
  unfold tlb_value.
  exists ma; auto.

  induction (ctxt_vadd_accessible ctxt va0).
  simpl in H_ma_map.
  induction (vadd_eq va va0); try (contradiction b; assumption).
  apply IHH_tlb_distinct; auto.
  apply IHH_tlb_distinct; auto.
Qed.

Lemma tlb_erase_distinct : forall (c:tlb),
  tlb_distinct c -> tlb_distinct (tlb_erase_fix c).
Proof.
  intros t H_t_distinct.
  induction H_t_distinct; try constructor.
  simpl.
  induction (ctxt_vadd_accessible ctxt va); auto.
  constructor; auto.
  intros [ma' H_ma'_map].
  apply H.
  exists ma'.
  apply tlb_erase_apply; auto.
Qed.

Lemma tlb_erase_accessible : forall (t:tlb) (va:vadd),
  os_accessible ctxt va ->
  map_apply vadd_eq t va = map_apply vadd_eq (tlb_erase_fix t) va.
Proof.
  intros t va H_va_accessible.
  induction t; auto.

  simpl.
  unfold os_accessible in H_va_accessible.
  induction (vadd_eq va (item_index a)).
  rewrite <- a0.
  rewrite H_va_accessible.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  contradiction b; auto.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  contradiction b; auto.
Qed.

Lemma tlb_erase_not_accessible : forall (t:tlb) (va:vadd),
  ~(os_accessible ctxt va) ->
  map_apply vadd_eq(tlb_erase_fix t) va = Error _ va.
Proof.
  intros t va H_va_not_accessible.
  induction t; auto.
  simpl.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt (item_index a)) as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt (item_index a)); [exists true | exists false]; auto.
  rewrite <- H_acc.
  induction acc; auto.
  simpl.
  induction (vadd_eq va (item_index a)); auto.
  unfold os_accessible in H_va_not_accessible.
  rewrite <- a0 in H_acc.
  elim H_va_not_accessible; auto.
Qed.

Lemma tlb_erase_included : forall (s:state),
  valid_state ctxt s -> tlb_included (state_erase s).
Proof.
  intros s H_s_valid.
  unfold tlb_included.
  intros va ma H_ma_map'.
  unfold state_erase in H_ma_map'; simpl in H_ma_map'.
  unfold tlb_erase in H_ma_map'.
  induction (os_ident_eq (st_active_os s) osi); try discriminate.
  assert (Value vadd ma = map_apply vadd_eq (st_tlb s) va) as H_ma_map.
  apply tlb_erase_apply; auto.
  generalize H_s_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_cache_valid [H_tlb_size
      [H_tlb_distinct H_tlb_included]]]]; auto.

  assert (os_accessible ctxt va) as H_va_accessible.
  unfold os_accessible.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va) as [acc H_acc_map].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  rewrite <- H_acc_map.
  induction acc; auto.
  rewrite tlb_erase_not_accessible in H_ma_map'.
  discriminate H_ma_map'.
  unfold os_accessible; intro H_accessible;
    rewrite <- H_acc_map in H_accessible; discriminate H_accessible.

  elim valid_state_included_tlb with ctxt s va ma; auto.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping, vadd_mapping.
  intros pt_addr [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]]
    [mp [pt_page H_pt_page_map H_mp_map H_ma_map_mem]].
  rewrite a in H_aos_map, H_p2m_map.
  unfold va_mapped_to_ma, os_current_vadd_mapping_madd,
    hypervisor_mapping, vadd_mapping.
  unfold state_erase; simpl.
  exists pt_addr.
  exists aos.
  rewrite oss_erase_osi; auto.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  exists (mp_erase mp).
  exists (Page (PT_c (mp_erase mp)) (Os_PO osi)); auto.
  elim pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp; auto.

  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with (st_active_os s) aos p2m
    as [pt_addr' H_pt_addr'_map [H_pt_addr_type H_pt_addr_owner]]; auto.
  rewrite a; auto.
  rewrite a; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner.
  rewrite <- a; auto.

  rewrite <- mp_erase_accessible; auto.
Qed.

Lemma cache_erase_drop : forall (s:state) (v:vadd),
  cache_erase_fix (map_drop vadd_eq (st_cache s) v) =
  map_drop vadd_eq (cache_erase_fix (st_cache s)) v.
Proof.
  intros s va.
  induction (st_cache s); auto.

  simpl.
  induction (vadd_eq va (item_index a)).
  (* va is the first one *)
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va) as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  induction acc.
  rewrite <- a0, <- H_acc; simpl.
  induction (vadd_eq va (item_index a)); try contradiction b; auto.
  rewrite <- a0, <- H_acc; auto.
  (* va is not the first one *)
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); try (contradiction b; auto; fail).
  rewrite IHc; auto.
Qed.

Lemma tlb_erase_drop : forall (s:state) (v:vadd),
  tlb_erase_fix (map_drop vadd_eq (st_tlb s) v) =
  map_drop vadd_eq (tlb_erase_fix (st_tlb s)) v.
Proof.
  intros s va.
  induction (st_tlb s); auto.

  simpl.
  induction (vadd_eq va (item_index a)).
  (* va is the first one *)
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va) as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va); [exists true | exists false]; auto.
  induction acc.
  rewrite <- a0, <- H_acc; simpl.
  induction (vadd_eq va (item_index a)); try contradiction b; auto.
  rewrite <- a0, <- H_acc; auto.
  (* va is not the first one *)
  simpl.
  induction (ctxt_vadd_accessible ctxt (item_index a)); auto.
  simpl.
  induction (vadd_eq va (item_index a)); try (contradiction b; auto; fail).
  rewrite IHt; auto.
Qed.

Definition osi_exists (s:state) : Prop :=
  exists system, Value _ system = map_apply os_ident_eq (st_oss s) osi.

(* The erasure of a valid state results in a valid state *)
Lemma erasure_valid : forall s:state,
  osi_exists s -> valid_state ctxt s ->
  valid_state ctxt (state_erase s).
Proof.
  intros s [system H_system_map] H_s_valid.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]].
  unfold valid_state.
  unfold state_erase; simpl.

  repeat split; simpl.

  unfold trusted_os_not_hypercall in *.
  intros o system' H_o_trusted H_system'_map.
  case (os_ident_eq o osi).
  intro H_o_eq_osi; subst o.
  rewrite oss_erase_osi in H_system'_map.
  elim H_trstos_nothypcall with osi system'; auto.
  intro H_o_ne_osi.
  rewrite oss_erase_not_osi in H_system'_map; auto;
    discriminate H_system'_map.
  
  unfold running_os_not_hypercall in *.
  exists system.
  rewrite oss_erase_osi; auto.
  intro H_activity.
  unfold activity_erase in H_activity.
  induction (os_ident_eq (st_active_os s) osi); try discriminate H_activity.
  rewrite a in H_running_nothypcall.
  destruct H_running_nothypcall as [system' H_system'_map H_running_nothypcall].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  apply H_running_nothypcall; auto.

  unfold valid_hyper_exec_mode.
  unfold exec_erase, activity_erase.
  intro H_activity.
  induction (os_ident_eq (st_active_os s) osi);
    induction (ctxt_oss ctxt osi); auto.
  rewrite H_activity; auto.

  unfold valid_trusted_os_exec_mode.
  unfold exec_erase.
  intro H_osi_trusted; rewrite H_osi_trusted; auto.

  unfold valid_untrusted_os_exec_mode.
  unfold exec_erase, activity_erase.
  induction (os_ident_eq (st_active_os s) osi).
  intros H_osi_untrusted H_activity.
  rewrite H_osi_untrusted.
  rewrite H_activity; auto.
  intros H_osi_untrusted H_abs; discriminate H_abs.

  unfold valid_hypervisor in *.
  intros o pa p2m H_p2m_map.
  induction (os_ident_eq o osi).
  subst o.
  rewrite hyper_erase_osi in H_p2m_map.
  destruct H_valid_hyper with osi pa p2m as [[ma H_ma_map H_ma_owner] | H_error]; auto.
  left; exists ma; auto.
  apply mem_erase_owner; auto.
  rewrite hyper_erase_not_osi in H_p2m_map; auto.
  discriminate H_p2m_map.
  
  unfold valid_virtual_mapping in *.
  intros o pt_addr mp pt_page H_pt_page_map H_pt_page_content
    H_pt_page_owner va ma H_ma_map.
  induction (os_ident_eq o osi).
  subst o.
  elim pt_mem_erase_old_mem with (st_memory s) pt_addr pt_page mp; auto.
  intros old_mp H_old_mp_erase [old_pt_page H_old_pt_page_map H_old_pt_page_content].
  destruct H_valid_virtmap with osi pt_addr old_mp old_pt_page va ma as
    [pg H_pg_map [H_accessible H_not_accessible]]; auto.
  rewrite <- mem_erase_old_owner with (st_memory s) pt_addr.
  unfold maddress_owner.
  rewrite <- H_old_pt_page_map; auto.
  unfold maddress_owner; rewrite <- H_pt_page_map; auto.
  apply mp_erase_old.
  rewrite <- H_old_mp_erase; auto.
  assert (os_accessible ctxt va \/ ~ os_accessible ctxt va) as [H_va_accessible | H_va_not_accessible].
  unfold os_accessible; induction (ctxt_vadd_accessible ctxt va); [left | right]; auto.
  (* va accessible *)
  assert (exists content, page_content pg = content) as [content H_pg_content].
  induction (page_content pg); [exists (RW_c o) | exists (PT_c m) | exists Other_c]; auto.
  induction content.
  exists pg; auto.
  rewrite rw_mem_erase_osi with (st_memory s) ma pg; auto.
  exists o; auto.
  exists (Page (PT_c (mp_erase m)) (Os_PO osi)).
  rewrite pt_mem_erase_osi with (st_memory s) ma pg m; auto.
  split; simpl; auto.
  intro H_abs; contradiction H_abs.
  exists pg; auto.
  apply other_mem_erase_osi; auto.
  (* va not accessible *)
  rewrite H_old_mp_erase in H_ma_map.
  rewrite mp_erase_not_accessible in H_ma_map; auto; discriminate H_ma_map.
  elim mem_erase_not_os_owner with (st_memory s) pt_addr pt_page o; auto.

  unfold valid_current_page in *.
  intros o os H_os_map p2m H_p2m_map.
  induction (os_ident_eq o osi).
  subst o.
  rewrite hyper_erase_osi in H_p2m_map.
  rewrite oss_erase_osi in H_os_map.
  rewrite <- H_system_map in H_os_map.
  injection H_os_map; intro; subst os; clear H_os_map.
  destruct H_valid_currpg with osi system p2m as [pt_addr H_pt_addr_map [H_pt_addr_ptype H_pt_addr_owner]]; auto.
  exists pt_addr; auto.
  unfold page_type, maddress_owner in *.
  assert (exists pt_page, Value _ pt_page = map_apply madd_eq (st_memory s) pt_addr) as [pt_page H_pt_page_map].
  induction (map_apply madd_eq (st_memory s) pt_addr); try discriminate H_pt_addr_ptype.
  exists v; auto.
  rewrite <- H_pt_page_map in H_pt_addr_ptype, H_pt_addr_owner.
  assert (exists mp, PT_c mp = page_content pt_page) as [mp H_pt_page_content].
  induction (page_content pt_page); try discriminate H_pt_addr_ptype; auto.
  exists m; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp; auto.
  (* o must be osi, this case is absurd *)
  rewrite hyper_erase_not_osi in H_p2m_map; try discriminate H_p2m_map; auto.
  
  unfold injective_hyper_mappings in *.
  intros o p2m H_p2m_map.
  induction (os_ident_eq o osi).
  subst o.
  rewrite hyper_erase_osi in H_p2m_map. 
  apply H_injective_hyper with osi; auto.
  rewrite hyper_erase_not_osi in H_p2m_map; try discriminate H_p2m_map; auto.

  unfold va_has_valid_pa in *.
  intros o pt_padd pt_addr pt_page mp va ma
    H_pt_page_map H_pt_addr_map H_pt_page_content H_ma_map H_va_accessible.
  unfold hypervisor_mapping_madd in *.
  destruct H_pt_addr_map as [p2m H_p2m_map H_pt_addr_map].
  induction (os_ident_eq o osi).
  subst o.
  rewrite hyper_erase_osi in H_p2m_map; auto.
  elim pt_mem_erase_old_mem with (st_memory s) pt_addr pt_page mp; auto.
  intros old_mp H_old_mp_erase [old_pt_page H_old_pt_page_map H_old_pt_page_content].
  destruct H_va_has_pa with osi pt_padd pt_addr old_pt_page old_mp va ma as [pa [p2m' H_p2m'_map H_ma_hyp_map]]; auto.
  exists p2m; auto.
  rewrite mp_erase_old with old_mp va ma; auto.
  rewrite <- H_old_mp_erase; auto.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists pa; auto.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  elim mem_erase_owner with (st_memory s) pt_addr.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  destruct H_valid_hyper with osi pt_padd p2m as [[pt_addr' H_pt_addr'_map H_pt_addr'_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; auto.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite hyper_erase_not_osi in H_p2m_map; auto; discriminate H_p2m_map.

  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_erase.
  induction (os_ident_eq (st_active_os s) osi).
  apply Le.le_trans with (map_size (st_cache s)); auto.
  apply cache_erase_less_elements.
  elim max_cache_pos; auto.

  destruct H_valid_cache as [H_cache_size [H_cache_distinct H_cache_included]].
  unfold cache_erase.
  induction (os_ident_eq (st_active_os s) osi).
  apply cache_erase_distinct; auto.
  constructor.
  apply cache_erase_included; auto.

  destruct H_valid_tlb as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  unfold tlb_erase.
  induction (os_ident_eq (st_active_os s) osi).
  apply Le.le_trans with (map_size (st_tlb s)); auto.
  apply tlb_erase_less_elements.
  elim max_tlb_pos; auto.

  destruct H_valid_tlb as [H_tlb_size [H_tlb_distinct H_tlb_included]].
  unfold tlb_erase.
  induction (os_ident_eq (st_active_os s) osi).
  apply tlb_erase_distinct; auto.
  constructor.
  apply tlb_erase_included; auto.
Qed.

(* The erasure of a (valid) state results in an equivalent state *)
Lemma erase_state_equiv : forall (s:state),
  valid_state ctxt s -> osi_exists s ->
  osi_equivalent_upto_aos s (state_erase s).
Proof.
  intros s H_s_valid H_osi_exists.
  assert (valid_state ctxt (state_erase s)) as H_es_valid.
  apply erasure_valid; auto.
  unfold state_erase, osi_equivalent_upto_aos; simpl.

  generalize H_s_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  rename H_valid_hyper into H_valid_hyper_s.
  unfold valid_hypervisor in H_valid_hyper_s.
  generalize H_es_valid; Intros_Valid_Hyp; 
    clear H_trstos_nothypcall H_running_nothypcall H_valid_exmd
      H_valid_osexmd H_valid_untosexmd H_valid_virtmap H_valid_currpg
      H_injective_hyper H_va_has_pa.
  rename H_valid_hyper into H_valid_hyper_es.
  unfold valid_hypervisor in H_valid_hyper_es.

  repeat split.
  
  (* Same osi data *)
  rewrite oss_erase_osi; auto.
  
  (* s eq_hyp (erase s) *)
  unfold osi_hyper_mapping_eq.
  unfold padd_page; simpl.
  intros pa pg [ma H_ma_map H_pg_map] [val H_pg_content].
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  exists ma.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  rewrite rw_mem_erase_osi with (st_memory s) ma pg; auto.
  destruct H_valid_hyper_s with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.
  rewrite <- H_ma_map in H_error; discriminate H_error.
  exists val; auto.

  (* (erase s) eq_hyp s *)
  unfold osi_hyper_mapping_eq.
  unfold padd_page; simpl.
  intros pa pg [ma H_ma_map H_pg_map] [val H_pg_content].
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  rewrite hyper_erase_osi in H_p2m_map; auto.
  exists ma.
  exists p2m; auto.
  apply rw_mem_erase_old_mem; try exists val; auto.
  destruct H_valid_hyper_es with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error].
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; auto.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  unfold state_erase in H_ma_owner; simpl in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.
  rewrite <- H_ma_map in H_error; discriminate H_error.
  
  (* s eq_pt (erase s) *)
  unfold osi_pt_pages_eq.
  unfold padd_page; simpl.
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pg_content.
  destruct H_pt_addr_map as [p2m H_p2m_map H_pt_addr_map].
  exists (Page (PT_c (mp_erase mp)) (Os_PO osi)).
  exists pt_addr.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  apply pt_mem_erase_osi with pt_page; auto.
  destruct H_valid_hyper_s with osi pt_padd p2m as [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  exists (mp_erase mp); auto.
  intros va pa ma pg H_ma_map H_pg_map H_va_accessible H_ma_ptype H_ma_hyp_map.
  destruct H_ma_hyp_map as [p2m' H_p2m'_map H_ma_hyp_map].
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists ma.
  rewrite <- mp_erase_accessible; auto.
  unfold hypervisor_mapping_madd.
  split.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  apply rw_mem_erase_osi; auto.
  destruct H_valid_hyper_s with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_hyp_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.
  rewrite <- H_ma_hyp_map in H_error; discriminate H_error.
  unfold page_type in H_ma_ptype.
  rewrite <- H_pg_map in H_ma_ptype.
  induction (page_content pg); try discriminate H_ma_ptype.
  exists o; auto.

  (* (erase s) eq_pt s *)
  unfold osi_pt_pages_eq.
  unfold padd_page; simpl.
  intros pt_padd pt_page mp [pt_addr H_pt_addr_map H_pt_page_map] H_pt_page_content.
  destruct H_pt_addr_map as [p2m H_p2m_map H_pt_addr_map].
  rewrite hyper_erase_osi in H_p2m_map.
  elim pt_mem_erase_old_mem with (st_memory s) pt_addr pt_page mp; auto.
  intros old_mp H_old_mp_erase [old_pt_page H_old_pt_page_map H_old_pt_page_content].
  exists old_pt_page.
  exists pt_addr; auto.
  unfold hypervisor_mapping_madd; exists p2m; auto.
  exists old_mp; auto.
  intros va pa ma pg H_ma_map H_pg_map H_va_accessible H_ma_ptype H_ma_hyp_map.
  destruct H_ma_hyp_map as [p2m' H_p2m'_map H_ma_hyp_map].
  rewrite hyper_erase_osi in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists ma.
  rewrite mp_erase_accessible; auto.
  rewrite <- H_old_mp_erase; auto.
  split.
  exists p2m; auto.  
  apply rw_mem_erase_old_mem; auto.
  destruct H_valid_hyper_es with osi pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error].
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; auto.
  rewrite <- H_ma_hyp_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  unfold state_erase in H_ma_owner; simpl in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.
  rewrite <- H_ma_hyp_map in H_error; discriminate H_error.
  unfold page_type in H_ma_ptype.
  rewrite <- H_pg_map in H_ma_ptype.
  induction (page_content pg); try discriminate H_ma_ptype.
  exists o; auto.
  destruct H_valid_hyper_es with osi pt_padd p2m as [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error].
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  unfold state_erase in H_pt_addr_owner; simpl in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
Qed.

Lemma os_running_same_aos : forall (s s':state) (a:action),
  valid_state ctxt s -> valid_exec ctxt s a s' -> ~(hyper_running s) ->
  st_active_os s = st_active_os s'.
Proof.
  intros s s' a H_s_valid H_valid_exec H_not_hyper_running.

  induction a;
    inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post);
      unfold post_action in H_post.

  (* silent *)
  unfold post_action, post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_aos System_Cache Active_OS.
  rewrite <- H_aos; try (intro; discriminate); auto.
  
  
  unfold post_read, post_all_read in H_post.
  unfold pre_action, pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running]].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Active_OS;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite H_cmp; auto.

  unfold pre_action, pre_read_hyper, hyper_running in H_pre.
  unfold hyper_running in H_not_hyper_running.
  destruct H_pre as (H_pre, H_act).
  rewrite H_act in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold post_write, post_all_write in H_post.
  elim H_post.
  intros ma H_va_mapped [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_eq_newmem].
  destruct H_s' as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s v pg'; auto.
  intros ma' H_ma'_map H_pg'_map_mem.
  rewrite H_s'; auto.

  unfold pre_action, pre_write_hyper, hyper_running in H_pre.
  unfold hyper_running in H_not_hyper_running.
  destruct H_pre as (H_act, H_pre).
  rewrite H_act in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_new_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrusted, (ma', H_hyp_map, (H_ptype, (H_ma'_owner, (H_os_acc, (H_hcall, H_activity)))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H_post as [ma1 H_ma1_hyp_map [pt_addr1
    H_pt_addr1_curr_map [newmem1 H_update1 [system1 H_system1_map H_s']]]].
  rewrite H_s'; auto.
  
  unfold pre_action, pre_new_hyper, hyper_running in H_pre.
  destruct H_pre as (H_waiting, H_pre).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_waiting in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_del_untrusted, hyper_running in H_pre.
  destruct H_pre as (H_os_untrtusted, (H_os_hcall, (system, H_system_map, (p2m, H_hyp_mapping,
    (pt_addr, H_pt_addr_map, (H_pre_all_del, (H_activity, H_accessible))))))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H_post as [pt_addr H_curr_map [newmem H_drop H_s']].
  rewrite H_s'; auto.

  unfold pre_action, pre_del_hyper, pre_all_del, hyper_running in H_pre.
  destruct H_pre as (pt_addr, H_curr_map, (H_pre', (H_activity, H_not_accessible))). 
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_switch, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_pre').
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
 
  destruct H_post as [os1 H_os1_map H_s'].
  rewrite H_s'; auto.

  unfold pre_action, pre_lswitch_untrusted, hyper_running in H_pre.
  destruct H_pre as [H_o0_untrusted [aos H_aos_map [H_hcall [H_act H_all_lswitch]]]].
  unfold hyper_running in H_not_hyper_running.
  rewrite H_act in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H_post as [aos1 H_aos1_map H_s1'].
  rewrite H_s1'; simpl; auto.
  
  destruct H_post with Active_OS as [[H_triv H_interp] | [H_abs H_interp]].
  unfold state_component_interp in H_interp; auto.
  contradiction H_abs; split; intro H_abs'; discriminate H_abs'.

  unfold pre_action, pre_chmod, hyper_running in H_pre.
  destruct H_pre as (H_activity, H_hcall).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  destruct H_post as [ma1 [pg1 H_pg1_map [H_pg1_owner
    [H_pg1_content [p2m1 H_post_hyp1 [os1 H_os1_map H_s1']]]]]].
  rewrite H_s1'; simpl; auto.
  destruct H_post as [p2m1 H_p2m1_map [ma1 H_ma1_hyp_map [system1 H_system1_map H_s1']]].
  rewrite H_s1'; simpl; auto.

  unfold pre_action, pre_page_pin_untrusted, pre_all_pin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.

  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin, hyper_running in H_pre.
  destruct H_pre as (H_ctxt, (H_hcall, (H_activity, H_pre'))).
  unfold hyper_running in H_not_hyper_running.
  rewrite H_activity in H_not_hyper_running; elim H_not_hyper_running; auto.
Qed.

Lemma same_osi_components_same_erase_state : forall (s s':state),
  osi_exists s' -> valid_state ctxt s' ->
  ((st_active_os s <> osi /\ st_active_os s' <> osi) \/
  (st_aos_activity s = OSA_waiting /\ st_aos_activity s' = OSA_waiting)) ->
  (map_apply os_ident_eq (st_oss s) osi = map_apply os_ident_eq (st_oss s') osi) ->
  (map_apply os_ident_eq (st_hypervisor s) osi = map_apply os_ident_eq (st_hypervisor s') osi) ->
  (mem_erase (st_memory s) = mem_erase (st_memory s')) ->

  same_state (state_erase s) (state_erase s').
Proof.
  intros s s' H_osi_exists H_s'_valid H_active_os H_oss H_hypervisor H_memory. 
  unfold same_state.
  exists (st_cache (state_erase s')); exists (st_tlb (state_erase s')); auto.

  assert (valid_state ctxt (state_erase s')) as H_s'_erase_valid.
  apply erasure_valid; auto.
  split; [idtac | split]; try (generalize H_s'_erase_valid; Intros_Valid_Hyp;
    destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]]; assumption).

  assert (activity_erase s = activity_erase s') as H_activity_erase.
  unfold activity_erase.
  destruct H_active_os as [[H_osi_not_active H_osi_not_active'] | [H_aos_waiting H_aos_waiting']].
  induction (os_ident_eq (st_active_os s) osi); try (contradiction a; auto); auto.
  induction (os_ident_eq (st_active_os s') osi); try (contradiction a; auto); auto.
  induction (os_ident_eq (st_active_os s) osi); induction (os_ident_eq (st_active_os s') osi);
    try rewrite H_aos_waiting; try rewrite H_aos_waiting'; auto.

  unfold state_components_assign, state_components_update.
  intros cmp.
  induction cmp; unfold state_component_interp, state_erase; simpl.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  apply same_oss_erase; auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  unfold exec_erase.
  induction (ctxt_oss ctxt osi); auto.
  rewrite <- H_activity_erase; auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  apply same_hyper_erase; auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.

  right.
  split.
  intro H_abs; elim H_abs; auto.
  split; auto.
  
  right.
  split.
  intro H_abs; elim H_abs; auto.
  split; auto.
Qed.

(* Actions of OSs different to osi *)
Definition other_osi_action (s:state) (a:action) : Prop :=
  (st_active_os s <> osi) /\
  ((st_aos_activity s = OSA_running /\ a = Silent) \/ (exists va, a = Read va) \/
  (exists va, exists val, a = Write va val) \/
  (exists va, exists pa, a = New_trusted va pa) \/
  (exists va, a = Del_trusted va) \/
  (exists pa, a = Lswitch_trusted pa) \/
  (exists h, a = Hcall h) \/ (a = Ret_ctrl) \/
  (exists pa, exists t, a = Page_pin_trusted pa t) \/
  (exists pa, a = Page_unpin_trusted pa)).

(* Actions of the hypervisor for itself, or for an OS different to osi *)
Definition hyper_action_not_osi (s:state) (a:action) : Prop :=
  ((st_aos_activity s = OSA_waiting /\ a = Silent) \/
   (exists va, a = Read_hyper va) \/
   (exists va, exists val, a = Write_hyper va val) \/
   (exists va, exists ma, a = New_hyper va ma) \/
   (exists va, a = Del_hyper va) \/
   (exists o, o<>osi /\ (exists va, exists pa, a = New_untrusted o va pa)) \/
   (exists o, o<>osi /\ (exists va, a = Del_untrusted o va)) \/
   (exists o, a = Switch o) \/
   (exists o, o<>osi /\ (exists pa, a = Lswitch_untrusted o pa)) \/
   (st_active_os s <> osi /\ a = Chmod) \/
   (exists o, o<>osi /\ (exists pa, exists t, a = Page_pin_untrusted o pa t)) \/
   (exists o, o<>osi /\ (exists pa, a = Page_unpin_untrusted o pa))).
   
Lemma other_os_action_aos_running : forall (s:state) (a:action),
  pre_action ctxt a s -> other_osi_action s a ->
  st_aos_activity s = OSA_running.
Proof.
  intros s a H_pre H_action.
  destruct H_action as [H_osi_not_active H_action].

  destruct H_action as [[H_activity H_silent] | [[va H_read] |
    [[va [ma H_write]] | [[va [pa H_new]] | [[va H_del] | [[pa H_lswitch] | [ [hc H_hcall] |
      [H_retctrl | [[pa [t H_pin]] | [pa H_unpin]]]]]]]]]]; subst a; unfold pre_action in H_pre; auto.

  unfold pre_read in H_pre.
  destruct H_pre as [H_pre [H_acc H_os_running]]; auto.
  unfold pre_write in H_pre.
  destruct H_pre as [H_os_running H_pre]; auto.
  unfold pre_new_trusted in H_pre.
  destruct H_pre as [ma H_ma_map [H_ma_ptype [H_ma_owner [H_acc [H_trst H_os_running]]]]]; auto.
  unfold pre_del_trusted in H_pre.
  destruct H_pre as [pt_addr H_pt_addr_map [H_pre [H_os_running [H_trst H_acc]]]]; auto.
  unfold pre_lswitch_trusted in H_pre.
  destruct H_pre as [H_os_running [H_trst H_pre]]; auto.
  unfold pre_hcall in H_pre.
  destruct H_pre as [H_os_running H_untrst]; auto.
  unfold pre_page_pin_trusted in H_pre.
  destruct H_pre as [H_os_running H_pre]; auto.
  unfold pre_page_unpin_trusted in H_pre.
  destruct H_pre as [H_os_running H_pre]; auto.
Qed.

Lemma valid_exec_osi_exists : forall (s s':state) (a:action),
  osi_exists s -> valid_state ctxt s -> valid_exec ctxt s a s' ->
  osi_exists s'.
Proof.
  intros s s' a [system H_system_map] H_s_valid H_valid_exec.
  unfold osi_exists in *.
  inversion_clear H_valid_exec as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).
  induction a; unfold pre_action, post_action in *.
  (* silent *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb
    [H_valid_cache [H_valid_tlb H_assign]]]].
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp.
  exists system; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.

  (* read *)
  unfold post_read, post_all_read in H_post.
  unfold pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running']].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp.
  exists system; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.

  (* read hyper *)
  unfold post_read_hyper, post_all_read in H_post.
  unfold pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v0; auto.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp.
  exists system; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.

  (* write *)
  unfold post_write, post_all_write in H_post.
  elim H_post.
  intros ma H_ma_map [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_newmem].
  destruct H_s' as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  rewrite H_s'; exists system; auto.

  (* write hyper *)
  unfold post_write_hyper, post_all_write in H_post.
  destruct H_post as [ma H_ma_map [newmem H_newmem
    [pg H_pg_map [new_cache H_new_cache [new_tlb H_new_tlb  H_s']]]]].
  rewrite H_s'; exists system; auto.
  unfold post_new_untrusted, post_all_new in H_post.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map [newmem H_newmem [system' H_system'_map H_s']]]].
  induction (os_ident_eq o osi).
  subst o; rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite H_s'; simpl.
  exists (OS None (os_current_page system)); rewrite lemma_map_add_apply; auto.
  rewrite H_s'; simpl.
  exists system; rewrite lemma_map_add_apply_distinct; auto.

  (* new untrusted *)
  unfold post_new_trusted, post_all_new in H_post.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map [newmem H_newmem
    [system' H_system'_map H_s']]]].
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi).
  rewrite a; rewrite lemma_map_add_apply.
  exists (OS None (os_current_page system)); auto.
  rewrite a in H_system'_map.
  rewrite <- H_system_map in H_system'_map; injection H_system'_map;
    intro; subst system'; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* new hyper *)
  unfold post_new_hyper in H_post.
  destruct H_post as [aos H_aos_map
    [pt_addr H_pt_addr_map [newmem H_newmem H_s']]].
  rewrite H_s'; exists system; auto.

  (* del untrusted *)
  unfold post_del_untrusted in H_post.
  destruct H_post as [system' H_system'_map
    [p2m H_p2m_map [pt_addr H_pt_addr_map
      [newmem H_newmem H_s']]]].
  rewrite H_s'; simpl.
  induction (os_ident_eq o osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* del trusted *)
  unfold post_del_trusted, post_all_del in H_post.
  destruct H_post as [pt_addr H_pt_addr_map [newmem H_newmem H_s']].
  rewrite H_s'; exists system; auto.

  (* del hyper *)
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as [pt_addr H_pt_addr_map [newmem H_newmem H_s']].
  rewrite H_s'; exists system; auto.

  (* switch *)
  unfold post_switch in H_post.
  rewrite H_post; exists system; auto.

  (* lswitch trusted *)
  unfold post_lswitch_trusted, post_all_lswitch in *.
  destruct H_post as [system' H_system'_map H_s'].
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi).
  rewrite a; rewrite lemma_map_add_apply.
  exists (OS None p); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* lswitch untrusted *)
  unfold post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as [system' H_system'_map H_s'].
  rewrite H_s'; simpl.
  induction (os_ident_eq o osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None p); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* hcall *)
  unfold post_hcall in H_post.
  destruct H_post as [system' H_system'_map H_s'].
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi).
  rewrite a, lemma_map_add_apply.
  exists (OS (Some h) (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* ret ctrl *)
  unfold post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update in H_post.
  elim H_post with Guest_OSs; clear H_post.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; exists system; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.

  (* chmod *)
  unfold post_chmod in H_post.
  assert (exists exmd, exists activity,
    state_components_assign Exec_Mode OS_Activity exmd activity s s').
  destruct H_post as [[H_trstd H_assign] | [H_untrstd H_assign]];
    [exists EXMD_svc | exists EXMD_usr]; exists OSA_running; auto.
  destruct H as [exmd [activity H_assign]].
  unfold state_components_assign, state_components_update in H_assign.
  elim H_assign with Guest_OSs; clear H_post.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; exists system; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.

  (* pin trusted *)
  unfold post_action, post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content [p2m H_p2m_map
    [system' H_system'_map H_s']]]]]].
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* unpin trusted *)
  unfold post_action, post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [system' H_system'_map H_s']]].
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* pin untrusted *)
  unfold post_page_pin_untrusted, post_all_pin in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content [p2m H_p2m_map
    [system' H_system'_map H_s']]]]]].
  rewrite H_s'; simpl.
  induction (os_ident_eq o osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.

  (* unpin untrusted *)
  unfold post_page_unpin_untrusted, post_all_unpin in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [system' H_system'_map H_s']]].
  rewrite H_s'; simpl.
  induction (os_ident_eq o osi).
  rewrite a, lemma_map_add_apply.
  exists (OS None (os_current_page system')); auto.
  rewrite lemma_map_add_apply_distinct; auto.
  exists system; auto.
Qed.
  

Lemma os_running_erase_valid : forall (s s':state) (a:action),
  osi_exists s -> valid_state ctxt s -> valid_exec ctxt s a s' ->
  other_osi_action s a ->
  valid_exec ctxt (state_erase s) Silent (state_erase s').
Proof.
  intros s s' a H_osi_exists H_s_valid H_valid_exec H_action.
  constructor.
  unfold pre_action, pre_silent; auto.

  unfold post_action, post_silent.
  generalize H_action; intro H_action'.
  unfold other_osi_action in H_action'.
  destruct H_action' as [H_osi_not_active H_action'].
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).

  assert (~ hyper_running s) as H_os_running.
  intro H_hyper_running.
  unfold hyper_running in H_hyper_running.
  rewrite other_os_action_aos_running with s a in H_hyper_running;
    try discriminate H_hyper_running; auto.

  apply same_osi_components_same_erase_state; auto.
  apply valid_exec_osi_exists with s a; auto.
  apply valid_state_invariant with s a; auto.

  elim os_running_same_aos with s s' a; auto.
  rewrite os_data_isolation with ctxt s s' a osi; auto.
  rewrite other_os_running_same_hyp_mapping with ctxt s s' a osi; auto.

  (* Memory does not change *)
  destruct H_action' as [[H_activity H_silent] | [[va H_read] |
    [[va [val H_write]] | [[va [pa H_new]] | [[va H_del] | [[pa H_lswitch] | [ [hc H_hcall] |
      [H_retctrl | [[pa [t H_pin]] | [pa H_unpin]]]]]]]]]]; subst a; unfold pre_action in H_pre; auto.

  unfold post_action, post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  unfold content_of_address.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_OneComp H_interp H_mem System_Cache System_Memory.
  rewrite <- H_mem; try (intro; discriminate); auto.
  
  unfold post_action, post_read, post_all_read in H_post.
  unfold pre_action, pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] [H_v_accessible H_os_running']].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.

  elim H_assign with System_Memory;
    try (intros [H_abs H_cmp]; contradiction H_abs;
      split; intro; discriminate).
  unfold state_component_interp; intros [H_triv H_cmp];
    rewrite H_cmp; auto.

  (* write action *)
  unfold post_action, post_write, post_all_write in H_post.
  elim H_post.
  intros ma H_ma_map [newmem H_update H_s'].
  unfold update_rw_page in H_update.
  destruct H_update as [pg H_pg_map H_newmem].
  destruct H_s' as [pg' H_pg'_map_cache
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]].
  elim valid_state_included_cache_mem with ctxt s va pg'; auto.
  intros ma' H_ma_map_mem H_pg'_map_mem.
  rewrite H_s'; simpl.
  rewrite H_newmem.
  unfold va_mapped_to_ma in H_ma_map.
  assert (ma = ma').
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma'.
  rewrite <- H_pg_map in H_pg'_map_mem.
  injection H_pg'_map_mem; intro; subst pg'; clear H_pg'_map_mem.
  destruct H_ma_map_mem as [pt_addr H_curr_pt [mp H_pt_page_content H_ma_map_mem]].
  unfold os_current_vadd_mapping_madd in H_curr_pt.
  destruct H_curr_pt as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  unfold hypervisor_mapping in H_p2m_map.
  unfold vadd_mapping in H_pt_page_content.
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  assert (page_owned_by pg <> Os_PO osi).
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_valid_virtmap with (st_active_os s) pt_addr mp pt_page va ma as
    [pg' H_pg'_map [H_accessible H_not_accessible]]; auto.
  destruct H_valid_hyper with (st_active_os s) (os_current_page system) p2m as
    [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite <- H_pg_map in H_pg'_map.
  injection H_pg'_map; intro; subst pg'; clear H_pg'_map.
  rewrite H_accessible.
  intro H_abs; injection H_abs; intro H_osi_active;
    contradiction H_osi_not_active; auto.
  unfold pre_write in H_pre.
  destruct H_pre as [H_os_running' [H_va_accessible  H_pre]]; auto.
  rewrite mem_erase_add_not_osi with (st_memory s) ma pg (Page (RW_c (Some val)) (page_owned_by pg)); auto.

  (* new action *)
  unfold post_action, post_new_trusted, post_all_new in H_post.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map [newmem H_newmem
    [system H_system_map H_s']]]].
  rewrite H_s'; simpl.
  unfold update_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  rewrite H_newmem.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in H_p2m_map.
  assert (page_owned_by pt_page <> Os_PO osi).
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_valid_hyper with (st_active_os s) (os_current_page system) p2m as
    [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  intro H_pt_page_owner.
  rewrite H_pt_addr_owner in H_pt_page_owner.
  injection H_pt_page_owner; intro; contradiction H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page (Page (PT_c (map_add vadd_eq oldmap va ma))
    (page_owned_by pt_page)); auto.
  (* del action *)
  unfold post_action, post_del_trusted, post_all_del in H_post.
  destruct H_post as [pt_addr H_pt_addr_map [newmem H_newmem H_s']].
  rewrite H_s'; simpl.
  unfold drop_va_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  rewrite H_newmem.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  unfold hypervisor_mapping in H_p2m_map.
  assert (page_owned_by pt_page <> Os_PO osi).
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_valid_hyper with (st_active_os s) (os_current_page system) p2m as
    [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  intro H_pt_page_owner.
  rewrite H_pt_addr_owner in H_pt_page_owner.
  injection H_pt_page_owner; intro; contradiction H_osi_not_active; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page (Page (PT_c (map_drop vadd_eq oldmap va))
    (page_owned_by pt_page)); auto.
  (* lswitch action *)
  unfold post_action, post_lswitch_trusted, post_all_lswitch in H_post.
  destruct H_post as [system H_system_map H_s'].
  rewrite H_s'; auto.
  (* hcall action *)
  unfold post_action, post_hcall in H_post.
  destruct H_post as [system H_system_map H_s'].
  rewrite H_s'; auto.
  (* ret ctrl action *)
  unfold post_action, post_ret_ctrl in H_post.
  unfold state_components_assign, state_components_update in H_post.
  elim H_post with System_Memory; clear H_post.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  (* pin action *)
  unfold post_action, post_page_pin_trusted, post_all_pin in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content [p2m H_p2m_map
    [system H_system_map H_s']]]]]].
  rewrite H_s'; simpl.
  unfold hypervisor_mapping in H_p2m_map.
  rewrite mem_erase_add_not_osi with (st_memory s) ma pg
    (match t with
       | PT => Page (PT_c (map_empty vadd madd)) (Os_PO (st_active_os s))
       | RW => Page (RW_c None) (Os_PO (st_active_os s))
     end); auto.
  rewrite H_pg_owner; intro H_abs; discriminate H_abs.
  induction t; simpl; intro H_abs; injection H_abs; intro H_abs'; contradiction H_osi_not_active; auto.
  (* unpin action *)
  unfold post_action, post_page_unpin_trusted, post_all_unpin in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [system H_system_map H_s']]].
  rewrite H_s'; simpl.
  unfold hypervisor_mapping in H_p2m_map.
  unfold pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running' [H_trst [system' H_system'_map [H_not_curr
    [p2m' H_p2m'_map [ma' H_ma'_map [H_ma_ptype H_no_va]]]]]]].
  rewrite <- H_system_map in H_system'_map;
    injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map;
    injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma_map in H_ma'_map;
    injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold page_type in H_ma_ptype.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); [exists v | contradiction H_ma_ptype]; auto.
  rewrite mem_erase_add_not_osi with (st_memory s) ma pg (Page Other_c No_Owner); auto.
  generalize H_s_valid.
  Intros_Valid_Hyp.
  destruct H_valid_hyper with (st_active_os s) pa p2m as
    [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  rewrite <- H_pg_map in H_ma_owner; auto.
  intro H_pt_page_owner.
  rewrite H_ma_owner in H_pt_page_owner.
  injection H_pt_page_owner; intro; contradiction H_osi_not_active; auto.
  rewrite <- H_ma_map in H_error; discriminate H_error.
  simpl; intro H_abs; discriminate H_abs.
Qed.

Lemma hyper_mapping_madd_owner : forall (s:state) (pa:padd) (ma:madd) (o:os_ident),
  valid_state ctxt s ->
  hypervisor_mapping_madd (st_hypervisor s) o pa ma->
  maddress_owner (st_memory s) ma = Os_PO o.
Proof.
  intros s pa ma o H_s_valid H_ma_map.

  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_valid_hyper with o pa p2m as
    [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  assumption.
  rewrite <- H_ma_map in H_error; discriminate H_error.
Qed.

Lemma accessible_va_mapped_to_ma_owner : forall (s:state) (va:vadd) (ma:madd) (o:os_ident),
  valid_state ctxt s ->
  os_accessible ctxt va -> va_mapped_to_ma s o va ma ->
  maddress_owner (st_memory s) ma = Os_PO o.
Proof.
  intros s va ma o H_s_valid H_va_accessible H_ma_map.

  unfold va_mapped_to_ma in H_ma_map.
  destruct H_ma_map as [pt_addr H_curr_pt [mp H_pt_page_content H_ma_map]].
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_curr_pt.
  destruct H_curr_pt as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  unfold vadd_mapping in H_pt_page_content.
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_valid_virtmap with o pt_addr mp pt_page va ma as
    [pg H_pg_map [H_accessible H_not_accessible]]; auto.
  destruct H_valid_hyper with o (os_current_page system) p2m as
    [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  unfold maddress_owner.
  rewrite <- H_pg_map.
  apply H_accessible; auto.
Qed.
  
Lemma not_accessible_va_mapped_to_ma_owner : forall (s:state) (va:vadd) (ma:madd) (o:os_ident),
  valid_state ctxt s ->
  ~ (os_accessible ctxt va) -> va_mapped_to_ma s o va ma ->
  maddress_owner (st_memory s) ma = Hyp_PO.
Proof.
  intros s va ma o H_s_valid H_va_not_accessible H_ma_map.

  unfold va_mapped_to_ma in H_ma_map.
  destruct H_ma_map as [pt_addr H_curr_pt [mp H_pt_page_content H_ma_map]].
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_curr_pt.
  destruct H_curr_pt as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  unfold vadd_mapping in H_pt_page_content.
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_valid_virtmap with o pt_addr mp pt_page va ma as
    [pg H_pg_map [H_accessible H_not_accessible]]; auto.
  destruct H_valid_hyper with o (os_current_page system) p2m as
    [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  unfold maddress_owner.
  rewrite <- H_pg_map.
  apply H_not_accessible; auto.
Qed.


Lemma hyper_action_erase_valid : forall (s s':state) (a:action),
  osi_exists s -> valid_state ctxt s -> valid_exec ctxt s a s' ->
  hyper_action_not_osi s a ->
  valid_exec ctxt (state_erase s) Silent (state_erase s').
Proof.
  intros s s' a H_osi_exists H_s_valid H_valid_exec H_action.
  constructor.
  unfold pre_action, pre_silent; auto.

  unfold post_action, post_silent.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).

  assert (osi_exists s') as H_osi_exists_s'.
  apply valid_exec_osi_exists with s a; auto.
  assert (valid_state ctxt s') as H_s'_valid.
  apply valid_state_invariant with s a; auto.

  unfold hyper_action_not_osi in H_action.
  destruct H_action as [[H_waiting H_silent] | [[va H_read_hyp] |
    [[va [val H_write_hyp]] | [[va [ma H_new_hyp]] |
      [[va H_del_hyp] | [[o [H_o_ne_osi [va [pa H_new_untr]]]] | 
    [[o [H_o_ne_osi [va H_del_untr]]] | [[o H_switch] | [[o [H_o_ne_osi [pa H_lswitch]]] | 
    [[H_osi_not_active H_chmod] | [[o [H_o_ne_osi [pa [t H_pin_untr]]]] | 
      [o [H_o_ne_osi [pa H_unpin_untr]]]]]]]]]]]]]]; subst a; unfold pre_action in H_pre;
  unfold post_action in H_post.

  (* Silent *)
  unfold post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid
    H_assign]]]]; auto.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_exec_mode System_Cache
    Active_OS Guest_OSs Exec_Mode.
  Simp_Inv_ThreeComps H_interp H_activity H_hyper H_mem System_Cache
    OS_Activity Hypervisor_SC System_Memory.

  apply same_osi_components_same_erase_state; auto.
  rewrite <- H_aos; try (intro; discriminate);
    rewrite <- H_activity; try (intro; discriminate); auto.
  rewrite <- H_oss; try (intro; discriminate); auto.
  rewrite <- H_hyper; try (intro; discriminate); auto.
  rewrite <- H_mem; try (intro; discriminate); auto.

  unfold post_read_hyper, post_all_read in H_post.
  unfold pre_action, pre_read_hyper, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_map H_ptype]] H_hyper_running].
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  unfold page_type in H_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  destruct H_post with ma pg as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply valid_state_included_cache_cache with ctxt ma; auto.
  unfold state_components_assign, state_components_update in H_assign.

  apply same_osi_components_same_erase_state; auto.
  right.
  elim H_assign with OS_Activity.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.
  unfold hyper_running in H_hyper_running.
  elim H_assign with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.
  elim H_assign with Hypervisor_SC.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.
  elim H_assign with System_Memory.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro; discriminate.

  (* write_hyper action *)
  unfold post_write_hyper, post_all_write in H_post.
  destruct H_post as [ma H_ma_map [newmem H_newmem
    [pg H_pg_map [new_cache H_new_cache [new_tlb H_new_tlb  H_s']]]]].
  unfold update_rw_page in H_newmem.
  destruct H_newmem as [pg' H_pg'_map H_newmem].
  assert (pg=pg').
  elim valid_state_included_cache_mem with ctxt s va pg; auto.
  intros ma' H_ma'_map H_pg_map_mem.
  assert (ma=ma').
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma'.
  rewrite <- H_pg_map_mem in H_pg'_map; injection H_pg'_map; auto.
  subst pg'.

  apply same_osi_components_same_erase_state; try (rewrite H_s'; auto; fail); auto.
  right; rewrite H_s'; simpl.
  unfold pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running; auto.
  rewrite H_s'; simpl.
  rewrite H_newmem; simpl.
  assert (page_owned_by pg <> Os_PO osi).
  intro H_pg_owner.
  unfold pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running [ma' H_ma'_map [H_ma_ptype H_ma_owner]]].
  rewrite <- va_mapped_same_ma with s va ma ma' in H_ma_owner; auto.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg'_map in H_ma_owner.
  rewrite H_ma_owner in H_pg_owner.
  discriminate H_pg_owner.
  apply valid_state_included_tlb_mem with ctxt; auto.
  rewrite mem_erase_add_not_osi with (st_memory s) ma pg (Page (RW_c (Some val)) (page_owned_by pg)); auto.

  (* new_hyper action *)
  unfold post_new_hyper in H_post.
  destruct H_post as [aos H_aos_map [pt_addr H_pt_addr_map [newmem H_newmem H_s']]].
  unfold update_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  apply same_osi_components_same_erase_state; try (rewrite H_s'; auto; fail); auto.
  right.
  rewrite H_s'; simpl.
  unfold pre_new_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite H_s'; simpl.
  rewrite H_newmem; simpl.
  assert (page_owned_by pt_page = Os_PO (st_active_os s)) as H_pt_page_owner.
  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [aos' H_aos'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_aos_map in H_aos'_map.
  injection H_aos'_map; intro; subst aos'; clear H_aos'_map.
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_pt_addr_ptype H_pt_addr_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  case (os_ident_eq (st_active_os s) osi).
  intro H_osi_active.
  unfold pre_new_hyper in H_pre.
  destruct H_pre as [H_hyper_running [H_ma_ptype [H_ma_owner [H_exists_p2m H_va_not_accessible]]]].
  rewrite pt_mem_erase_add_not_acc_osi with (st_memory s) va pt_addr ma pt_page oldmap; auto.
  rewrite <- H_osi_active; auto.
  intro H_osi_not_active.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page
    (Page (PT_c (map_add vadd_eq oldmap va ma)) (page_owned_by pt_page)); auto.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_osi_not_active; auto.
  simpl.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_osi_not_active; auto.

  (* del_hyper action *)
  unfold post_del_hyper, post_all_del in H_post.
  destruct H_post as [pt_addr H_pt_addr_map [newmem H_newmem H_s']].
  unfold drop_va_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_del_hyper in H_pre.
  destruct H_pre as [pt_addr' H_pt_addr'_map [H_pre [H_hyper_running H_va_not_accessible]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite H_newmem; simpl.
  assert (page_owned_by pt_page = Os_PO (st_active_os s)) as H_pt_page_owner.
  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  destruct H_valid_currpg with (st_active_os s) aos p2m as
    [pt_addr' H_pt_addr'_map [H_pt_addr_ptype H_pt_addr_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  case (os_ident_eq (st_active_os s) osi).
  intro H_osi_active.
  unfold pre_new_hyper in H_pre.
  destruct H_pre as [pt_addr' H_pt_addr'_map [H_pre [H_hyper_running H_va_not_accessible]]].
  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd, hypervisor_mapping in *.
  destruct H_pt_addr_map as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  destruct H_pt_addr'_map as [aos' H_aos'_map [p2m' H_p2m'_map H_pt_addr'_map]].
  rewrite <- H_aos_map in H_aos'_map.
  injection H_aos'_map; intro; subst aos'; clear H_aos'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold pre_all_del, vadd_mapping in H_pre.
  destruct H_pre as [oldmap' [pt_page' H_pt_page'_map H_pt_page'_content] [ma H_ma_map]].
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page'_content.
  injection H_pt_page'_content; intro; subst oldmap'; clear H_pt_page'_content.
  rewrite pt_mem_erase_drop_osi with (st_memory s) va pt_addr pt_page oldmap; auto.
  rewrite <- H_osi_active; auto.
  intro H_osi_not_active.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page
    (Page (PT_c (map_drop vadd_eq oldmap va)) (page_owned_by pt_page)); auto.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_osi_not_active; auto.
  simpl.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_osi_not_active; auto.
  (* new_untrusted action *)
  unfold post_new_untrusted, post_all_new in H_post.
  destruct H_post as [ma H_ma_map [pt_addr H_pt_addr_map [newmem H_newmem [system H_system_map H_s']]]].
  unfold update_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_new_untrusted in H_pre.
  destruct H_pre as [H_o_untrst [ma' H_ma'_map [H_ma_ptype
    [H_ma_owner [H_va_accessible [H_hcall H_hyper_running]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite H_newmem; simpl.
  assert (page_owned_by pt_page = Os_PO o) as H_pt_page_owner.
  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  destruct H_valid_currpg with o system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_addr_ptype H_pt_addr_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page
    (Page (PT_c (map_add vadd_eq oldmap va ma)) (page_owned_by pt_page)); auto.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_o_ne_osi; auto.
  simpl.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_o_ne_osi; auto.
  (* del_untrusted action *)
  unfold post_del_untrusted in H_post.
  destruct H_post as [system H_system_map [p2m H_p2m_map [pt_addr H_pt_addr_map [newmem H_newmem H_s']]]].
  unfold drop_va_pt_page in H_newmem.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pt_page_content H_newmem]].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_del_untrusted in H_pre.
  destruct H_pre as [H_o_untrst [H_hcall [system' H_system'_map
    [p2m' H_p2m'_map [pt_addr' H_pt_addr'_map
      [H_pre [H_hyper_running H_va_accessible]]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite H_newmem; simpl.
  assert (page_owned_by pt_page = Os_PO o) as H_pt_page_owner.
  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with o system p2m as
    [pt_addr' H_pt_addr'_map [H_pt_addr_ptype H_pt_addr_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite mem_erase_add_not_osi with (st_memory s) pt_addr pt_page
    (Page (PT_c (map_drop vadd_eq oldmap va)) (page_owned_by pt_page)); auto.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_o_ne_osi; auto.
  simpl.
  intro H_pt_page_owner'.
  rewrite H_pt_page_owner in H_pt_page_owner'.
  injection H_pt_page_owner'; intro.
  contradiction H_o_ne_osi; auto.
  (* switch action *)
  unfold post_switch in H_post.
  apply same_osi_components_same_erase_state; auto; rewrite H_post; simpl; auto.
  right.
  unfold pre_switch in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  (* lswitch action *)
  unfold post_lswitch_untrusted, post_all_lswitch in H_post.
  destruct H_post as [system H_system_map H_s'].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_lswitch_untrusted in H_pre.
  destruct H_pre as [H_o_untrst [system' H_system'_map [H_hcall [H_hyper_running H_pre]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  (* chmod action *)
  unfold post_chmod in H_post.
  assert (exists exmd, exists activity,
    state_components_assign Exec_Mode OS_Activity exmd activity s s').
  destruct H_post as [[H_trstd H_assign] | [H_untrstd H_assign]];
    [exists EXMD_svc | exists EXMD_usr]; exists OSA_running; auto.
  destruct H as [exmd [activity H_assign]].
  unfold state_components_assign, state_components_update in H_assign.
  apply same_osi_components_same_erase_state; auto.
  left.
  elim H_assign with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  elim H_assign with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  elim H_assign with Hypervisor_SC.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  elim H_assign with System_Memory.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  (* pin_untrusted action *)
  unfold post_page_pin_untrusted, post_all_pin in H_post.
  destruct H_post as [ma [pg H_pg_map [H_pg_owner [H_pg_content [p2m H_p2m_map
    [system H_system_map H_s']]]]]].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_page_pin_untrusted in H_pre.
  destruct H_pre as [H_untrst [H_hcall [H_hyper_running H_pre]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  elim mem_erase_add_not_osi with (st_memory s) ma pg (new_page (Some t) (Os_PO o)); auto.
  intro H_pg_owner'.
  rewrite H_pg_owner in H_pg_owner'; discriminate H_pg_owner'.
  induction t; simpl;
    intro H_abs; injection H_abs; intro H_abs'; contradiction H_o_ne_osi; auto.
  (* unpin_untrusted action *)
  unfold post_page_unpin_untrusted, post_all_unpin in H_post.
  destruct H_post as [p2m H_p2m_map [ma H_ma_map [system H_system_map H_s']]].
  apply same_osi_components_same_erase_state; auto; rewrite H_s'; simpl; auto.
  right.
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as [H_untrst [H_hcall [H_hyper_running H_pre]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  rewrite lemma_map_add_apply_distinct; auto.
  unfold pre_page_unpin_untrusted in H_pre.
  destruct H_pre as [H_untrst [H_hcall [H_hyper_running H_pre]]].
  unfold pre_all_unpin in H_pre.
  destruct H_pre as [system' H_system'_map [H_pa_ne_curr [p2m' H_p2m'_map [ma' H_ma'_map [H_ma_ptype H_no_va]]]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in *.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold page_type in H_ma_ptype.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma);
    [exists v | contradiction H_ma_ptype]; auto.
  elim mem_erase_add_not_osi with (st_memory s) ma pg (new_page None No_Owner); auto.
  intro H_pg_owner.
  contradiction H_o_ne_osi.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_valid_hyper with o pa p2m as [[ma' H_ma'_map H_ma_owner] | H_error]; auto.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold maddress_owner in H_ma_owner.
  rewrite <- H_pg_map in H_ma_owner; auto.
  rewrite H_pg_owner in H_ma_owner; injection H_ma_owner; auto.
  rewrite <- H_ma_map in H_error; discriminate H_error.
  simpl; intro H_abs; discriminate H_abs.
Qed.


Lemma read_osi_erase_valid : forall (s s': state) (va:vadd),
  valid_state ctxt s -> pre_read va ctxt s -> post_read va s s' ->
  st_active_os s = osi ->
  pre_read va ctxt (state_erase s) /\ post_read va (state_erase s) (state_erase s').
Proof.
  intros s s' va H_s_valid.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].
  intros H_pre H_post H_osi_active.
  
  unfold pre_action, pre_read, pre_all_read in H_pre.
  destruct H_pre as [[ma [H_ma_mapped H_ma_type]] [H_va_accessible H_osi_running]].
  
  generalize H_ma_mapped; intro H_ma_mapped'.
  unfold va_mapped_to_ma in H_ma_mapped.
  destruct H_ma_mapped as [pt_addr H_curr_pt [mp H_mp_map H_ma_map]].
  unfold os_current_vadd_mapping_madd in *; simpl.
  destruct H_curr_pt as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  unfold hypervisor_mapping in H_p2m_map.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as [pt_page H_pt_page_map H_pt_page_content].
  assert (page_owned_by pt_page = Os_PO osi) as H_pt_page_owner.
  destruct H_valid_hyper with osi (os_current_page aos) p2m
    as [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error].
  rewrite <- H_osi_active; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  assert (va_mapped_to_ma (state_erase s) (st_active_os (state_erase s)) va ma)
    as H_ma_map_erase.
  unfold vadd_mapping.
  exists pt_addr.
  exists aos.
  unfold state_erase; simpl.
  rewrite oss_erase_osi; rewrite <- H_osi_active; auto.
  exists p2m; auto.
  unfold hypervisor_mapping.
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; rewrite <- H_osi_active; auto.
  exists (mp_erase  mp).
  unfold vadd_mapping.
  exists (Page (PT_c (mp_erase mp)) (Os_PO osi)); auto.
  unfold state_erase; simpl.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp; auto.
  rewrite <- mp_erase_accessible; auto.

  unfold pre_action, pre_read, pre_all_read.
  repeat split; auto.
  exists ma; split; auto.
  unfold page_type in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_type.
  exists v; auto.
  rewrite <- H_pg_map in H_ma_type.
  unfold state_erase; simpl.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  elim accessible_va_mapped_to_ma_owner with s va ma osi; auto.
  unfold maddress_owner; rewrite <- H_pg_map; auto.
  rewrite <- H_osi_active; auto.
  induction (page_content pg); try discriminate.
  exists o; auto.
  unfold os_running in *.
  unfold state_erase, activity_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.

  unfold post_action, post_read, post_all_read in *.
  intros ma' pg' H_ma'_map H_pg'_map.
  assert (Some ma' = va_mapped_to_ma_cache s va) as H_ma'_map_old.
  unfold va_mapped_to_ma_cache in *.
  unfold state_erase in H_ma'_map; simpl in H_ma'_map.
  assert (exists it, it = map_apply vadd_eq (tlb_erase s) va) as [it H_it_map].
  induction (map_apply vadd_eq (tlb_erase s) va).
  exists (Value vadd v); auto.
  exists (Error madd e); auto.
  rewrite <- H_it_map in H_ma'_map.
  induction it.
  injection H_ma'_map; intro; subst v; clear H_ma'_map.
  unfold tlb_erase in H_it_map.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- tlb_erase_accessible in H_it_map; auto.
  rewrite <- H_it_map; auto.
  unfold tlb_erase in H_it_map.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- tlb_erase_accessible in H_it_map; auto.
  rewrite <- H_it_map.
  rewrite oss_erase_osi in H_ma'_map.
  rewrite <- a, <- H_aos_map in H_ma'_map.
  rewrite <- H_aos_map.
  rewrite a, hyper_erase_osi in H_ma'_map; auto.
  rewrite <- a, <- H_p2m_map in H_ma'_map.
  rewrite <- H_p2m_map.
  rewrite <- H_pt_addr_map in H_ma'_map.
  rewrite <- H_pt_addr_map.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp in H_ma'_map; auto.
  simpl in H_ma'_map.
  rewrite <- mp_erase_accessible in H_ma'_map; auto.
  rewrite <- H_pt_page_map.
  rewrite <- H_pt_page_content.
  rewrite <- H_ma_map in H_ma'_map.
  rewrite <- H_ma_map; auto.

  assert (Some pg' = va_mapped_to_pg_cache s va) as H_pg'_cache_map_old.
  unfold va_mapped_to_pg_cache in H_pg'_map.
  rewrite <- H_ma'_map in H_pg'_map.
  unfold va_mapped_to_pg_cache.
  rewrite <- H_ma'_map_old.
  unfold state_erase, cache_erase in H_pg'_map; simpl in H_pg'_map.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- cache_erase_accessible in H_pg'_map; auto.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_it_map in H_pg'_map.
  rewrite <- H_it_map.
  induction it; auto.
  assert (ma = ma').
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma'.

  assert (Value _ pg' = map_apply madd_eq (mem_erase (st_memory s)) ma) as H_pg'_map'.
  induction (map_apply madd_eq (mem_erase (st_memory s)) ma); try discriminate.
  injection H_pg'_map; intro; subst v; auto.

  unfold page_type in H_ma_type.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); try discriminate.
  exists v; auto.
  rewrite <- H_pg_map in H_ma_type.
  assert (exists val, RW_c val = page_content pg) as [val H_pg_type].
  induction (page_content pg); try discriminate.
  exists o; auto.
  rewrite rw_mem_erase_osi with (st_memory s) ma pg in H_pg_map; auto.
  rewrite <- H_pg_map.
  rewrite <- H_pg'_map'; auto.

  elim accessible_va_mapped_to_ma_owner with s va ma osi; auto.
  unfold maddress_owner; simpl.
  rewrite <- H_pg_map; auto.
  rewrite <- a; auto.
  exists val; auto.  

  destruct H_post with ma' pg' as [new_cache H_cache_add
    [new_tlb H_tlb_add H_assign]]; auto.
  unfold state_components_assign, state_components_update in H_assign.

  exists (st_cache (state_erase s')).
  unfold cache_add in *.
  destruct H_cache_add as [H_pg'_map_new [H_old_mapping [H_new_size H_new_distinct]]].
  repeat split.
  unfold state_erase; simpl.
  unfold cache_erase.

  elim H_assign with Active_OS.
  intros [H_triv H_aos].
  unfold state_component_interp in H_aos.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- cache_erase_accessible; auto.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv'] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache.
  rewrite H_new_cache; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  intros va' pg'' H_va_ne_va' H_pg''_new_map.

  unfold state_erase, cache_erase in *.
  simpl.
  simpl in H_pg''_new_map.
  elim H_assign with Active_OS;
    try (intros [H_abs H_interp]; contradiction H_abs; split;
      intro H_abs'; discriminate H_abs').
  intros [H_triv H_aos].
  unfold state_component_interp in H_aos.
  rewrite <- H_aos in H_pg''_new_map.
  clear H_triv.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv'] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache.
  rewrite H_new_cache in H_pg''_new_map; auto.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va') as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va'); [exists true | exists false]; auto.
  induction acc.
  rewrite <- cache_erase_accessible in H_pg''_new_map; auto.
  rewrite <- cache_erase_accessible; auto.
  unfold os_accessible; auto.
  unfold os_accessible; auto.  

  rewrite cache_erase_not_accessible in H_pg''_new_map; try discriminate.
  unfold os_accessible.
  intro H_abs; rewrite <- H_acc in H_abs; discriminate.

  unfold state_erase, cache_erase; simpl.
  elim H_assign with Active_OS;
    try (intros [H_abs H_interp]; contradiction H_abs; split;
      intro H_abs'; discriminate H_abs').
  intros [H_triv H_aos].
  unfold state_component_interp in H_aos.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv'] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache.
  rewrite H_new_cache.
  apply Le.le_trans with (map_size new_cache); auto.
  apply cache_erase_less_elements.  


  unfold state_erase, cache_erase; simpl.
  elim H_assign with Active_OS;
    try (intros [H_abs H_interp]; contradiction H_abs; split;
      intro H_abs'; discriminate H_abs').
  intros [H_triv H_aos].
  unfold state_component_interp in H_aos.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_Cache.
  intros [[H_abs H_triv'] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_cache.
  rewrite H_new_cache.
  apply cache_erase_distinct; auto.

  elim H_assign with Active_OS;
    try (intros [H_abs H_interp]; contradiction H_abs; split;
      intro H_abs'; discriminate H_abs').
  intros [H_triv H_aos].
  unfold state_component_interp in H_aos.
  clear H_triv.

  exists (st_tlb (state_erase s')).
  unfold tlb_add in *.
  destruct H_tlb_add as [H_ma'_map_new [H_old_mapping [H_new_size H_new_distinct]]].
  repeat split.
  unfold state_erase; simpl.
  unfold tlb_erase.

  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- tlb_erase_accessible; auto.
  elim H_assign with System_TLB.
  intros [[H_triv' H_abs] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_tlb.
  rewrite H_new_tlb; auto.
  
  intros va' ma'' H_va_ne_va' H_ma''_new_map.
  unfold state_erase, tlb_erase in *.
  simpl.
  simpl in H_ma''_new_map.
  rewrite <- H_aos in H_ma''_new_map.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_TLB.
  intros [[H_triv' H_abs] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_tlb.
  rewrite H_new_tlb in H_ma''_new_map; auto.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va') as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va'); [exists true | exists false]; auto.
  induction acc.
  rewrite <- tlb_erase_accessible in H_ma''_new_map; auto.
  rewrite <- tlb_erase_accessible; auto.
  unfold os_accessible; auto.
  unfold os_accessible; auto.  

  rewrite tlb_erase_not_accessible in H_ma''_new_map; try discriminate.
  unfold os_accessible.
  intro H_abs; rewrite <- H_acc in H_abs; discriminate.

  unfold state_erase, tlb_erase; simpl.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_TLB.
  intros [[H_triv' H_abs] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_tlb.
  rewrite H_new_tlb.
  apply Le.le_trans with (map_size new_tlb); auto.
  apply tlb_erase_less_elements.  

  unfold state_erase, tlb_erase; simpl.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with System_TLB.
  intros [[H_triv' H_abs] H_interp]; contradiction H_abs; auto.
  intros [H_triv' [H_new_cache H_new_tlb]].
  unfold state_component_interp in H_new_tlb.
  rewrite H_new_tlb.
  apply tlb_erase_distinct; auto.

  assert (activity_erase s = activity_erase s') as H_activity.
  unfold activity_erase.
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  elim H_assign with OS_Activity;
    try (intros [H_abs H_interp]; contradiction H_abs; split;
      intro H_abs'; discriminate H_abs').
  intros [H_triv H_activity].
  unfold state_component_interp in H_activity; auto.

  unfold state_components_assign, state_components_update.
  intro cmp; elim H_assign with cmp.
  intros [[H_cmp_ne_cache H_cmp_ne_tlb] H_interp].
  unfold state_component_interp; induction cmp;
    simpl; simpl in H_interp;
      try (left; repeat split; try (intro; discriminate);
        rewrite <- H_interp; auto).
  left; repeat split; try (intro; discriminate).
  unfold exec_erase.
  induction (ctxt_oss ctxt osi); auto.
  rewrite <- H_activity.
  induction (activity_erase s); auto.

  left; repeat split; try (intro; discriminate); auto.
  right; repeat split.
  intros [H_abs H_triv]; contradiction H_abs; auto.
  right; repeat split.
  intros [H_triv H_abs]; contradiction H_abs; auto.
  
  intros [H_cmp_ne_cache_tlb [H_cache H_tlb]].
  unfold state_component_interp in H_cache, H_tlb.
  right; auto.
Qed.


(* Mover a Cache_invariant *)
Lemma fix_cache_invalid_ma : forall (s:state) (c:cache) (va:vadd) (ma:madd),
  valid_state ctxt s -> va_mapped_to_ma s (st_active_os s) va ma ->
  map_valid_index_fix vadd_eq (fix_cache_synonym s c ma) va = false.
Proof.
  intros s c va ma H_s_valid H_ma_mapped.
  induction c; auto.

  simpl.
  induction a.
  assert (exists opt, opt = va_mapped_to_ma_cache s item_index) as [opt H_opt_map].
  induction (va_mapped_to_ma_cache s item_index).
  exists (Some a); auto.
  exists None; auto.
  rewrite <- H_opt_map.
  induction opt; auto.
  induction (madd_eq ma a); auto.
  simpl.
  induction (vadd_eq item_index va); auto.
  subst item_index.
  contradiction b.
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
Qed.

Lemma fix_cache_erase_accessible : forall (s:state) (va:vadd) (ma:madd),
  valid_state ctxt s ->
  st_active_os s = osi ->
  os_accessible ctxt va ->

  map_apply vadd_eq (fix_cache_synonym (state_erase s) (st_cache (state_erase s)) ma) va =
  map_apply vadd_eq (fix_cache_synonym s (st_cache s) ma) va.
Proof.
  intros s va ma H_s_valid H_osi_active H_va_accessible.
  unfold state_erase at 2; unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.

  induction (st_cache s); auto.
  simpl.
  assert (exists va', exists pg', a0 = Item va' pg') as [va' [pg' H_a0_value]].
  destruct a0.
  exists item_index; exists item_info; auto.
  rewrite H_a0_value; simpl.
  assert (exists opt, opt = va_mapped_to_ma_cache s va) as [opt H_opt_map].
  induction (va_mapped_to_ma_cache s va); [exists (Some a1) | exists None]; auto.

  assert (forall v, os_accessible ctxt v ->
    va_mapped_to_ma_cache s v = va_mapped_to_ma_cache (state_erase s) v)
    as H_v_map_erase.
  intros v H_v_accessible.
  generalize H_opt_map; intro H_a_map.
  unfold va_mapped_to_ma_cache.
  unfold va_mapped_to_ma_cache in H_opt_map.
  unfold state_erase; simpl.
  unfold tlb_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- tlb_erase_accessible; auto.
  induction (map_apply vadd_eq (st_tlb s) v); auto.
  rewrite oss_erase_osi.
  rewrite H_osi_active.
  assert (exists aos_it, aos_it = map_apply os_ident_eq (st_oss s) osi) as [aos H_aos_map].
  induction (map_apply os_ident_eq (st_oss s) osi); try discriminate.
  exists (Value os_ident v0); auto.
  exists (Error os e0); auto.
  rewrite <- H_aos_map; rewrite H_osi_active, <- H_aos_map in H_opt_map.
  induction aos; auto.
  assert (exists p2m, p2m = map_apply os_ident_eq (st_hypervisor s) osi) as [p2m H_p2m_map].
  induction (map_apply os_ident_eq (st_hypervisor s) osi).
  exists (Value os_ident v1); auto.
  exists (Error _ e0); auto.
  induction p2m.
  rewrite hyper_erase_osi.
  rewrite <- H_p2m_map; rewrite <- H_p2m_map in H_opt_map.
  assert (exists pt_addr, pt_addr = map_apply padd_eq v1 (os_current_page v0)) as [pt_addr H_pt_addr_map].
  induction (map_apply padd_eq v1 (os_current_page v0)).
  exists (Value padd v2); auto.
  exists (Error _ e0); auto.
  induction pt_addr.
  rewrite <- H_pt_addr_map; rewrite <- H_pt_addr_map in H_opt_map.
  assert (exists pt_page, pt_page = map_apply madd_eq (st_memory s) v2) as [pt_page H_pt_page_map].
  induction (map_apply madd_eq (st_memory s) v2).
  exists (Value _ v3); auto.
  exists (Error _ e0); auto.
  rewrite <- H_pt_page_map in H_opt_map.
  induction pt_page.

  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with osi v0 v1 as [pt_addr H_pt_addr_map' [H_pt_type H_pt_owner]]; auto.
  unfold page_type in H_pt_type.
  rewrite <- H_pt_addr_map in H_pt_addr_map'; injection H_pt_addr_map';
    intro; subst pt_addr; clear H_pt_addr_map'.
  rewrite <- H_pt_page_map in H_pt_type.
  assert (exists mp, PT_c mp = page_content v3) as [mp H_mp_map].
  induction (page_content v3); try discriminate.
  exists m; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) v2 v3 mp; auto.
  simpl.
  
  rewrite <- H_mp_map in H_opt_map.
  rewrite <- H_pt_page_map.
  rewrite <- H_mp_map.
  rewrite <- mp_erase_accessible; auto.
  elim hyper_mapping_madd_owner with s (os_current_page v0) v2 osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  exists v1; auto.
  generalize H_s_valid; Intros_Valid_Hyp.
  unfold valid_current_page in H_valid_currpg.
  destruct H_valid_currpg with osi v0 v1
    as [pt_addr H_pt_addr_map' [H_pt_type H_pt_owner]]; auto.
  rewrite <- H_pt_addr_map in H_pt_addr_map'; injection H_pt_addr_map';
    intro; subst pt_addr; clear H_pt_addr_map'.
  unfold page_type in H_pt_type.
  rewrite <- H_pt_page_map in H_pt_type; discriminate H_pt_type.
  rewrite <- H_pt_addr_map; auto.
  rewrite hyper_erase_osi.
  rewrite <- H_p2m_map; auto.

  
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va') as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va'); [exists true | exists false]; auto.
  assert (exists opt2, opt2 = va_mapped_to_ma_cache s va') as [opt2 H_opt2_map].
  induction (va_mapped_to_ma_cache s va'); [exists (Some a1) | exists None]; auto.
  rewrite <- H_opt2_map.
  induction opt2.
  rewrite <- H_acc.
  induction acc; simpl.
  (* va' accessible *)
  rewrite <- H_v_map_erase with va'; unfold os_accessible; auto.
  rewrite <- H_opt2_map.
  induction (madd_eq ma a1); auto.
  simpl.
  induction (vadd_eq va va'); auto.
  induction (madd_eq ma a1); auto.
  simpl.
  induction (vadd_eq va va'); auto.
  subst va'.
  rewrite H_va_accessible in H_acc; discriminate H_acc.

  rewrite <- H_acc.
  induction acc; auto.
  simpl.
  rewrite <- H_v_map_erase with va'; unfold os_accessible; auto.
  rewrite <- H_opt2_map; auto.
Qed.


Lemma write_osi_erase_valid : forall (s s': state) (va:vadd) (val:value),
  valid_state ctxt s -> st_active_os s = osi ->
  pre_write va val ctxt s -> post_write va val s s' ->
  (pre_write va val ctxt (state_erase s) /\ post_write va val (state_erase s) (state_erase s')).
Proof.
  intros s s' va val H_s_valid.
  generalize H_s_valid; Intros_Valid_Hyp.
  destruct H_va_has_pa as [H_va_has_pa [H_cache_valid H_tlb_valid]].
  intros H_osi_active H_pre H_post.
  unfold pre_write in H_pre.
  destruct H_pre as [H_os_running [H_va_accessible [ma H_ma_mapped H_ma_type]]].
  
  generalize H_ma_mapped; intro H_ma_mapped'.
  unfold va_mapped_to_ma in H_ma_mapped.
  destruct H_ma_mapped as [pt_addr H_curr_pt [mp H_mp_map H_ma_map]].
  unfold os_current_vadd_mapping_madd in *; simpl.
  destruct H_curr_pt as [aos H_aos_map [p2m H_p2m_map H_pt_addr_map]].
  unfold hypervisor_mapping in H_p2m_map.
  unfold vadd_mapping in H_mp_map.
  destruct H_mp_map as [pt_page H_pt_page_map H_pt_page_content].
  assert (page_owned_by pt_page = Os_PO osi) as H_pt_page_owner.
  destruct H_valid_hyper with osi (os_current_page aos) p2m
    as [[pt_addr' H_pt_addr'_map H_pt_addr_owner] | H_error].
  rewrite <- H_osi_active; auto.
  unfold maddress_owner in H_pt_addr_owner.
  rewrite <- H_pt_addr_map in H_pt_addr'_map; injection H_pt_addr'_map;
    intro; subst pt_addr'; clear H_pt_addr'_map.
  rewrite <- H_pt_page_map in H_pt_addr_owner; auto.
  rewrite <- H_pt_addr_map in H_error; discriminate H_error.
  assert (va_mapped_to_ma (state_erase s) (st_active_os (state_erase s)) va ma) as H_ma_map_erase.
  unfold vadd_mapping.
  exists pt_addr.
  exists aos.
  unfold state_erase; simpl.
  rewrite oss_erase_osi; rewrite <- H_osi_active; auto.
  exists p2m; auto.
  unfold hypervisor_mapping.
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; rewrite <- H_osi_active; auto.
  exists (mp_erase mp).
  unfold vadd_mapping.
  exists (Page (PT_c (mp_erase mp)) (Os_PO osi)); auto.
  unfold state_erase; simpl.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page mp; auto.
  rewrite <- mp_erase_accessible; auto.
  split.

  unfold pre_write.
  repeat split; auto.
  unfold os_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  exists ma; auto.
  unfold page_type in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_type.
  exists v; auto.
  rewrite <- H_pg_map in H_ma_type.
  unfold state_erase; simpl.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  elim accessible_va_mapped_to_ma_owner with s va ma osi; auto.
  unfold maddress_owner; rewrite <- H_pg_map; auto.
  rewrite <- H_osi_active; auto.
  induction (page_content pg); try discriminate.
  exists o; auto.

  unfold post_write, post_all_write in *.
  destruct H_post as [ma' H_ma_map_old [newmem H_newmem [pg H_pg_map
    [new_cache H_new_cache [new_tlb H_new_tlb H_s']]]]].
  assert (Value _ pg = map_apply madd_eq (st_memory s) ma) as H_pg_map_mem.
  elim valid_state_included_cache_mem with ctxt s va pg; auto.
  intros ma'' H_ma''_map H_pg_map'.
  replace ma with ma''; auto.
  apply va_mapped_same_ma with s va; auto.
  assert (ma=ma').
  apply va_mapped_same_ma with s va; auto.
  apply valid_state_included_tlb_mem with ctxt; auto.
  subst ma'.
  assert (Some ma = va_mapped_to_ma_cache (state_erase s) va) as H_ma_map'.
  apply valid_state_included_tlb_cache with ctxt; auto.
  apply erasure_valid; auto.
  unfold osi_exists; exists aos; rewrite <- H_osi_active; auto.

  unfold page_type in H_ma_type.
  rewrite <- H_pg_map_mem in H_ma_type.
  assert (exists val', RW_c val' = page_content pg) as [val' H_pg_type].
  induction (page_content pg); try discriminate.
  exists o; auto.

  exists ma; auto.
  exists (mem_erase newmem); auto.
  unfold update_rw_page in *.
  destruct H_newmem as [pg' H_pg'_map H_newmem].
  rewrite <- H_pg_map_mem in H_pg'_map; injection H_pg'_map;
    intro; subst pg'; clear H_pg'_map.
  unfold state_erase; simpl.
  assert (page_owned_by pg = Os_PO osi) as H_pg_owner.
  elim accessible_va_mapped_to_ma_owner with s va ma osi; auto.
  unfold maddress_owner; rewrite <- H_pg_map_mem; auto.
  unfold os_accessible; auto.
  rewrite <- H_osi_active; auto.
  exists pg.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  induction (page_content pg); try discriminate.
  exists o; auto.
  rewrite H_newmem.
  rewrite rw_mem_erase_add_osi; auto.
  simpl; exists (Some val); auto.

  exists pg.
  unfold va_mapped_to_pg_cache in *.
  assert (exists it, it = map_apply vadd_eq (st_cache s) va) as [it H_it_map].
  induction (map_apply vadd_eq (st_cache s) va).
  exists (Value vadd v); auto.
  exists (Error page e); auto.
  rewrite <- H_ma_map'.
  rewrite <- H_it_map in H_pg_map.
  unfold state_erase, cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- cache_erase_accessible; auto.
  induction it; rewrite <- H_it_map; auto.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  elim accessible_va_mapped_to_ma_owner with s va ma osi; auto.
  unfold maddress_owner; rewrite <- H_pg_map_mem; auto.
  rewrite <- H_osi_active; auto.
  exists val'; rewrite H_pg_type; auto.

  exists (cache_erase_fix new_cache).
  unfold cache_add in *; simpl.
  destruct H_new_cache as [H_pg'_map_new [H_old_mapping [H_new_size H_new_distinct]]].
  repeat split.
  rewrite <- cache_erase_accessible; auto.

  intros va' pg' H_va'_ne_va H_pg'_map.
  assert (ctxt_vadd_accessible ctxt va' = true) as H_va'_accesible.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va') as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va'); [exists true | exists false]; auto.
  induction acc; auto.
  rewrite cache_erase_not_accessible in H_pg'_map; try discriminate; auto.
  unfold os_accessible.
  intro H_abs; rewrite <- H_acc in H_abs; discriminate.
  
  rewrite H_old_mapping with va' pg'; auto.
  rewrite <- fix_cache_erase_accessible; auto.
  rewrite <- cache_erase_accessible in H_pg'_map; auto.
  apply Le.le_trans with (map_size new_cache); auto.
  apply cache_erase_less_elements.
  apply cache_erase_distinct; auto.

  exists (tlb_erase_fix new_tlb).
  unfold tlb_add in *; simpl.
  destruct H_new_tlb as [H_ma'_map_new [H_old_mapping [H_new_size H_new_distinct]]].
  repeat split.
  rewrite <- tlb_erase_accessible; auto.

  intros va' ma' H_va'_ne_va H_ma'_map.
  assert (ctxt_vadd_accessible ctxt va' = true) as H_va'_accesible.
  assert (exists acc, acc = ctxt_vadd_accessible ctxt va') as [acc H_acc].
  induction (ctxt_vadd_accessible ctxt va'); [exists true | exists false]; auto.
  induction acc; auto.
  rewrite tlb_erase_not_accessible in H_ma'_map; try discriminate; auto.
  unfold os_accessible.
  intro H_abs; rewrite <- H_acc in H_abs; discriminate.
  
  rewrite H_old_mapping with va' ma'; auto.
  unfold tlb_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  rewrite <- tlb_erase_accessible; auto.
  rewrite <- tlb_erase_accessible in H_ma'_map; auto.
  apply Le.le_trans with (map_size new_tlb); auto.
  apply tlb_erase_less_elements.
  apply tlb_erase_distinct; auto.

  rewrite H_s'; unfold state_erase; simpl.
  unfold cache_erase, tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
Qed.


(* Solve a trivial (solvable by auto) disjunction *)
Ltac triv_disj :=
  repeat (try (left; auto; fail); right); fail.

Lemma erasure_step : forall (s s':state) (a:action),
  osi_exists s -> valid_state ctxt s -> valid_exec ctxt s a s' ->
  valid_exec ctxt (state_erase s) (action_erase s a) (state_erase s').
Proof.
  intros s s' a H_osi_exists H_s_valid H_valid_exec.
  generalize H_s_valid; Intros_Valid_Hyp.
  generalize H_valid_exec; intro H_valid_exec'.
  inversion_clear H_valid_exec' as (s_vexec, s'_vexec, act_vexec, H_pre, H_post).

  assert (osi_exists s') as H_osi_exists_s'.
  apply valid_exec_osi_exists with s a; auto.
  assert (valid_state ctxt s') as H_s'_valid.
  apply valid_state_invariant with s a; auto.

  induction a.
  (* Silent action *)
  assert (action_erase s Silent = Silent) as H_action.
  unfold action_erase.
  case (st_aos_activity); case (os_ident_eq (st_active_os s) osi); auto.
  rewrite H_action.
  constructor.
  unfold pre_action, pre_silent; auto.

  unfold post_action, post_silent, same_state in H_post.
  destruct H_post as [new_cache [new_tlb [H_cache'_valid [H_tlb'_valid H_assign]]]].

  unfold post_action, post_silent.
  unfold state_components_assign, state_components_update in H_assign.
  assert (forall cmp, cmp <> System_Cache -> cmp <> System_TLB ->
    state_component_interp s cmp = state_component_interp s' cmp) as H_interp.
  intros cmp H_cmp_not_cache H_cmp_not_tlb.
  elim H_assign with cmp; auto.
  intros [H_triv H_interp]; auto.
  intros [H_abs H_interp].
  contradiction H_abs; split; auto.
  Simp_Inv_ThreeComps H_interp H_aos H_oss H_exec_mode System_Cache
    Active_OS Guest_OSs Exec_Mode.
  Simp_Inv_ThreeComps H_interp H_activity H_hyper H_mem System_Cache
    OS_Activity Hypervisor_SC System_Memory.

  unfold same_state.
  exists (st_cache (state_erase s')); exists (st_tlb (state_erase s')).
  assert (valid_state ctxt (state_erase s')) as H_s'_erase_valid.
  apply erasure_valid; auto.
  split; [idtac | split]; try (generalize H_s'_erase_valid;
    intros (H_trstos_nothypcall', (H_running_nothypcall',
      (H_valid_exmd', (H_valid_osexmd', (H_valid_untosexmd',
        (H_valid_hyper', (H_valid_virtmap', (H_valid_currpg',
          (H_injective_hyper', (H_va_has_pa', (H_cache_valid, H_tlb_valid)))))))))));
    destruct H_va_has_pa as [H_va_has_pa [H_valid_cache H_valid_tlb]]; assumption).

  assert (activity_erase s = activity_erase s') as H_activity_erase.
  unfold activity_erase.
  rewrite <- H_aos; try (intro; discriminate).
  rewrite <- H_activity; try (intro; discriminate).
  induction (os_ident_eq (st_active_os s) osi); auto.

  unfold state_components_assign, state_components_update.
  intro cmp.
  induction cmp; unfold state_component_interp, state_erase; simpl.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  apply same_oss_erase; auto.
  rewrite <- H_oss; try (intro; discriminate); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  unfold exec_erase.
  induction (ctxt_oss ctxt osi); auto.
  rewrite <- H_activity_erase; auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  apply same_hyper_erase; auto.
  rewrite <- H_hyper; try (intro; discriminate); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  rewrite <- H_mem; try (intro; discriminate); auto.

  right.
  split.
  intro H_abs; elim H_abs; auto.
  split; auto.
  
  right.
  split.
  intro H_abs; elim H_abs; auto.
  split; auto.

  (* Read action *)
  unfold action_erase.
  unfold pre_action in H_pre.
  generalize H_pre; intro H_pre'.
  unfold  pre_read, pre_all_read in H_pre'.
  destruct H_pre' as [[ma [H_ma_mapped H_ma_type]] [H_v_accessible H_os_running]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi is active *)
  constructor.
  unfold pre_action.
  elim read_osi_erase_valid with s s' v; auto.
  elim read_osi_erase_valid with s s' v; auto.
  (* osi is not active *)
  apply os_running_erase_valid with (Read v); auto.
  split; auto.
  assert (exists va, Read v = Read va); [exists v; auto |triv_disj].

  (* Read hyper action *)
  replace (action_erase s (Read_hyper v)) with Silent.
  apply hyper_action_erase_valid with (Read_hyper v); auto.
  assert (exists va, Read_hyper v = Read_hyper va);
    [exists v; auto | triv_disj].
  unfold action_erase.
  unfold pre_action, pre_read_hyper in H_pre.
  destruct H_pre as [H_pre H_hyper_running].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.

  (* Write action *)
  unfold action_erase.
  unfold pre_action in H_pre.
  generalize H_pre; intro H_pre'.
  unfold  pre_write in H_pre'.
  destruct H_pre' as [H_os_running [H_va_accessible H_pre']].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi is active *)
  constructor.
  unfold pre_action.
  elim write_osi_erase_valid with s s' v v0; auto.
  elim write_osi_erase_valid with s s' v v0; auto.
  (* osi not active *)
  apply os_running_erase_valid with (Write v v0); auto.
  split; auto.
  assert (exists va, exists val, Write v v0 = Write va val);
    [exists v; exists v0; auto | triv_disj].
  
  (* Write hyper action *)
  replace (action_erase s (Write_hyper v v0)) with Silent.
  apply hyper_action_erase_valid with (Write_hyper v v0); auto.
  assert (exists va, exists val, Write_hyper v v0 = Write_hyper va val);
    [exists v; exists v0; auto | triv_disj].
  unfold action_erase.
  unfold pre_action, pre_write_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.

  (* New untrusted action *)
  unfold action_erase.
  unfold pre_action, pre_new_untrusted in H_pre.
  destruct H_pre as [H_os_untrst [ma H_ma_map [H_ma_ptype [H_ma_owner [H_v_accessible [H_hcall H_hyper_running]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  induction (os_ident_eq o osi).
  (* osi action *)
  subst o.
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_new_untrusted.
  split; auto.
  exists ma.
  unfold state_erase; simpl.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  repeat split; auto.
  unfold page_type in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_type.
  exists v0; auto.
  contradiction H_ma_ptype; auto.
  rewrite <- H_pg_map in H_ma_ptype.
  unfold state_erase; simpl.
  assert (exists content, page_content pg = content) as [content H_pg_content].
  induction (page_content pg);
    [exists (RW_c o) | exists (PT_c m) | contradiction H_ma_ptype]; auto.
  induction content.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  exists o; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pg m; auto.
  simpl; intro H_abs; discriminate H_abs.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  rewrite H_pg_content in H_ma_ptype.
  contradiction H_ma_ptype; auto.
  unfold state_erase; simpl.
	apply mem_erase_owner; auto.
  unfold get_os_hypercall in *.
  destruct H_hcall as [system H_system_map H_hcall].
  exists system; auto.
  unfold state_erase; simpl.
  rewrite oss_erase_osi; auto.
  unfold hyper_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); auto.
  (* post-condition *)
  unfold post_action, post_new_untrusted, post_all_new in *.
  destruct H_post as [ma' H_ma'_map [pt_addr H_pt_addr_map [newmem H_newmem [system H_system_map H_s']]]].
  rewrite <- hyp_mapped_same_maddr with s osi p ma ma' in H_newmem; auto.
  clear ma' H_ma'_map.
  exists ma.
  unfold state_erase; simpl.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  exists pt_addr.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in *.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  exists system.
  unfold state_erase; simpl.
  rewrite oss_erase_osi; auto.
  exists p2m; auto.
  unfold state_erase; simpl.
  rewrite hyper_erase_osi; auto.
  unfold state_erase; simpl.
  exists (mem_erase newmem).
  unfold update_pt_page in *.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pg_content H_newmem]].
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)).
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  exists (mp_erase oldmap); auto.
  simpl.
  rewrite <- mp_erase_add_osi; auto.
  rewrite <- pt_mem_erase_add_osi with (st_memory s) pt_addr (map_add vadd_eq oldmap v ma).
  rewrite H_newmem.
  replace (page_owned_by pt_page) with (Os_PO osi); auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  exists system.
  rewrite oss_erase_osi; auto.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  
  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS None (os_current_page system))))
      as H_oss; try (rewrite H_s'; simpl; rewrite oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase newmem)) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = map_drop vadd_eq (cache_erase s) v).
  intro H_cache.
  cut ((tlb_erase s') = map_drop vadd_eq (tlb_erase s) v).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; unfold tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite tlb_erase_drop; auto.
  rewrite H_s'; unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite cache_erase_drop; auto.

  (* Not osi action *)
  apply hyper_action_erase_valid with (New_untrusted o v p); auto.
  unfold hyper_action_not_osi.
  do 5 right; left.
  exists o; split; auto.
  exists v; exists p; auto.
  
  (* New trusted action *)
  unfold action_erase.
  unfold pre_action, pre_new_trusted in H_pre.
  destruct H_pre as [ma H_ma_map [H_ma_ptype [H_ma_owner [H_v_accessible [H_aos_trst H_os_running]]]]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_new_trusted.
  exists ma.
  unfold state_erase; simpl.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  exists p2m; auto.
  rewrite a in H_p2m_map.
  rewrite hyper_erase_osi; auto.
  repeat split; auto.
  unfold page_type in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_type.
  exists v0; auto.
  contradiction H_ma_ptype; auto.
  rewrite <- H_pg_map in H_ma_ptype.
  unfold state_erase; simpl.
  assert (exists content, page_content pg = content) as [content H_pg_content].
  induction (page_content pg);
    [exists (RW_c o) | exists (PT_c m) | contradiction H_ma_ptype]; auto.
  induction content.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  rewrite a in H_ma_map; auto.
  exists o; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pg m; auto.
  simpl; intro H_abs; discriminate H_abs.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  rewrite a in H_ma_map; auto.
  rewrite H_pg_content in H_ma_ptype.
  contradiction H_ma_ptype; auto.
  unfold state_erase; simpl.
  rewrite a in H_ma_owner.
	apply mem_erase_owner; auto.
  unfold trusted_aos in *.
  unfold state_erase; simpl.
  rewrite <- a; auto.
  unfold os_running.
  unfold state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  (* post-condition *)
  unfold post_action, post_new_trusted, post_all_new in *.
  destruct H_post as [ma' H_ma'_map [pt_addr H_pt_addr_map [newmem H_newmem [system H_system_map H_s']]]].
  rewrite a in H_ma'_map.
  rewrite <- hyp_mapped_same_maddr with s osi p ma ma' in H_newmem; auto.
  clear ma' H_ma'_map.
  exists ma.
  unfold state_erase; simpl.
  unfold hypervisor_mapping_madd in *.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  exists p2m; auto.
  rewrite a in H_p2m_map.
  rewrite hyper_erase_osi; auto.
  exists pt_addr.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in *.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  exists system.
  unfold state_erase; simpl.
  rewrite a in H_system_map.
  rewrite oss_erase_osi; auto.
  exists p2m; auto.
  unfold state_erase; simpl.
  rewrite a in H_p2m_map.
  rewrite hyper_erase_osi; auto.
  unfold state_erase; simpl.
  exists (mem_erase newmem).
  unfold update_pt_page in *.
  destruct H_newmem as [pt_page H_pt_page_map [oldmap H_pg_content H_newmem]].
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)).
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold os_current_vadd_mapping_madd, hypervisor_mapping in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping_madd.
  rewrite a in H_p2m_map.
  exists p2m; auto.
  exists (mp_erase oldmap); auto.
  simpl.
  rewrite <- mp_erase_add_osi; auto.
  rewrite <- pt_mem_erase_add_osi with (st_memory s) pt_addr (map_add vadd_eq oldmap v ma).
  rewrite H_newmem.
  replace (page_owned_by pt_page) with (Os_PO osi); auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  destruct H_pt_addr_map as [system' H_system'_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping_madd.
  rewrite a in H_p2m_map.
  exists p2m; auto.
  exists system.
  rewrite a in H_system_map.
  rewrite oss_erase_osi; auto.

  induction (os_ident_eq osi osi); try contradiction b; auto.
  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS None (os_current_page system))))
      as H_oss; try (rewrite H_s'; simpl; rewrite a, oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase newmem)) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = map_drop vadd_eq (cache_erase s) v).
  intro H_cache.
  cut ((tlb_erase s') = map_drop vadd_eq (tlb_erase s) v).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; unfold tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite a1.
  destruct (os_ident_eq osi osi); try contradiction n; auto.
  rewrite tlb_erase_drop; auto.
  rewrite H_s'; unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite a1.
  destruct (os_ident_eq osi osi); try contradiction n; auto.
  rewrite cache_erase_drop; auto.
  rewrite <- a; auto.
  (* osi not active *)
  apply os_running_erase_valid with (New_trusted v p); auto.
  split; auto.
  assert (exists va, exists pa, New_trusted v p = New_trusted va pa);
    [exists v; exists p; auto | triv_disj].

  (* New hyper action *)
  replace (action_erase s (New_hyper v m)) with Silent.
  apply hyper_action_erase_valid with (New_hyper v m); auto.
  assert (exists va, exists ma, New_hyper v m = New_hyper va ma);
    [exists v; exists m; auto | triv_disj].
  unfold action_erase.
  unfold pre_action, pre_new_hyper in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.

  (* Del untrusted action *)
  unfold action_erase.
  unfold pre_action, pre_del_untrusted, pre_all_del in H_pre.
  destruct H_pre as [H_os_untrst [H_hcall 
    [system H_system_map [p2m H_p2m_map [pt_addr H_pt_addr_map [[oldmap H_pt_page_content [ma H_ma_map]] 
      [H_hyper_running H_v_accessible]]]]]]].
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  induction (os_ident_eq o osi).
  (* osi action *)
  subst o.
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_del_untrusted, pre_all_del.
  split; auto.
  split.
  unfold get_os_hypercall in *.
  destruct H_hcall as [system' H_system'_map H_hcall].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold state_erase; simpl.
  exists system; auto.
  rewrite oss_erase_osi; auto.
  unfold state_erase; simpl.  
  exists system.
  rewrite oss_erase_osi; auto.
  exists p2m.
  unfold hypervisor_mapping in *.
  rewrite hyper_erase_osi; auto.
  exists pt_addr; auto.
  split.
  exists (mp_erase oldmap).
  unfold vadd_mapping in *.
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)); auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  exists ma; auto.
  rewrite <- mp_erase_accessible; auto.
  unfold hyper_running; simpl.
  split; auto.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); auto.
  (* post-condition *)
  unfold post_action, post_del_untrusted in *.
  destruct H_post as [system' H_system'_map [p2m' H_p2m'_map [pt_addr' H_pt_addr'_map [newmem H_newmem H_s']]]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold drop_va_pt_page in H_newmem.
  destruct H_newmem as [pt_page' H_pt_page'_map [oldmap' H_pt_page'_content H_newmem]].
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page'_content.
  injection H_pt_page'_content; intro; subst oldmap'; clear H_pt_page'_content.
  unfold state_erase; simpl.
  exists system.
  rewrite oss_erase_osi; auto.
  exists p2m.
  unfold hypervisor_mapping.
  rewrite hyper_erase_osi; auto.
  exists pt_addr; auto.
  exists (mem_erase newmem).
  unfold drop_va_pt_page.
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)).
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  exists (mp_erase oldmap); auto.
  rewrite <- mp_erase_drop_osi_acc; auto.
  simpl.
  rewrite <- pt_mem_erase_add_osi with (st_memory s) pt_addr (map_drop vadd_eq oldmap v).
  rewrite H_newmem.
  replace (page_owned_by pt_page) with (Os_PO osi); auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  
  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS None (os_current_page system))))
      as H_oss; try (rewrite H_s'; simpl; rewrite oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase newmem)) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = map_drop vadd_eq (cache_erase s) v).
  intro H_cache.
  cut ((tlb_erase s') = map_drop vadd_eq (tlb_erase s) v).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; unfold tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite tlb_erase_drop; auto.
  rewrite H_s'; unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite cache_erase_drop; auto.
  (* Not osi action *)
  apply hyper_action_erase_valid with (Del_untrusted o v); auto.
  unfold hyper_action_not_osi.
  do 6 right; left.
  exists o; split; auto.
  exists v; auto.

  (* Del trusted action *)
  unfold action_erase.
  unfold pre_action, pre_del_trusted, pre_all_del in H_pre.
  destruct H_pre as [pt_addr H_pt_addr_map [[oldmap H_pt_page_content [ma H_ma_map]]
    [H_os_running [H_aos_trst H_v_accessible]]]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_del_trusted, pre_all_del.
  unfold state_erase; simpl.
  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd,
    hypervisor_mapping in *.
  destruct H_pt_addr_map as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  rewrite a in H_p2m_map.
  unfold vadd_mapping in *.
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  exists pt_addr.
  exists system; simpl.
  rewrite oss_erase_osi.
  rewrite <- a; auto.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  repeat split; auto.
  exists (mp_erase oldmap).
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)); auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  exists ma; auto.
  rewrite <- mp_erase_accessible; auto.
  unfold os_running; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold trusted_aos; simpl.
  rewrite <- a; auto.
  (* post-condition *)
  unfold post_action, post_del_trusted, post_all_del in *.
  destruct H_post as [pt_addr' H_pt_addr'_map [newmem H_newmem H_s']].
  unfold current_vadd_mapping_madd, os_current_vadd_mapping_madd,
    hypervisor_mapping in *.
  destruct H_pt_addr_map as [system H_system_map [p2m H_p2m_map H_pt_addr_map]].
  destruct H_pt_addr'_map as [system' H_system'_map [p2m' H_p2m'_map H_pt_addr'_map]].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_pt_addr_map in H_pt_addr'_map.
  injection H_pt_addr'_map; intro; subst pt_addr'; clear H_pt_addr'_map.
  unfold state_erase; simpl.
  exists pt_addr.
  exists system.
  rewrite a in H_system_map.
  rewrite oss_erase_osi; auto.
  exists p2m; auto.
  rewrite a in H_p2m_map.
  rewrite hyper_erase_osi; auto.
  exists (mem_erase newmem).
  unfold drop_va_pt_page in *.
  unfold vadd_mapping in *.
  destruct H_pt_page_content as [pt_page H_pt_page_map H_pt_page_content].
  destruct H_newmem as [pt_page' H_pt_page'_map [oldmap' H_pt_page'_content H_newmem]].
  rewrite <- H_pt_page_map in H_pt_page'_map.
  injection H_pt_page'_map; intro; subst pt_page'; clear H_pt_page'_map.
  rewrite <- H_pt_page_content in H_pt_page'_content.
  injection H_pt_page'_content; intro; subst oldmap'; clear H_pt_page'_content.
  exists (Page (PT_c (mp_erase oldmap)) (Os_PO osi)).
  rewrite <- pt_mem_erase_osi with (st_memory s) pt_addr pt_page oldmap; auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold hypervisor_mapping_madd.
  rewrite a in H_p2m_map.
  exists p2m; auto.
  exists (mp_erase oldmap); auto.
  simpl.
  rewrite <- mp_erase_drop_osi_acc; auto.
  rewrite <- pt_mem_erase_add_osi with (st_memory s) pt_addr (map_drop vadd_eq oldmap v).
  rewrite H_newmem.
  replace (page_owned_by pt_page) with (Os_PO osi); auto.
  elim hyper_mapping_madd_owner with s (os_current_page system) pt_addr osi; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  unfold os_current_vadd_mapping_madd in H_pt_addr_map.
  unfold hypervisor_mapping_madd.
  rewrite a in H_p2m_map.
  exists p2m; auto.

  assert (st_oss (state_erase s') = (oss_erase (st_oss s))) as H_oss;
    try (rewrite H_s'; simpl; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase newmem)) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = map_drop vadd_eq (cache_erase s) v).
  intro H_cache.
  cut ((tlb_erase s') = map_drop vadd_eq (tlb_erase s) v).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; unfold tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite tlb_erase_drop; auto.
  rewrite H_s'; unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite cache_erase_drop; auto.

  (* osi is not active *)
  apply os_running_erase_valid with (Del_trusted v); auto.
  split; auto.
  assert (exists va, Del_trusted v = Del_trusted va);
    [exists v; auto | triv_disj].

  (* Del hyper action *)
  replace (action_erase s (Del_hyper v)) with Silent.
  apply hyper_action_erase_valid with (Del_hyper v); auto.
  assert (exists va, Del_hyper v = Del_hyper va);
    [exists v; auto | triv_disj].
  unfold action_erase.
  unfold pre_action, pre_del_hyper in H_pre.
  destruct H_pre as [pt_addr H_curr [H_pre [H_hyper_running H_acc]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.

  (* Switch action *)
  replace (action_erase s (Switch o)) with Silent.
  apply hyper_action_erase_valid with (Switch o); auto.
  assert (exists os, Switch o = Switch os);
    [exists o; auto | triv_disj].
  unfold action_erase.
  unfold pre_action, pre_switch in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running; auto.

  (* Lswitch trusted action *)
  unfold action_erase.
  unfold pre_action, pre_lswitch_trusted, pre_all_lswitch in H_pre.
  destruct H_pre as [H_os_running [H_aos_trst [ma H_ma_map [H_ma_ptype H_ma_owner]]]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_lswitch_trusted, pre_all_lswitch.
  repeat split.
  unfold os_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold trusted_aos in *.
  rewrite a in H_aos_trst.
  unfold state_erase; simpl; auto.
  unfold hypervisor_mapping_madd.
  rewrite a in H_p2m_map.
  unfold state_erase; simpl.
  exists ma.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  assert (exists pt_page, Value _ pt_page = map_apply madd_eq (st_memory s) ma) as [pt_page H_pt_page_map].
  unfold page_type in H_ma_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_ptype.
  exists v; auto.
  unfold page_type, maddress_owner in *.
  rewrite <- H_pt_page_map in H_ma_ptype, H_ma_owner.
  rewrite a in H_ma_owner.
  assert (exists oldmap, PT_c oldmap = page_content pt_page) as [oldmap H_pt_page_content].
  induction (page_content pt_page); try discriminate H_ma_ptype.
  exists m; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pt_page oldmap; auto.
  (* post-condition *)
  unfold post_action, post_lswitch_trusted, post_all_lswitch in *.
  destruct H_post as [system H_system_map H_s'].
  unfold state_erase; simpl.
  rewrite a in H_system_map.
  exists system.
  rewrite oss_erase_osi; auto.

  induction (os_ident_eq osi osi); try contradiction b; auto.
  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS None p)))
      as H_oss; try (rewrite H_s'; simpl; rewrite a, oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase (st_memory s))) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = cache_flush (cache_erase s)).
  intro H_cache.
  cut ((tlb_erase s') = tlb_flush (tlb_erase s)).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; simpl.
  rewrite a.
  unfold tlb_flush.
  unfold tlb_erase; simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  rewrite H_s'; simpl.
  rewrite a.
  unfold cache_flush.
  unfold cache_erase; simpl.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  (* osi not active *)
  apply os_running_erase_valid with (Lswitch_trusted p); auto.
  split; auto.
  assert (exists pa, Lswitch_trusted p = Lswitch_trusted pa);
    [exists p; auto | triv_disj].
  
  (* Lswitch untrusted action *)
  unfold action_erase.
  unfold pre_action, pre_lswitch_untrusted, pre_all_lswitch in H_pre.
  destruct H_pre as [H_aos_untrst [system H_system_map [H_hcall
    [H_hyper_running [ma H_ma_map [H_ma_ptype H_ma_owner]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  induction (os_ident_eq o osi).
  (* osi action *)
  subst o.
  unfold hypervisor_mapping_madd in H_ma_map.
  destruct H_ma_map as [p2m H_p2m_map H_ma_map].
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_lswitch_untrusted, pre_all_lswitch.
  repeat split; auto.
  unfold state_erase; simpl.
  exists system.
  rewrite oss_erase_osi; auto.
  repeat split; auto.
  unfold hyper_running; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); auto.
  exists ma.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  assert (exists pt_page, Value _ pt_page = map_apply madd_eq (st_memory s) ma) as [pt_page H_pt_page_map].
  unfold page_type in H_ma_ptype.
  induction (map_apply madd_eq (st_memory s) ma); try discriminate H_ma_ptype.
  exists v; auto.
  unfold page_type, maddress_owner in *.
  rewrite <- H_pt_page_map in H_ma_ptype, H_ma_owner.
  assert (exists oldmap, PT_c oldmap = page_content pt_page) as [oldmap H_pt_page_content].
  induction (page_content pt_page); try discriminate H_ma_ptype.
  exists m; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pt_page oldmap; auto.
  (* post-condition *)
  unfold post_action, post_lswitch_untrusted, post_all_lswitch in *.
  destruct H_post as [system' H_system'_map H_s'].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold state_erase; simpl.
  exists system.
  rewrite oss_erase_osi; auto.

  induction (os_ident_eq osi osi); try contradiction b; auto.
  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS None p)))
      as H_oss; try (rewrite H_s'; simpl; rewrite oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_aos_activity;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase (st_memory s))) as H_memory;
    try (rewrite H_s'; auto; fail).
  cut ((cache_erase s') = cache_flush (cache_erase s)).
  intro H_cache.
  cut ((tlb_erase s') = tlb_flush (tlb_erase s)).
  intro H_tlb.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.

  rewrite H_s'; simpl.
  unfold tlb_flush.
  unfold tlb_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite H_s'; simpl.
  unfold cache_flush.
  unfold cache_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  (* not osi action *)
  apply hyper_action_erase_valid with (Lswitch_untrusted o p); auto.
  unfold hyper_action_not_osi.
  do 8 right; left.
  exists o; split; auto.
  exists p; auto.

  (* Hcall action *)
  unfold action_erase.
  unfold pre_action, pre_hcall in H_pre.
  destruct H_pre as [H_os_running H_aos_untrst].
  unfold os_running in H_os_running.
  unfold trusted_aos in H_aos_untrst.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_hcall.
  split.
  unfold os_running.
  unfold state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold trusted_aos.
  intro H_abs.
  contradiction H_aos_untrst.
  unfold state_erase in H_abs; simpl in H_abs.
  rewrite a; auto.
  (* post-condition *)
  unfold post_action, post_hcall in *.
  destruct H_post as [system H_system_map H_s'].
  rewrite a in H_system_map.
  unfold state_erase; simpl.
  exists system.
  rewrite oss_erase_osi; auto.

  assert (st_oss (state_erase s') =
    (map_add os_ident_eq (oss_erase (st_oss s)) osi (OS (Some h) (os_current_page system))))
      as H_oss; try (rewrite H_s'; simpl; rewrite a, oss_erase_add_osi; auto; fail).
  assert ((exec_erase s') = EXMD_svc) as H_exec_mode.
  rewrite H_s'; simpl.
  unfold exec_erase, activity_erase; simpl.
  induction (ctxt_oss ctxt osi); auto.
  induction (os_ident_eq (st_active_os s) osi); auto.

  assert ((activity_erase s') = OSA_waiting) as H_aos_activity.
  rewrite H_s'; unfold activity_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  assert ((hyper_erase (st_hypervisor s')) = (hyper_erase (st_hypervisor s))) as H_hypervisor;
    try (rewrite H_s'; auto; fail).
  assert ((mem_erase (st_memory s')) = (mem_erase (st_memory s))) as H_memory;
    try (rewrite H_s'; auto; fail).
  assert ((cache_erase s') = (cache_erase s)) as H_cache;
    try (rewrite H_s'; auto; fail).
  assert ((tlb_erase s') = (tlb_erase s)) as H_tlb;
    try (rewrite H_s'; auto; fail).

  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_aos_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.
  (* osi not active *)
  apply os_running_erase_valid with (Hcall h); auto.
  unfold other_osi_action.
  split; auto.
  right; right; right; right;
    right; right; left.
  exists h; auto.

  (* Ret_ctrl action *)
  unfold action_erase.
  unfold pre_action, pre_ret_ctrl in H_pre.
  unfold os_running in H_pre.
  rewrite H_pre.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_ret_ctrl, os_running.
  unfold state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  (* post-condition *)
  unfold post_action, post_ret_ctrl,
    state_components_assign, state_components_update in *.
  intro cmp.
  assert (activity_erase s' = OSA_waiting) as H_activity.
  unfold activity_erase.
  elim H_post with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp.
  rewrite a.
  induction (os_ident_eq osi osi); try contradiction b; auto.
  elim H_post with OS_Activity.
  intros [[H_triv' H_abs] H_interp']; contradiction H_abs; auto.
  intros [H_triv' [H_exec_mode H_activity]].
  unfold state_component_interp in *; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  assert (st_active_os s' = osi) as H_aos_s'.
  elim H_post with Active_OS.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  assert (exec_erase s' = EXMD_svc) as H_exec_mode.
  unfold exec_erase.
  rewrite H_activity.
  induction (ctxt_oss ctxt osi); auto.
  induction cmp; simpl.
  left; repeat split; try (intro H_abs; discriminate H_abs); auto.
  left; repeat split; try (intro H_abs; discriminate H_abs).
  elim H_post with Guest_OSs.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  right.
  split; try (intros [H_abs H_triv]; contradiction H_abs); auto.
  right.
  split; try (intros [H_triv H_abs]; contradiction H_abs); auto.
  left.
  split.
  split; intro H_abs; discriminate H_abs.
  elim H_post with Hypervisor_SC.
  intros  [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  left.
  split.
  split; intro H_abs; discriminate H_abs.
  elim H_post with System_Memory.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  rewrite H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  left.
  split.
  split; intro H_abs; discriminate H_abs.
  elim H_post with System_Cache.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  unfold cache_erase.
  rewrite H_aos_s', a, <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  left.
  elim H_post with System_TLB.
  intros [H_triv H_interp].
  unfold state_component_interp in H_interp.
  unfold tlb_erase.
  rewrite H_aos_s', a, <- H_interp; auto.
  intros [H_abs H_interp]; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  
  (* osi not active *)
  apply os_running_erase_valid with (Ret_ctrl); auto.
  unfold other_osi_action.
  split; [auto | triv_disj].

  (* Chmod action *)
  induction (os_ident_eq (st_active_os s) osi).
  (* osi is active *)
  unfold pre_action, pre_chmod in H_pre.
  destruct H_pre as [H_hyper_running H_no_hcall].
  unfold hyper_running in H_hyper_running.
  unfold action_erase; rewrite H_hyper_running.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  constructor.
  unfold pre_action, pre_chmod.
  split.
  unfold hyper_running.
  unfold state_erase, activity_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold get_os_hypercall in *.
  unfold state_erase; simpl.
  destruct H_no_hcall as [system H_system_map H_no_hcall].
  rewrite a in H_system_map.
  rewrite oss_erase_osi; auto.
  exists system; auto.
  (* post action *)
  unfold post_action, post_chmod in *.
  unfold trusted_aos in *.
  rewrite a in H_post.
  destruct H_post as [[H_trst H_assign] | [H_untrst H_assign]];
    [left | right]; split; auto.
  unfold state_components_assign, state_components_update in *; intro cmp;
    destruct H_assign with cmp as [[[H_cmp1 H_cmp2] H_interp] | [H_cmp H_interp]];
      [left | right]; split; auto.
  assert (st_active_os s' = osi) as H_aos_s'.
  elim H_assign with Active_OS.
  intros [H_triv H_interp'].
  unfold state_component_interp in H_interp'.
  rewrite <- H_interp'; auto.
  intros [H_abs H_interp']; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.

  induction cmp; try (contradiction H_cmp1; auto; fail);
    try (contradiction H_cmp2; auto; fail); unfold state_component_interp in *;
      simpl; auto; try (rewrite H_interp); auto.
  unfold cache_erase.
  rewrite H_aos_s', a, <- H_interp; auto.
  unfold tlb_erase.
  rewrite H_aos_s', a, <- H_interp; auto.

  unfold state_component_interp in *; simpl.
  assert (activity_erase s' = OSA_running) as H_activity.
  unfold activity_erase.
  elim H_assign with Active_OS.
  intros [H_triv H_aos].
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  destruct H_interp; auto.
  intros [H_abs H_interp']; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  induction cmp; try (contradiction H_cmp; split; intro H_abs; discriminate H_abs); simpl;
    unfold exec_erase; rewrite H_activity; rewrite H_trst; auto.
  unfold state_components_assign, state_components_update in *; intro cmp;
    destruct H_assign with cmp as [[[H_cmp1 H_cmp2] H_interp] | [H_cmp H_interp]];
      [left | right]; split; auto.
  assert (st_active_os s' = osi) as H_aos_s'.
  elim H_assign with Active_OS.
  intros [H_triv H_interp'].
  unfold state_component_interp in H_interp'.
  rewrite <- H_interp'; auto.
  intros [H_abs H_interp']; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  induction cmp; try (contradiction H_cmp1; auto; fail);
    try (contradiction H_cmp2; auto; fail); unfold state_component_interp in *;
      simpl; auto; try (rewrite H_interp); auto.
  unfold cache_erase.
  rewrite H_aos_s', a, <- H_interp; auto.
  unfold tlb_erase.
  rewrite H_aos_s', a, <- H_interp; auto.

  unfold state_component_interp in *; simpl.
  assert (activity_erase s' = OSA_running) as H_activity.
  unfold activity_erase.
  elim H_assign with Active_OS.
  intros [H_triv H_aos].
  rewrite <- H_aos.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  destruct H_interp; auto.
  intros [H_abs H_interp']; contradiction H_abs; split; intro H_abs'; discriminate H_abs'.
  induction cmp; try (contradiction H_cmp; split; intro H_abs; discriminate H_abs); simpl;
    unfold exec_erase; rewrite H_activity; split; auto;
      induction (ctxt_oss ctxt osi); auto; contradiction H_untrst; auto.
  (* osi is not active *)
  replace (action_erase s Chmod) with Silent.
  apply hyper_action_erase_valid with Chmod; try triv_disj; auto.
  unfold action_erase.
  unfold pre_action, pre_chmod in H_pre.
  destruct H_pre as [H_hyper_running H_pre].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  induction (os_ident_eq (st_active_os s) osi); auto; try contradiction b; auto.

  (* Page pin trusted action *)
  unfold action_erase.
  unfold pre_action, pre_page_pin_trusted, pre_all_pin in H_pre.
  destruct H_pre as [H_os_running [H_aos_trst [[p2m H_p2m_map H_error]
    [ma [pg H_pg_map [H_pg_owner H_pg_content]]]]]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_page_pin_trusted, pre_all_pin.
  repeat split.
  unfold os_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold trusted_aos in *.
  rewrite a in H_aos_trst.
  unfold state_erase; auto.
  rewrite a in H_p2m_map.
  unfold state_erase; simpl.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  unfold state_erase; simpl.
  exists ma; auto.
  exists (Page Other_c No_Owner).
  apply other_mem_erase_not_osi with pg; auto.
  rewrite H_pg_owner; intro H_abs; discriminate H_abs.
  simpl; split; auto.
  (* post-condition *)
  unfold post_action, post_page_pin_trusted, post_all_pin in *.
  destruct H_post as [ma' [pg' H_pg'_map [H_pg'_owner [H_pg'_content 
    [p2m' H_p2m'_map [system H_system_map H_s']]]]]].
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  exists ma'.
  exists (Page Other_c No_Owner).
  rewrite other_mem_erase_not_osi with (st_memory s) ma' pg'; auto.
  rewrite H_pg'_owner; intro H_abs; discriminate H_abs.
  repeat split; auto.
  exists p2m.
  unfold hypervisor_mapping.
  unfold state_erase; simpl.
  rewrite a in H_p2m_map.
  rewrite hyper_erase_osi; auto.
  exists system.
  unfold state_erase; simpl.
  rewrite a in H_system_map.
  rewrite oss_erase_osi; auto.

  assert ((oss_erase (st_oss s')) = (map_add os_ident_eq (oss_erase (st_oss s)) osi
    (OS None (os_current_page system)))) as H_oss;
      try (rewrite H_s'; simpl; rewrite a, oss_erase_add_osi; auto; fail).
  assert ((activity_erase s') = (activity_erase s)) as H_activity;
    try (rewrite H_s'; auto; fail).
  assert ((exec_erase s') = (exec_erase s)) as H_exec_mode;
    try (rewrite H_s'; auto; fail).
  assert ((hyper_erase (st_hypervisor s')) = (map_add os_ident_eq (hyper_erase (st_hypervisor s)) osi
    (map_add padd_eq p2m p ma'))) as H_hypervisor;
      try (rewrite H_s'; simpl; rewrite a, hyper_erase_add_osi; auto; fail).
  assert ((mem_erase (st_memory s')) = (map_add madd_eq (mem_erase (st_memory s)) ma'
    match p0 with
      | PT =>
        Page (PT_c (map_empty vadd madd)) (Os_PO osi)
      | RW => Page (RW_c None) (Os_PO osi)
    end)) as H_memory.
  rewrite H_s'; simpl.
  induction p0.
  rewrite a, pt_mem_erase_add_osi; auto.
  rewrite a, rw_mem_erase_add_osi; auto.
  exists None; auto.
  assert ((cache_erase s') = (cache_erase s)) as H_cache;
    try (rewrite H_s'; auto; fail).
  assert ((tlb_erase s') = (tlb_erase s)) as H_tlb;
    try (rewrite H_s'; auto; fail).

  unfold state_erase; simpl.
  rewrite <- H_oss.
  rewrite <- H_exec_mode.
  rewrite <- H_activity.
  rewrite <- H_hypervisor.
  rewrite <- H_memory.
  rewrite <- H_cache.
  rewrite <- H_tlb; auto.
  (* osi not active *)
  apply os_running_erase_valid with (Page_pin_trusted p p0); auto.
  unfold other_osi_action.
  split; auto.
  right; right; right; right;
    right; right; right; right; left.
  exists p; exists p0; auto.

  (* Page unpin trusted *)
  unfold action_erase.
  unfold pre_action, pre_page_unpin_trusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_os_running [H_aos_trst [system H_system_map
    [H_p_not_curr [p2m H_p2m_map [ma H_ma_map [H_ma_ptype H_no_va]]]]]]].
  unfold os_running in H_os_running.
  rewrite H_os_running.
  induction (os_ident_eq (st_active_os s) osi).
  (* osi action *)
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_page_unpin_trusted, pre_all_unpin.
  repeat split.
  unfold os_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); try contradiction b; auto.
  unfold trusted_aos in *.
  rewrite a in H_aos_trst.
  unfold state_erase; auto.
  rewrite a in H_p2m_map.
  rewrite a in H_system_map.
  unfold state_erase; simpl.
  exists system.
  rewrite oss_erase_osi; auto.
  split; auto.
  unfold hypervisor_mapping in *.
  exists p2m.
  rewrite hyper_erase_osi; auto.
  exists ma; auto.
  split.
  unfold page_type in *.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as
    [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma);
    try (contradiction H_ma_ptype; auto; fail).
  exists v; auto.
  rewrite <- H_pg_map in H_ma_ptype.
  assert (page_owned_by pg = Os_PO osi) as H_pg_owner.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  assert (exists content, content = page_content pg) as [content H_pg_content].
  induction (page_content pg);
    [exists (RW_c o) | exists (PT_c m) | exists Other_c]; auto.
  induction content.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  exists o; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pg m; auto.
  simpl.
  intro H_abs; discriminate H_abs.
  rewrite <- H_pg_content in H_ma_ptype; contradiction H_ma_ptype; auto.
  unfold no_va_mapped_to_ma in *.
  simpl.
  intros pt_addr pt_page va mp H_pt_page_map H_pt_page_owner H_pt_page_content H_va_to_ma.
  elim pt_mem_erase_old_mem with (st_memory s) pt_addr pt_page mp; auto.
  intros oldmap H_oldmap_erase [old_pt_page H_old_pt_page_map H_old_pt_apge_content].
  destruct H_no_va with pt_addr old_pt_page va oldmap; auto.
  rewrite a.
  rewrite <- mem_erase_old_owner with (st_memory s) pt_addr.
  unfold maddress_owner.
  rewrite <- H_old_pt_page_map; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  rewrite H_oldmap_erase in H_va_to_ma.
  rewrite mp_erase_old with oldmap va ma; auto.
  (* post-condition *)
  unfold post_action, post_page_unpin_trusted, post_all_unpin in *.
  destruct H_post as [p2m' H_p2m'_map [ma' H_ma'_map [system' H_system'_map H_s']]].
  unfold hypervisor_mapping in *.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold state_erase; simpl.
  rewrite a in H_p2m_map.
  rewrite a in H_system_map.
  exists p2m.
  rewrite hyper_erase_osi; auto.
  exists ma; auto.
  exists system.
  rewrite oss_erase_osi; auto.
  rewrite <- oss_erase_add_osi.
  assert (activity_erase s = activity_erase s') as H_activity.
  unfold activity_erase.
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite <- H_activity.
  replace (exec_erase s') with (exec_erase s).
  replace (hyper_erase (st_hypervisor s')) with (map_add os_ident_eq (hyper_erase (st_hypervisor s)) osi
        (map_drop padd_eq p2m p)).
  replace (mem_erase (st_memory s')) with (map_add madd_eq (mem_erase (st_memory s)) ma
      (Page Other_c No_Owner)).
  assert ((cache_erase s') = (cache_erase s)) as H_cache;
    try (rewrite H_s'; auto; fail).
  assert ((tlb_erase s') = (tlb_erase s)) as H_tlb;
    try (rewrite H_s'; auto; fail).

  rewrite H_cache.
  rewrite H_tlb; auto.
  rewrite H_s'; simpl.
  rewrite a; auto.
  
  rewrite H_s'; simpl.
  rewrite other_mem_erase_add_osi; auto.
  rewrite H_s'; simpl.
  rewrite <- hyper_erase_add_osi.
  rewrite <- a; auto.
  unfold exec_erase.
  unfold trusted_aos in H_aos_trst.
  rewrite a in H_aos_trst.
  rewrite H_aos_trst; auto.
  apply os_running_erase_valid with (Page_unpin_trusted p); auto.
  unfold other_osi_action.
  split; auto.
  right; right; right; right;
    right; right; right; right; right.
  exists p; auto.

  (* Page pin untrusted action *)
  unfold action_erase.
  unfold pre_action, pre_page_pin_untrusted, pre_all_pin in H_pre.
  destruct H_pre as [H_aos_untrst [H_hcall [H_hyper_running [[p2m H_p2m_map H_error]
    [ma [pg H_pg_map [H_pg_owner H_pg_content]]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  unfold get_os_hypercall in H_hcall.
  destruct H_hcall as [system H_system_map H_hcall].
  induction (os_ident_eq o osi).
  (* osi action *)
  subst o.
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_page_pin_untrusted, pre_all_pin.
  repeat split; auto.
  unfold get_os_hypercall.
  exists system; auto.
  unfold state_erase; simpl.
  rewrite oss_erase_osi; auto.
  unfold hyper_running, state_erase; simpl.
  unfold activity_erase.
  induction (os_ident_eq (st_active_os s) osi); auto.
  unfold state_erase; simpl.
  exists p2m; auto.
  rewrite hyper_erase_osi; auto.
  unfold state_erase; simpl.
  exists ma; auto.
  exists (Page Other_c No_Owner).
  apply other_mem_erase_not_osi with pg; auto.
  rewrite H_pg_owner; intro H_abs; discriminate H_abs.
  simpl; split; auto.
  (* post-condition *)
  unfold post_action, post_page_pin_untrusted, post_all_pin in *.
  destruct H_post as [ma' [pg' H_pg'_map [H_pg'_owner [H_pg'_content
    [p2m' H_p2m'_map [system' H_system'_map H_s']]]]]].
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  unfold state_erase; simpl.
  exists ma'.
  exists (Page Other_c No_Owner).
  rewrite other_mem_erase_not_osi with (st_memory s) ma' pg'; auto.
  rewrite H_pg'_owner; intro H_abs; discriminate H_abs.
  repeat split; auto.
  exists p2m.
  unfold hypervisor_mapping.
  rewrite hyper_erase_osi; auto.
  exists system.
  rewrite oss_erase_osi; auto.
  rewrite <- oss_erase_add_osi.
  assert (activity_erase s = activity_erase s') as H_activity.
  unfold activity_erase.
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite <- H_activity.
  replace (exec_erase s') with (exec_erase s).
  replace (hyper_erase (st_hypervisor s')) with
    (map_add os_ident_eq (hyper_erase (st_hypervisor s)) osi
        (map_add padd_eq p2m p ma')).
  replace (mem_erase (st_memory s')) with
    (map_add madd_eq (mem_erase (st_memory s)) ma'
      match p0 with
        | PT => Page (PT_c (map_empty vadd madd)) (Os_PO osi)
        | RW => Page (RW_c None) (Os_PO osi)
      end).
  replace (cache_erase s') with (cache_erase s).
  replace (tlb_erase s') with (tlb_erase s).
  rewrite H_s'; simpl.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  auto.
  unfold tlb_erase.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; simpl.
  induction p0.
  rewrite pt_mem_erase_add_osi; auto.
  rewrite rw_mem_erase_add_osi; auto.
  exists None; auto.
  rewrite <- hyper_erase_add_osi.
  rewrite H_s'; auto.
  unfold exec_erase.
  rewrite H_aos_untrst; rewrite H_activity; auto.
  apply hyper_action_erase_valid with (Page_pin_untrusted o p p0); auto.
  unfold hyper_action_not_osi.
  right; right; right; right;
    right; right; right; right;
      right; right; left.
  exists o; split; auto; exists p; exists p0; auto.
  
  (* Page unpin untrusted action *)
  unfold action_erase.
  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin in H_pre.
  destruct H_pre as [H_aos_untrst [H_hcall [H_hyper_running
    [system H_system_map [H_p_not_curr [p2m H_p2m_map
      [ma H_ma_map [H_ma_ptype H_no_va]]]]]]]].
  unfold hyper_running in H_hyper_running.
  rewrite H_hyper_running.
  unfold get_os_hypercall in H_hcall.
  destruct H_hcall as [system' H_system'_map H_hcall].
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  unfold hypervisor_mapping in H_p2m_map.
  induction (os_ident_eq o osi).
  (* osi action *)
  subst o.
  constructor.
  (* pre-condition *)
  unfold pre_action, pre_page_unpin_untrusted, pre_all_unpin.
  unfold state_erase; simpl.
  repeat split; auto.
  unfold get_os_hypercall.
  exists system; auto.
  simpl.
  rewrite oss_erase_osi; auto.
  unfold hyper_running, activity_erase; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  exists system; auto.
  rewrite oss_erase_osi; auto.
  split; auto.
  exists p2m.
  unfold hypervisor_mapping.
  rewrite hyper_erase_osi; auto.
  exists ma; auto.
  unfold page_type in H_ma_ptype.
  assert (exists pg, Value _ pg = map_apply madd_eq (st_memory s) ma) as
    [pg H_pg_map].
  induction (map_apply madd_eq (st_memory s) ma);
    try (contradiction H_ma_ptype; auto; fail).
  exists v; auto.
  rewrite <- H_pg_map in H_ma_ptype.
  assert (page_owned_by pg = Os_PO osi) as H_pg_owner.
  elim hyper_mapping_madd_owner with s p ma osi; auto.
  unfold maddress_owner.
  rewrite <- H_pg_map; auto.
  unfold hypervisor_mapping_madd.
  exists p2m; auto.
  assert (exists content, content = page_content pg) as [content H_pg_content].
  induction (page_content pg);
    [exists (RW_c o) | exists (PT_c m) | exists Other_c]; auto.
  rewrite <- H_pg_content in H_ma_ptype.
  split.
  unfold page_type.
  induction content.
  rewrite <- rw_mem_erase_osi with (st_memory s) ma pg; auto.
  rewrite <- H_pg_content; auto.
  exists o; auto.
  rewrite <- pt_mem_erase_osi with (st_memory s) ma pg m; auto.
  contradiction H_ma_ptype; auto.
  unfold no_va_mapped_to_ma in *.
  simpl.
  intros pt_addr pt_page va mp H_pt_page_map H_pt_page_owner H_pt_page_content H_va_to_ma.
  elim pt_mem_erase_old_mem with (st_memory s) pt_addr pt_page mp; auto.
  intros oldmap H_oldmap_erase [old_pt_page H_old_pt_page_map H_old_pt_apge_content].
  destruct H_no_va with pt_addr old_pt_page va oldmap; auto.
  rewrite <- mem_erase_old_owner with (st_memory s) pt_addr.
  unfold maddress_owner.
  rewrite <- H_old_pt_page_map; auto.
  unfold maddress_owner.
  rewrite <- H_pt_page_map; auto.
  rewrite H_oldmap_erase in H_va_to_ma.
  rewrite mp_erase_old with oldmap va ma; auto.
  (* post-condition *)
  unfold post_action, post_page_unpin_untrusted, post_all_unpin in *.
  destruct H_post as [p2m' H_p2m'_map [ma' H_ma'_map [system' H_system'_map H_s']]].
  unfold hypervisor_mapping in H_p2m'_map.
  rewrite <- H_p2m_map in H_p2m'_map.
  injection H_p2m'_map; intro; subst p2m'; clear H_p2m'_map.
  rewrite <- H_system_map in H_system'_map.
  injection H_system'_map; intro; subst system'; clear H_system'_map.
  rewrite <- H_ma_map in H_ma'_map.
  injection H_ma'_map; intro; subst ma'; clear H_ma'_map.
  unfold state_erase; simpl.
  exists p2m.
  unfold hypervisor_mapping.
  rewrite hyper_erase_osi; auto.
  exists ma; auto.
  exists system.
  rewrite oss_erase_osi; auto.
  rewrite <- oss_erase_add_osi.
  assert (activity_erase s = activity_erase s') as H_activity.
  unfold activity_erase.
  rewrite H_s'; simpl.
  induction (os_ident_eq (st_active_os s) osi); auto.
  rewrite <- H_activity.
  replace (exec_erase s') with (exec_erase s).
  replace (hyper_erase (st_hypervisor s')) with (map_add os_ident_eq (hyper_erase (st_hypervisor s)) osi
        (map_drop padd_eq p2m p)).
  replace (mem_erase (st_memory s')) with (map_add madd_eq (mem_erase (st_memory s)) ma
      (Page Other_c No_Owner)).
  replace (cache_erase s') with (cache_erase s).
  replace (tlb_erase s') with (tlb_erase s).
  rewrite H_s'; simpl; auto.
  rewrite H_s'; auto.
  rewrite H_s'; auto.
  rewrite H_s'; simpl.
  rewrite other_mem_erase_add_osi; auto.
  rewrite H_s'; simpl.
  rewrite <- hyper_erase_add_osi; auto.
  unfold exec_erase.
  rewrite H_aos_untrst; auto.
  rewrite <- H_activity.
  induction (activity_erase s); auto.
  apply hyper_action_erase_valid with (Page_unpin_untrusted o p); auto.
  unfold hyper_action_not_osi.
  right; right; right; right;
    right; right; right; right;
      right; right; right.
  exists o; split; auto; exists p; auto.
Qed.

End Erasure.



